/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.repository.common;

import java.net.URI;

/**
 * ContentMovedException is thrown when the content of a given resource has been
 * moved permanently to a different location. Throwing this exception will result 
 * in 301 HTPP Status code being set as the response to a HTTP Request.
 * The Location Header of the HTTP Response will contain the new location of the 
 * resource, which can be used to perform any action on.
 * 
 */
@SuppressWarnings("serial")
public class ContentMovedException extends TeamRepositoryException {

    /**
     * New Content URI
     */
    private URI newContentURI;
    
    /**
     * Exception used to indicate that the content has been moved permanently. Throwing this 
     * exception will result in a 301 HTTP Status code being set on the response with Location 
     * header set to the given URI during the creation of the exception.
     *  
     * @param message a message that should be displayed when this error is thrown.
     *  
     * @param newContentURI A URI to where the content has been moved to. Must be absolute.
     * Must not be <code>null</code>.
     */
    public ContentMovedException(String message, final URI newContentURI) {
        super(message);
        if (newContentURI == null) {
            throw new IllegalArgumentException("content URI cannot be null."); //$NON-NLS-1$
        }
        if (!newContentURI.isAbsolute()) {
            throw new IllegalArgumentException("content URI must be absolute."); //$NON-NLS-1$
        }
        this.newContentURI = newContentURI;
    }
    
    /**
     * Return a new URI to the content. 
     * 
     * @return A URI to where the content has been moved to. It is absolute URI and will never
     *         be <code>null</code>.
     */
    public URI getContentURI() {
        return newContentURI;
    }
    
}
