/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.repository.common;

import java.net.URL;
import java.util.Map;

import org.eclipse.core.runtime.IPath;

import com.ibm.team.repository.common.internal.util.InternalTeamPlatform;

/**
 * This class contains helper methods aimed at finding files in bundles. This
 * class can be used regardless if OSGi plugin is available and will work in
 * plain java clients.
 * <p>
 * The class is not intended to be subclassed or instantiated by clients.
 * </p>
 * <p>
 * For argument substitution to work under plain java clients, the following
 * system variables must be set: <code>osgi.nl</code>, <code>osgi.os</code>,
 * <code>osgi.ws</code>, <code>osgi.arch</code>.
 * </p>
 * 
 * @see org.eclipse.core.runtime.FileLocator for more documentation on argument
 *      substitution.
 * 
 * @since 0.5
 */
public final class FileLocator {

    /**
     * Prevent this class from being instantiated.
     */
    private FileLocator() {
        // do nothing
    }

    /**
     * Returns a URL for the given path in the given bundle. Returns
     * <code>null</code> if the URL could not be computed or created.
     * <p>
     * find looks for this path in given bundle and any attached fragments.
     * <code>null</code> is returned if no such entry is found. Note that
     * there is no specific order to the fragments.
     * </p>
     * 
     * @param bundleSymbolicName
     *            the bundle's symbolic name in which to search; must not be
     *            <code>null</code>
     * @param path
     *            file path relative to plug-in installation location
     * @return a URL for the given path or <code>null</code>. The actual form
     *         of the returned URL is not specified.
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public static URL find(String bundleSymbolicName, IPath path) {
        return InternalTeamPlatform.getDefault().find(bundleSymbolicName, path,
                null);
    }

    /**
     * Returns a URL for the given path in the given bundle. Returns
     * <code>null</code> if the URL could not be computed or created.
     * <p>
     * find looks for this path in given bundle and any attached fragments.
     * <code>null</code> is returned if no such entry is found. Note that
     * there is no specific order to the fragments.
     * </p>
     * 
     * @param bundleSymbolicName
     *            the bundle's symbolic name in which to search; must not be
     *            <code>null</code>
     * @param path
     *            file path relative to plug-in installation location
     * @param override
     *            map of override substitution arguments to be used for any
     *            $arg$ path elements. The map keys correspond to the
     *            substitution arguments (eg. "$nl$" or "$os$"). If the map is
     *            <code>null</code>, or does not contain the required
     *            substitution argument, the default is used.
     * @return a URL for the given path or <code>null</code>. The actual form
     *         of the returned URL is not specified.
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public static URL find(String bundleSymbolicName, IPath path,
            Map<String, String> override) {
        return InternalTeamPlatform.getDefault().find(bundleSymbolicName, path,
                override);
    }

    /**
     * Returns all URLs for the given path in the given bundle. Returns empty
     * array if the URL could not be computed or created.
     * <p>
     * find looks for this path in given bundle and any attached fragments.
     * Empty array is returned if no such entry is found. Note that there is no
     * specific order to the fragments.
     * </p>
     * 
     * @param bundleSymbolicName
     *            the bundle's symbolic name in which to search; must not be
     *            <code>null</code>
     * @param path
     *            file path relative to plug-in installation location
     * @return all matching URLs for the given path; never <code>null</code>.
     *         The actual form of the returned URLs is not specified.
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public static URL[] findAll(String bundleSymbolicName, IPath path) {
        return InternalTeamPlatform.getDefault().findAll(bundleSymbolicName,
                path, null);
    }

    /**
     * Returns all URLs for the given path in the given bundle. Returns empty
     * array if the URL could not be computed or created.
     * <p>
     * find looks for this path in given bundle and any attached fragments.
     * Empty array is returned if no such entry is found. Note that there is no
     * specific order to the fragments.
     * </p>
     * 
     * @param bundleSymbolicName
     *            the bundle's symbolic name in which to search; must not be
     *            <code>null</code>
     * @param path
     *            file path relative to plug-in installation location
     * @param override
     *            map of override substitution arguments to be used for any
     *            $arg$ path elements. The map keys correspond to the
     *            substitution arguments (eg. "$nl$" or "$os$"). If the map is
     *            <code>null</code>, or does not contain the required
     *            substitution argument, the default is used.
     * @return all matching URLs for the given path; never <code>null</code>.
     *         The actual form of the returned URLs is not specified.
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public static URL[] findAll(String bundleSymbolicName, IPath path,
            Map<String, String> override) {
        return InternalTeamPlatform.getDefault().findAll(bundleSymbolicName,
                path, override);
    }
}
