/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010, 2015. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.repository.common;

import com.ibm.team.repository.common.model.RepositoryPackage;

/**
 * Represents an auditable item. Auditable items exist in multiple states in a
 * repository. There is a global notion of the current state of an auditable
 * item; states other than the current are used for an audit trail.
 * <p>
 * Item handles and items are passive data objects that are retrieved or created
 * while the client is logged in to a repository. These objects remain valid if
 * the client logs out.
 * </p>
 * <p>
 * [TODO Issue: Deleting auditable items. Currently the client API does allow
 * auditable items to be deleted. This keeps the sync story nice and simple, but
 * it does mean that items will accumulate over time with no way of cleaning
 * things up. Imagine how many contributor workspaces there would be in the
 * Eclipse Platform team after 5 years of development. Kinda messy, to be sure.
 * The big problem with actually deleting auditable items is that the audit
 * trail would disappear too, which would be bad. This could be dealt with by
 * dumping a journal of deleted item state elsewhere, just to satisfy the
 * auditors. On the other hand, if we added a "marked as deleted - hide from end
 * user" flag to auditable items, we could counter the clutter while keeping the
 * story otherwise the same as it is now.
 * 
 * MarkCC: I think the answer here is to distinguish between user operations and
 * administrative operations; and administrative operations are not part of the
 * client API. Removing an auditable from a *user* perspective doesn't actually
 * delete it from the repository; it leaves it in the repository, but marks it
 * as having no current state - just the log of previous states. In a separate
 * administrators API, we should include a set of "pruning" operations, which
 * can actually remove auditables and unneeded version sequences (i.e., version
 * sequences from a workspace which was discarded, and which are no longer of
 * interest.)
 * 
 * jeem: It's not clear yet whether there will ever be any unsophisticated
 * clients for the client API. I believe we can get away with telling clients
 * that there are these "deleted" items which they should suppress when
 * presenting to end users (and maybe providing them with an AccessProfile that
 * suppresses deleted items from the results). This would be simpler than having
 * a separate administrators API.]
 * </p>
 * <p>
 * [TODO Issue: Transactions for auditable items. Currently the client API does
 * not allow a client to create or modify multiple auditable items and then save
 * the new states in a single transaction. Consider adding a bulk
 * save(Collection <IAuditable>) to go with the bulk fetchCurrent(Collection
 * <IAuditable>, AccessProfile). Might entail adding a multi-status object to
 * RepositoryException so that failure info from the save can be communicated
 * back to the client.
 * 
 * MarkCC: This is definitely important, and not just for auditables. For the
 * extensibility story, we need to be able to take collections of mutating
 * operations and have them applied to the repository as a single atomic unit.
 * This is sort of like a changeset - a gathering of changes into a single unit
 * which is transmitted to the server and either accepted as a whole, or
 * discarded. Perhaps we should expand changeset to include auditable updates?]
 * </p>
 * <p>
 * [TODO Issue: Locking for auditable items. Currently the client API does not
 * allow a client to lock or reserve an auditable item that they are about to
 * update. In situations where there may be a number of users vying to update
 * the same item simultaneously, it may become difficult for users to get work
 * done. The problem is magnified if there are large auditable item states where
 * each user retrieves and updates just a small subpart, because the platform
 * will have to treat these as conflicts even if there is an easy and automatic
 * way to handle most of them. The schemas for the standard auditable items are
 * all dead easy; it's likely that work items will push on this matter (cf.
 * Bugzilla "Mid-air collision"). We should be on the lookout for signs that we
 * will need to add locking sooner rather than later.]
 * </p>
 * <p>
 * [TODO Issue: Since save() only operates on a working copy, it cannot be used
 * to replicate a state between repositories while preserving the state id. This
 * means that we need to revisit the story of how the client prepares to go
 * offline and how the resync when they come back online.]
 * </p>
 * 
 * 
 * @since 0.5
 * @NoImplement This interface is not intended to be implemented by clients.
 */
public interface IAuditable extends IAuditableHandle, IManagedItem {

    /*
     * Property names supported by all auditable item types.
     */

    /**
     * The predecessor state property.
     * 
     * @see #getPredecessorState()
     */
    public static final String PREDECESSOR_STATE_PROPERTY = RepositoryPackage.eINSTANCE
            .getAuditable_Predecessor().getName();

    /**
     * The merge predecessor state property.
     * 
     * @see #getMergePredecessorState()()
     */
    public static final String MERGE_PREDECESSOR_STATE_PROPERTY = RepositoryPackage.eINSTANCE
            .getAuditable_MergePredecessor().getName();

    /**
     * Returns the primary predecessor state of this auditable item state. The
     * predecessor state is another state of the same audtiable item. When
     * <code>IRepositoryItemService.fetchAllStates(IAuditableHandle)</code> is
     * used to retrieve all the known states of the item, the predecessor (and
     * merge predecessor) states of each state can be used to construct the
     * version graph.
     * <p>
     * The initial state of an item has no predecessor state.
     * </p>
     * 
     * @return the predecessor state, or <code>null</code> if none
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public IAuditableHandle getPredecessorState();

    /**
     * Returns the secondary merge predecessor state of this auditable item
     * state. The merge predecessor state is another state of the same auditable
     * item, and is always used in conjuction with the main predecessor state.
     * When <code>IRepositoryItemService.fetchAllStates(IAuditableHandle)</code>
     * is used to retrieve all the known states of the item, the predecessor
     * (and merge predecessor) states of each state can be used to construct the
     * version graph.
     * <p>
     * The initial state of an item has no predecessor state.
     * </p>
     * <p>
     * Design note: Without loss of generality, 0..2 predecessors is sufficient.
     * Most states have exactly 1 predecessor. The initial state has 0
     * predecessors. And a state that comes about from a 3-way merge has 2
     * predecessors, tops.
     * </p>
     * 
     * @return the merge predecessor state, or <code>null</code> if none
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public IAuditableHandle getMergePredecessorState();
}
