/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.repository.common;

import java.util.List;

import com.ibm.team.repository.common.model.RepositoryPackage;

/**
 * An auditable baseline is a collection of specific states of a group of
 * auditable items. The members of the collection may be of different item
 * types, but all belong to the same repository.
 * 
 * @since 0.5
 * @NoImplement This interface is not intended to be implemented by clients.
 */
public interface IAuditableBaseline extends IAuditableBaselineHandle, IAuditable {

    /**
     * Auditable baseline item type.
     */
    public static final IItemType ITEM_TYPE = IItemType.IRegistry.INSTANCE
            .getItemType(
                    RepositoryPackage.eINSTANCE.getAuditableBaseline().getName(),
                    RepositoryPackage.eNS_URI);

    /*
     * Property names supported by the auditable baseline item type.
     */

    /**
     * The name property.
     * 
     * @see #getName()
     */
    public static final String NAME_PROPERTY = RepositoryPackage.eINSTANCE
            .getAuditableBaseline_Name().getName();

    /**
     * The owner property.
     * 
     * @see #getOwner()
     */
    public static final String OWNER_PROPERTY = RepositoryPackage.eINSTANCE
            .getAuditableBaseline_Owner().getName();

    /**
     * The member states property.
     * 
     * @see #getMemberStates()
     */
    public static final String MEMBER_STATES_PROPERTY = RepositoryPackage.eINSTANCE
            .getAuditableBaseline_Members().getName();

    /**
     * Get the name of the baseline.
     * 
     * @return the name of the baseline; never <code>null</code>
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public String getName();

    /**
     * Set the name of the baseline.
     * 
     * @param name
     *            the new name for the baseline; must not be <code>null</code>
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public void setName(String name);

    /**
     * Get the owner of the baseline.
     * 
     * @return the owner of the baseline; never <code>null</code>
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public IContributorHandle getOwner();

    /**
     * Set the owner of the baseline.
     * 
     * @param owner
     *            the new owner of the baseline; must not be <code>null</code>
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public void setOwner(IContributorHandle owner);

    /**
     * Get the set of the auditable states that are included in the baseline.
     * This property of auditable baselines is immutable.
     * 
     * @return Unmodifiable list (element type: {@link IAuditableHandle} s.t.
     *         {@link IAuditableHandle#hasStateId()} is <code>true</code>) of
     *         states captured in the baseline; no duplicates; no particular
     *         order; never <code>null</code>
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public List getMemberStates();

}
