/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.repository.common;

import java.util.List;

/**
 * Expresses the differences and similarities between two auditable baselines.
 * These two baselines are identified as the "prior" baseline and the "after"
 * baseline, although no particular time ordering is assumed.
 * 
 * @since 0.5
 * @NoImplement This interface is not intended to be implemented by clients.
 */
public interface IBaselineComparison {

    /**
     * Get the items added to the after baseline that were absent in the prior
     * baseline.
     * 
     * @return List (element type: {@link IAuditableHandle} s.t.
     *         {@link IAuditableHandle#hasStateId()} is <code>true</code>) of
     *         added items; no duplicates; no particular order; never
     *         <code>null</code>
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public List getAddedItems();

    /**
     * Get the items removed from the after baseline that were present in the
     * prior baseline.
     * 
     * @return List (element type: {@link IAuditableHandle} s.t.
     *         {@link IAuditableHandle#hasStateId()} is <code>true</code>) of
     *         removed items; no duplicates; no particular order; never
     *         <code>null</code>
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public List getRemovedItems();

    /**
     * Get the items that are present <em>in the same state</em> in both
     * baselines.
     * 
     * @return List (element type: {@link IAuditableHandle} s.t.
     *         {@link IAuditableHandle#hasStateId()} is <code>true</code>) of
     *         identical item states; no duplicates; no particular order; never
     *         <code>null</code>
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public List getIdenticalItems();

    /**
     * Get the items that are present <em>in different states</em> in both
     * baselines.
     * 
     * @return List (element type: {@link IBaselineMemberComparison}) of
     *         items whose state changed between baselines; no duplicates; no
     *         particular order; never <code>null</code>
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public List getChangedItems();

}
