/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010, 2013. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.repository.common;

import java.sql.Timestamp;
import java.util.List;

import com.ibm.team.repository.common.model.ChangeEvent;
import com.ibm.team.repository.common.model.Item;
import com.ibm.team.repository.common.model.RepositoryPackage;

/**
 * Represents a modification to the repository.
 * 
 * The repository records an event for each modification made to it, and
 * associates that event with the repository object modified.
 * 
 * @since 0.5
 * @NoImplement This interface is not intended to be implemented by clients.
 */
public interface IChangeEvent extends ISimpleItem, IChangeEventHandle {

    /**
     * ChangeEvent item type.
     */
    public static final IItemType ITEM_TYPE = IItemType.IRegistry.INSTANCE
            .getItemType(
                    RepositoryPackage.eINSTANCE.getChangeEvent().getName(),
                    RepositoryPackage.eNS_URI);

    /**
     * The author property.
     * 
     * @see #getAuthor()
     */
    public static final String AUTHOR_PROPERTY = RepositoryPackage.eINSTANCE
            .getChangeEvent_EventAuthor().getName();

    /**
     * The category property.
     * 
     * @see #getCategory()
     */
    public static final String CATEGORY_PROPERTY = RepositoryPackage.eINSTANCE
            .getChangeEvent_EventCategory().getName();

    /**
     * The contributors property.
     * 
     * @see #getContributors()
     */
    public static final String CONTRIBUTORS_PROPERTY = RepositoryPackage.eINSTANCE
            .getChangeEvent_EventContributors().getName();

    /**
     * The description property.
     * 
     * @see #getDescription()
     */
    public static final String DESCRIPTION_PROPERTY = RepositoryPackage.eINSTANCE
            .getChangeEvent_EventDescription().getName();

    /**
     * The expiration date property.
     * 
     * @see #getExpiration()
     * 
     */
    public static final String EXPIRATION_PROPERTY = RepositoryPackage.eINSTANCE
            .getChangeEvent_EventExpiration().getName();
    
    /**
     * The item property.
     * 
     * @see #getItem()
     */
    public static final String ITEM_PROPERTY = RepositoryPackage.eINSTANCE
            .getChangeEvent_Item().getName();

    /**
     * The packageNS property.
     * 
     * @see #getPackageNS()
     */
    public static final String OWNERNS_PROPERTY = RepositoryPackage.eINSTANCE
            .getChangeEvent_EventOwnerNS().getName();

    /**
     * The priority property.
     * 
     * @see #getPriority()
     */
    public static final String PRIORITY_PROPERTY = RepositoryPackage.eINSTANCE
            .getChangeEvent_EventPriority().getName();

    /**
     * The process area property.
     * 
     * @see #getProcessArea()
     */
    public static final String PROCESS_AREA_PROPERTY = RepositoryPackage.eINSTANCE
            .getChangeEvent_EventProcessArea().getName();

    /**
     * The time property.
     * 
     * @see #getTime()
     */
    public static final String TIME_PROPERTY = RepositoryPackage.eINSTANCE
            .getChangeEvent_EventTime().getName();

    /**
     * The title property.
     * 
     * @see #getTitle()
     */
    public static final String TITLE_PROPERTY = RepositoryPackage.eINSTANCE
            .getChangeEvent_EventTitle().getName();

    /**
     * Returns a handle to the {@link Item} that was modified.
     * 
     * @return A handle to the modified {@link Item}. Can be <code>null</code>.
     * 
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public IItemHandle getItem();

    /**
     * Sets the reference to the item to be associated with this ChangeEvent.
     * 
     * @param itemHandle
     *            A handle to the item associated with this ChangeEvent. May be
     *            <code>null</code>.
     * 
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public void setItem(IItemHandle itemHandle);

    /**
     * Returns the namespace of the component that created this event, e.g.
     * {@link RepositoryPackage#eNS_URI}.
     * 
     * @return The namespace of the package that created this event. Can be
     *         <code>null</code>.
     * 
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public String getOwnerNamespace();

    /**
     * Sets the namespace of the component that created this event, e.g.
     * {@link RepositoryPackage#eNS_URI}.
     * 
     * @param namespace
     *            The namespace of the package that created this event. May be
     *            <code>null</code>.
     * 
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public void setOwnerNamespace(String namespace);

    /**
     * Returns the date/time associated with this repository modification.
     * 
     * @return The date/time associated with this repository modification. If
     * this item was retrieved from the repository, it will never be <code>null</code>.
     * If non-<code>null</code>, the value will be that as set by {@link #setTime(Timestamp)}
     * (if non-null), or the repository time at the time this item was created.
     * 
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public Timestamp getTime();

    /**
     * Sets the date/time associated with this repository modification.
     * 
     * @param time The date/time associated with this repository modification.
     *  May be <code>null</code>.  If <code>null</code> and this item is saved,
     *  then the value set will be that of the last modified date of the item
     *  associated with this change event.  If there is no item, then the value
     *  set will be that of the current system time of the repository to which
     *  it's saved.
     * 
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public void setTime(Timestamp time);

    /**
     * Returns a handle to the contributor that created this repository
     * modification.
     * 
     * @return A handle to the contributor that created this repository
     *         modification. Can be <code>null</code>.
     * 
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public IContributorHandle getAuthor();

    /**
     * Sets the contributor that created this repository modification.
     * 
     * @param author
     *            A handle to the contributor that created this repository
     *            modification. Can be <code>null</code>.
     * 
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public void setAuthor(IContributorHandle author);

    /**
     * Returns the title or headline for this event.
     * 
     * @return The title for the event. Can be <code>null</code>.
     * 
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public String getTitle();

    /**
     * Sets the title or headline for this event.
     * 
     * @param title
     *            The title for the event. Can be <code>null</code>.
     * 
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public void setTitle(String title);

    /**
     * Returns the description of this event.
     * 
     * @return The description of this event. Can be <code>null</code>.
     * 
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public String getDescription();

    /**
     * Sets the description of this event.
     * 
     * @param description
     *            The description of this event. Can be <code>null</code>.
     * 
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public void setDescription(String description);

    /**
     * Returns the expiration date of this event.
     * 
     * @return The expiration date of this event. Can be <code>null</code>.
     * 
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public Timestamp getExpiration();
    
    /**
     * Sets the expiration date of this event. If unset, an expiration date will
     * be set by the server during the saving of this ChangeEvent, determined by
     * the server configuration.
     * 
     * @param expiration
     *            The expiration of this event. Can be <code>null</code>.
     * 
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public void setExpiration(Timestamp expiration);
    
    /**
     * Returns the list of contributors that participated in the creation of
     * this event. Typically the author ({@link #getAuthor()}) is not included
     * in this value, but the exact definition of this field value is per the
     * component package that created it. The order of the contributors in the
     * list represents the order of the contributors as they were initially
     * saved.
     * 
     * @return An unmodifiable list of {@link IContributorHandle} contributors
     *         that participated in the creation of this event. Never
     *         <code>null</code>.
     * 
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public List getContributors();

    /**
     * Sets the list of contributors that participated in the creation of this
     * event. Typically the author ({@link #getAuthor()}) is not included in
     * this value, but the exact definition of this field value is per the
     * component package that created it. The order of the contributors in the
     * list represents the order of the contributors as they will be saved (and
     * the order that will be present in subsequent requests to
     * {@link #getContributors()}).
     * 
     * @param contributors
     *            A collection of {@link IContributorHandle} contributors that
     *            participated in the creation of this event. The list of
     *            contributors replaces any contributors that may have been set
     *            on this instance. May be <code>null</code>. If
     *            <code>null</code>, the list of contributors on this
     *            instance is cleared. The list is not modified.
     * 
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public void setContributors(List contributors);

    /**
     * Returns the string representing the category of the event. The domain of
     * values of this string is defined by the component/package that created
     * the event.
     * 
     * @return The category of the ChangeEvent. Can be <code>null</code>.
     * 
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public String getCategory();

    /**
     * Sets the string representing the category of the event. The domain of
     * values of this string is defined by the component/package that created
     * the event.
     * 
     * @param category
     *            The category of the ChangeEvent. Can be <code>null</code>.
     * 
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public void setCategory(String category);

    /**
     * Returns the string representing the priority of the event. The domain of
     * values of this string is defined by the component/package that created
     * the event.
     * 
     * @return The priority of the ChangeEvent. Can be <code>null</code>.
     * 
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public String getPriority();

    /**
     * Sets the string representing the priority of the event. The domain of
     * values of this string is defined by the component/package that created
     * the event.
     * 
     * @param priority
     *            The priority of the ChangeEvent. Can be <code>null</code>.
     * 
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public void setPriority(String priority);

    /**
     * Returns a reference to the process area that this ChangeEvent is associated
     * to.
     * 
     * While the signature of this method types an {@link IAuditableHandle},
     * the actual type of handle returned will be that of the process area that
     * is associated to this event.
     * 
     * @return  A handle to the process area that
     * this ChangeEvent is associated to.  Will be <code>null</code> if this
     * event has no such association.
     */
    public IAuditableHandle getProcessArea();
    
    /**
     * Sets the process area to be associated to this event.
     * 
     * @param processArea A handle to a process area.  May be <code>null</code>.
     * There is no verification, either client or server-side, that the referenced
     * auditable handle refers to a process area handle type.
     */
    public void setProcessArea(IAuditableHandle processArea);
    
    /**
     * Extends this ChangeEvent with the property named <code>key</code> with
     * the value <code>value</code>. Note that if this event is to be
     * rendered via the feed service, this extension will be rendered as a
     * custom item element in the syndication feed, and so the value of
     * <code>key</code> must be suitable for use as an XML element name (else
     * it will be rejected).
     * <p>
     * NOTE: only (short) string exceptions are supported for a change event -
     * other types supported by {@link IExtensibleItem} are NOT supported for
     * change events.
     */
    public void setStringExtension(String key, String value);
    
    /**
     * Gets the extension with the key, if it exists
     * <p>
     * NOTE: only (short) string exceptions are supported for a change event -
     * other types supported by {@link IExtensibleItem} are NOT supported for
     * change events.
     * @param key
     * @return
     */
    public String getStringExtension(String key);
    

    /**
     * Instance creation factory for initializing new {@link IChangeEvent}
     * instances from {@link IItem item-derived} types.
     * 
     * @NoImplement Clients must not implement this interface.
     */
    public interface Factory {

        /**
         * Instantiates an {@link IChangeEvent} instance from the referenced
         * <code>item</code>.
         * 
         * The item, state, type name, type namespace, package, author and time
         * field values are initialized from the referenced item - the time
         * derived from the last modified time of the referenced item. The item is
         * not persisted to any repository, and the server is not contacted.
         * 
         * @param item
         *            The item to initialize the ChangeEvent from. May be
         *            <code>null</code>.
         * 
         * @return The initialized ChangeEvent. Never <code>null</code>. The
         *         instantiated ChangeEvent is not persisted to the repository.
         * 
         * @ShortOp This is a short operation; it may block only momentarily;
         *          safe to call from a responsive thread.
         */
        public IChangeEvent newInstance(IItem item);
    }

    /**
     * Global factory instance for the creation of {@link IChangeEvent}
     * instances via the {@link Factory#newInstance(IItem)}, initialized with
     * the referenced <code>item</code>.
     * 
     * @see Factory#newInstance(IItem)
     */
    public static final Factory FACTORY = new Factory() {

        public IChangeEvent newInstance(IItem item) {

            ChangeEvent event = (ChangeEvent)IChangeEvent.ITEM_TYPE.createItem();
            if (item == null)
                return event;

            event.setItem(item);
            event.setEventOwnerNS(item.getItemType().getNamespaceURI());
            event.setEventAuthor(item.getModifiedBy());
            event.setTime(((Item)item).getModified());
            
            // set change event context id to be the same as item context
            if (item.getContextId() != null) {
                event.setContextId(item.getContextId());
            }

            return event;
        }
    };

}
