/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.repository.common;

import com.ibm.team.repository.common.utils.HashCode;

/**
 * Represents a description of content that is stored in the repository. The
 * actual contents (e.g. the bytes) are stored and retrieved using an
 * <code>IContentManager</code> and this description as the identity of the
 * bytes.
 * 
 * @since 0.5
 * @NoImplement This interface is not intended to be implemented by clients.
 */
public interface IContent {

    /**
     * Common Character encoding constants.  Other encodings may also be used.
     * 
     * @see #getCharacterEncoding()
     */
    public static final String ENCODING_US_ASCII = "us-ascii"; //$NON-NLS-1$
    
    public static final String ENCODING_UTF_8 = "UTF-8"; //$NON-NLS-1$
    
    public static final String ENCODING_UTF_16BE = "UTF-16BE"; //$NON-NLS-1$
    
    public static final String ENCODING_UTF_16LE = "UTF-16LE"; //$NON-NLS-1$

    /**
     * Common Content type constant (value &quot;text/plain&quot;) for text/plain.
     * 
     * @see #getContentType()
     */
    public static final String CONTENT_TYPE_TEXT = "text/plain"; //$NON-NLS-1$
    
    /**
     * Common Content type constant (value &quot;application/xml&quot;) for application/xml.
     * 
     * @see #getContentType()
     */
    public static final String CONTENT_TYPE_XML = "application/xml"; //$NON-NLS-1$    

    /**
     * Common Content type constant (value &quot;application/unknown&quot;) for
     * arbitrary binary content.
     * 
     * @see #getContentType()
     */
    public static final String CONTENT_TYPE_UNKNOWN = "application/unknown"; //$NON-NLS-1$
    
    /**
     * Returns the unique id of this content. The content id gives contents
     * their identity.
     * 
     * @return the content id; never <code>null</code>
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public UUID getContentId();

    /**
     * Returns the content type identifier that describes the nature of the
     * content. The identifier format follows the Internet Media Types
     * specification for media types (e.g. text/plain, application/unknown) as
     * defined in s RFC 2046.
     * <p>
     * The encoding will be specified separately by {@link #getCharacterEncoding()};
     * 
     * @return the content type identifier that describes the nature of the
     *         content; never <code>null</code>
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public String getContentType();

    /**
     * Returns the character encoding for the content. This is an IANA character
     * set encoding name such as "8859_1", "us-ascii", "iso-8859-1", "UTF8", or
     * "shift_jis". The official registry of "charset" (i.e., character
     * encoding) names, with references to documents defining their meanings, is
     * kept by IANA at http://www.iana.org/assignments/character-sets.
     * 
     * @return the character encoding for this content or <code>null</code> if
     *         unknown or not applicable to this content type.
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public String getCharacterEncoding();

    /**
     * Returns the length of the content referenced by this content handle.
     * This length is the raw size of the content.  The content
     * lineDelimiters have been converted according to {@link #getLineDelimiter()}
     * For content with {@link IContent#LINE_DELIMITER_PLATFORM}, the line delimiter
     * used during transmission is {@link IContent#LINE_DELIMITER_LF}.
     * 
     * @return the length of the content references by this content handle after converting
     * the line delimiters (as required).
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public long getRawLength();
    
    /**
     * Returns the checksum value for the content. This is a simple measure to
     * allow data integrity checks on the contents.  This is a checksum value
     * for contents after any line delimiter translations that may have taken
     * place prior to transmission.  This may not be the checksum of the
     * original contents requested to be stored if the line delimiters were converted.
     * This may not be the checksum of the contents retrieved from <code>IContentManager</code>
     * if line delimiters are translated.
     * 
     * @return the checksum value.  May be <code>null</code> if not known
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public HashCode getRawHashCode();

    /**
     * Returns the length of the content referenced by this content handle
     * after any conversion required by {@link #getLineDelimiter()}
     * In most cases this length will be the same as the {@link #getTransimissionContentLength()} 
     * In the case of content with {@link IContent#LINE_DELIMITER_PLATFORM} line delimiters,
     * the length may be different if the platform uses "CRLF" as the line delimiter.
     * In this case, the length may be longer than the transmission length
     * depending on the number of lines in the content.
     * 
     * @return the estimated length of the content referenced by this content handle
     * after conversion for agreement with the {@link #getLineDelimiter()}.
     * If this content's encoding is not known to the current platform, then this just
     * an estimation.
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public long getEstimatedConvertedLength();
    
    /**
     * The preferred line delimiter to be used when the content is retrieved.
     * Valid values include {@link #LINE_DELIMITER_PLATFORM}
     * {@link #LINE_DELIMITER_NONE}, {@link #LINE_DELIMITER_LF},
     * {@link #LINE_DELIMITER_CRLF}, {@link #LINE_DELIMITER_CR}
     * 
     * @returns the line delimiter for the content.
     * The default is {@link #LINE_DELIMITER_NONE}
     */
    public LineDelimiter getLineDelimiter();
}
