/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.repository.common;

import java.sql.Timestamp;
import java.util.List;

/**
 * @since 0.5
 */
public interface IContributorLicenseType extends ILicenseType {

    /**
     * Returns a short name of this CAL type.
     * 
     * @return a short name of this CAL type.
     */
    public String getName();
    
    /**
     * Returns the number of CALs of this type that have been assigned
     * to users.  This number might exceed the result of {@link getTotal()}
     * if the server is in a bad state and too many are assigned. 
     * 
     * @return the number of CALs that have been assigned
     */
    public int getUsed();
    
    /**
     * Returns the maximum number of CALs of this type that could ever 
     * be present on this server.  If this value is greater than the
     * result of {@link getTotal()} then more CALs can be acquired for this
     * server. If the result is -1 then an unlimited number of
     * CALs of this type can be present on this server.
     * 
     * @return the maximum number of CALs of this type that could ever 
     *         be present on this server or -1 if an unlimited number of
     *         CALs of this type can be present on this server
     */
    public int getMax();
    
    /**
     * Returns the maximum number of CALs of this type that could be assigned
     * to users without purchasing more.  If this number is greater than the result of 
     * {@link getUsed()} then some CALs are available to be assigned to a user
     * without purchasing any more.  If the result is -1 then an unlimited number of
     * CALs of this type can be assigned.
     * 
     * @return the total number of CALs that are available to this server or -1 if 
	 *         an unlimited number of CALs of this type can be assigned
     */
    public int getTotal();
    
    /**
     * Returns the number of CALs that have been included in the policy.  This
     * is the smallest number of CALs can can be present on the server. If the 
     * result is -1 then an unlimited number of CALs of this type have been
     * included in the license policy.
     * 
     * @return the number of CALs that are included in the policy or -1 if an 
     *         unlimited number of CALs of this type have been included in the 
	 *         license policy
     */
    public int getFree();
    
    /**
     * Returns true if an assignment of this CAL type is counted towards the
     * the server limit for the maximum number of users.  If this method returns
     * false, then a Contributor who has only this CAL assigned will not count 
     * towards the maximum number of licensed users for the server.  
     *   
     * @return true if this CAL type affects the total licensed user count
     */
    public boolean isCountedInServerLimit();

    /**
     * Returns true if this CAL type should only be assigned to functional
     * users. 
     * 
     * @return true if this CAL type should only be assigned to functional
     * users. 
     */
    public boolean isFunctional();
    
    /**
     * Returns true if this CAL type is a floating CAL type.
     * 
     * @return true if this CAL type is a floating CAL type.
     * @since 0.7
     */
    public boolean isFloating();
    
    public String getVersion();

    public Timestamp getUnactivatedTrialExpirationTime();

    public boolean hasUnactivatedTrial();

    public List<IContributorLicenseType> getKeys();
    
    public boolean isPurchaseCountLocked();

    String getActivationNotificationMessage();

    String getActivationNotificationURL();
    
}
