/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.repository.common;

import java.util.List;

import com.ibm.team.repository.common.model.RepositoryPackage;

/**
 * Represents a contributor known to the repository.
 * 
 * @since 0.7
 * @NoImplement This interface is not intended to be implemented by clients.
 */
public interface IContributorRecord extends IContributorRecordHandle, IAuditable {

    /**
     * Contributor item type.
     */
    public static final IItemType ITEM_TYPE = IItemType.IRegistry.INSTANCE
            .getItemType(
                    RepositoryPackage.eINSTANCE.getContributorRecord().getName(),
                    RepositoryPackage.eNS_URI);

    /*
     * Property names supported by the contributor item type.
     */

    /**
     * The name property.
     * 
     * @see #getName()
     */
    public static final String NAME_PROPERTY = RepositoryPackage.eINSTANCE
            .getContributorRecord_Name().getName();

    /**
     * The email address property.
     * 
     * @see #getEmailAddress()
     */
    public static final String EMAIL_ADDRESS_PROPERTY = RepositoryPackage.eINSTANCE
            .getContributorRecord_EmailAddress().getName();

    /**
     * The userIds property.
     * 
     * @see #getUserIds()
     */
    public static final String USERIDS_PROPERTY = RepositoryPackage.eINSTANCE
            .getContributorRecord_UserIds().getName();
    
    /**
     * The identities property.
     * 
     * @see #getIdentities()
     */
    public static final String IDENTITIES_PROPERTY = RepositoryPackage.eINSTANCE
            .getContributorRecord_Identities().getName();

    /**
     * The details property.
     * 
     * @see #getDetails()
     */
    public static final String DETAILS_PROPERTY = RepositoryPackage.eINSTANCE
            .getContributorRecord_Details().getName();
    
    /**
     * The archived property
     * 
     * @see #isArchived()
     */
    public static final String ARCHIVED_PROPERTY = RepositoryPackage.eINSTANCE
            .getContributorRecord_Archived().getName();

    /**
     * Returns the name.
     * 
     * @return the name; never <code>null</code>
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public String getName();

    /**
     * Sets the name.
     * 
     * @param name
     *            the name; must not be <code>null</code>
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public void setName(String name);

    /**
     * Returns the list of user ids for this contributor. A user id is typically used
     * for authentication purposes while the name is simply for display
     * purposes.
     * 
     * @return the list of userids; never <code>null</code>
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public List<IContributorUserId> getUserIds();

    /**
     * Returns the list of identities for this contributor. An identity is typically used
     * to reference this contributor by other items.
     * 
     * @return the list of identities; never <code>null</code>
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public List<IContributorIdentity> getIdentities();
    
    /**
     * Returns the email address.
     * 
     * @return the email address; never <code>null</code>
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public String getEmailAddress();

    /**
     * Sets the email address.
     * 
     * @param email
     *            the email address; must not be <code>null</code>
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public void setEmailAddress(String email);
    
    /**
     * Returns the contributor details.
     * 
     * @return the contributor details or <code>null</code>
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public IContributorDetailsHandle getDetails();

    /**
     * Sets the contributor details.
     * 
     * @param details
     *            the contributor details or <code>null</code>
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public void setDetails(IContributorDetailsHandle details);
    
    /**
     * Indicates if this contributor has been archived. Archived contributors
     * will be hidden from some UIs and cannot login to the repository.
     * 
     * @return boolean
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public boolean isArchived();
    
    /**
     * Sets the archived state of the contributor. Archived contributors will be
     * hidden from some UIs and cannot login to the repository.
     * 
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
     public void setArchived(boolean archived);
}