/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010, 2012. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.repository.common;

import java.math.BigDecimal;
import java.sql.Timestamp;
import java.util.Set;

import com.ibm.team.repository.common.model.RepositoryPackage;
import com.ibm.team.repository.common.query.ast.IDynamicQueryModel;

/**
 * <p>
 * IExtensibleItem defines methods available on items that allow it's state to
 * be augmented with additional data, referred to as extensions. This is
 * indicated by the model annotation
 * <code>teamClass->allowsStateExtensions</code> (default is false). When a
 * component "opts-in" to extensibility for an item type, they should add this
 * interface to the hierarchy of the facade class for their item (e.g.,
 * <code>IWorkItem extends IExtensibleItem</code>).
 * </p>
 * <p>
 * The data types permitted for extensions are a subset of the data types
 * allowed for "built-in" modeled attributes. Separate getters and setters exist
 * for each supported extension type.
 * <p>
 * <b>Design notes:</b> Fetching a complete (extensible) item will fetch all
 * it's state extensions. When fetching a partial item and state extensions are
 * desired, specify the property {@link #ALL_STATE_EXTENSIONS_PROPERTY}.
 * Currently for partial item fetches, state extensions are all or nothing; we
 * will be adding API in the future to fetch partial items with a subset of
 * extensions.
 * 
 * For any key, only one value may exist as a state extension. It is not legal
 * to have extensions with the same key and different value types.
 * </p>
 * <p>
 * The component that declares the item type controls whether its item states
 * can be augmented with extensions. If a component chooses to permit extensions
 * of its item type, it controls the extension keys that are allowed, and their
 * corresponding types. It does this in virtue of providing the only (service)
 * methods that can create and update items of that type, in which it is free to
 * lay down (and enforce) whatever restrictions it chooses.
 * </p>
 * <p>
 * Here are some important design considerations to think about when deciding
 * whether to allow state extensions on an item type:
 * <ul>
 * <li>Extensions persist until explicitly deleted. Once anyone saves a state
 * of an item X with an extension with key K and value V, then all subsequent
 * states of X will also have the extension K=V. This situation won't change
 * until someone explicitly retrieves X, creates a working copy, deletes the
 * extension for key K, and saves the working copy back as a new state of X.</li>
 * <li>A single item state can carry any number of extensions. The footprint of
 * an item state's extensions (in memory, in the repository, on the wire) grows
 * roughly linearly with the number of extensions.</li>
 * <li>When a complete item is retrieved, the keys and values of all its
 * extensions are automatically included, even if the party doing the retrieving
 * is uninterested/unaware of these extensions.</li>
 * <li>In order to access any of an item's extensions, it is necessary to
 * retrieve the keys and values of all of the item state's extensions.</li>
 * </ul>
 * Because of these considerations, state extensions should be used carefully
 * and sparingly. If a component writer decides to support state extension for
 * its item type, they should restrict the mechanism so that it is does not
 * become a serious problem at some point.
 */
public interface IExtensibleItem extends IItem {

    /**
     * Special property that indicates state extensions should be included in
     * partial item fetches, regardless of the value type
     */
    public static String ALL_STATE_EXTENSIONS_PROPERTY = "com.ibm.team.repository.item.allStateExtensions"; //$NON-NLS-1$

    /**
     * property used for dynamic queries of String extensions; generally when
     * the query structure is known at development time dynamic queries should
     * not be used.
     * 
     * @see IDynamicQueryModel#getReference(String)
     */
    public static final String STRING_EXTENSIONS_QUERY_PROPERTY = RepositoryPackage.eINSTANCE.getItem_StringExtensions().getName();

    /**
     * property used for dynamic queries of medium string extensions; generally
     * when the query structure is known at development time dynamic queries
     * should not be used.
     * 
     * @see IDynamicQueryModel#getReference(String)
     */
    public static final String MEDIUM_STRING_EXTENSIONS_QUERY_PROPERTY = RepositoryPackage.eINSTANCE.getItem_MediumStringExtensions().getName();

    /**
     * property used for dynamic queries of large string extensions; generally
     * when the query structure is known at development time dynamic queries
     * should not be used.
     * 
     * @see IDynamicQueryModel#getReference(String)
     */
    public static final String LARGE_STRING_EXTENSIONS_QUERY_PROPERTY = RepositoryPackage.eINSTANCE.getItem_LargeStringExtensions().getName();

    /**
     * property used for dynamic queries of int extensions; generally when the
     * query structure is known at development time dynamic queries should not
     * be used.
     * 
     * @see IDynamicQueryModel#getReference(String)
     */
    public static final String INT_EXTENSIONS_QUERY_PROPERTY = RepositoryPackage.eINSTANCE.getItem_IntExtensions().getName();

    /**
     * property used for dynamic queries of long extensions; generally when the
     * query structure is known at development time dynamic queries should not
     * be used.
     * 
     * @see IDynamicQueryModel#getReference(String)
     */
    public static final String LONG_EXTENSIONS_QUERY_PROPERTY = RepositoryPackage.eINSTANCE.getItem_LongExtensions().getName();

    /**
     * property used for dynamic queries of BigDecimal extensions; generally
     * when the query structure is known at development time dynamic queries
     * should not be used.
     * 
     * @see IDynamicQueryModel#getReference(String)
     */
    public static final String BIG_DECIMAL_EXTENSIONS_QUERY_PROPERTY = RepositoryPackage.eINSTANCE.getItem_BigDecimalExtensions().getName();

    /**
     * property used for dynamic queries of boolean extensions; generally when
     * the query structure is known at development time dynamic queries should
     * not be used.
     * 
     * @see IDynamicQueryModel#getReference(String)
     */
    public static final String BOOLEAN_EXTENSIONS_QUERY_PROPERTY = RepositoryPackage.eINSTANCE.getItem_BooleanExtensions().getName();

    /**
     * property used for dynamic queries of timestamp extensions; generally when
     * the query structure is known at development time dynamic queries should
     * not be used.
     * 
     * @see IDynamicQueryModel#getReference(String)
     */
    public static final String TIMESTAMP_EXTENSIONS_QUERY_PROPERTY = RepositoryPackage.eINSTANCE.getItem_TimestampExtensions().getName();

    /**
     * Returns the String value of the extension on this item state for the
     * given key, or <code>null</code> if the item state has no string
     * extension with the given key.
     * <p>
     * Use {@link #setStringExtension(String, String)} on an item working copy
     * to add or update string extensions. Use
     * {@link #unsetStringExtension(String)} to completely remove the extension
     * key/value pair from the item.
     * <p>
     * A return value of <code>null</code> implies one of the following:
     * <ul>
     * <li>The item does not have a state extension with that key
     * <li>The item has a state extension with that key, with a value of
     * <code>null</code>. This can be checked using
     * {@link #isStringExtensionSet(String)}
     * </ul>
     * 
     * @param key
     *            the state extension key to be used for accessing and storing
     *            the state extension value; must be a qualified name starting
     *            with the component id identifying which component is
     *            responsible for the given state extension; must not be
     *            <code>null</code> or an empty string
     * @return the value of this state extension, which may be <code>null</code>,
     *         or <code>null</code> if this item state doesn't have a state
     *         extension with the given key
     * @throws UnsupportedOperationException
     *             if the item type does not support state extensions
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public String getStringExtension(String key);

    /**
     * Sets the String value of the state extension on this item state as
     * identified by the given key. The item state must be a working copy.
     * <p>
     * String extensions are defined as SMALL strings, and thus the size of the
     * values is bounded by {@link IItem#MAX_SMALL_STRING_BYTES}. This will be
     * validated when an item working copy is saved.
     * <p>
     * When dealing with working copies created from partial item states,
     * specify the {@link #ALL_STATE_EXTENSIONS_PROPERTY} property when fetching
     * the item state from the repository, to ensure that all of its string
     * extensions are retrieved. If this is not done, you cannot add, update, or
     * remove state extensions of type String on the working copy based on that
     * item state.
     * </p>
     * <p>
     * 
     * @param key
     *            the state extension key to be used for accessing and storing
     *            the state extension value; must be a qualified name starting
     *            with by the component id identifying which component is
     *            responsible for the given state extension; must not be
     *            <code>null</code> or the empty string
     * @param value
     *            the value of the state extension, bounded by
     *            {@link IItem#MAX_SMALL_STRING_BYTES}; null is allowed
     * @throws UnsupportedOperationException
     *             if the item type does not support state extensions
     * @see IItemType#allowsStateExtensions()
     * @see #unsetStringExtension(String)
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public void setStringExtension(String key, String value);

    /**
     * Removes the state extension of type String from this item state as
     * identified by the given key. The item state must be a working copy.
     * <p>
     * When dealing with working copies created from partial item states,
     * specify the {@link #ALL_STATE_EXTENSIONS_PROPERTY} property when fetching
     * the item state from the repository, to ensure that all of its string
     * extensions are retrieved. If this is not done, you cannot add, update, or
     * remove state extensions of type String on the working copy based on that
     * item state.
     * </p>
     * <p>
     * 
     * @param key
     *            the state extension key to be used for accessing and storing
     *            the state extension value; must be a qualified name starting
     *            with by the component id identifying which component is
     *            responsible for the given state extension; must not be
     *            <code>null</code> or the empty string
     * @throws UnsupportedOperationException
     *             if the item type does not support state extensions
     * @see IItemType#allowsStateExtensions()
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public void unsetStringExtension(String extensionKey);

    /**
     * Answers whether a String extension with the given key exists on the item.
     * <p>
     * When dealing with working copies created from partial item states,
     * specify the {@link #ALL_STATE_EXTENSIONS_PROPERTY} property when fetching
     * the item state from the repository, to ensure that all of its string
     * extensions are retrieved. If this is not done, you cannot add, update, or
     * remove state extensions of type String on the working copy based on that
     * item state.
     * </p>
     * <p>
     * 
     * @param key
     *            the state extension key to be used for accessing and storing
     *            the state extension value; must be a qualified name starting
     *            with by the component id identifying which component is
     *            responsible for the given state extension; must not be
     *            <code>null</code> or the empty string
     * @return true if the string extension exists on the item; otherwise false *
     * @throws UnsupportedOperationException
     *             if the item type does not support state extensions
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public boolean isStringExtensionSet(String extensionKey);

    /**
     * Returns the (medium) String value of the extension on this item state for
     * the given key, or <code>null</code> if the item state has no medium
     * string extension with the given key.
     * <p>
     * Use {@link #setMediumStringExtension(String, String)} on an item working
     * copy to add or update medium string extensions. Use
     * {@link #unsetMediumStringExtension(String)} to completely remove the
     * extension key/value pair from the item.
     * <p>
     * A return value of <code>null</code> implies one of the following:
     * <ul>
     * <li>The item does not have a state extension with that key
     * <li>The item has a state extension with that key, with a value of
     * <code>null</code>. This can be checked using
     * {@link #isMediumStringExtensionSet(String)}
     * </ul>
     * 
     * @param key
     *            the state extension key to be used for accessing and storing
     *            the state extension value; must be a qualified name starting
     *            with the component id identifying which component is
     *            responsible for the given state extension; must not be
     *            <code>null</code> or an empty string
     * @return the value of this state extension, which may be <code>null</code>,
     *         or <code>null</code> if this item state doesn't have a state
     *         extension with the given key
     * @throws UnsupportedOperationException
     *             if the item type does not support state extensions
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public String getMediumStringExtension(String key);

    /**
     * Sets the medium String value of the state extension on this item state as
     * identified by the given key. The item state must be a working copy.
     * <p>
     * Medium String extensions are defined as strings with the
     * <code>dbStringSize</code> annotation set to <code>LARGE</code>, and
     * thus the size of the values is bounded by
     * {@link IItem#MAX_MEDIUM_STRING_BYTES}. This will be validated when an
     * item working copy is saved.
     * <p>
     * When dealing with working copies created from partial item states,
     * specify the {@link #ALL_STATE_EXTENSIONS_PROPERTY} property when fetching
     * the item state from the repository, to ensure that all of its medium
     * string extensions are retrieved. If this is not done, you cannot add,
     * update, or remove state extensions of type (medium) String on the working
     * copy based on that item state.
     * </p>
     * <p>
     * 
     * @param key
     *            the state extension key to be used for accessing and storing
     *            the state extension value; must be a qualified name starting
     *            with by the component id identifying which component is
     *            responsible for the given state extension; must not be
     *            <code>null</code> or the empty string
     * @param value
     *            the value of the state extension, bounded by
     *            {@link IItem#MAX_MEDIUM_STRING_BYTES}; null is allowed
     * @throws UnsupportedOperationException
     *             if the item type does not support state extensions
     * @see IItemType#allowsStateExtensions()
     * @see #unsetMediumStringExtension(String)
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public void setMediumStringExtension(String key, String value);

    /**
     * Removes the state extension of type (medium) String from this item state
     * as identified by the given key. The item state must be a working copy.
     * <p>
     * When dealing with working copies created from partial item states,
     * specify the {@link #ALL_STATE_EXTENSIONS_PROPERTY} property when fetching
     * the item state from the repository, to ensure that all of its medium
     * string extensions are retrieved. If this is not done, you cannot add,
     * update, or remove state extensions of type (medium) String on the working
     * copy based on that item state.
     * </p>
     * <p>
     * 
     * @param key
     *            the state extension key to be used for accessing and storing
     *            the state extension value; must be a qualified name starting
     *            with by the component id identifying which component is
     *            responsible for the given state extension; must not be
     *            <code>null</code> or the empty string
     * @throws UnsupportedOperationException
     *             if the item type does not support state extensions
     * @see IItemType#allowsStateExtensions()
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public void unsetMediumStringExtension(String extensionKey);

    /**
     * Answers whether a (medium) String extension with the given key exists on
     * the item.
     * <p>
     * When dealing with working copies created from partial item states,
     * specify the {@link #ALL_STATE_EXTENSIONS_PROPERTY} property when fetching
     * the item state from the repository, to ensure that all of its medium
     * string extensions are retrieved. If this is not done, you cannot add,
     * update, or remove state extensions of type (medium) String on the working
     * copy based on that item state.
     * </p>
     * <p>
     * 
     * @param key
     *            the state extension key to be used for accessing and storing
     *            the state extension value; must be a qualified name starting
     *            with by the component id identifying which component is
     *            responsible for the given state extension; must not be
     *            <code>null</code> or the empty string
     * @return true if the (medium) string extension exists on the item;
     *         otherwise false
     * @throws UnsupportedOperationException
     *             if the item type does not support state extensions
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public boolean isMediumStringExtensionSet(String extensionKey);

    /**
     * Returns the (large) String value of the extension on this item state for
     * the given key, or <code>null</code> if the item state has no large
     * string extension with the given key.
     * <p>
     * Use {@link #setLargeStringExtension(String, String)} on an item working
     * copy to add or update large string extensions. Use
     * {@link #unsetLargeStringExtension(String)} to completely remove the
     * extension key/value pair from the item.
     * <p>
     * A return value of <code>null</code> implies one of the following:
     * <ul>
     * <li>The item does not have a state extension with that key
     * <li>The item has a state extension with that key, with a value of
     * <code>null</code>. This can be checked using
     * {@link #isLargeStringExtensionSet(String)}
     * </ul>
     * 
     * @param key
     *            the state extension key to be used for accessing and storing
     *            the state extension value; must be a qualified name starting
     *            with the component id identifying which component is
     *            responsible for the given state extension; must not be
     *            <code>null</code> or an empty string
     * @return the value of this state extension, which may be <code>null</code>,
     *         or <code>null</code> if this item state doesn't have a state
     *         extension with the given key
     * @throws UnsupportedOperationException
     *             if the item type does not support state extensions
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public String getLargeStringExtension(String key);

    /**
     * Sets the large String value of the state extension on this item state as
     * identified by the given key. The item state must be a working copy.
     * <p>
     * Large String extensions are defined as strings with the
     * <code>dbStringSize</code> annotation set to <code>LARGE</code>, and
     * thus the size of the values is bounded by
     * {@link IItem#MAX_LARGE_STRING_BYTES}. This will be validated when an
     * item working copy is saved.
     * <p>
     * When dealing with working copies created from partial item states,
     * specify the {@link #ALL_STATE_EXTENSIONS_PROPERTY} property when fetching
     * the item state from the repository, to ensure that all of its large
     * string extensions are retrieved. If this is not done, you cannot add,
     * update, or remove state extensions of type (large) String on the working
     * copy based on that item state.
     * </p>
     * <p>
     * 
     * @param key
     *            the state extension key to be used for accessing and storing
     *            the state extension value; must be a qualified name starting
     *            with by the component id identifying which component is
     *            responsible for the given state extension; must not be
     *            <code>null</code> or the empty string
     * @param value
     *            the value of the state extension, bounded by
     *            {@link IItem#MAX_LARGE_STRING_BYTES}; null is allowed
     * @throws UnsupportedOperationException
     *             if the item type does not support state extensions
     * @see IItemType#allowsStateExtensions()
     * @see #unsetLargeStringExtension(String)
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public void setLargeStringExtension(String key, String value);

    /**
     * Removes the state extension of type (large) String from this item state
     * as identified by the given key. The item state must be a working copy.
     * <p>
     * When dealing with working copies created from partial item states,
     * specify the {@link #ALL_STATE_EXTENSIONS_PROPERTY} property when fetching
     * the item state from the repository, to ensure that all of its large
     * string extensions are retrieved. If this is not done, you cannot add,
     * update, or remove state extensions of type (large) String on the working
     * copy based on that item state.
     * </p>
     * <p>
     * 
     * @param key
     *            the state extension key to be used for accessing and storing
     *            the state extension value; must be a qualified name starting
     *            with by the component id identifying which component is
     *            responsible for the given state extension; must not be
     *            <code>null</code> or the empty string
     * @throws UnsupportedOperationException
     *             if the item type does not support state extensions
     * @see IItemType#allowsStateExtensions()
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public void unsetLargeStringExtension(String extensionKey);

    /**
     * Answers whether a (large) String extension with the given key exists on
     * the item.
     * <p>
     * When dealing with working copies created from partial item states,
     * specify the {@link #ALL_STATE_EXTENSIONS_PROPERTY} property when fetching
     * the item state from the repository, to ensure that all of its large
     * string extensions are retrieved. If this is not done, you cannot add,
     * update, or remove state extensions of type (large) String on the working
     * copy based on that item state.
     * </p>
     * <p>
     * 
     * @param key
     *            the state extension key to be used for accessing and storing
     *            the state extension value; must be a qualified name starting
     *            with by the component id identifying which component is
     *            responsible for the given state extension; must not be
     *            <code>null</code> or the empty string
     * @return true if the (large) string extension exists on the item;
     *         otherwise false
     * @throws UnsupportedOperationException
     *             if the item type does not support state extensions
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public boolean isLargeStringExtensionSet(String extensionKey);

    /**
     * Returns the int value of the extension on this item state for the given
     * key, or <code>null</code> if the item state has no int extension with
     * the given key.
     * <p>
     * Use {@link #setIntExtension(String, int)} on an item working copy to add
     * or update int extensions. Use {@link #unsetIntExtension(String)} to
     * completely remove the extension key/value pair from the item.
     * <p>
     * A return value of <code>0</code> implies one of the following:
     * <ul>
     * <li>The item does not have a state extension with that key
     * <li>The item has a state extension with that key, with a value of
     * <code>0</code>. This can be checked using
     * {@link #isIntExtensionSet(String)}
     * </ul>
     * 
     * @param key
     *            the state extension key to be used for accessing and storing
     *            the state extension value; must be a qualified name starting
     *            with the component id identifying which component is
     *            responsible for the given state extension; must not be
     *            <code>null</code> or an empty string
     * @return the value of this state extension, which may be <code>0</code>,
     *         or <code>0</code> if this item state doesn't have a state
     *         extension with the given key
     * @throws UnsupportedOperationException
     *             if the item type does not support state extensions
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public int getIntExtension(String key);

    /**
     * Sets the int value of the state extension on this item state as
     * identified by the given key. The item state must be a working copy.
     * <p>
     * When dealing with working copies created from partial item states,
     * specify the {@link #ALL_STATE_EXTENSIONS_PROPERTY} property when fetching
     * the item state from the repository, to ensure that all of its int
     * extensions are retrieved. If this is not done, you cannot add, update, or
     * remove state extensions of type int on the working copy based on that
     * item state.
     * </p>
     * <p>
     * 
     * @param key
     *            the state extension key to be used for accessing and storing
     *            the state extension value; must be a qualified name starting
     *            with by the component id identifying which component is
     *            responsible for the given state extension; must not be
     *            <code>null</code> or the empty string
     * @param value
     *            the value of the state extension
     * @throws UnsupportedOperationException
     *             if the item type does not support state extensions
     * @see IItemType#allowsStateExtensions()
     * @see #unsetIntExtension(String)
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public void setIntExtension(String key, int value);

    /**
     * Removes the state extension of type int from this item state as
     * identified by the given key. The item state must be a working copy.
     * <p>
     * When dealing with working copies created from partial item states,
     * specify the {@link #ALL_STATE_EXTENSIONS_PROPERTY} property when fetching
     * the item state from the repository, to ensure that all of its int
     * extensions are retrieved. If this is not done, you cannot add, update, or
     * remove state extensions of type int on the working copy based on that
     * item state.
     * </p>
     * <p>
     * 
     * @param key
     *            the state extension key to be used for accessing and storing
     *            the state extension value; must be a qualified name starting
     *            with by the component id identifying which component is
     *            responsible for the given state extension; must not be
     *            <code>null</code> or the empty string
     * @throws UnsupportedOperationException
     *             if the item type does not support state extensions
     * @see IItemType#allowsStateExtensions()
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public void unsetIntExtension(String extensionKey);

    /**
     * Answers whether an int extension with the given key exists on the item.
     * <p>
     * When dealing with working copies created from partial item states,
     * specify the {@link #ALL_STATE_EXTENSIONS_PROPERTY} property when fetching
     * the item state from the repository, to ensure that all of its int
     * extensions are retrieved. If this is not done, you cannot add, update, or
     * remove state extensions of type int on the working copy based on that
     * item state.
     * </p>
     * <p>
     * 
     * @param key
     *            the state extension key to be used for accessing and storing
     *            the state extension value; must be a qualified name starting
     *            with by the component id identifying which component is
     *            responsible for the given state extension; must not be
     *            <code>null</code> or the empty string
     * @return true if the int extension exists on the item; otherwise false
     * @throws UnsupportedOperationException
     *             if the item type does not support state extensions
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public boolean isIntExtensionSet(String extensionKey);

    /**
     * Returns the long value of the extension on this item state for the given
     * key, or <code>null</code> if the item state has no long extension with
     * the given key.
     * <p>
     * Use {@link #setLongExtension(String, long)} on an item working copy to
     * add or update long extensions. Use {@link #unsetLongExtension(String)} to
     * completely remove the extension key/value pair from the item.
     * <p>
     * A return value of <code>0</code> implies one of the following:
     * <ul>
     * <li>The item does not have a state extension with that key
     * <li>The item has a state extension with that key, with a value of
     * <code>null</code>. This can be checked using
     * {@link #isLongExtensionSet(String)}
     * </ul>
     * 
     * @param key
     *            the state extension key to be used for accessing and storing
     *            the state extension value; must be a qualified name starting
     *            with the component id identifying which component is
     *            responsible for the given state extension; must not be
     *            <code>null</code> or an empty string
     * @return the value of this state extension, which may be <code>0</code>,
     *         or <code>0</code> if this item state doesn't have a state
     *         extension with the given key
     * @throws UnsupportedOperationException
     *             if the item type does not support state extensions
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public long getLongExtension(String key);

    /**
     * Sets the long value of the state extension on this item state as
     * identified by the given key. The item state must be a working copy.
     * <p>
     * When dealing with working copies created from partial item states,
     * specify the {@link #ALL_STATE_EXTENSIONS_PROPERTY} property when fetching
     * the item state from the repository, to ensure that all of its long
     * extensions are retrieved. If this is not done, you cannot add, update, or
     * remove state extensions of type long on the working copy based on that
     * item state.
     * </p>
     * <p>
     * 
     * @param key
     *            the state extension key to be used for accessing and storing
     *            the state extension value; must be a qualified name starting
     *            with by the component id identifying which component is
     *            responsible for the given state extension; must not be
     *            <code>null</code> or the empty string
     * @param value
     *            the value of the state extension
     * @throws UnsupportedOperationException
     *             if the item type does not support state extensions
     * @see IItemType#allowsStateExtensions()
     * @see #unsetLongExtension(String)
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public void setLongExtension(String key, long value);

    /**
     * Removes the state extension of type long from this item state as
     * identified by the given key. The item state must be a working copy.
     * <p>
     * When dealing with working copies created from partial item states,
     * specify the {@link #ALL_STATE_EXTENSIONS_PROPERTY} property when fetching
     * the item state from the repository, to ensure that all of its long
     * extensions are retrieved. If this is not done, you cannot add, update, or
     * remove state extensions of type long on the working copy based on that
     * item state.
     * </p>
     * <p>
     * 
     * @param key
     *            the state extension key to be used for accessing and storing
     *            the state extension value; must be a qualified name starting
     *            with by the component id identifying which component is
     *            responsible for the given state extension; must not be
     *            <code>null</code> or the empty string
     * @throws UnsupportedOperationException
     *             if the item type does not support state extensions
     * @see IItemType#allowsStateExtensions()
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public void unsetLongExtension(String extensionKey);

    /**
     * Answers whether a long extension with the given key exists on the item.
     * <p>
     * When dealing with working copies created from partial item states,
     * specify the {@link #ALL_STATE_EXTENSIONS_PROPERTY} property when fetching
     * the item state from the repository, to ensure that all of its long
     * extensions are retrieved. If this is not done, you cannot add, update, or
     * remove state extensions of type long on the working copy based on that
     * item state.
     * </p>
     * <p>
     * 
     * @param key
     *            the state extension key to be used for accessing and storing
     *            the state extension value; must be a qualified name starting
     *            with by the component id identifying which component is
     *            responsible for the given state extension; must not be
     *            <code>null</code> or the empty string
     * @return true if the long extension exists on the item; otherwise false
     * @throws UnsupportedOperationException
     *             if the item type does not support state extensions
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public boolean isLongExtensionSet(String extensionKey);

    /**
     * Returns the BigDecimal value of the extension on this item state for the
     * given key, or <code>null</code> if the item state has no BigDecimal
     * extension with the given key.
     * <p>
     * Use {@link #setBigDecimalExtension(String, BigDecimal)} on an item
     * working copy to add or update BigDecimal extensions. Use
     * {@link #unsetBigDecimalExtension(String)} to completely remove the
     * extension key/value pair from the item.
     * <p>
     * A return value of <code>null</code> implies one of the following:
     * <ul>
     * <li>The item does not have a state extension with that key
     * <li>The item has a state extension with that key, with a value of
     * <code>null</code>. This can be checked using
     * {@link #isBigDecimalExtensionSet(String)}
     * </ul>
     * 
     * @param key
     *            the state extension key to be used for accessing and storing
     *            the state extension value; must be a qualified name starting
     *            with the component id identifying which component is
     *            responsible for the given state extension; must not be
     *            <code>null</code> or an empty string
     * @return the value of this state extension, which may be <code>null</code>,
     *         or <code>null</code> if this item state doesn't have a state
     *         extension with the given key
     * @throws UnsupportedOperationException
     *             if the item type does not support state extensions
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public BigDecimal getBigDecimalExtension(String key);

    /**
     * Sets the BigDecimal value of the state extension on this item state as
     * identified by the given key. The item state must be a working copy.
     * <p>
     * When dealing with working copies created from partial item states,
     * specify the {@link #ALL_STATE_EXTENSIONS_PROPERTY} property when fetching
     * the item state from the repository, to ensure that all of its BigDecimal
     * extensions are retrieved. If this is not done, you cannot add, update, or
     * remove state extensions of type BigDecimal on the working copy based on
     * that item state.
     * </p>
     * <p>
     * 
     * @param key
     *            the state extension key to be used for accessing and storing
     *            the state extension value; must be a qualified name starting
     *            with by the component id identifying which component is
     *            responsible for the given state extension; must not be
     *            <code>null</code> or the empty string
     * @param value
     *            the value of the state extension; null is allowed
     * @throws UnsupportedOperationException
     *             if the item type does not support state extensions
     * @see IItemType#allowsStateExtensions()
     * @see #unsetBigDecimalExtension(String)
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public void setBigDecimalExtension(String key, BigDecimal value);

    /**
     * Removes the state extension of type BigDecimal from this item state as
     * identified by the given key. The item state must be a working copy.
     * <p>
     * When dealing with working copies created from partial item states,
     * specify the {@link #ALL_STATE_EXTENSIONS_PROPERTY} property when fetching
     * the item state from the repository, to ensure that all of its BigDecimal
     * extensions are retrieved. If this is not done, you cannot add, update, or
     * remove state extensions of type BigDecimal on the working copy based on
     * that item state.
     * </p>
     * <p>
     * 
     * @param key
     *            the state extension key to be used for accessing and storing
     *            the state extension value; must be a qualified name starting
     *            with by the component id identifying which component is
     *            responsible for the given state extension; must not be
     *            <code>null</code> or the empty string
     * @throws UnsupportedOperationException
     *             if the item type does not support state extensions
     * @see IItemType#allowsStateExtensions()
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public void unsetBigDecimalExtension(String extensionKey);

    /**
     * Answers whether a BigDecimal extension with the given key exists on the
     * item.
     * <p>
     * When dealing with working copies created from partial item states,
     * specify the {@link #ALL_STATE_EXTENSIONS_PROPERTY} property when fetching
     * the item state from the repository, to ensure that all of its BigDecimal
     * extensions are retrieved. If this is not done, you cannot add, update, or
     * remove state extensions of type BigDecimal on the working copy based on
     * that item state.
     * </p>
     * <p>
     * 
     * @param key
     *            the state extension key to be used for accessing and storing
     *            the state extension value; must be a qualified name starting
     *            with by the component id identifying which component is
     *            responsible for the given state extension; must not be
     *            <code>null</code> or the empty string
     * @return true if the BigDecimal extension exists on the item; otherwise
     *         false
     * @throws UnsupportedOperationException
     *             if the item type does not support state extensions
     * 
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public boolean isBigDecimalExtensionSet(String extensionKey);

    /**
     * Returns the boolean value of the extension on this item state for the
     * given key, or <code>null</code> if the item state has no boolean
     * extension with the given key.
     * <p>
     * Use {@link #setBooleanExtension(String, boolean)} on an item working copy
     * to add or update boolean extensions. Use
     * {@link #unsetBooleanExtension(String)} to completely remove the extension
     * key/value pair from the item.
     * <p>
     * A return value of <code>false</code> implies one of the following:
     * <ul>
     * <li>The item does not have a state extension with that key
     * <li>The item has a state extension with that key, with a value of
     * <code>null</code>. This can be checked using
     * {@link #isBooleanExtensionSet(String)}
     * </ul>
     * 
     * @param key
     *            the state extension key to be used for accessing and storing
     *            the state extension value; must be a qualified name starting
     *            with the component id identifying which component is
     *            responsible for the given state extension; must not be
     *            <code>null</code> or an empty string
     * @return the value of this state extension, which may be
     *         <code>false</code>, or <code>false</code> if this item state
     *         doesn't have a state extension with the given key or if the item
     *         type does not allow state extensions *
     * @throws UnsupportedOperationException
     *             if the item type does not support state extensions
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public boolean getBooleanExtension(String key);

    /**
     * Sets the boolean value of the state extension on this item state as
     * identified by the given key. The item state must be a working copy.
     * <p>
     * When dealing with working copies created from partial item states,
     * specify the {@link #ALL_STATE_EXTENSIONS_PROPERTY} property when fetching
     * the item state from the repository, to ensure that all of its boolean
     * extensions are retrieved. If this is not done, you cannot add, update, or
     * remove state extensions of type boolean on the working copy based on that
     * item state.
     * </p>
     * <p>
     * 
     * @param key
     *            the state extension key to be used for accessing and storing
     *            the state extension value; must be a qualified name starting
     *            with by the component id identifying which component is
     *            responsible for the given state extension; must not be
     *            <code>null</code> or the empty string
     * @param value
     *            the value of the state extension
     * @throws UnsupportedOperationException
     *             if the item type does not support state extensions
     * @see IItemType#allowsStateExtensions()
     * @see #unsetBooleanExtension(String)
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public void setBooleanExtension(String key, boolean value);

    /**
     * Removes the state extension of type boolean from this item state as
     * identified by the given key. The item state must be a working copy.
     * <p>
     * When dealing with working copies created from partial item states,
     * specify the {@link #ALL_STATE_EXTENSIONS_PROPERTY} property when fetching
     * the item state from the repository, to ensure that all of its boolean
     * extensions are retrieved. If this is not done, you cannot add, update, or
     * remove state extensions of type boolean on the working copy based on that
     * item state.
     * </p>
     * <p>
     * 
     * @param key
     *            the state extension key to be used for accessing and storing
     *            the state extension value; must be a qualified name starting
     *            with by the component id identifying which component is
     *            responsible for the given state extension; must not be
     *            <code>null</code> or the empty string
     * @throws UnsupportedOperationException
     *             if the item type does not support state extensions
     * @see IItemType#allowsStateExtensions()
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public void unsetBooleanExtension(String extensionKey);

    /**
     * Answers whether a booelan extension with the given key exists on the
     * item.
     * <p>
     * When dealing with working copies created from partial item states,
     * specify the {@link #ALL_STATE_EXTENSIONS_PROPERTY} property when fetching
     * the item state from the repository, to ensure that all of its boolean
     * extensions are retrieved. If this is not done, you cannot add, update, or
     * remove state extensions of type boolean on the working copy based on that
     * item state.
     * </p>
     * <p>
     * 
     * @param key
     *            the state extension key to be used for accessing and storing
     *            the state extension value; must be a qualified name starting
     *            with by the component id identifying which component is
     *            responsible for the given state extension; must not be
     *            <code>null</code> or the empty string
     * @return true if the boolean extension exists on the item; otherwise false *
     * @throws UnsupportedOperationException
     *             if the item type does not support state extensions
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     * 
     */
    public boolean isBooleanExtensionSet(String extensionKey);

    /**
     * Returns the timestamp value of the extension on this item state for the
     * given key, or <code>null</code> if the item state has no timestamp
     * extension with the given key.
     * <p>
     * Use {@link #setTimestampExtension(String, Timestamp)} on an item working
     * copy to add or update timestamp extensions. Use
     * {@link #unsetTimestampExtension(String)} to completely remove the
     * extension key/value pair from the item.
     * <p>
     * A return value of <code>null</code> implies one of the following:
     * <ul>
     * <li>The item does not have a state extension with that key
     * <li>The item has a state extension with that key, with a value of
     * <code>null</code>. This can be checked using
     * {@link #isTimestampExtensionSet(String)}
     * </ul>
     * 
     * @param key
     *            the state extension key to be used for accessing and storing
     *            the state extension value; must be a qualified name starting
     *            with the component id identifying which component is
     *            responsible for the given state extension; must not be
     *            <code>null</code> or an empty string
     * @return the value of this state extension, which may be <code>null</code>,
     *         or <code>null</code> if this item state doesn't have a state
     *         extension with the given key
     * @throws UnsupportedOperationException
     *             if the item type does not support state extensions
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public Timestamp getTimestampExtension(String key);

    /**
     * Sets the timestamp value of the state extension on this item state as
     * identified by the given key. The item state must be a working copy.
     * <p>
     * When dealing with working copies created from partial item states,
     * specify the {@link #ALL_STATE_EXTENSIONS_PROPERTY} property when fetching
     * the item state from the repository, to ensure that all of its timestamp
     * extensions are retrieved. If this is not done, you cannot add, update, or
     * remove state extensions of type timestamp on the working copy based on
     * that item state.
     * </p>
     * <p>
     * 
     * @param key
     *            the state extension key to be used for accessing and storing
     *            the state extension value; must be a qualified name starting
     *            with by the component id identifying which component is
     *            responsible for the given state extension; must not be
     *            <code>null</code> or the empty string
     * @param value
     *            the value of the state extension; null is allowed
     * @throws UnsupportedOperationException
     *             if the item type does not support state extensions
     * @see IItemType#allowsStateExtensions()
     * @see #unsetTimestampExtension(String)
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public void setTimestampExtension(String key, Timestamp value);

    /**
     * Removes the state extension of type timestamp from this item state as
     * identified by the given key. The item state must be a working copy.
     * <p>
     * When dealing with working copies created from partial item states,
     * specify the {@link #ALL_STATE_EXTENSIONS_PROPERTY} property when fetching
     * the item state from the repository, to ensure that all of its timestamp
     * extensions are retrieved. If this is not done, you cannot add, update, or
     * remove state extensions of type timestamp on the working copy based on
     * that item state.
     * </p>
     * <p>
     * 
     * @param key
     *            the state extension key to be used for accessing and storing
     *            the state extension value; must be a qualified name starting
     *            with by the component id identifying which component is
     *            responsible for the given state extension; must not be
     *            <code>null</code> or the empty string
     * @throws UnsupportedOperationException
     *             if the item type does not support state extensions
     * @see IItemType#allowsStateExtensions()
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public void unsetTimestampExtension(String extensionKey);

    /**
     * Answers whether a timestamp extension with the given key exists on the
     * item.
     * <p>
     * When dealing with working copies created from partial item states,
     * specify the {@link #ALL_STATE_EXTENSIONS_PROPERTY} property when fetching
     * the item state from the repository, to ensure that all of its timestamp
     * extensions are retrieved. If this is not done, you cannot add, update, or
     * remove state extensions of type timestamp on the working copy based on
     * that item state.
     * </p>
     * <p>
     * 
     * @param key
     *            the state extension key to be used for accessing and storing
     *            the state extension value; must be a qualified name starting
     *            with by the component id identifying which component is
     *            responsible for the given state extension; must not be
     *            <code>null</code> or the empty string
     * @return true if the timestamp extension exists on the item; otherwise
     *         false
     * @throws UnsupportedOperationException
     *             if the item type does not support state extensions
     * 
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public boolean isTimestampExtensionSet(String extensionKey);

    /**
     * Answer a set of all the keys for which extensions are set on this item,
     * for all data types. The item must be complete or have been fetched with
     * the {@link #ALL_STATE_EXTENSIONS_PROPERTY} property.
     * 
     * @return set of extension keys for extensions that have been set; will be
     *         empty if the item is partial and
     *         {@link #ALL_STATE_EXTENSIONS_PROPERTY} has not been requested
     * @throws UnsupportedOperationException
     *             if the item type does not support state extensions
     * 
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public Set<String> getExtensionKeys();

}
