/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.repository.common;

/**
 * Common protocol for helper objects.
 * 
 * @since 0.5
 * @NoImplement Clients must not implement this interface.
 */
public interface IHelper {

    /**
     * Returns the type of this helper object.
     * 
     * @return the type; never <code>null</code>
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public IHelperType getHelperType();
    
    /**
     * Returns an object that adapts this item to the given class. The result
     * can be cast to <code>clazz</code>. Depending on the type of item
     * involved, the result may or may not be identical to the receiver.
     * <p>
     * This method is used to get at the special API for the item when possible.
     * </p>
     * <p>
     * Pass <code>EObject.class</code> to get reflective access to the EMF
     * underlying object. From the EMF object one can navigate to the EMF class
     * to discover all structural features of this kind of object. Use
     * <code>EObject.eGet/eSet(EStructuralFeature)</code> to access the
     * attributes on the item in a generic manner.
     * </p>
     * <p>
     * The ground rules for accessing the underlying EMF objects:
     * <ul>
     
     * <li>The name of the EMF class of the object bears the name acknowledged
     * in the item type. That is, <code>this.eClass().getName()</code> matches
     * <code>this.getHelperType().getName()</code>[ + &quot;
     * <code>Handle</code> &quot;]. Similarly,
     * <code>this.eClass().getPackage().getNsURI()</code> matches
     * <code>this.getHelperType().getNamespaceURI()</code>.</li>
     * <li>The exact set of structural features inherited from the EMF classes
     * are internal, undocumented, and subject to
     * change. While a client is free to display these structural features for
     * debug purposes, they must not make assumptions about their continued
     * existence or usage.</li>
     * <li>The client must not explicitly modify (<code>eSet</code>) any of
     * the structural features inherited from the EMF classes. 
     * These are reserved for use by the implementation
     * of the Client API.</li>
     * <li>The client must not explicitly modify (<code>eSet</code>) meta
     * models or install new ones. The meta models are read-only and reserved
     * for use by the implementation of the Client API.</li>
     * <li>As always, everything in Java packages with &quot;internal&quot; in
     * the name is strictly off limits.</li>
     * </ul>
     * </p>
     * <p>
     * Implementation note: There is no requirement that the nice API provided
     * for an extended item type bear any direct relationship to the underlying
     * EMF objects.
     * </p>
     * 
     * @param clazz
     *            the desired Java class; must not be <code>null</code>
     * @return an instance of <code>clazz</code>, or <code>null</code> if
     *         this item could not be so adapted
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public Object getAdapter(Class clazz);


}
