/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.repository.common;

import com.ibm.team.repository.common.internal.util.ComponentRegistry;

/**
 * Represents a type of helper object found on the server.
 * 
 * @since 0.5
 * @NoImplement This interface is not intended to be implemented by clients.
 */
public interface IHelperType extends IType {

    /**
     * Creates a new helper object of this type.
     * 
     * @return the created object or <code>null</code> if this helper type is
     *         an invalid helper type whose corresponding instance class is not
     *         a helper class.
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public IHelper createHelper();

    /**
     * Static registry used to locate helper types.
     * <p>
     * Example usage:
     * 
     * <pre>
     *                public interface ISomeClass extends ISomeClassHandle {
     *                public static final IHelperType HELPER_TYPE = IHelperType.IRegistry.INSTANCE
     *                      .getHelperType(SomePackage.eINSTANCE.getSomeClass().getName(),
     *                              SomePackage.eNS_URI);
     * </pre>
     * 
     * </p>
     */
    public static interface IRegistry {

        /**
         * The one instance of the <code>IRegistry</code> object.
         */
        public static IRegistry INSTANCE = ComponentRegistry.INSTANCE;

        /**
         * Locates a helper type given its name and its model URI.
         * 
         * @param name
         *            The helper type's name; must not be <code>null</code>
         * @param modelUri
         *            the helper type's model URI; must not be
         *            <code>null</code>
         * @return the located helper type or <code>null</code> if the helper
         *         type name and model URI pair was not located in the
         *         helper type registry
         * @ShortOp This is a short operation; it may block only momentarily;
         *          safe to call from a responsive thread.
         */
        public IHelperType getHelperType(String name, String modelUri);

        /**
         * Locates all helper types.  Abstract types will be excluded.
         * 
         * @return all helper types; never <code>null</code>
         * @ShortOp This is a short operation; it may block only momentarily;
         *          safe to call from a responsive thread.
         */
        public IHelperType[] getAllHelperTypes();
        
        /**
         * Locates all helper types.  
         * 
         * @param includeAbstractTypes indicates whether abstract types should 
         *          be included in the returned collection
         * 
         * @return all helper types; never <code>null</code>
         * @ShortOp This is a short operation; it may block only momentarily;
         *          safe to call from a responsive thread.
         */
        public IHelperType[] getAllHelperTypes(boolean includeAbstractTypes);
        
    }
}
