/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010, 2014. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.repository.common;

import java.sql.Timestamp;
import java.util.Date;

import com.ibm.team.repository.common.internal.util.ItemUtil;
import com.ibm.team.repository.common.internal.validation.IInternalItemValidator;
import com.ibm.team.repository.common.model.RepositoryPackage;

/**
 * Common protocol for simple, auditable, and versionable item states alike.
 * <p>
 * Item handles and items are passive data objects that are retrieved or created
 * while the client is logged in to a repository. These objects remain valid if
 * the client logs out.
 * </p>
 * <p>
 * Item states, the data for an item, can be augmented with additional
 * open-ended data extensions, called <b>state extensions</b>. State extensions
 * are part of the item state and are captured in the item states as they change
 * over time. Components must "opt-in" to allowing state extensions for a given
 * item type in order for an item state to be augmented. A state extension is a
 * key-value pair, where the key is a distinctive string prefixed by the
 * component id identifying which component is responsible for the given state
 * extension. The values are limited to type <code>String</code> which must be
 * non-null string and limited to 250 characters.
 * 
 * Specifying the {@link #com.ibm.team.repository.common.IItem.STATE_EXTENSIONS2_PROPERTY} property of an item state
 * when fetching an item from the repository retrieves all of its state
 * extensions.
 * </p>
 * 
 * @since 0.5
 * @NoImplement Clients must not implement this interface.
 */
public interface IItem extends IItemHandle {

    /**
     * Max size (250) in UTF-8 encoded bytes for a small string property. This
     * is the default value for all string properties, unless overriden using
     * the <code>teamAttribute->dbStringSize</code> annotation.
     * 
     * @see IInternalItemValidator for utility to validate item properties
     */
    public static final long MAX_SMALL_STRING_BYTES = ItemUtil.SMALL_STRING_BYTES;
    

    /**
     * Max size (1000) in UTF-8 encoded bytes for a medium string property. A medium
     * string property is a string property where the model attribute is
     * annotated with the <code>teamAttribute->dbStringSize=MEDIUM</code> annotation.
     * 
     * @see IInternalItemValidator for utility to validate item properties
     */
    public static final long MAX_MEDIUM_STRING_BYTES = ItemUtil.MEDIUM_STRING_BYTES;

    /**
     * Max size in UTF-8 encoded bytes for a large string property. A large
     * string property is a string property where the model attribute is
     * annotated with the <code>teamAttribute->dbStringSize=LARGE</code> annotation.
     * 
     * @see IInternalItemValidator for utility to validate item properties
     */
    public static final long MAX_LARGE_STRING_BYTES = ItemUtil.LARGE_STRING_BYTES;
    
    /**
     * The item id property.
     * 
     * @see IItemHandle#getItemId()
     */
    public static final String ITEM_ID_PROPERTY = RepositoryPackage.eINSTANCE
            .getItemHandle_ItemId().getName();

    /**
     * The modified by property.
     * 
     * @see #getModifiedBy()
     */
    public static final String MODIFIED_BY_PROPERTY = RepositoryPackage.eINSTANCE
            .getItem_ModifiedBy().getName();

    /**
     * The modified property.
     * 
     * @see #getModified()
     */
    public static final String MODIFIED_PROPERTY = RepositoryPackage.eINSTANCE
            .getItem_Modified().getName();
       
    /**
     * The state id property.
     * 
     * @see IItemHandle#getStateId()
     */
    public static final String STATE_ID_PROPERTY = RepositoryPackage.eINSTANCE
            .getItemHandle_StateId().getName();
    
    /**
     * The context id property
     * 
     * @see #getContextId()
     * @since 0.7
     */
    public static final String CONTEXT_ID_PROPERTY = RepositoryPackage.eINSTANCE
            .getItem_ContextId().getName();
        
    /**
     * Returns whether this item has historical information about the origins of
     * this state (modified and modified by). Note that <code>hasHistory</code>
     * implies <code>isState</code>, but is independent of
     * <code>isFullState</code>.
     * 
     * @return <code>true</code> if the item has state, and <code>false</code>
     *         otherwise
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public boolean hasHistory();

    /**
     * Returns the contributor responsible for creating this state of the item.
     * <p>
     * A newly-created item may or may not carry the contributor's id. The initial
     * state of an item bears the item id of the contributor that created the
     * item; other states bear the item id of the contributor that made the
     * change that resulted in that state.
     * </p>
     * 
     * @return the contributor responsible for the state, or <code>null</code>
     *         if unknown
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public IContributorHandle getModifiedBy();

    /**
     * Returns the time at which this item state was created.
     * <p>
     * A newly-created item may or may not carry the modified property.
     * Moreover, the modified property is only retrieved when history is
     * requested. Clients that want to see these must use
     * <code>allStates(LARGE)</code> in order to get meaningful values that
     * they can depend on.
     * </p>
     * 
     * @return the time at which the state was created, or <code>null</code>
     *         if unknown
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public Date modified();

    /**
     * Returns whether this item is complete or partial. A complete item has all
     * of its properties set, whereas a partial item may have some of its
     * properties not set.
     * 
     * @return <code>true</code> if the item is complete, and
     *         <code>false</code> if this item is partial
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public boolean isComplete();

    /**
     * Returns whether the given property of this item is set to a particular
     * value.
     * 
     * @param propertyName
     *            the property name; must must not be <code>null</code>
     * @return <code>true</code> if the property is set, and
     *         <code>false</code> if the property is unset. 
     * @throws IllegalArgumentException if the item does not have such a property        
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public boolean isPropertySet(String propertyName);

    /**
     * Returns whether this item state can be modified.
     * <p>
     * Items created via the <code>IItemType#createItem()</code> start off as
     * working copies. Items created independently are not working copies.
     * States retrieved from the repository are not working copies. Use
     * {@link #getWorkingCopy()}to obtain a working copy from a retrieved or
     * independently created item.
     * </p>
     * <p>
     * Working copies do not have state ids ({@link IItemHandle#hasStateId()}
     * returns <code>false</code>).
     * </p>
     * <p>
     * Implementation note: This flag is strictly a property of the in-memory
     * object.
     * </p>
     * 
     * @return <code>true</code> if this state is a working copy, and
     *         <code>false</code> otherwise
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public boolean isWorkingCopy();


    /**
     * Returns a working copy of this state that can be modified. Returns the
     * receiver if it is already a working copy.
     * <p>
     * Items created via the <code>IItemType#createItem()</code> start off as
     * working copies. Items created independently are not working copies.
     * States retrieved from the repository are not working copies. This method
     * must be called to obtain a working copy from a retrieved item or a new
     * item that was created independently; the client must not modify states
     * that are not working copies.
     * </p>
     * <p>
     * Modifying the value of a property that is currently unset will cause an
     * <code>IllegalStateException</code> to be thrown.
     * </p>
     * 
     * @return a working copy; never <code>null</code>
     * @see #isWorkingCopy()
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public IItem getWorkingCopy();
    
	/**
	 * Returns whether this item is a redacted copy of the state. A redacted
	 * copy has has certain sensitive information removed.
	 * <p>
	 * Implementation note: This flag is strictly a property of the in-memory
	 * object.
	 * </p>
	 * 
	 * @return <code>true</code> if this state is a redacted copy, and
	 *         <code>false</code> otherwise
	 * @ShortOp This is a short operation; it may block only momentarily; safe
	 *          to call from a responsive thread.
	 */
    public boolean isRedactedCopy();
    
	/**
	 * Returns a redacted copy of this state. Returns the receiver if it is
	 * already a redacted copy.
	 * <p>
	 * If this item type does not support redaction,
	 * {@link IItemType#isRedactable()} returns false, this method will throw an
	 * {@link UnsupportedOperationException}.
	 * </p>
	 * 
	 * @return a redacted copy; never <code>null</code>
	 * @see #isRedactedCopy()
	 * @ShortOp This is a short operation; it may block only momentarily; safe
	 *          to call from a responsive thread.
	 * @throws UnsupportedOperationException
	 *             if this item type does not support redaction
	 *             (IItemType#isRedactable() returns false)
	 */
    public IItem getRedactedCopy();

    /**
     * Creates a new item handle with the same item id and item type as this
     * item. state id if applicable will be set to null.
     * 
     * @return a new item handle to this item, with the same item id and item
     *         type; state id if applicable will be set to null.
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public IItemHandle getItemHandle();
    
    /**
     * Creates a new item handle with the same item id, state id and item type
     * as this item.
     * 
     * @return a new item handle to this item, with the same item id, state id
     *         and item type
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public IItemHandle getStateHandle();
    
    /**
     * Get the context id of the item.  This is used to compute read access
     * permission for protected types.
     * @return the context id of the item
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     * @since 0.7
     * @see com.ibm.team.repository.service.permission
     */
    public UUID getContextId();
    
    /**
     * Set the contextId for the given item.  This is used to compute read
     * access permission for protected types.
     * @param contextId context id of the item; can be <code>null</code>
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     * @since 0.7
     * @see com.ibm.team.repository.service.permission
     */
    public void setContextId(UUID contextId);

    /**
     * Request a specific stateId when saving a new state of this working copy.
     * 
     * For auditable items, the requested stateId must be different from the
     * stateId for all historical states of all auditable items.  For simple
     * items, the stateId must be different from the current stateId.  If these
     * uniqueness rules are not followed, saving the item will fail with a
     * TeamRepositoryException.
     * 
     * If you use UUID.generate() to produce a stateId, you are guaranteed to
     * get a unique one.
     * 
     * @param stateId requested state id
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     * @throws IllegalStateException if this is not a working copy
     */
    public void setRequestedStateId(UUID stateId);

    /**
     * Get the previously requested state id used when saving a new state of
     * this working copy
     * @return the requested state id, or <code>null</code> if no state id has
     * been requested
     * @see #setRequestedStateId()
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public UUID getRequestedStateId();

    /**
     * Request a specific modified timestamp when saving a new state of this working copy.
     * 
     * @param modified requested modified timestamp
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     * @throws IllegalStateException if this is not a working copy
     */
    public void setRequestedModified(Timestamp modified);

    /**
     * Get the previously requested modified timestamp used when saving a new state of
     * this working copy
     * @return the requested modified timestamp, or <code>null</code> if no modified timestamp has
     * been requested
     * @see #setRequestedModified()
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public Timestamp getRequestedModified();
}
