/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010, 2014. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.repository.common;

/**
 * Common protocol for auditable and versionable item handles alike. A handle is
 * used to reference a particular item, or, in some forms, to reference a
 * particular state of a particular item.
 * <p>
 * Item handles and items can only be created or retrieved from a repository
 * while the client is logged in. When the client logs out, all objects that
 * they obtained while logged in must be considered invalid. It is unspecified
 * whether these objects come back to life, so to speak, if the client logs back
 * in.
 * </p>
 * <p>
 * Speaking informally, <b>items </b> are objects that are stored in the team
 * repository. Each item is something that has a unique identity, a fixed item
 * type, and some states. Each item is identified by its <b>item id </b>(a
 * 128-bit universally unique identifier, or UUID). Each item has a fixed
 * <b>item type </b>. A few item types are built in to the foundations of the
 * team repository. The team repository can be extended with the definition of
 * additional types of items. The <b>state </b> of an item is the data that it
 * recorded for that item. The state of an item can change over time. Recording
 * the state of an item is what takes up the space in the repository.
 * </p>
 * <p>
 * These concepts are mapped to API interfaces as a pair of interface
 * hierarchies, one for item handles and the other for the item states.
 * <ul>
 * <li>An {@link IItemHandle}is an item handle, and records an item id and
 * type. These object can also carry a state id, in which case they are state
 * handles. All item handles are typed.</li>
 * <li>Use {@link #sameItemId(IItemHandle)}to compare whether two item handles
 * refer to the same item in virtue of having the same item id.</li>
 * <li>An {@link IAuditableHandle}is a handle to an item of one of the
 * auditable item types. This is a subtype of <code>IItemHandle</code>.
 * {@link #isAuditable()} returns <code>true</code> only for auditable item
 * types. Any <code>IItemHandle</code> for an auditable type can be safely
 * downcast to <code>IAuditableHandle</code>. </li>
 * <li>An {@link IUnmanagedItemHandle}is a handle to an item of one of the
 * versionable item types. This is a subtype of <code>IItemHandle</code>.
 * {@link #isUnmanaged()} returns <code>true</code> only for versionable
 * item types. Any <code>IItemHandle</code> for a versionable type can be
 * safely downcast to <code>IVersionableHandle</code>.</li>
 * <li>Each built-in item type has its own specific handle interface; e.g.,
 * <code>IContributorHandle</code>. Any <code>IItemHandle</code> can be
 * safely downcast to its particular handle interface.</li>
 * <li>An {@link IItem}is an item state, and records the all the attributes
 * and properties of an item. <code>IItem</code> extends
 * <code>IItemHandle</code>, reflecting the fact that a state also carries
 * the item id, item type, and state id. This allows an item state to be used
 * where only an item handle is expected/required.</li>
 * <li>An {@link IAuditable}is a state of an item of one of the auditable item
 * types. This is a subtype of <code>IItem</code> and
 * <code>IAuditableHandle</code>. Any <code>IItem</code> for an auditable
 * type can be safely downcast to <code>IAuditable</code>.</li>
 * <li>An {@link IUnmanagedItem}is a state of an item of one of the versionable
 * item types. This is a subtype of <code>IItem</code> and
 * <code>IVersionableHandle</code>. Any <code>IItem</code> for a
 * versionable type can be safely downcast to <code>IVersionable</code>.
 * </li>
 * <li>Each built-in item type has its own specific state interface; e.g.,
 * <code>IContributor</code>. Any <code>IItem</code> can be safely downcast
 * to its particular state interface.</li>
 * <li>Objects implementing this interface are generally not canonical, and
 * cannot be reliably compared with <code>==</code>.</li>
 * </ul>
 * Note that although a state can always be cast as a handle, the converse is
 * not generally true: a handle need not be a full state. Only handles that
 * answer <code>true</code> to {@link IItemHandle#hasFullState()}can be
 * safely downcast to state interfaces.
 * </p>
 * <p>
 * Design note: Note that <code>equals()</code> has no special semantics for
 * item handles (i.e., it's just Java object identity). This means that
 * expression like <code>team.getMembers().contains(contributor)</code> do not
 * work as you might like. Use <code>sameItemId()</code> when searching for
 * duplicate items with the same id as a given item.
 * </p>
 * <p>
 * Terminology: There is obviously a close relationship between an item handle
 * and the item that it refers to. Where there is no risk of confusion, we drop
 * the word &quot;handle&quot; and simply talk of items directly.
 * </p>
 * 
 * @since 0.5
 * @NoImplement This interface is not intended to be implemented by clients.
 */
public interface IItemHandle {

    /**
     * Returns the item id of this item. The item id gives items their identity.
     * 
     * @return the item id; never <code>null</code>
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public UUID getItemId();
    
    /**
     * Returns whether this item has a state id and is therefore capable of
     * being used to reference to particular state of the item in the
     * repository; in other words, can also be used as a state handle.
     * 
     * @return <code>true</code> if this is a state handle, and
     *         <code>false</code> if it just has enough information to
     *         identify the item but not an particular state
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public boolean hasStateId();

    /**
     * Returns the id of a state of this item. The state id is used within the
     * repository to identify the different states of an item.
     * <p>
     * A newly-created item may or may not have a state id; clients must not
     * make assumptions either way.
     * </p>
     * 
     * @return the state id, or <code>null</code> if none
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public UUID getStateId();
    
    /**
     * Returns whether this item and the given item are objects representing the
     * same state of the same item in the repository (that is, both have the
     * same item id). If one of the other does not have a state id, this method
     * returns <code>false</code>. Also returns <code>false</code> if
     * <code>other</code> is <code>null</code>.
     * <p>
     * Note: This method cannot be meaningfully used to compare working copies,
     * since working copies have no state id.
     * </p>
     * 
     * @param other
     *            the other item, or <code>null</code>
     * @return <code>true</code> if the items have the same item ids and
     *         non-null state ids, and <code>false</code> if
     *         <code>other</code> is <code>null</code> or the items have
     *         different item ids, different state ids, or no state ids
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public boolean sameStateId(IItemHandle other);


    /**
     * Returns the type of this item.
     * 
     * @return the item type; never <code>null</code>
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public IItemType getItemType();

    /**
     * Returns the origin of this item handle. The component client
     * implementation is responsible for tagging its items with the proper
     * origin. The item origin should have sufficient information to allow
     * refreshing/refetching the item.
     * <p>
     * The component client API methods that tag items with an origin are
     * responsible for documenting that fact.
     * </p>
     * 
     * @return the origin of this item or <code>null</code> if unknown
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public Object getOrigin();

    /**
     * Returns whether this item handle is immutable.
     * 
     * @return <code>true</code> if this is an immutable item handle, and
     *         <code>false</code> otherwise
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public boolean isImmutable();

    /**
     * Marks this item handle as being immutable. Calling this method will not
     * enforce immutability of the item handle. One must call {@link #protect()}
     * to enforce the immutability of the item handle.
     * <p>
     * Once, immutability is enforced, calling any method that modifies the item
     * handle, will cause an <code>IllegalStateException</code> to be thrown.
     * </p>
     * 
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public void makeImmutable();

    /**
     * Enforces the immutability and working copy rules on items and item
     * handles.
     * 
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     * @see #makeImmutable()
     * @see IItem#getWorkingCopy()
     */
    public void protect();

    /**
     * Returns whether this item and the given item are objects representing the
     * same item in the repository (that is, both have the same item id). Also
     * returns <code>false</code> if <code>other</code> is <code>null</code>.
     * 
     * @param other
     *            the other item, or <code>null</code>
     * @return <code>true</code> if the items have the same item ids, and
     *         <code>false</code> if <code>other</code> is <code>null</code>
     *         or the items have different item ids
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public boolean sameItemId(IItemHandle other);

    /**
     * Returns whether this represents a full state of the item. A full state is
     * one where some of the state's property values are available to the
     * client. The client can modify a full state and save it back. If an object
     * answers <code>true</code> to this method, {@link #getFullState()}will
     * return non- <code>null</code>; equivalently, this object can be safely
     * cast to a state ({@link IItem}).
     * <p>
     * A newly-created item is also considered to be a full state which the
     * client happens to be in the process of formulating from scratch.
     * </p>
     * 
     * @return <code>true</code> if the item is a full state, and
     *         <code>false</code> otherwise
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public boolean hasFullState();

    /**
     * Returns the object representing the full state of the item, or
     * <code>null</code> if this is just a handle. The handle can have a state
     * id and is still only considered to be a handle to the referenced current
     * state in the repository and not a full state. Note that server/repository
     * is not contacted; a result is only available if the state had already
     * been retrieved (or the item is in the process of being created).
     * <p>
     * If the item handle has an associated full state, the item returned is
     * always identical to the receiver (the receiver is an object that
     * implements both <code>IItemHandle</code> and <code>IItem</code>).
     * </p>
     * 
     * @return the full state of the item, or <code>null</code> if this item
     *         is just a handle without state information (other than perhaps a
     *         state id)
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public IItem getFullState();

    /**
     * Returns whether this item handle is for an auditable type of item. This
     * method returns <code>true</code> if and only if this item handle can be
     * successfully cast to {@link IAuditableHandle}.
     * 
     * @return <code>true</code> if this is an auditable item, and
     *         <code>false</code> otherwise
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public boolean isAuditable();

    /**
     * Returns whether this item handle is for an unmanaged type of item. This
     * method returns <code>true</code> if and only if this item handle can be
     * successfully cast to {@link IUnmanagedItemHandle}.
     * 
     * @return <code>true</code> if this is a versionable item, and
     *         <code>false</code> otherwise
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public boolean isUnmanaged();

    /**
     * Returns whether this item handle is for a simple type of item. This
     * method returns <code>true</code> if and only if this item handle can be
     * successfully cast to {@link ISimpleItemHandle}.
     * 
     * @return <code>true</code> if this is a simple item, and
     *         <code>false</code> otherwise
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public boolean isSimple();

    /**
     * Returns whether this item handle is for an item that is configuration
     * aware.  This method returns <code>true</code> if and only if this item
     * handle can be successfully cast to {@link IConfigurationAwareItemHandle}.
     * 
     * @return <code>true</code> if this is a configuration aware item, and
     *         <code>false</code> otherwise
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public boolean isConfigurationAware();

    /**
     * Returns an object that adapts this item to the given class. The result
     * can be cast to <code>clazz</code>. Depending on the type of item
     * involved, the result may or may not be identical to the receiver.
     * <p>
     * This method is used to get at the special API for the item when possible.
     * </p>
     * <p>
     * Pass <code>EObject.class</code> to get reflective access to the EMF
     * underlying object. From the EMF object one can navigate to the EMF class
     * to discover all structural features of this kind of object. Use
     * <code>EObject.eGet/eSet(EStructuralFeature)</code> to access the
     * attributes on the item in a generic manner.
     * </p>
     * <p>
     * The ground rules for accessing the underlying EMF objects:
     * <ul>
     * <li>The client may cast an object of an auditable item type to
     * <code>IAuditableHandle</code>; if it has full state, it may be cast to
     * <code>IAuditable</code>.</li>
     * <li>The client may cast an object of a versionable item type to
     * <code>IVersionableHandle</code>; if it has full state, it may be cast
     * to <code>IVersionable</code>.</li>
     * <li>When there are locally-available client side representations for a
     * particular item type, the client may cast an object of that item type to
     * the corresponding Java handle type documented with that item type. For
     * instance, a contributor item can be cast to
     * <code>IContributorHandle</code>; and, if it has full state, it can be
     * cast to <code>IContributor</code>.</li>
     * <li>The name of the EMF class of the object bears the name acknowledged
     * in the item type. That is, <code>this.eClass().getName()</code> matches
     * <code>this.getItemType().getName()</code>[ + &quot;
     * <code>Handle</code> &quot;]. Similarly,
     * <code>this.eClass().getPackage().getNsURI()</code> matches
     * <code>this.getItemType().getNamespaceURI()</code>.</li>
     * <li>The exact set of structural features inherited from the EMF classes
     * named &quot;ItemHandle&quot;, &quot;Item&quot;, &quot;Auditable&quot;,
     * and &quot;SimpleItem&quot; are internal, undocumented, and subject to
     * change. While a client is free to display these structural features for
     * debug purposes, they must not make assumptions about their continued
     * existence or usage.</li>
     * <li>The client must not explicitly modify (<code>eSet</code>) any of
     * the structural features inherited from the EMF classes named
     * &quot;ItemHandle&quot;, &quot;Item&quot;, &quot;Auditable&quot;, or
     * &quot;SimpleItem&quot;. These are reserved for use by the implementation
     * of the Client API.</li>
     * <li>The client must not explicitly modify (<code>eSet</code>) meta
     * models or install new ones. The meta models are read-only and reserved
     * for use by the implementation of the Client API.</li>
     * <li>As always, everything in Java packages with &quot;internal&quot; in
     * the name is strictly off limits.</li>
     * </ul>
     * </p>
     * <p>
     * Implementation note: There is no requirement that the nice API provided
     * for an extended item type bear any direct relationship to the underlying
     * EMF objects.
     * </p>
     * 
     * @param clazz
     *            the desired Java class; must not be <code>null</code>
     * @return an instance of <code>clazz</code>, or <code>null</code> if
     *         this item could not be so adapted
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public Object getAdapter(Class clazz);

    /**
     * Returns whether the given object is the same item as this item.
     * <p>
     * This method is implemented as Java object identity (==) for all item
     * handles (and items).
     * </p>
     * <p>
     * Design note: equals() is defined this way due in part to EMF restrictions
     * on the underlying objects (EMF objects cannot redefine equals).
     * </p>
     * 
     * @param o
     *            {@inheritDoc}
     * @return {@inheritDoc}
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public boolean equals(Object o);

    /**
     * Returns an approximation of the size in bytes of this object plus the
     * objects it references.
     * 
     * @return the size in bytes of this object plus the objects it references.
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public long size();

}
