/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010, 2015. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.repository.common;

import com.ibm.team.repository.common.internal.ItemTypeRegistry;
import com.ibm.team.repository.common.query.ast.IDynamicItemQueryModel;

/**
 * Represents a type of item found on the server. The items types include both
 * auditable items (teams, contributors, workspaces, etc.), versionable items
 * (files, folders) and simple items.
 * 
 * @since 0.5
 * @NoImplement This interface is not intended to be implemented by clients.
 */
public interface IItemType extends IType {

	/* Known violators of @NoImplement rule:
	 * CCM: com.ibm.team.build.common.tests/src/com/ibm/team/build/internal/common/helper/tests/PropertyHelperTests.java */

    /**
     * Constants for Read Access Policy. 
     * @since 0.7
     * @see com.ibm.team.repository.service.permission
     */
    public static enum READ_ACCESS_POLICY {
		/** All items of this type can be accessed by all users. */
		PUBLIC, 
		/** The accessibility of an item is determined by its contextId */
		PROTECTED, 
		/** The accessibility of an item is determined by a corresponding
		 * {@link com.ibm.team.repository.service.IComponentPermissionParticipant}.
		 * The use of this read access policy is discouraged for performance
		 * and complexity reasons. */
		APPLICATION_MANAGED,
		/** No items of this type can be accessed by any user. */
		INTERNAL,
    };

    /**
     * Returns whether this item type is an auditable item type.
     * 
     * @return <code>true</code> if this is an auditable item type, and
     *         <code>false</code> otherwise
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public boolean isAuditable();

    /**
     * Returns whether this item type is a versionable item type.
     * 
     * @return <code>true</code> if this is a versionable item type, and
     *         <code>false</code> otherwise
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public boolean isUnmanaged();

    /**
     * Returns whether this item type is a simple item type.
     * 
     * @return <code>true</code> if this is a simple item type, and
     *         <code>false</code> otherwise
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public boolean isSimple();

	/**
	 * Returns whether this item type is a configuration aware type.
	 *
	 * @return <code>true</code> if this is a configuration aware type, and
	 *         <code>false</code> otherwise
	 * @ShortOp This is a short operation; it may block only momentarily; safe
	 *          to call from a responsive thread.
	 */
	public boolean isConfigurationAware();

    /**
     * Returns whether items of this type can be read by everyone.
     * @return true if the item type has a Public read access policy;
     *         false otherwise
     * @since 0.7
     */
    public boolean isPublic();
    
    /**
     * Returns whether items of this type is protected and can be accessed only
     * by the members of the item's context
     * @return true if the item type has a protected read access policy;
     *         false otherwise
     * @since 0.7
     */
    public boolean isProtected();
    
    /**
     * Returns whether read access to the items of this type is managed by the
     * application
     * @return true if the item type has an application managed read access policy;
     *         false otherwise
     * @since 0.7
     */
    public boolean isApplicationManaged();
    
    /**
     * Returns whether read access to the items of this type is internal
     * 
     * @return true if the item type has an internal access policy;
     *         false otherwise
     * @since 0.7
     */
    public boolean isInternal();
    
	/**
	 * Returns whether items of this type support redaction.
	 * 
	 * @return true if the item type supports redaction; false otherwise
	 */
	public boolean isRedactable();

	/**
	 * Returns whether items of this type can be stored in the common
	 * configuration.
	 * 
	 * @return true if items of this type can be stored in the common
	 * configuration; false otherwise
	 */
	public boolean isStoreableInCommonConfiguration();

    /**
     * Creates a new item of this item type. The resulting item can be edited
     * and saved to the repository as opposed to a read-only item.
     * 
     * @return the created item or <code>null</code> if this item type is an
     *         invalid item type whose corresponding instance class is not an
     *         item class.
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public IItem createItem();

    /**
     * Creates a new item of this item type. The resulting item can be edited
     * and saved to the repository as opposed to a read-only item. Method will
     * also assign the given origin to the item.
     * 
     * @param origin
     *            The item's origin; must not be <code>null</code>
     * @return the created item or <code>null</code> if this item type is an
     *         invalid item type whose corresponding instance class is not an
     *         item class.
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public IItem createItem(Object origin);

    /**
     * Creates a new item handle of this item type, item id, and optional state
     * id. The result records the item type, item id, and state if supplied.
     *
     * There is no validation on the parameters.  This method will succeed
     * whether or not there is an existing item with the itemId, and will even
     * succeed if there is an existing item with the itemId of a different type.
     *
     * @param itemId
     *            the item id; must not be <code>null</code>
     * @param optionalStateId
     *            the state id, or <code>null</code> if none
     * @return the created item handle or <code>null</code> if this item type
     *         is an invalid item type whose corresponding instance class is not
     *         an item class.
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public IItemHandle createItemHandle(UUID itemId, UUID optionalStateId);

    /**
     * Creates a new item handle of this item type, item id, and optional state
     * id. The result records the item type, item id, and state if supplied.
     * Method will also assign the given origin to the item handle.
     * 
     * There is no validation on the parameters.  This method will succeed
     * whether or not there is an existing item with the itemId, and will even
     * succeed if there is an existing item with the itemId of a different type.
     *
     * @param origin
     *            The item handle's origin; must not be <code>null</code>
     * @param itemId
     *            the item id; must not be <code>null</code>
     * @param optionalStateId
     *            the state id, or <code>null</code> if none
     * @return the created item handle or <code>null</code> if this item type
     *         is an invalid item type whose corresponding instance class is not
     *         an item class.
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public IItemHandle createItemHandle(Object origin, UUID itemId,
            UUID optionalStateId);

    /**
     * Answers whether state extensions are allowed for this item type.
     * <p>
     * The component that declares the item type controls whether its item
     * states can be augmented with state extensions (default: not allowed).
     * </p>
     * 
     * @return <code>true</code> if state extensions are allowed for this item
     *         type; otherwise <code>false</code>
     * @see IExtensibleItem
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public boolean allowsStateExtensions();
    
    /**
     * Gets dynamic version of the root query model associated with this
     * item type. Used for dynamic query building.  Note that this is only needed when
     * the static query model for an item type is not visible, for instance because 
     * of an inverted or soft dependency on another model.  Generally, static query 
     * models should be used whenever they are visible and the types/properties are known
     * at compile time.  Note also that there are no API contracts regarding dynamic APIs - 
     * model objects may change shape, queryable properties, etc.
     * 
     * @return a dynamic query model for this item type
     * @ShortOp This is a short operation; it may block only momentarily;
     *          safe to call from a responsive thread.
     */
     public IDynamicItemQueryModel getQueryModel();

    
    /**
     * Static registry used to locate item types.
     * <p>
     * Example usage:
     * 
     * <pre>
     *              public interface ISomeClass extends ISomeClassHandle {
     *              public static final IItemType ITEM_TYPE = IItemType.IRegistry.INSTANCE
     *                    .getItemType(SomePackage.eINSTANCE.getSomeClass().getName(),
     *                            SomePackage.eNS_URI);
     * </pre>
     * 
     * </p>
     */
    public static interface IRegistry {

        /**
         * The one instance of the <code>IRegistry</code> object.
         */
        public static IRegistry INSTANCE = new ItemTypeRegistry();

        /**
         * Locates an item type given its name and its namespace URI.
         * 
         * @param name
         *            The item type's name; must not be <code>null</code>
         * @param namespaceURI
         *            the item type's namespace URI; must not be
         *            <code>null</code>
         * @return the located item type or <code>null</code> if the item type
         *         name and namespace URI pair was not located in the item type
         *         registry
         * @ShortOp This is a short operation; it may block only momentarily;
         *          safe to call from a responsive thread.
         */
        public IItemType getItemType(String name, String namespaceURI);

        /**
         * Locates all item types.  Abstract types will be excluded.
         * 
         * @return all item types; never <code>null</code>
         * @ShortOp This is a short operation; it may block only momentarily;
         *          safe to call from a responsive thread.
         */
        public IItemType[] getAllItemTypes();
        
        /**
         * Locates all local item types.  Abstract types will be excluded.
         * 
         * @return all item types; never <code>null</code>
         * @ShortOp This is a short operation; it may block only momentarily;
         *          safe to call from a responsive thread.
         */
        public IItemType[] getLocalItemTypes();        
        
        /**
         * Locates all item types.
         * 
         * @param includeAbstractTypes indicates whether abstract types should 
         *          be included in the returned collection
         * @return all item types; never <code>null</code>
         * @ShortOp This is a short operation; it may block only momentarily;
         *          safe to call from a responsive thread.
         */
        public IItemType[] getAllItemTypes(boolean includeAbstractTypes);

               
    }
}
