/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010, 2012. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.repository.common;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import com.ibm.team.repository.common.model.ILicensesInfo;

public interface ILicenseAdminService {

	/**
	 * Revokes the license for the given operation.  This does not
	 * make any permanent changes to the license, so the main use 
	 * is for testing.
	 * 
	 * @param operationId the operation to revoke the license for (cannot be null)
	 */
	public void revokeLicense(String operationId);
	
	/**
	 * Restores a previously revoked license.  If the operation was
	 * not allowed before the revoke call then calling this method
	 * will not grant the operation.  No permanent changes are made
	 * the the license files, so the main use is for testing.
	 * 
	 * @param operationId the operation to revoke the license for (cannot be null)
	 */
	public void restoreLicense(String operationId);
	
	/**
	 * Returns types of known CAL licenses, not include default licenses.
	 * 
	 * @return a list of know license types.
	 */
	public IContributorLicenseType[] getLicenseTypes() throws TeamRepositoryException;
	
	/**
	 * Assigns a license to a contributor. The caller must have admin permissions.
	 * 
	 * @param contributor the contributor to get the license (cannot be null)
	 * @param licenseId the license to assign (cannot be null)
	 * @throws TeamRepositoryException if the assignments fails, such as
	 *         when there are no more licenses available to assign
	 * @throws IllegalArgumentException if the passed in license id does not 
	 * correspond to a valid license type. 
	 */
	public void assignLicense(IContributorHandle contributor, String licenseId) throws TeamRepositoryException;
	
	/**
	 * Unassigns a license from a contributor. The caller must have admin permissions.
	 * 
	 * @param contributor the contributor to lose the license (cannot be null)
	 * @param licenseId the license to unassign (cannot be null)
	 * @throws TeamRepositoryException if something goes wrong
	 */
	public void unassignLicense(IContributorHandle contributor, String licenseId) throws TeamRepositoryException;

	/**
	 * Unassigns all licenses from a contributor. The caller must have admin permissions. 
	 * 
	 * @param contributor the contributor to lose all licenses (cannot be null)
	 * @throws TeamRepositoryException if something goes wrong
	 */
	public void unassignAllLicenses(IContributorHandle contributor) throws TeamRepositoryException;
	
	/**
	 * Returns the set of assigned licenses for a given contributor.
	 * 
	 * @param contributor the contributor to find the licenses for (cannot be null).
	 * @return the set of license ids currently assigned to the contributor; Never null.
	 * @throws TeamRepositoryException
	 */
	public String[] getAssignedLicenses(IContributorHandle contributor) throws TeamRepositoryException;
	
	/**    
	 * Updates the number of licenses that have been purchased. This method
	 * will fail if the new maximum exceeds the maximum licenses allowed by the current
	 * server license. The caller must have admin permissions.
	 * 
	 * @param licenseId the license id to update (cannot be null)
	 * @param max the new maximum licenses.
	 * @throws TeamRepositoryException if anything goes wrong
	 */
	public void setLicensePurchaseCount(String licenseId, Integer newMax) throws TeamRepositoryException;
	
	/**
	 * Adds the license to the list of stored licenses.  The new license can be 
	 * either a server or contributor license.
	 * 
	 * @param licenseFileContents the contents of the license file
	 * @throws TeamRepositoryException if anything goes wrong
	 */
	public void addLicense(byte[] licenseFileContents) throws TeamRepositoryException;
	
    /**
     * Adds the set of licenses to the server and returns the state of the
     * server after the add.  If submit is false then the licenses are not 
     * really added and the return value is the calculated value of what 
     * the server state <em>would</em> look like if submit had been true
     * 
     * @param localFiles the licenses to be added
     * @param submit whether to apply the licenses or not
     * @throws TeamRepositoryException if anything goes wrong
     */	
	public ILicensesInfo addLicenses(String[] localFiles, boolean submit) throws TeamRepositoryException;
	
	/**
     * Removes the license to the list of stored licenses.  The new license can be 
     * either a server or contributor license.
     * 
     * @param licenseFileContents the contents of the license file
     * @throws TeamRepositoryException if anything goes wrong
     */
    public void removeLicense(byte[] licenseFileContents) throws TeamRepositoryException;
    
    /**
     * Removes a license by its id
     * 
     * @param id
     * @throws TeamRepositoryException
     */
    public void removeLicenseById(String id) throws TeamRepositoryException;
	
	public IServerLicenseType getServerLicenseType() throws TeamRepositoryException;
	
	/**
	 * Returns a list of contributors who have been assigned the given license.
	 * 
	 * @param licenseId the license id to search for
	 * @return a list of contributors who have been assigned the given license
	 * @throws TeamRepositoryException if anything goes wrong
	 */
	public IContributorHandle[] getLicensedContributors(String licenseId) throws TeamRepositoryException;
	
    /**
     * Is valid license
     * @return
     * @throws TeamRepositoryException
     */
    public boolean isValidLicense(String licenseId) throws TeamRepositoryException;

    /**
     * Returns the license agreement text for passed in license id.
     * 
     * @param licenseId
     * @return
     * @throws TeamRepositoryException
     */
    public String getLicenseAgreement(String licenseId) throws TeamRepositoryException;

    /**
     * Reassigns the users from the source license to the target license. The
     * caller must have admin permissions.
     * 
     * @param contributors
     *            the array contributor to get the license (cannot be null)
     * @param licenseId
     *            the license to assign (cannot be null)
     * @throws TeamRepositoryException
     *             if the assignments fails, such as when there are no more
     *             licenses available to assign
     * @throws IllegalArgumentException
     *             if the passed in license id does not correspond to a valid
     *             license type.
     */
    public BatchResult reassignLicense(String sourceLicenseId, String targetLicenseId, IContributorHandle... contributors) throws TeamRepositoryException;

    /**
     * Assigns a license to a user and returns an object representing the result
     * of the assignment.
     * 
     * @param handle
     *            The contributor being assigned a license
     * @param licenseId
     *            The license to assign
     * @return The result of the assignment
     * @throws TeamRepositoryException
     *             If an error occurs during assignment
     * @since 4.0.1
     */
    public ILicenseAssignmentResult assignLicenseWithResult(IContributorHandle handle, String licenseId) throws TeamRepositoryException;

    /**
     * Keeps track of any failures. The uuid of the handle is keyed to a result
     * object rather than the IContributor itsself. This allows for retrieval of
     * the result even if the original handle is not around
     */
    class BatchResult {

        private Map<UUID, ResultMapEntry> resultMap = new HashMap<UUID, ResultMapEntry>();

        public Set<IContributorHandle> getFailed() {
            Set<IContributorHandle> handles = new HashSet<IContributorHandle>();
            for (ResultMapEntry r : resultMap.values())
                handles.add(r.handle);
            return handles;
        }

        public String getError(IContributorHandle handle) {
            ResultMapEntry result = resultMap.get(handle.getItemId());
            return result != null ? result.message : null;
        }

        public void failure(IContributorHandle handle, String message) {
            if (message != null && message.length() > 0)
                resultMap.put(handle.getItemId(), new ResultMapEntry(handle, message));
        }
    }

    class ResultMapEntry {
        IContributorHandle handle = null;
        String message = ""; //$NON-NLS-1$

        public ResultMapEntry(IContributorHandle handle, String message) {
            this.handle = handle;
            this.message = message;
        }
    }
}