/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.repository.common;

public interface ILicenseService {

    /**
     * Asserts that the current user has a license to perform the operation
     * requested.
     * 
     * @param operationId
     *            the operation to check
     * @throws LicenseNotGrantedException
     *             if no license is available for the operation
     */
    public void assertLicense(String operationId) throws LicenseNotGrantedException;

    /**
     * Asserts that the current user has a license to perform the operation
     * requested. Attempts to auto-assign a license if not assigned and if an
     * appropriate one is available.
     * 
     * @param operationId
     * @param autoAssign
     * @throws LicenseNotGrantedException
     */
    public void assertLicenseAndAssign(String operationId) throws LicenseNotGrantedException;

    /**
     * Checks that the current user has a license to perform the listed
     * operation ids. The result of this call is an array with the same number
     * of elements as the given operationIds array. Each entry in the array is
     * null if the corresponding operation id is allowed, and a String
     * containing the reason for failure if the operation is not allowed.
     * 
     * @param operationIds
     *            the operations to check
     * @return an array of results that corresponds to the array of requested
     *         operation ids
     */
    public String[] checkLicenses(String[] operationIds);

    /**
     * Checks that the passed in user has a license to perform the listed
     * operation ids. The result of this call is an array with the same number
     * of elements as the given operationIds array. Each entry in the array is
     * null if the corresponding operation id is allowed, and a String
     * containing the reason for failure if the operation is not allowed.
     * 
     * @param opIds
     *            the operations to check
     * @param contributor
     *            the contributor to check the ids for
     * @return an array of results that corresponds to the array of requested
     *         operation ids
     */
    public String[] checkLicensesForContributor(String[] opIds, IContributorHandle contributor);

    /**
     * Asserts that the given Contributor has a license to perform the operation
     * requested. This methods should rarely be used, and using
     * {@link #assertLicense(String)} is preferred whenever possible.
     * 
     * @param operationId
     *            the operation to check. Must not be null.
     * @param contributor
     *            the contributor to use for the check. Must not be null.
     * @throws LicenseNotGrantedException
     *             if no license is available for the operation
     */
    public void assertLicenseForContributor(String operationId, IContributorHandle contributor) throws LicenseNotGrantedException;

    /**
     * Asserts that the given Contributor has a license to perform the operation
     * requested. Attempts to auto-assign a license if not assigned and if an
     * appropriate one is available. This methods should rarely be used, and
     * using {@link #assertLicense(String)} is preferred whenever possible.
     * 
     * @param operationId
     *            the operation to check. Must not be null.
     * @param contributor
     *            the contributor to use for the check. Must not be null.
     * @param autoAssign
     *            whether ot not to auto-assign a license.
     * @throws LicenseNotGrantedException
     *             if no license is available for the operation
     */
    public void assertLicenseAndAssignForContributor(String operationId, IContributorHandle contributor) throws LicenseNotGrantedException;


    
}