/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.repository.common;

/**
 * Represents the relative synchronization of two states of an auditable item
 * and provides the information needs to reconcile any differences.
 * 
 * @since 0.5
 * @NoImplement This interface is not intended to be implemented by clients.
 */
public interface IReconcileReport {

    /**
     * Kind constant indicating nothing to reconcile.
     * 
     * @see #getKind()
     */
    public static final int NO_CHANGE = 0;

    /**
     * Kind constant indicating an outgoing item creation.
     * 
     * @see #getKind()
     */
    public static final int OUTGOING_CREATE = 1;

    /**
     * Kind constant indicating an outgoing update, unopposed by any sort of
     * incoming change.
     * 
     * @see #getKind()
     */
    public static final int OUTGOING_UPDATE = 2;

    /**
     * Kind constant indicating an incoming update, unopposed by any sort of
     * outgoing change.
     * 
     * @see #getKind()
     */
    public static final int INCOMING_UPDATE = 3;

    /**
     * Kind constant indicating an incoming item creation.
     * 
     * @see #getKind()
     */
    public static final int INCOMING_CREATE = 4;

    /**
     * Kind constant indicating an outgoing update in conflict with an incoming
     * update.
     * 
     * @see #getKind()
     */
    public static final int CONFLICTING_UPDATES = 5;

    /**
     * Returns the relative synchronization of two states of an auditable item.
     * 
     * @return one of the kind constants declared on this interface
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public int getKind();

    /**
     * Returns the state of the item in the outgoing stream. This state may or
     * may not be a working copy.
     * 
     * @return the outgoing item state; never <code>null</code>
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public IAuditable getOutgoing();

    /**
     * Returns the state of the item in the incoming stream. This state may or
     * may not be a working copy.
     * 
     * @return the incoming item state; never <code>null</code>
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public IAuditable getIncoming();

    /**
     * Returns the common base state shared between the incoming stream and
     * outgoing streams. This is the state at which the streams diverged. This
     * state is never a working copy. It is not applicable for creations.
     * 
     * @return the base item state, or <code>null</code> if none
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public IAuditable getBase();

    /**
     * Returns the new state of the item which the client should modify to
     * represent the result of merging the incoming changes with the outgoing
     * changes. Always a working copy, and likely an alias for either the
     * incoming or outgoing state. This is only applicable for conflicting
     * updates.
     * 
     * @return the merge item working copy, or <code>null</code> if none
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public IAuditable getMergeResult();

}
