/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.repository.common;

import com.ibm.team.repository.common.internal.validation.IInternalItemValidator;

/**
 * Represents a type of object found on the server. Types include auditable
 * items (teams, contributors, workspaces, etc.), versionable items (files,
 * folders), simple items and helpers . Since the repository schema is
 * extensible, the set of types is open-ended.
 * <p>
 * Within a repository, type objects are singletons and could be compared for
 * equality using <code>equals()</code> or <code>==</code>.
 * 
 * @since 0.5
 * @NoImplement This interface is not intended to be implemented by clients.
 */
public interface IType {

    /**
     * @see #getMaxSize(String)
     * @deprecated - This will be removed in 0.6 M3 - for validation use {@link IPropertyConstraintValidator}
     */
    public final static long NO_SIZE = -1;

    /**
     * Returns the name of this type. A type is given a name by the party that a
     * declares that type.
     * <p>
     * For built-in types, the names are the same as the ones found in the
     * corresponding API interfaces minus the 'I' prefix. For example, the name
     * of built-in contributor item type is &quot;Contributor&quot;,
     * corresponding to the interface <code>IContributor</code>.
     * </p>
     * <p>
     * The name of the type is also the base name for the EMF classes used to
     * represent objects of this type.
     * 
     * @return the name of the class; never <code>null</code>
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public String getName();

    /**
     * Returns the namespace URI for this type. The namespace URI is used to
     * ensure that types declared by different providers are distinct even if
     * they should happen to have the same name.
     * <p>
     * The namespace URI of the type is also that of the EMF package that
     * contains the EMF classes used to represent objects of this type.
     * </p>
     * 
     * @return the namespace URI for this type; never <code>null</code>
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public String getNamespaceURI();

    /**
     * Returns the guaranteed maximum size of the given property.
     * <p>
     * For properties of type <code>String</code>, this method will return
     * the guaranteed maximum number of characters allowed in the property
     * value. For properties of type other than string, this method will return
     * the constant value <code>NO_SIZE</code>.
     * </p>
     * 
     * @param propertyName
     *            the property; must not be <code>null</code>
     * @return guaranteed maximum size of property
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     * @deprecated {@link IInternalItemValidator} should be
     *             used instead. This method is too inaccurate for reasonable
     *             string validation. This will be removed in 0.6 M3
     */
    public long getMaxSize(String propertyName);

    /**
     * Returns if the given string can be persisted for the given property. This
     * method is a more precise form of the {@link} #getMaxSize(String)}
     * 
     * @param propertyName
     *            the property; must not be <code>null</code>
     * @param value
     *            the string that should be measured
     * @return <code>true</code> if the string is small enough to fit under
     *         the given property
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     * @deprecated use {@link #validateStringProperty(String, String)}. This
     *             will be removed in 0.6 M3
     */
    public boolean isStringSmallEnough(String propertyName, String value);

    /**
     * Returns an object that adapts this type to the given class. The result
     * can be cast to <code>clazz</code>.
     * <p>
     * This method is used to get at the special API for objects of this type
     * when possible.
     * </p>
     * Pass <code>EPackage.class</code> to get reflective access to the
     * underlying EMF package object (the one for the given
     * <code>getNamespaceURI()</code>). From the EMF package one can navigate
     * to the EMF classes within it.
     * </p>
     * </p>
     * Pass <code>EClass.class</code> to get reflective access to the
     * underlying EMF class object.
     * </p>
     * 
     * @param clazz
     *            the desired Java class; must not be <code>null</code>
     * @return an instance of <code>clazz</code>, or <code>null</code> if
     *         this type could not be so adapted
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public Object getAdapter(Class clazz);

    /**
     * Indicates if this type represents an abstract base class in the component
     * model
     * 
     * @return true if abstract; otherwise false
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public boolean isAbstract();
}
