/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.repository.common;

import com.ibm.team.repository.common.internal.util.ComponentRegistry;

/**
 * Represents a type of data transfer object (DTO) found on the server. DTOs are
 * not persisted in the repository.
 * 
 * @since 0.6
 * @NoImplement This interface is not intended to be implemented by clients.
 */
public interface IVirtualType extends IType {

    /**
     * Creates a new virtual object of this type.
     * 
     * @return the created object or <code>null</code> if this type is an
     *         invalid type whose corresponding instance class is not a
     *         virtual class.
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public IVirtual createVirtual();
    
    /**
     * Static registry used to locate virtual types.
     * <p>
     * Example usage:
     * 
     * <pre>
     *              public interface ISomeClass extends ISomeClassHandle {
     *              public static final IVirtualType TYPE = IVirtualType.IRegistry.INSTANCE
     *                    .getVirtualType(SomePackage.eINSTANCE.getSomeClass().getName(),
     *                            SomePackage.eNS_URI);
     * </pre>
     * 
     * </p>
     */
    public static interface IRegistry {

        /**
         * The one instance of the <code>IRegistry</code> object for virtual types.
         */
        public static IRegistry INSTANCE = ComponentRegistry.INSTANCE;

        /**
         * Locates an virtual type given its name and its namespace.
         * 
         * @param name
         *            The virtual type's name; must not be <code>null</code>
         * @param namespace
         *            the virtual type's namespace; must not be
         *            <code>null</code>
         * @return the located virtual type or <code>null</code> if the virtual type
         *         name and namespace pair was not located in the virtual type
         *         registry
         * @ShortOp This is a short operation; it may block only momentarily;
         *          safe to call from a responsive thread.
         */
        public IVirtualType getVirtualType(String name, String namespace);

        /**
         * Locates all virtual types.  Abstract types will be excluded.
         * 
         * @return all virtual types; never <code>null</code>
         * @ShortOp This is a short operation; it may block only momentarily;
         *          safe to call from a responsive thread.
         */
        public IVirtualType[] getAllVirtualTypes();
        
        /**
         * Locates all virtual types.
         * 
         * @param includeAbstractTypes indicates whether abstract types should 
         *          be included in the returned collection
         * @return all virtual types; never <code>null</code>
         * @ShortOp This is a short operation; it may block only momentarily;
         *          safe to call from a responsive thread.
         */
        public IVirtualType[] getAllVirtualTypes(boolean includeAbstractTypes);

    }
}
