/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.repository.common.internal.marshal;

import java.io.InputStream;
import java.io.OutputStream;
import java.nio.charset.Charset;
import java.util.Map;

import org.eclipse.emf.ecore.EObject;

import com.ibm.team.repository.common.transport.internal.services.DataArg;
import com.ibm.team.repository.common.transport.internal.services.DataArgType;

/**
 * Marshaller to serialize / deserialize an object
 * @model abstract="true"
 */
public interface Marshaller extends EObject {

	/**
	 * String describing the version of marshalling currently supported.
	 */
	static final public String MarshallingVersion = "0.2"; //$NON-NLS-1$
    
    /**
     * Get the type of Marshaller. Marshaller type can be used to determine whether
     * a marshaller can marshal the passed in object.
     * @model
     */
    MarshallerType getMarshallerType();
    
    /**
     * Sets the value of the '{@link com.ibm.team.repository.common.internal.marshal.Marshaller#getMarshallerType <em>Marshaller Type</em>}' attribute.
     * <!-- begin-user-doc -->
     * <!-- end-user-doc -->
     * @param value the new value of the '<em>Marshaller Type</em>' attribute.
     * @see com.ibm.team.repository.common.internal.marshal.MarshallerType
     * @see #getMarshallerType()
     * @generated
     */
    void setMarshallerType(MarshallerType value);

    /**
     * Supported Type of the Marshaller. The marshaller can marshal an object of the supported type.
     * @model type="java.lang.Class" 
     */
    Class getSupportedType();
    
    /**
     * Sets the value of the '{@link com.ibm.team.repository.common.internal.marshal.Marshaller#getSupportedType <em>Supported Type</em>}' attribute.
     * <!-- begin-user-doc -->
     * <!-- end-user-doc -->
     * @param value the new value of the '<em>Supported Type</em>' attribute.
     * @see #getSupportedType()
     * @generated
     */
    void setSupportedType(Class value);

    /**
     * Get the supported data arg type. The marshaller can be used to marshal the supported
     * data arg type
     * @model parameters=""
     */
    DataArgType getSupportedDataArgType();

    /**
     * Registration name of the marshaller. The registration name-marshaller combination is 
     * registered with the marshaller factory. The consumer of the marshaller framework
     * must request the appropriate marshaller from the marshaller factory
     * @model parameters=""
     */
    String getRegistrationName();
    
    /**
     * Can the Marshaller marshal the object?
     * @model parameters="java.lang.Object"
     */
    boolean canMarshal(Object object);

    /**
     * Can Marshal an object of the type?
     * @model parameters="java.lang.Class"
     */
    boolean canMarshal(Class type);
    
    /**
     * Marshal the object to output stream
     */
    void marshalObjectToOutputStream(Object object, OutputStream outputStream) throws MarshallingException;
    
    /**
     * @param object The object to serialize. Must not be <code>null</code>.
     * 
     * @param outputStream The output stream to write to. Must not be null.
     * 
     * @param charset The charset to encode the response as. May be <code>null</code>.
     * If <code>null</code>, "UTF-8" is used.
     * 
     * @throws MarshallingException Thrown on error.
     */
    void marshalObjectToOutputStream(Object object, OutputStream outputStream, 
            Charset charset) throws MarshallingException;

	/**
	 * Marshal the object to output stream
	 * @param object The object to serialize. Must not be <code>null</code>.
	 * @param outputStream The output stream to write to. Must not be null.
	 * @param charset The charset to encode the response as. May be <code>null</code>.
	 * If <code>null</code>, "UTF-8" is used.
	 * @param componentConfiguration Version information about the entity receiving
	 * the marshaled object.  Must not be <code>null</code>.
	 * @throws MarshallingException Thrown on error.
	 */
	void marshalObjectToOutputStream(Object object, OutputStream outputStream, 
			Charset charset, Map<String, String> componentConfiguration) throws MarshallingException;

    /**
     * Marshal the object to a String
     * @param object
     * @return
     * @throws MarshallingException
     */
    String marshalObjectToString(Object object) throws MarshallingException;
    
    /**
     * This method is meant for JSON serialization and must be used for marshaling 
     * EOBject attributes
    
     * @param object 
     *              Object to be led
     * @return marshaled value
     * @throws MarshallingException if the object cannot be marshalled
     */
    Object marshalAttributeValue(Object object) throws MarshallingException;
    
    /**
     * Marshal the passed in object
     * @model parameters="java.lang.Object" type="com.ibm.team.repository.common.transport.internal.services.DataArg"
     */
    DataArg marshalObjectToDataArg(Object object);
    
    /**
     * Demarshal the input stream to an Object. 
     * 
     * @param inputStream 
     *              The input stream to read and deserialize from. Must not be 
     *              <code>null</code>.
     * @return de-serialized object
     * @throws MarshallingException if there are problems demarshalling the 
     * inputstream             
     */
    Object demarshalInputStreamToObject(InputStream inputStream) throws MarshallingException;
    
    /**
     * Demarshal the input stream to an Object. This varies from 
     * {@link #demarshalInputStreamToObject(InputStream)} only in that the 
     * decoding character set may be specified.
     * 
     * @param inputStream 
     *                  The input stream to read and deserialize from. Must
     *                  not be <code>null</code>.
     * @param charset   The character set used to decode the input stream contents. 
     *                  May be <code>null</code>, in which case the default charset is used.
     * 
     * @return The array of de-marshaled objects. Never <code>null</code>.
     * @throws MarshallingException Thrown on marshaling errors.
     */
    Object demarshalInputStreamToObject(InputStream inputStream, Charset charset)
        throws MarshallingException;

    /**
     * Demarshal the data arg to Object
     */
    Object demarshalDataArgToObject(DataArg dataArg, Class type);
}