/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010, 2014. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.repository.common.internal.marshal.impl;

import java.util.Map;

import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;

import com.ibm.team.repository.common.internal.marshal.MarshalFactory;
import com.ibm.team.repository.common.internal.marshal.Marshaller;
import com.ibm.team.repository.common.internal.marshal.MarshallerType;
import com.ibm.team.repository.common.internal.marshal.MarshallingException;
import com.ibm.team.repository.common.internal.marshal.MarshallingExceptionReasonCode;
import com.ibm.team.repository.common.internal.marshal.util.MarshallerUtil;
import com.ibm.team.repository.common.model.RepositoryPackage;
import com.ibm.team.repository.common.serialize.internal.nls.Messages;
import com.ibm.team.repository.common.util.NLS;

public abstract class AbstractEObjectMarshaller extends ComplexTypeMarshallerImpl {
    
    protected static final String EMF_TYPE_NAME = "org.eclipse.emf.EObject";  //$NON-NLS-1$
    protected static final String EOBJECT_REGISTRATION_NAME = "org.eclipse.emf.ecore.EObject"; //$NON-NLS-1$
    
    protected static final String EXTENDED_METADATA_ANNOTATION = "http:///org/eclipse/emf/ecore/util/ExtendedMetaData"; //$NON-NLS-1$
    protected static final String NAME_ANNOTATION_KEY = "name"; //$NON-NLS-1$
    protected static final String NAMESPACE_ANNOTATION_KEY = "namespace"; //$NON-NLS-1$
    
    // Option to save type information. The default value is Boolean.FALSE 
    public static final String OPTION_SAVE_TYPE_INFORMATION = "SAVE_TYPE_INFORMATION"; //$NON-NLS-1$
    
    protected boolean saveTypeInformation = false;
    
    protected AbstractEObjectMarshaller(Class clazz, MarshallerType marshallerType) {
        super();
        setSupportedType(clazz);
        setMarshallerType(marshallerType);
    }
    
    protected AbstractEObjectMarshaller(Class clazz, MarshallerType marshallerType, Map options) {
        super();
        setSupportedType(clazz);
        setMarshallerType(marshallerType);
        if (Boolean.TRUE.equals(options.get(OPTION_SAVE_TYPE_INFORMATION))) {
            saveTypeInformation = true;
        }
    }
    @Override
    public String getRegistrationName() {
        return EOBJECT_REGISTRATION_NAME;
    }
    
    @Override
    public boolean canMarshal(Object object) {
        return object instanceof EObject;
    }
    
    /**
     * Can marshal an item of the passed in class?
     */
    @Override
    public boolean canMarshal(Class type) {
        return EObject.class.isAssignableFrom(type);
    }
    
    /**
     * Can Marshal an object of the given type
     * @throws MarshallingException 
     */
    public boolean canMarshal(String typeName) throws MarshallingException {
        return typeName.equals(EMF_TYPE_NAME);
    }
    
    protected String getQualifiedEClassName(EClass eClass) {
        return getQualifiedEClassName(eClass, eClass.getEPackage().getNsURI());
    }
    
    /**
     * @deprecated use {@link #getQualifiedEClassName(EObject, Map)} instead 
     */
    @Deprecated
    protected String getQualifiedEClassName(EObject eObject) {
        return  getQualifiedEClassName(eObject.eClass(), eObject.eClass().getEPackage().getNsURI());
    }
    
    /**
     * Utility method for getting the namespace prefix qualified class name of an object's eclass.
     * @param object an EMF object
     * @return the string containing the namespace-prefix qualified class name of the eobject
     */
    protected String getQualifiedEClassName(EObject eObject, Map<String, String> nsURIToNsPrefixMap) {
        return getQualifiedEClassName(eObject.eClass(), nsURIToNsPrefixMap);
    }
    
    protected String getQualifiedEClassName(EClass eClass, Map<String, String> nsURIToNsPrefixMap) {
        String nsURI = eClass.getEPackage().getNsURI();
        String nsPrefix = nsURIToNsPrefixMap.get(nsURI);
        if (nsPrefix == null) {
            // reference type references a super type not present in the nsURIToNsPrefix map
            nsPrefix = eClass.getEPackage().getNsPrefix();
        }
        return getQualifiedEClassName(eClass, nsPrefix);
    }
    
    /**
     * Get the qualified eClass Name given the EClass and namespace prefix
     */
    protected String getQualifiedEClassName(EClass eClass, String nsPrefix) {
        return nsPrefix + ':' + eClass.getName(); 
    }
    
    protected String getQualifieldEClassName(EObject eObject, String name) {
        return eObject.eClass().getEPackage().getNsPrefix() + ':' + name;
    }
    
    protected boolean shouldMarshalAsXMLAttribute(EAttribute attribute) {
        return attribute == RepositoryPackage.eINSTANCE.getItemHandle_ItemId();
    }
    
    protected boolean shouldAddTypeInformation(EObject eObject, String containingReferenceType, 
            Map nsUriToNsPrefix) {
        String eObjectReferenceType = getQualifiedEClassName(eObject, nsUriToNsPrefix);
        return saveTypeInformation || !eObjectReferenceType.equals(containingReferenceType);
    }
    
    protected boolean shouldBeMarshalled(EObject eObject, EStructuralFeature structuralFeature) {
        return MarshallerUtil.shouldBeMarshalled(eObject, structuralFeature);
    }
    
    protected String getStringValueUsingFactory(EObject eObject, EAttribute attribute, Object objectValue) {
        return eObject.eClass().getEPackage().getEFactoryInstance().convertToString(
                attribute.getEAttributeType(), objectValue);
    }
    
    protected String serializeEAttributeValue(EObject eObject, EAttribute attribute) throws MarshallingException {
        Object objectValue = eObject.eGet(attribute);
        if (objectValue == null) {
            return null;
        }
        
        Class classZ = attribute.getEType().getInstanceClass();
        String attributeValueAsString = null;
        if (classZ == null) {
            // Instance class is null for dynamic EMF. Use EMF Factory to determine the string value
            attributeValueAsString = getStringValueUsingFactory(eObject, attribute, objectValue);
        } else {
            Marshaller marshaller = MarshalFactory.eINSTANCE.getMarshaller(classZ);
            if (marshaller != null) {
                attributeValueAsString = marshaller.marshalObjectToString(objectValue);
            } else if (attribute.getEType().getName().equals(Object.class.getSimpleName())) {
                attributeValueAsString = getStringValueUsingFactory(eObject, attribute, objectValue);
            } else {
                String message = createCannotMarshalAttributeMessage(eObject, attribute);
                throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_MARSHAL_LITERAL, message);
            }
        }
        return attributeValueAsString;
    }
    
    protected final String createCannotMarshalAttributeMessage(EObject eObject, EAttribute attribute) {
        Class classZ = attribute.getEType().getInstanceClass();
        String typeName = classZ == null ? attribute.getEType().getName() : classZ.getName();
        return NLS.bind(Messages.getServerString("AbstractEObjectMarshaller.MarshallerNotFound"),  //$NON-NLS-1$
                eObject.eClass().getName(), attribute.getName(), typeName);
    }
    
    protected String getReferenceName(EObject eObject, EStructuralFeature structuralFeature) {
        if (structuralFeature.getEAnnotations().size() > 0) {
            String nameAnnotationValue = MarshallerUtil.getAnnotation(structuralFeature, 
                        EXTENDED_METADATA_ANNOTATION, NAME_ANNOTATION_KEY);
            String namespaceAnnotationValue = MarshallerUtil.getAnnotation(structuralFeature, 
                        EXTENDED_METADATA_ANNOTATION, NAMESPACE_ANNOTATION_KEY);
            if (nameAnnotationValue != null && namespaceAnnotationValue != null) {
                return getQualifieldEClassName(eObject, nameAnnotationValue);
            }
        }
        return structuralFeature.getName();
    }
    
    protected String createHREFs(EObject eObject, EReference reference) {
        EClassifier eClassifier = (EClassifier)eObject.eGet(reference);
        StringBuffer hRefURI = new StringBuffer("href = \""); //$NON-NLS-1$
        hRefURI.append(eClassifier.getEPackage().getNsURI());
        hRefURI.append("#//"); //$NON-NLS-1$
        hRefURI.append(eClassifier.getName());
        hRefURI.append("\""); //$NON-NLS-1$
        return hRefURI.toString();
    }
}