/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010, 2014. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.repository.common.internal.marshal.impl;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.io.Writer;
import java.nio.charset.Charset;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EDataType;
import org.eclipse.emf.ecore.EFactory;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.EcorePackage;
import org.eclipse.emf.ecore.util.FeatureMap.Entry;

import com.ibm.team.repository.common.UUID;
import com.ibm.team.repository.common.internal.marshal.MarshalFactory;
import com.ibm.team.repository.common.internal.marshal.Marshaller;
import com.ibm.team.repository.common.internal.marshal.MarshallerType;
import com.ibm.team.repository.common.internal.marshal.MarshallingException;
import com.ibm.team.repository.common.internal.marshal.MarshallingExceptionReasonCode;
import com.ibm.team.repository.common.internal.marshal.util.MarshallerUtil;
import com.ibm.team.repository.common.internal.marshal.util.NumberHelper;
import com.ibm.team.repository.common.internal.marshal.util.XMLConstructionUtil;
import com.ibm.team.repository.common.internal.util.ItemUtil;
import com.ibm.team.repository.common.json.JSONArray;
import com.ibm.team.repository.common.json.JSONObject;
import com.ibm.team.repository.common.model.ItemHandle;
import com.ibm.team.repository.common.model.RepositoryPackage;
import com.ibm.team.repository.common.serialize.internal.nls.Messages;
import com.ibm.team.repository.common.transport.HttpUtil.CharsetEncoding;
import com.ibm.team.repository.common.util.NLS;

/**
 * Class to marshal an EObject to a JSON Object 
 */
public class EObjectJSONMarshaller extends AbstractEObjectMarshaller {
    
    public static final String QUALIFIED_CLASS_NAME = "_eQualifiedClassName"; //$NON-NLS-1$
    public static final String ITEM_ID = "itemId"; //$NON-NLS-1$
    public static final String STATE_ID = "stateId"; //$NON-NLS-1$
    
    /**
     * Create a EObjectJSONMarshaller
     */
    public EObjectJSONMarshaller() {
        super(EObjectJSONMarshaller.class, MarshallerType.JSON_LITERAL);
    }
    
    @Deprecated
    public void marshalObjectToOutputStream(Object object, OutputStream outputStream) throws MarshallingException {
        marshalObjectToOutputStream(object, outputStream, CharsetEncoding.UTF8.toCharset());
    }

    public void marshalObjectToOutputStream(Object object, OutputStream outputStream,
            Charset charset) throws MarshallingException {
        
        if (charset == null)
            charset = CharsetEncoding.UTF8.toCharset();
            
        Writer writer = null;
        try {
            writer = XMLConstructionUtil.createWriter(outputStream, charset);
        } catch (final UnsupportedEncodingException e1) {
            throw new MarshallingException(e1);
        }
        
        marshalObject(object, writer);
    }

    private void marshalObject(Object object, Writer writer) throws MarshallingException {
        try {
            final Object result = marshalObjectToJSON(object);
            if (result instanceof JSONObject) {
                ((JSONObject)result).serialize(writer);
            } else if (result instanceof JSONArray) {
                ((JSONArray)result).serialize(writer);
            }
            writer.flush();
        } catch (final IOException e) {
            throw new MarshallingException(e);
        }
    }

    /**
     * Marshal the object to output stream.  Assumed to be either a collection,
     * or an EObject.
     */
    public Object marshalObjectToJSON(Object object) throws MarshallingException {
        
        // TODO: canMarshal will return TRUE only for an instance of EObject. 
        // Todd, we need to either add a new method to handle collection / change
        // canMarshal method to return true on a collection of EObjects
        
        // Handle collections of objects.
        if (object instanceof Collection) {
            return marshalObjectCollection((Collection) object);
        }
        
        // TODO: Add a comment on why we are assuming the passed in object is an
        // instance of EObject
        
        final EObject eObject = (EObject)object;
        final JSONObject jsonObject = new JSONObject();
        final EClass eClass = eObject.eClass();
        
        // Add qualified class name attribute
        final EPackage ePackage = eClass.getEPackage();
        final String qualifiedClassName = getQualifiedEClassName(eClass, ePackage.getNsURI());
        addAttributeToJSONObject(jsonObject, QUALIFIED_CLASS_NAME, qualifiedClassName);
        
        // Add the item type attribute, if appropriate
        if (RepositoryPackage.eINSTANCE.getItemHandle().isSuperTypeOf(eClass)) {
            addAttributeToJSONObject(jsonObject, JSONConstants.TYPENAME, 
                    ItemUtil.typeToName(ItemUtil.itemTypeFor(eClass)));
        }
        
        // Marshal attributes 
        marshalAttributes(eObject, eClass, jsonObject); 
        
        // Marshal references (containment and aggregation references)
        marshalReferences(eObject, eClass, jsonObject);
        
        return jsonObject;
    }
    
    private Object marshalObjectCollection(Collection collection) throws MarshallingException {
        final JSONArray jsonArray = new JSONArray();
        for (final Object obj : collection) {
            if (!(obj instanceof EObject)) {
                final String message = constructEObjectMarshallingOnlySupportedMessage(obj);
                throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL, message);
            }
            jsonArray.add(marshalObjectToJSON(obj));
        }
        return jsonArray;
    }
    
    private String constructEObjectMarshallingOnlySupportedMessage(Object obj) {
        
        final String message = MessageFormat.format("Only EObjects are supported as collection members.  Member is of type \"{0}\"", //$NON-NLS-1$
                obj.getClass());
        return message;
    }
    
    /**
     * Marshal Attributes present in EObject
     */
    protected void marshalAttributes(EObject eObject, EClass eClass, JSONObject jsonObject) throws MarshallingException {
        for (final EAttribute attribute : eClass.getEAllAttributes()) { 
            if (shouldBeMarshalled(eObject, attribute)) {
                marshalAttribute(eObject, attribute, jsonObject);
            }
        }
    }
    
    protected void marshalAttribute(EObject eObject, EAttribute attribute, JSONObject jsonObject) 
            throws MarshallingException {
        
        if (attribute.isMany()) {
            marshalManyValuedAttribute(eObject, attribute, jsonObject);
        } else {
            final Object attributeValue = serializeEAttributeValue2(eObject, attribute);
            addAttributeToJSONObject(jsonObject, attribute.getName(), attributeValue);
        }
    }
    
    protected Object serializeEAttributeValue2(EObject eObject, EAttribute attribute) throws MarshallingException {
        Object objectValue = eObject.eGet(attribute);
        if (objectValue == null) {
            return null;
        }
        
        Class classZ = attribute.getEType().getInstanceClass();
        Object attributeValue = null;
        if (classZ == null) {
            // Instance class is null for dynamic EMF. Use EMF Factory to determine the string value
            attributeValue = getStringValueUsingFactory(eObject, attribute, objectValue);
        } else {
            Marshaller marshaller = MarshalFactory.eINSTANCE.getMarshaller(classZ);
            if (marshaller != null) {
                attributeValue = marshaller.marshalAttributeValue(objectValue);
            } else if (attribute.getEType().getName().equals(Object.class.getSimpleName())) {
                attributeValue = getStringValueUsingFactory(eObject, attribute, objectValue);
            } else {
                String message = createCannotMarshalAttributeMessage(eObject, attribute);
                throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_MARSHAL_LITERAL, message);
            }
        }
        return attributeValue;
    }
    
    private void marshalManyValuedAttribute(EObject eObject, EAttribute attribute,  
            JSONObject jsonObject) throws MarshallingException {
        
        final Object objectValue = eObject.eGet(attribute);
        if (objectValue == null) {
            addAttributeToJSONObject(jsonObject, attribute.getName(), null);
            return;
        }
        
        final Class classZ = attribute.getEType().getInstanceClass();
        final Marshaller marshaller = MarshalFactory.eINSTANCE.getMarshaller(classZ);
        
        if (objectValue instanceof List) {
            final JSONArray jsonArray = new JSONArray();
            for (final Object attributeValue : (List)objectValue) {
                if (classZ != Entry.class) {
                    Object value = null;
                    if (attributeValue != null)
                        value = marshaller.marshalAttributeValue(attributeValue);
                    jsonArray.add(value);
                } 
            }
            addAttributeToJSONObject(jsonObject, attribute.getName(), jsonArray);
        }
    }
    
    /**
     * Marshal referenced items or helpers
     */
    protected void marshalReferences(EObject eObject, EClass eClass, JSONObject jsonObject) 
            throws MarshallingException {
        final Iterator i = eClass.getEAllReferences().iterator();
        while (i.hasNext()) {
            final EReference reference = (EReference) i.next();
            if (MarshallerUtil.shouldBeMarshalled(eObject, reference)) {
                // the reference should be marshaled
                if(MarshallerUtil.shouldConsiderAsContainment(reference)) {
                    marshalContainmentReferences(eObject, reference, jsonObject);
                } else {
                    marshalAssociationReferences(eObject, reference, jsonObject);
                }
            }
        }
    }
    
    /**
     * Marshal Association references
     */
    private void marshalAssociationReferences(EObject eObject, EReference reference, 
            JSONObject jsonObject) {
        
        if (reference.isMany()) {
            marshalMultiValuedAssociationReferences(eObject, reference, jsonObject);
        } else {
            marshalSingleValuedAssociationReference(eObject, reference, jsonObject);
        }
    }
    
    /**
     * Marshal single valued association reference 
     */
    private void marshalSingleValuedAssociationReference(EObject eObject, 
        EReference reference, JSONObject jsonObject) {
        
        if (!marshalEClassifiersUsingHREFs(eObject, reference, jsonObject)) {
            final Object objectValue = eObject.eGet(reference);
            if (objectValue == null) {
                addAttributeToJSONObject(jsonObject, reference.getName(), null);
            } else if (objectValue instanceof ItemHandle) {
                final JSONObject referenceJSONObject = marshalItemHandleAssociationReference((ItemHandle)objectValue);
                addAttributeToJSONObject(jsonObject, reference.getName(), referenceJSONObject);
            }
        }
    }
    
    private boolean marshalEClassifiersUsingHREFs(EObject eObject, EReference reference, 
            JSONObject jsonObject) {
        
        if (reference.getEReferenceType().equals(EcorePackage.eINSTANCE.getEDataType()) ||
            reference.getEReferenceType().equals(EcorePackage.eINSTANCE.getEClass()) ||
            reference.getEReferenceType().equals(EcorePackage.eINSTANCE.getEEnum())) {
          
             // use hrefURI to reference the EClassifier
             final String hrefURI = createHREFs(eObject, reference);
             addAttributeToJSONObject(jsonObject, reference.getName(), hrefURI);
             return true;
        }
        return false;
    }
    
    private void marshalMultiValuedAssociationReferences(EObject eObject, 
        EReference reference, JSONObject jsonObject) {
        
        final List objectValues = (List)eObject.eGet(reference);
        
        if (objectValues == null) {
            addAttributeToJSONObject(jsonObject, reference.getName(), null);
            return;
        }
        
        final JSONArray jsonArray = new JSONArray();
        for (final Object objectValue : objectValues) {
            if (objectValue != null && objectValue instanceof ItemHandle) {
                final JSONObject referenceJSONObject = marshalItemHandleAssociationReference((ItemHandle)objectValue);
                jsonArray.add(referenceJSONObject);
            }
        }
        addAttributeToJSONObject(jsonObject, reference.getName(), jsonArray);
    }
    
    @SuppressWarnings("deprecation")
	protected JSONObject marshalItemHandleAssociationReference(ItemHandle itemHandle) {
            
        final JSONObject referenceJSONObject = new JSONObject();
        final UUID itemId = itemHandle.getItemId();
        final UUID stateId = itemHandle.getStateId();
        final String qualifiedNameOfInstance = getQualifiedEClassName((EObject)itemHandle);
        
        addAttributeToJSONObject(referenceJSONObject, QUALIFIED_CLASS_NAME, qualifiedNameOfInstance);

        addAttributeToJSONObject(referenceJSONObject, JSONConstants.TYPENAME, 
                ItemUtil.typeToName(itemHandle.getItemType()));

        if (itemId != null) {
            addAttributeToJSONObject(referenceJSONObject, ITEM_ID, itemId.getUuidValue()); 
        } 
        if (stateId != null) {
            addAttributeToJSONObject(referenceJSONObject, STATE_ID, stateId.getUuidValue());
        }
        
        return referenceJSONObject;
    }
        
    /**
     * Marshal the containment references to outputstream
     */
    private void marshalContainmentReferences(EObject eObject, EReference reference, 
        JSONObject jsonObject) throws MarshallingException {
        
        if (reference.isMany()) {
            marshalMultiValuedContainmentReferences(eObject, reference, jsonObject);
        } else {
            marshalSingleValuedContainmentReference(eObject, reference, jsonObject);
        }
    }
    
    /**
     * Marshal single valued containment reference
     */
    private void marshalSingleValuedContainmentReference(EObject eObject, 
        EReference reference, JSONObject jsonObject) throws MarshallingException {
        
        final Object objectValue = eObject.eGet(reference);
        if (objectValue == null) {
            addAttributeToJSONObject(jsonObject, getReferenceName(null, reference), null);
            return;
        }
        if (marshalEClassifiersUsingHREFs(eObject, reference, jsonObject)) {
            return;
        }
        final Object containedJSONObject = marshalObjectToJSON(objectValue); 
        addAttributeToJSONObject(jsonObject, getReferenceName((EObject)objectValue, reference), containedJSONObject);
    }
    
    /**
     * Marshal multi-valued containment references
     */
    private void marshalMultiValuedContainmentReferences(EObject eObject, 
        EReference reference, JSONObject jsonObject) throws MarshallingException {
        
        final List objectValues = (List)eObject.eGet(reference);
        if (objectValues == null) {
            addAttributeToJSONObject(jsonObject, reference.getName(), null);
            return;
        }
        final JSONArray jsonArray = new JSONArray();
        for (final Object object : objectValues) {
            final EObject eObjectValue = (EObject) object;
            final Object containedObject = marshalObjectToJSON(eObjectValue);
            jsonArray.add(containedObject);
        }
        addAttributeToJSONObject(jsonObject, reference.getName(), jsonArray);
    }
    
    /* (non-Javadoc)
     * @see com.ibm.team.repository.common.marshal.Marshaller#marshalObjectToString(java.lang.Object)
     */
    public String marshalObjectToString(Object object) throws MarshallingException {
        return marshalObjectToJSON(object).toString();
    }

    /* (non-Javadoc)
     * @see com.ibm.team.repository.common.marshal.Marshaller#demarshalInputStreamToObjectArray(java.io.InputStream)
     */
    public Object demarshalInputStreamToObject(InputStream inputStream) throws MarshallingException {
        
        final InputStreamReader reader = createInputStreamReader(inputStream, null);
        return demarshalInputStream(reader);
    }

    /* (non-Javadoc)
     * @see com.ibm.team.repository.common.internal.marshal.Marshaller#demarshalInputStreamToObject(java.io.InputStream, java.nio.charset.Charset)
     */
    public Object demarshalInputStreamToObject(InputStream inputStream, Charset charset)
            throws MarshallingException {
        final InputStreamReader reader = createInputStreamReader(inputStream, charset);
        return demarshalInputStream(reader);
    }

    /**
     * Creates an object instance from the referenced input stream.
     * 
     * @param reader The input reader. Must not be <code>null</code>.
     * 
     * @return The deserialized item. Never <code>null</code>.
     * 
     * @throws MarshallingException Thrown on any error.
     */
    private Object demarshalInputStream(InputStreamReader reader) throws MarshallingException {
        
        Object obj = null;
        final String JSON_OBJECT = "JSON Object"; //$NON-NLS-1$
        
        try {
            obj = JSONObject.parseAny(reader);
        } catch (final IOException e) {
            final String msg = NLS.bind(Messages.getServerString("EObjectJSONMarshaller.ParseError"), JSON_OBJECT, e.getMessage()); //$NON-NLS-1$
            throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL, msg, e);
        }
        
        if (obj instanceof JSONObject)
            return objToItem((JSONObject)obj);
        else if (obj instanceof JSONArray)
            return objsToItem((JSONArray)obj);
        else {
            final String msg = NLS.bind(Messages.getServerString("EObjectJSONMarshaller.UnsupportedTypeInStream"),  JSON_OBJECT, obj.getClass()); //$NON-NLS-1$
            throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL, msg);
        }
    }

    /**
     * Returns a collection of Items deserialized from the referenced JSON object
     * array.
     * 
     * @param objs The array of JSONObjects. Must not be <code>null</code>.
     * 
     * @return The collection of IItems. Never <code>null</code>.
     * 
     * @throws MarshallingException Thrown on demarshalling errors.
     */
    private Collection objsToItem(JSONArray objs) throws MarshallingException {
        
        final ArrayList retVal = new ArrayList(objs.size());
        for (final Object obj : objs) {
            if (!(obj instanceof JSONObject)) {
                final String OBJECT = "JSON object"; //$NON-NLS-1$
                final String ARRAY = "JSON array"; //$NON-NLS-1$
                final String msg = NLS.bind(Messages.getServerString("EObjectJSONMarshaller.UnsupportedTypeInArray"), OBJECT, ARRAY, obj.getClass()); //$NON-NLS-1$
                throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL, msg);
            }
            retVal.add(objToItem((JSONObject)obj));
        }
        
        return retVal;
    }
    
    /**
     * Converts the given JSON object to a Jazz item.
     * @param obj 
     *              The JSON object to convert. Must not be <code>null</code>, and
     *              must contain a reference to the item type ({@link JSONConstants#TYPENAME}).
     *              
     * @return The item instance. Never <code>null</code>.
     * 
     * @throws MarshallingException if the JSON object cannot be converted
     * to an item.
     */
    private EObject objToItem(JSONObject obj) throws MarshallingException {
        
        if (obj == null) {
            throw new IllegalArgumentException("Object must not be null"); //$NON-NLS-1$
        }
        return convertJSONToEObject(obj);
    }
    
    private EObject convertJSONToEObject(JSONObject obj) throws MarshallingException {
        final String qualifiedClassName = (String)obj.get(QUALIFIED_CLASS_NAME);
        final String namespaceURI = getNamespaceURI(qualifiedClassName);
        final String typeName = getTypeName(qualifiedClassName);
        
        final EPackage ePackage = EPackage.Registry.INSTANCE.getEPackage(namespaceURI);
        if (ePackage == null) {
        	throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL, NLS.bind(Messages.getServerString("JSONDeserializer_03.ErrorMissingPackage"), namespaceURI)); //$NON-NLS-1$
        }
        
        final EClass eClass = (EClass)ePackage.getEClassifier(typeName);
        
        final EObject eObject = ePackage.getEFactoryInstance().create(eClass);
        
        Map<String, EStructuralFeature> attributeMappings = generateAttributeMappings(eObject);
        
        for (final Object keyObject : obj.keySet()) {
            final String key = (String) keyObject;
            if (key.equals(JSONConstants.TYPENAME) || key.equals(QUALIFIED_CLASS_NAME)) {
                continue;
            }
            fillEObjectFromJSON(eObject, attributeMappings.get(key), obj.get(key));
        }
        
        return eObject;
    }
    
    
    /** JSONified Attributes aren't necessarily given their attribute name as 
     * an identifier. Finds the attribute name used on serialization.
     */
    private Map<String, EStructuralFeature> generateAttributeMappings(
			EObject object) {
    	HashMap<String, EStructuralFeature> result = new HashMap<String, EStructuralFeature>();
    	
    	for (EStructuralFeature f : object.eClass().getEAllStructuralFeatures()) {
    		result.put(getReferenceName(object, f), f);
    	}
    	
    	return result;
	}

	private String getNamespaceURI(String qualifiedClassName) {
        final int index = qualifiedClassName.lastIndexOf(":"); //$NON-NLS-1$
        return qualifiedClassName.substring(0, index);
    }
    
    private String getTypeName(String qualifiedClassName) {
        final int index = qualifiedClassName.lastIndexOf(":"); //$NON-NLS-1$
        return qualifiedClassName.substring(index + 1);
    }
    
    /**
     * Set the attribute/reference specified by <code>attrName</code> on the referenced
     * <code>item</code>.  If the attribute identified by attrName is not a feature
     * of the item, this method has no effect on item.
     * 
     * @param item The item to set the attribute on. Must not be <code>null</code>.
     * 
     * @param structuralFeature The name of the attribute to set. Must not be <code>null</code>.
     * 
     * @param attrValue The value of the attribute to set. May be <code>null</code>.
     * 
     * @throws MarshallingException Thrown if there is an error converting the
     * JSON string to an attribute or reference.
     */
    private void fillEObjectFromJSON(EObject eObject, EStructuralFeature feature, Object attrValue) throws MarshallingException {
        
        if (feature == null)
            throw new IllegalArgumentException("Attribute name must not be null"); //$NON-NLS-1$
        
        // Don't demarshal the immutable flag if present; will cause feature loading failures later.
        if (isImmutableFeature(feature)) {
            return;
        }
        
        if (attrValue == null) {
            eObject.eSet(feature, null);
            return;
        }
        
        if (feature instanceof EAttribute) {
            if (feature.isMany()) {
                // TODO: JSONArray is an ArrayList. Why are we reading all the
                // elements and adding it to another ArrayList.
				final JSONArray jArray = (JSONArray) attrValue;
                final ArrayList list = new ArrayList(jArray.size());
				for (int i = 0; i < jArray.size(); i++) {
					Object obj = determineValue(feature, jArray.get(i));
                    list.add(obj);
                }
                eObject.eSet(feature, list);
            } else {
                final EDataType dataType = ((EAttribute)feature).getEAttributeType();
                final EFactory eFactory = dataType.getEPackage().getEFactoryInstance();
                // TODO: Would it work for Timestamp. we might need to get the appropriate
                // marshaller and get the demarshalled value
                final Object obj = eFactory.createFromString(dataType, attrValue.toString());
                eObject.eSet(feature, obj);
            }
        } else if (feature instanceof EReference) {
            if (feature.isMany()) {
                final JSONArray jArray = (JSONArray)attrValue;
                final ArrayList list = new ArrayList(jArray.size());
                for (int i = 0; i < jArray.size(); i++) {
                    final JSONObject obj = (JSONObject)jArray.get(i);
                    final EObject ref = convertJSONToEObject(obj);
                    list.add(ref);
                }
                eObject.eSet(feature, list);
            } else {
                final EObject ref = convertJSONToEObject((JSONObject)attrValue);
                eObject.eSet(feature, ref);
            }
        } else {
            final String msg = NLS.bind(Messages.getServerString("EObjectJSONMarshaller.UnexpectedFeatureType"), feature.getName()); //$NON-NLS-1$
            throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL, msg);
        }
    }

	/**
	 * Determines the correct value from a JSON value to assign to the EMF
	 * feature.
	 * 
	 * 
	 * @param feature
	 * @param object
	 * @return
	 */
	private Object determineValue(EStructuralFeature feature, Object value) {
		// Fix for WI 117935. Primitive data types from JSON object causes array
		// store
		// exception
		// if the target type is not assignable from the source. This attempts
		// to convert
		// the value into the correct type to be assigned.
		Class targetClass = feature.getEType().getInstanceClass();
		Class sourceClass = value.getClass();
		// If the target/source do not match , we have to convert it to the
		// target, or try at least
		if (!targetClass.isAssignableFrom(sourceClass)) {
			if (targetClass.equals(int.class) || targetClass.equals(Integer.class)) {
				value = NumberHelper.getInteger(value);
			} else if (targetClass.equals(long.class) || targetClass.equals(Long.class)) {
				value = NumberHelper.getLong(value);
			} else if (targetClass.equals(short.class) || targetClass.equals(Short.class)) {
				value = NumberHelper.getShort(value);
			} else if (targetClass.equals(float.class) || targetClass.equals(Float.class)) {
				value = NumberHelper.getFloat(value);
			} else if (targetClass.equals(double.class) || targetClass.equals(Double.class)) {
				value = NumberHelper.getDouble(value);
			} else if (targetClass.equals(byte.class) || targetClass.equals(Byte.class)) {
				value = NumberHelper.getByte(value);
			}
		}
		return value;
	}



	private boolean isImmutableFeature(EStructuralFeature feature) {
        return feature.equals(RepositoryPackage.eINSTANCE.getItemHandle_Immutable());
    }
    
    /**
     * Returns an {@link InputStreamReader} for the referenced inputStream,
     * optionally set to the specified charset.
     * 
     * @param inputStream The input stream. Must not be <code>null</code>.
     * 
     * @param charset The charset encoding. May be <code>null</code>. If null
     * the reader uses the default charset encoding.
     * 
     * @return The created reader. Never <code>null</code>.
     */
    private InputStreamReader createInputStreamReader(InputStream inputStream, Charset charset) {
        if (inputStream == null) {
            throw new IllegalArgumentException("inputStream must not be null"); //$NON-NLS-1$
        }
        
        InputStreamReader reader = null;
        if (charset != null) {
            reader = new InputStreamReader(inputStream, charset);
        } else {
            reader = new InputStreamReader(inputStream);
        }
        
        return reader;
    }
    
    /**
     * Add the attribute key/value pair to JSON Object
     */
    private void addAttributeToJSONObject(JSONObject jsonObject, String key, Object value) {
        jsonObject.put(key, value);
    }
}