/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010, 2015. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/

package com.ibm.team.repository.common.internal.marshal.impl;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.io.Writer;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.EcorePackage;
import org.eclipse.emf.ecore.util.FeatureMap.Entry;
import org.eclipse.emf.ecore.xmi.XMLParserPool;
import org.eclipse.emf.ecore.xmi.impl.XMLParserPoolImpl;

import com.ibm.team.repository.common.UUID;
import com.ibm.team.repository.common.internal.marshal.MarshalFactory;
import com.ibm.team.repository.common.internal.marshal.Marshaller;
import com.ibm.team.repository.common.internal.marshal.MarshallerType;
import com.ibm.team.repository.common.internal.marshal.MarshallingException;
import com.ibm.team.repository.common.internal.marshal.MarshallingExceptionReasonCode;
import com.ibm.team.repository.common.internal.marshal.util.MarshallerUtil;
import com.ibm.team.repository.common.internal.marshal.util.XMLConstructionUtil;
import com.ibm.team.repository.common.internal.marshal.util.XMLNamespaceUtil;
import com.ibm.team.repository.common.internal.util.ComponentConfigurationRegistry;
import com.ibm.team.repository.common.internal.util.ComponentRegistry;
import com.ibm.team.repository.common.internal.util.IComponentElementDescriptor;
import com.ibm.team.repository.common.internal.util.ItemUtil;
import com.ibm.team.repository.common.model.ItemHandle;
import com.ibm.team.repository.common.model.ItemType;
import com.ibm.team.repository.common.transport.HttpUtil.CharsetEncoding;

/**
 * Class to serialize an EObject to a XML File
 */
public class EObjectMarshaller extends AbstractEObjectMarshaller {

    public static final String UTF8_ENCODING = CharsetEncoding.UTF8.toString();
    
    protected static final int DEMARSHAL_CHUNK_SIZE = 2048;
    
    protected static final EClass ECLASS_TYPE = EcorePackage.eINSTANCE.getEClass();
    protected static final EClass EDATA_TYPE = EcorePackage.eINSTANCE.getEDataType();
    protected static final EClass EENUM_TYPE = EcorePackage.eINSTANCE.getEEnum(); 
    
    // Option to represent attributes as XML Element. The default value is Boolean.TRUE
    public static final String OPTION_USE_XML_ELEMENT_TO_REPRESENT_ATTRIBUTE = "USE_XML_ELEMENT_TO_REPRESENT_ATTRIBUTE"; //$NON-NLS-1$
    
    protected boolean useXmlElementToRepresentAttribute = true;
    
    // though the comments in the EMF 2.1 impl of XMLParserPoolImpl claim it is not
    // thread safe, it is.  See newsgroup comment:
    // http://www.eclipse.org/newsportal/article.php?id=14887&group=eclipse.tools.emf#14887
    static protected final XMLParserPool xmlParserPool = new XMLParserPoolImpl();
    
    private static final Map parserFeatures = new HashMap(11);
    static {
    	// When validation is on, DOCTYPES and external DTDS are always processed.
    	// So schema validation is also turned off while parsing the request.
        parserFeatures.put("http://xml.org/sax/features/validation", Boolean.FALSE); //$NON-NLS-1$
        parserFeatures.put("http://apache.org/xml/features/validation/schema", Boolean.FALSE); //$NON-NLS-1$
        parserFeatures.put("http://apache.org/xml/features/disallow-doctype-decl", Boolean.TRUE); //$NON-NLS-1$
        parserFeatures.put("http://apache.org/xml/features/nonvalidating/load-dtd-grammar", Boolean.FALSE); //$NON-NLS-1$
        parserFeatures.put("http://apache.org/xml/features/nonvalidating/load-external-dtd", Boolean.FALSE); //$NON-NLS-1$
    }
    
    public EObjectMarshaller() {
        super(EObject.class, MarshallerType.EOBJECT_LITERAL);
    }
    
    public EObjectMarshaller(Class clazz, MarshallerType marshallerType) {
        super(clazz, marshallerType);
    }
    
    /**
     * Create a EObjectMarshaller with the passed in option map
     * @param options 
     */
    public EObjectMarshaller(Map options) {
        super(EObject.class, MarshallerType.EOBJECT_LITERAL, options);
        if (Boolean.FALSE.equals(options.get(OPTION_USE_XML_ELEMENT_TO_REPRESENT_ATTRIBUTE))) {
            useXmlElementToRepresentAttribute = false;
        }
    }
    
    public EObjectMarshaller(Class clazz, MarshallerType marshallerType, Map options) {
        super(clazz, marshallerType, options);
        if (Boolean.FALSE.equals(options.get(OPTION_USE_XML_ELEMENT_TO_REPRESENT_ATTRIBUTE))) {
            useXmlElementToRepresentAttribute = false;
        }
    }
        
    /**
     * Marshals the given object to the specified output stream.
     * @param object 
     *              The object to serialize. Must not be <code>null</code>.
     * @param outputStream 
     *              The stream to write to. Must not be <code>null</code>.
     * @throws MarshallingException if there are problems marshalling the given
     * object
     * @deprecated use {@link #marshalObjectToOutputStream(Object, OutputStream, Charset)}
     */
    @Deprecated
    public void marshalObjectToOutputStream(Object object, OutputStream outputStream) throws MarshallingException {
        marshalObjectToOutputStream(object, outputStream, CharsetEncoding.UTF8.toCharset());
    }
    
    /**
     * Marshals the given object to the specified output stream.
     * @param object 
     *              The object to serialize. Must not be <code>null</code>.
     * @param outputStream 
     *              The stream to write to. Must not be <code>null</code>.
     * @param charset 
     *              The charset to encode the response as. May be <code>null</code>.
     *              If <code>null</code>, the default char set is used to marshal
     *              the given object
     * @throws MarshallingException if there are problems marshalling the given
     * object
     */
	public void marshalObjectToOutputStream(Object object, OutputStream outputStream,
			Charset charset) throws MarshallingException {
		marshalObjectToOutputStream(object, outputStream, charset, null);
	}

    public void marshalObjectToOutputStream(Object object, OutputStream outputStream,
    		Charset charset, Map<String, String> clientConfig) throws MarshallingException {
        if (charset == null)
            charset = CharsetEncoding.UTF8.toCharset();
        
        try {
            Writer writer = XMLConstructionUtil.createWriter(outputStream, charset);
            marshalObjectToOutputStreamUsingWriter(object, writer, charset, clientConfig);
            writer.flush();
        } catch (UnsupportedEncodingException unsupportedEncodingException) {
            throw new MarshallingException(unsupportedEncodingException);
        } catch (IOException e) {
            throw new MarshallingException(e);
        }
    }
    
    /**
     * Marshal the object to output stream
     */
    protected void marshalObjectToOutputStreamUsingWriter(Object object, 
            Writer writer, Charset charset, Map<String, String> clientConfig) throws MarshallingException {
        
        int depth = 0;
        EObject eObject = (EObject)object;
        EClass eClass = eObject.eClass();
        
        EPackage ePackage = eClass.getEPackage();
        String qualifiedClassName = getQualifiedEClassName(eClass, ePackage.getNsPrefix());
        
        // Write XML Header and the Class name
        writeXMLHeader(writer, charset);
        
        marshalContainedObjectToOutputStreamUsingWriter(eObject, qualifiedClassName, 
                writer, null, null, clientConfig, depth);
    }
    
    protected void writeXMLHeader(Writer writer, Charset charset) throws MarshallingException {
        XMLConstructionUtil.writeXMLHeader(writer, charset);   
        XMLConstructionUtil.writeNewLineCharacter(writer);
    }
    
    /**
     * Marshal the contained objects to output stream. 
     */
    protected void marshalContainedObjectToOutputStreamUsingWriter(EObject eObject,
            String elementName, Writer writer, String containingReferenceType,
            Map nsURIToNsPrefixMap, Map<String, String> clientConfig, int depth)
    		throws MarshallingException { 
        
        EClass eClass = eObject.eClass();
        
        // Write Element Name
        XMLConstructionUtil.writeElementName(writer, elementName);
        
        if (depth == 0) {
            // Write namespace information for top level object
            Map nsPrefixToNsURIMap = XMLNamespaceUtil.getReferencedNamespaces(eObject);
            XMLConstructionUtil.writeNamespaceInformation(writer, nsPrefixToNsURIMap);
            nsURIToNsPrefixMap = XMLNamespaceUtil.constructNsURIToNsPrefixMap(nsPrefixToNsURIMap);
        } 
        
        // Add object type attribute if necessary
        addTypeAttribute(eObject, containingReferenceType, writer, nsURIToNsPrefixMap, depth);
       
        // Marshal primitive types as XML attributes 
        serializeEAttributesAsXMLAttributes(eObject, eClass, writer, clientConfig);
        marshalEmptyMultiValuedReferencesAsAttributes(eObject, eClass, writer, clientConfig);
        XMLConstructionUtil.writeEndElementTag(writer);
        
        // Marshal Primitive types as XML elements
        serializeEAttributesAsXMLElements(eObject, eClass, writer, clientConfig, depth + 1);  
        
        // Marshal references (containment and aggregation references)
        marshalReferences(eObject, eClass, writer, nsURIToNsPrefixMap, clientConfig, depth + 1);
        
        // Write End Element tag
        XMLConstructionUtil.writeEndElementName(writer, elementName, depth);
    }
    
    protected void addTypeAttribute(EObject eObject, String containingReferenceType,
            Writer writer, Map<String, String> nsUriToNsPrefix, int depth) throws MarshallingException {
        if (depth > 0) {
            String eObjectReferenceType = getQualifiedEClassName(eObject, nsUriToNsPrefix);
            if (saveTypeInformation || !eObjectReferenceType.equals(containingReferenceType)) { 
                XMLConstructionUtil.writeAttribute(writer, "xsi:type", eObjectReferenceType); //$NON-NLS-1$
            }
        }
    }
    
    /**
     * Serialize the Ettributes as XML attributes
     */
    protected void serializeEAttributesAsXMLAttributes(EObject eObject, EClass eClass, 
            Writer writer, Map<String, String> clientConfig) throws MarshallingException {
        for (EAttribute attribute : eClass.getEAllAttributes()) {
            if (shouldBeMarshalled(eObject, attribute, clientConfig)) {
                if (shouldMarshalAsXMLAttribute(attribute)) {
                   serializeEAttributeAsXMLAtribute(eObject, attribute, writer); 
                }
            }
        }
    }
    
    protected void serializeEAttributeAsXMLAtribute(EObject eObject, EAttribute attribute, 
            Writer writer) throws MarshallingException {
        String marshalledValue = serializeEAttributeValue(eObject, attribute);
        if (marshalledValue != null) {
            XMLConstructionUtil.writeAttribute(writer, attribute.getName(), marshalledValue);
        }
    }
    
    @Override
    protected boolean shouldMarshalAsXMLAttribute(EAttribute attribute) {
        if (!useXmlElementToRepresentAttribute) { 
            return true;
        }
        return super.shouldMarshalAsXMLAttribute(attribute);
    }
    
    /**
     * Marshal Empty Reference Lists as attributes
     */
    protected void marshalEmptyMultiValuedReferencesAsAttributes(EObject eObject, EClass eClass, 
            Writer writer, Map<String, String> clientConfig) throws MarshallingException {
        
        for (EReference reference : eClass.getEAllReferences()) {
            if (shouldBeMarshalled(eObject, reference, clientConfig)) {
                if (isEmptyListReference(eObject, reference)) {
                    XMLConstructionUtil.writeEmptyAttribute(reference.getName(), writer);
                }
            }
        }
    }
    
    private boolean isEmptyListReference(EObject eObject, EReference reference) {
        if (!reference.isMany()) {
            return false;
        }
        List objectValues = (List)eObject.eGet(reference);
        return objectValues != null && objectValues.size() == 0;
    }
    
    protected void serializeEAttributesAsXMLElements(EObject eObject, EClass eClass, 
            Writer writer, Map<String, String> clientConfig, int depth) throws MarshallingException {
        
        for (EAttribute attribute : eClass.getEAllAttributes()) {
            if (shouldBeMarshalled(eObject, attribute, clientConfig)) {
                if (!shouldMarshalAsXMLAttribute(attribute)) {
                    if (attribute.isMany()) {
                        serializeManyValuedEAttributesAsXMLElements(eObject, attribute, writer, depth);
                    } else {
                        String serializedValue = serializeEAttributeValue(eObject, attribute);
                        serializeAsXMLElement(writer, attribute, serializedValue, depth);
                    }
                }
            }
        }
    }
    
    private void serializeAsXMLElement(Writer writer, EAttribute attribute, 
            String serializedValue, int depth) throws MarshallingException {
        if (serializedValue == null) {
        	marshalNullElement(attribute.getName(), writer, depth);
        } else {
            XMLConstructionUtil.writeElement(writer, attribute.getName(), serializedValue, depth);
        }
    }
    
    protected void serializeManyValuedEAttributesAsXMLElements(EObject eObject, EAttribute attribute,
            Writer writer, int depth) throws MarshallingException {
        
        Object objectValue = eObject.eGet(attribute);
        if (objectValue == null) {
        	marshalNullElement(attribute.getName(), writer, depth);
            return;
        }
        
        Class classZ = attribute.getEType().getInstanceClass();
        Marshaller marshaller = MarshalFactory.eINSTANCE.getMarshaller(classZ);
        
        if (objectValue instanceof List) {
            List objectValueList = (List)objectValue;
            for (Object attributeValue : objectValueList) {
                if (classZ != Entry.class ) {
                	String attributeValueAsString;
                	if (attributeValue != null) {
                		attributeValueAsString = marshaller.marshalObjectToString(attributeValue);
                	} else {
                		attributeValueAsString = null;
                	}
                
                    serializeAsXMLElement(writer, attribute, attributeValueAsString, depth);
                }
            }
        }
    }
    
    protected void marshalReferences(EObject eObject, EClass eClass, Writer writer, 
            Map nsUriToNsPrefixMap, Map<String, String> clientConfig, int depth)
    		throws MarshallingException {
        
        for (EReference reference : eClass.getEAllReferences()) {
            if (shouldBeMarshalled(eObject, reference, clientConfig)) {
                if (reference.isMany()) {
                    marshalMultiValuedReference(eObject, reference, writer, nsUriToNsPrefixMap, clientConfig, depth);
                } else {
                    marshalSingleValuedReference(eObject, reference, writer, nsUriToNsPrefixMap, clientConfig, depth);
                }
            }
        }
    }

	protected boolean shouldBeMarshalled(EObject object, EStructuralFeature structuralFeature,
			Map<String, String> clientConfig) {
		if (shouldBeMarshalled(object, structuralFeature) == false)
			return false;
		String nsURI = structuralFeature.getEContainingClass().getEPackage().getNsURI();
		IComponentElementDescriptor desc = getComponentDescriptor(nsURI);
		String clientVersion = getClientVersionForComponentIfDifferent(desc, clientConfig);
		if (clientVersion == null)
			return true;
		else
			/* We could check whether this feature has had a string size change
			 * between the client and server version here, and explicitly fail
			 * if it does.  But I don't think it is worth the peformance hit.
			 * Hopefully we can trust component owners to not violate this rule. */
			return desc.getModelElementDescriptor(nsURI).attributeExistedInVersion(
					structuralFeature.getEContainingClass().getInstanceClassName(),
					structuralFeature.getName(), clientVersion);
	}

	protected IComponentElementDescriptor getComponentDescriptor(String nsURI) {
		if (ComponentRegistry.INSTANCE.isValidComponent(nsURI) == true)
			return ComponentRegistry.INSTANCE.getComponentDescriptorForModelUri(nsURI);
		if (ComponentRegistry.INSTANCE.getComponentDescriptorForDtoModelUri(nsURI) != null)
			return ComponentRegistry.INSTANCE.getComponentDescriptorForDtoModelUri(nsURI);
		return null;
	}

	// Returns null if the client version is the same as the server version for
	// the given component description.  Otherwise, returns a string representing
	// the client version.
	protected String getClientVersionForComponentIfDifferent(
			IComponentElementDescriptor desc, Map<String, String> clientConfig) {
		if (clientConfig == null)
			return null;
		if (desc == null)
			return null;
		String componentId = desc.getComponentConfigurationAttribute();
		if (componentId == null)
			return null;
		String serverVersion = ComponentConfigurationRegistry.INSTANCE.getComponentConfigurationsVersionMap().
			get(componentId);
		String clientVersion = clientConfig.get(componentId);
		if ((serverVersion == null) || (clientVersion == null))
			return null;
		if (clientVersion.equals(serverVersion))
			return null;
		return clientVersion;
	}

	/**
     * Marshal single valued reference
     */
    protected void marshalSingleValuedReference(EObject eObject, EReference reference, 
        Writer writer, Map nsUriToNsPrefixMap, Map<String, String> clientConfig, int depth) throws MarshallingException {
        
        if (!marshalEClassifiersUsingHREFs(eObject, reference, writer, depth)) {
            Object objectValue = eObject.eGet(reference);
            if (objectValue == null) {
                // the reference is set but the value is null. Add xsi:nil attribute
                marshalNullElement(reference.getName(), writer, depth);
            } else if (MarshallerUtil.shouldConsiderAsContainment(reference)) {
                // Marshal single valued containment reference
                marshalContainedReference(reference, objectValue, writer, nsUriToNsPrefixMap, clientConfig, depth);
            } else if (objectValue instanceof ItemHandle) {
                // marshal single valued shared reference
                marshalItemHandleAssociationReference(reference, (ItemHandle)objectValue, writer, nsUriToNsPrefixMap, depth);
            }
        }
    }

    /**
     * Marshal EClassifiers using HREFs
     */
    protected boolean marshalEClassifiersUsingHREFs(EObject eObject, EReference reference, 
            Writer writer, int depth) throws MarshallingException {
        
        if (reference.getEReferenceType().equals(EDATA_TYPE) || reference.getEReferenceType().equals(ECLASS_TYPE) ||
                reference.getEReferenceType().equals(EENUM_TYPE)) {
             // use hrefURI to reference the EClassifier
             String hrefURI = createHREFs(eObject, reference);
             XMLConstructionUtil.writeDataType(writer, reference.getName(), hrefURI, depth);
             return true;
        }
        return false;
    }
    
    /**
     * Marshal an ItemHandle association reference
     */
    protected void marshalItemHandleAssociationReference(EReference reference,    
        ItemHandle itemHandle, Writer writer, Map nsUriToNsPrefixMap, int depth) throws MarshallingException {
        
        XMLConstructionUtil.writeNewLineWithTabCharacters(writer, depth);
        UUID itemId = itemHandle.getItemId();
        UUID stateId = itemHandle.getStateId();
        String qualifiedNameOfInstance = getQualifiedEClassName(
                ItemUtil.getHandleEClass((ItemType)itemHandle.getItemType()), nsUriToNsPrefixMap);
        String qualifiedNameOfReferenceType = getQualifiedEClassName(ItemUtil.getTargetClassIfFacade(reference), nsUriToNsPrefixMap);
        
        if (saveTypeInformation) {
            // save xsi:type information for all references
            XMLConstructionUtil.writeReference(writer, reference.getName(),
                itemId != null? itemId.getUuidValue() : null, 
                stateId != null? stateId.getUuidValue() : null, qualifiedNameOfInstance);
        } else {
            // save xsi:type information if instance class and declared reference type are different
            XMLConstructionUtil.writeReference(writer, reference.getName(),
                itemId != null? itemId.getUuidValue() : null, 
                stateId != null? stateId.getUuidValue() : null,
                qualifiedNameOfInstance.equals(qualifiedNameOfReferenceType) ? null : qualifiedNameOfInstance);
        }
    }
    
    /**
     * Marshal Contained Reference to XML
     */
    private void marshalContainedReference(EReference reference, Object objectValue, 
            Writer writer, Map nsUriToNsPrefixMap, Map<String, String> clientConfig,
            int depth) throws MarshallingException {
        
        String referenceTypeName = getQualifiedEClassName(ItemUtil.getTargetClassIfFacade(
                reference), nsUriToNsPrefixMap);
        
        XMLConstructionUtil.writeNewLineWithTabCharacters(writer, depth);
        marshalContainedObjectToOutputStreamUsingWriter((EObject)objectValue, 
                getReferenceName((EObject)objectValue, reference), 
                writer, referenceTypeName, nsUriToNsPrefixMap, clientConfig, depth);
    }
    
    protected void marshalMultiValuedReference(EObject eObject, EReference reference, 
        Writer writer, Map nsUriToNsPrefixMap, Map<String, String> clientConfig, int depth) throws MarshallingException {
        
        List objectValues = (List)eObject.eGet(reference);
        if(objectValues != null && objectValues.size() != 0) {
            if (MarshallerUtil.shouldConsiderAsContainment(reference)) {
                // marshal contained objects
                String referenceTypeName = getQualifiedEClassName(ItemUtil.getTargetClassIfFacade(
                        reference), nsUriToNsPrefixMap);
                Iterator i = objectValues.iterator();
                while (i.hasNext()) {
                    EObject objectValue = (EObject)i.next();
                    XMLConstructionUtil.writeNewLineWithTabCharacters(writer, depth);
                    marshalContainedObjectToOutputStreamUsingWriter(objectValue, 
                        getReferenceName(objectValue, reference), writer, 
                        referenceTypeName, nsUriToNsPrefixMap, clientConfig, depth);
                } 
            } else { 
                XMLConstructionUtil.writeNewLineCharacter(writer);
                Iterator i = objectValues.iterator();
                while (i.hasNext()) {
                    Object objectValue = i.next();
                    if (objectValue != null && objectValue instanceof ItemHandle) {
                        marshalItemHandleAssociationReference(reference, (ItemHandle)objectValue, 
                                writer, nsUriToNsPrefixMap, depth);
                    }
                }
                XMLConstructionUtil.writeNewLineCharacter(writer);
            }
        }
    }
        
    private void marshalNullElement(String elementName, Writer writer, int depth) throws MarshallingException {
    	XMLConstructionUtil.writeNewLineWithTabCharacters(writer, depth);
        XMLConstructionUtil.writeNullElement(writer, elementName);
    }

    /* (non-Javadoc)
     * @see com.ibm.team.repository.common.marshal.Marshaller#marshalObjectToString(java.lang.Object)
     */
    public String marshalObjectToString(Object object) throws MarshallingException {
        ByteArrayOutputStream byteArrayOutputStream = new ByteArrayOutputStream();
        marshalObjectToOutputStream(object, byteArrayOutputStream);
        try {
            return byteArrayOutputStream.toString(UTF8_ENCODING);
        } catch (UnsupportedEncodingException e) {
            // UTF-8 encoding not supported, use the default 
            return byteArrayOutputStream.toString();
        }
    }

    /* (non-Javadoc)
     * @see com.ibm.team.repository.common.marshal.Marshaller#demarshalInputStreamToObjectArray(java.io.InputStream)
     */
	public Object demarshalInputStreamToObject(InputStream inputStream) throws MarshallingException {
        if (inputStream == null) {
            throw new IllegalArgumentException("inputStream must not be null"); //$NON-NLS-1$
        }
		//Read the input stream in chunks of DEMARSHAL_CHUNK_SIZE and then
		//combine the chunks into one byte[].
		ArrayList<byte[]> readChunks = new ArrayList<byte[]>();
		int totalBytesRead = 0;
		try {
			int lastRead = 0;
			int bytesRead = 0;
			while (lastRead != -1) {
				byte[] chunk = new byte[DEMARSHAL_CHUNK_SIZE];
				for (bytesRead = 0; bytesRead < DEMARSHAL_CHUNK_SIZE;) {
					lastRead = inputStream.read(chunk, bytesRead, DEMARSHAL_CHUNK_SIZE - bytesRead);
					if (lastRead == -1) {
						break;
					}

					bytesRead += lastRead;
				}

				readChunks.add(chunk);
				totalBytesRead += bytesRead;
			}
		} catch (IOException e) {
			throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL, e.getMessage(), e);
		}

		byte[] document = new byte[totalBytesRead];

		//Each chunk should be size DEMARSHAL_CHUNK_SIZE except for maybe the last one.
		int remainingBytes = totalBytesRead;
		for (int i = 0; i < readChunks.size(); i++) {
			System.arraycopy(readChunks.get(i), 0, document, i * DEMARSHAL_CHUNK_SIZE, remainingBytes > DEMARSHAL_CHUNK_SIZE ? DEMARSHAL_CHUNK_SIZE : remainingBytes);
			remainingBytes -= DEMARSHAL_CHUNK_SIZE;
		}

		return internalFastDemarshal(document);
	}
    
    /* (non-Javadoc)
     * @see com.ibm.team.repository.common.internal.marshal.Marshaller#demarshalInputStreamToObject(java.io.InputStream, java.nio.charset.Charset)
     */
	public Object demarshalInputStreamToObject(InputStream inputStream, Charset charset) throws MarshallingException {
        if (inputStream == null) {
            throw new IllegalArgumentException("inputStream must not be null"); //$NON-NLS-1$
        }
		if (charset == null || charset.equals(CharsetEncoding.UTF8.toCharset()) || charset.equals(Charset.forName("US_ASCII"))) //$NON-NLS-1$
			return demarshalInputStreamToObject(inputStream);

		InputStreamReader reader = new InputStreamReader(inputStream, charset);
		StringBuilder buffer = new StringBuilder(DEMARSHAL_CHUNK_SIZE);
		try {
			char[] chunk = new char[DEMARSHAL_CHUNK_SIZE];
			int charsRead = 0;
			while (charsRead != -1) {
				charsRead = reader.read(chunk, 0, DEMARSHAL_CHUNK_SIZE);
				if (charsRead != -1) {
					buffer.append(chunk, 0, charsRead);
				}
			}
		} catch (IOException e) {
			throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL, e.getMessage(), e);
		}

		try {
			byte[] document = buffer.toString().getBytes(UTF8_ENCODING);
			return internalFastDemarshal(document);
		} catch (UnsupportedEncodingException e) {
			throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL, e.getMessage(), e);
		}
	}
    
    protected Object internalFastDemarshal(byte[] bytes) throws MarshallingException {
    	//FastWebDemarshaller is a generated class, see FastDemarshaller.ppjava in the same package.
    	FastWebDemarshaller demarshaller = new FastWebDemarshaller(bytes);
    	return demarshaller.demarshal();
    }
}