/*******************************************************************************
 * 
 * Licensed Materials - Property of IBM
 * 
 * © Copyright IBM Corporation 2013, 2014. All Rights Reserved.
 * 
 * 
 * 
 * Note to U.S. Government Users Restricted Rights:
 * 
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.repository.common.internal.marshal.impl;

import java.io.UnsupportedEncodingException;
import java.math.BigDecimal;
import java.sql.Timestamp;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.eclipse.emf.common.util.AbstractEList;
import org.eclipse.emf.common.util.BasicEMap.Entry;
import org.eclipse.emf.common.util.Enumerator;
import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EDataType;
import org.eclipse.emf.ecore.EFactory;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.util.EcoreEMap;

import com.ibm.team.repository.common.UUID;
import com.ibm.team.repository.common.internal.marshal.MarshallingException;
import com.ibm.team.repository.common.internal.marshal.MarshallingExceptionReasonCode;
import com.ibm.team.repository.common.internal.marshal.util.MarshallerUtil;
import com.ibm.team.repository.common.internal.util.ItemUtil;
import com.ibm.team.repository.common.internal.util.SafeSimpleDataFormat;
import com.ibm.team.repository.common.model.ItemHandle;
import com.ibm.team.repository.common.model.RepositoryPackage;
import com.ibm.team.repository.common.serialize.internal.nls.Messages;
import com.ibm.team.repository.common.util.NLS;
import com.ibm.team.repository.common.utils.DateUtils;

/**
 * 
 * FastDemarshaller is a fast xml parser/demarshaller which is written using
 * 
 * C/C++ preprocessor macros to inline critical functions.
 * 
 * 
 * 
 * The preprocess.xml ant script should be used to generated the
 * FastDocumentDemarshaller
 * 
 * and FastWebDemarshaller java classes. These two classes differ in how they
 * expect
 * 
 * timestamps to be stored (as long or using the RFC3339 format).
 * 
 * 
 * 
 * Changes should be made in FastDemarshaller.ppjava and the preprocess.xml
 * script should
 * 
 * be used to regenerate the FastDocumentMarshaller and FastWebMarshaller. For
 * readability,
 * 
 * the two generated java classes should be formatted before committing changes.
 * 
 * The formatter to use is the one we use for generated model code (Repository
 * Model Generation).
 */
public class FastWebDemarshaller {
	private static Log _logger = LogFactory.getLog(FastWebDemarshaller.class);
	// Tolerance is enabled by default
	private static boolean isNamespaceToleranceEnabled = true;
	private static Set<String> corruptedNamespaces = new HashSet<String>();
	protected static final String EXTENDED_METADATA_ANNOTATION = "http:///org/eclipse/emf/ecore/util/ExtendedMetaData"; //$NON-NLS-1$
	protected static final String NAME_ANNOTATION_KEY = "name"; //$NON-NLS-1$
	static final EAttribute ITEM_ID_ATTR = RepositoryPackage.eINSTANCE.getItemHandle_ItemId();
	int wpos;
	final char[] buf;
	final SimpleDateFormat df = SafeSimpleDataFormat.DATE_FORMATS.get(0);
	final byte[] bytes;
	final int end1;
	int pos1;
	boolean foundClosingTag;
	boolean foundStartOfTag;
	boolean foundEndOfTag;
	boolean isHeader;
	boolean isEnd;
	String currentTag;
	final ArrayList<EObject> objectStack = new ArrayList<EObject>();
	final HashMap<String, EFactory> factories = new HashMap<String, EFactory>();

	public FastWebDemarshaller(byte[] bytes) {
		this.bytes = bytes;
		this.end1 = bytes.length;
		this.buf = new char[bytes.length];
	}

	String readTag() throws MarshallingException {
		foundEndOfTag = false;
		int myPos = this.pos1;
		final int myEnd = this.end1;
		final byte[] myBytes = this.bytes;
		if (foundClosingTag) {
			int c;
			{
				do {
					{
						if (((myPos)) == ((myEnd))) {
							throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
						}
						{
							(((c))) = (((myBytes)))[(((myPos)))++];
							if (((((c))) & 0x80) != 0) {
								int myPrivateRemaining;
								int checkMask;
								if (((((c))) & 0xE0) == 0xC0) {
									myPrivateRemaining = 1;
									(((c))) = (((c))) & 0x1F;
									checkMask = 0xFFFFFF80;
								} else if ((((c)) & 0xF0) == 0xE0) {
									myPrivateRemaining = 2;
									(((c))) = (((c))) & 0x0F;
									checkMask = 0xFFFFF800;
								} else if (((((c))) & 0xF8) == 0xF0) {
									myPrivateRemaining = 3;
									(((c))) = (((c))) & 0x07;
									checkMask = 0xFFFF0000;
								} else if (((((c))) & 0xFC) == 0xF8) {
									myPrivateRemaining = 4;
									(((c))) = (((c))) & 0x03;
									checkMask = 0xFFE00000;
								} else if (((((c))) & 0xFE) == 0xFC) {
									myPrivateRemaining = 5;
									(((c))) = (((c))) & 0x01;
									checkMask = 0xFC000000;
								} else {
									throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
								}
								int myPrivateNext;
								do {
									myPrivateNext = (((myBytes)))[(((myPos)))++];
									if ((myPrivateNext & 0xC0) != 0x80) {
										throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
									}
									(((c))) = ((((c))) << 6) | (myPrivateNext & 0x3F);
									myPrivateRemaining--;
								} while (myPrivateRemaining != 0);
								if (((((c))) & checkMask) == 0) {
									throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
								}
							}
						}
						;
					}
					;
				} while ((c) == ' ' || (c) == '\t' || (c) == '\r' || (c) == '\n');
			}
			;
			if (c != '>') {
				throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
			}
			isEnd = true;
			String tag = currentTag;
			currentTag = null;
			foundClosingTag = false;
			this.pos1 = myPos;
			return tag;
		}
		if (!foundStartOfTag) {
			int c;
			{
				do {
					if ((myPos) == (myEnd)) {
						(c) = -1;
						break;
					}
					{
						((c)) = ((myBytes))[((myPos))++];
						if ((((c)) & 0x80) != 0) {
							int myPrivateRemaining;
							int checkMask;
							if ((((c)) & 0xE0) == 0xC0) {
								myPrivateRemaining = 1;
								((c)) = ((c)) & 0x1F;
								checkMask = 0xFFFFFF80;
							} else if (((c) & 0xF0) == 0xE0) {
								myPrivateRemaining = 2;
								((c)) = ((c)) & 0x0F;
								checkMask = 0xFFFFF800;
							} else if ((((c)) & 0xF8) == 0xF0) {
								myPrivateRemaining = 3;
								((c)) = ((c)) & 0x07;
								checkMask = 0xFFFF0000;
							} else if ((((c)) & 0xFC) == 0xF8) {
								myPrivateRemaining = 4;
								((c)) = ((c)) & 0x03;
								checkMask = 0xFFE00000;
							} else if ((((c)) & 0xFE) == 0xFC) {
								myPrivateRemaining = 5;
								((c)) = ((c)) & 0x01;
								checkMask = 0xFC000000;
							} else {
								throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
							}
							int myPrivateNext;
							do {
								myPrivateNext = ((myBytes))[((myPos))++];
								if ((myPrivateNext & 0xC0) != 0x80) {
									throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
								}
								((c)) = (((c)) << 6) | (myPrivateNext & 0x3F);
								myPrivateRemaining--;
							} while (myPrivateRemaining != 0);
							if ((((c)) & checkMask) == 0) {
								throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
							}
						}
					}
					;
				} while ((c) == ' ' || (c) == '\t' || (c) == '\r' || (c) == '\n');
			}
			;
			if (c == -1) {
				this.pos1 = myPos;
				return null;
			}
			if (c != '<') {
				throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
			}
		}
		foundStartOfTag = false;
		int c;
		{
			do {
				{
					if (((myPos)) == ((myEnd))) {
						throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
					}
					{
						(((c))) = (((myBytes)))[(((myPos)))++];
						if (((((c))) & 0x80) != 0) {
							int myPrivateRemaining;
							int checkMask;
							if (((((c))) & 0xE0) == 0xC0) {
								myPrivateRemaining = 1;
								(((c))) = (((c))) & 0x1F;
								checkMask = 0xFFFFFF80;
							} else if ((((c)) & 0xF0) == 0xE0) {
								myPrivateRemaining = 2;
								(((c))) = (((c))) & 0x0F;
								checkMask = 0xFFFFF800;
							} else if (((((c))) & 0xF8) == 0xF0) {
								myPrivateRemaining = 3;
								(((c))) = (((c))) & 0x07;
								checkMask = 0xFFFF0000;
							} else if (((((c))) & 0xFC) == 0xF8) {
								myPrivateRemaining = 4;
								(((c))) = (((c))) & 0x03;
								checkMask = 0xFFE00000;
							} else if (((((c))) & 0xFE) == 0xFC) {
								myPrivateRemaining = 5;
								(((c))) = (((c))) & 0x01;
								checkMask = 0xFC000000;
							} else {
								throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
							}
							int myPrivateNext;
							do {
								myPrivateNext = (((myBytes)))[(((myPos)))++];
								if ((myPrivateNext & 0xC0) != 0x80) {
									throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
								}
								(((c))) = ((((c))) << 6) | (myPrivateNext & 0x3F);
								myPrivateRemaining--;
							} while (myPrivateRemaining != 0);
							if (((((c))) & checkMask) == 0) {
								throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
							}
						}
					}
					;
				}
				;
			} while ((c) == ' ' || (c) == '\t' || (c) == '\r' || (c) == '\n');
		}
		;
		isEnd = false;
		isHeader = false;
		final char[] myBuf = this.buf;
		int mywPos = this.wpos;
		int start = mywPos;
		while (true) {
			if (c == '/') {
				if (foundClosingTag) {
					throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
				}
				if (isHeader) {
					throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
				}
				foundClosingTag = true;
				if (mywPos == start) {
					isEnd = true;
					{
						do {
							{
								if (((myPos)) == ((myEnd))) {
									throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
								}
								{
									(((c))) = (((myBytes)))[(((myPos)))++];
									if (((((c))) & 0x80) != 0) {
										int myPrivateRemaining;
										int checkMask;
										if (((((c))) & 0xE0) == 0xC0) {
											myPrivateRemaining = 1;
											(((c))) = (((c))) & 0x1F;
											checkMask = 0xFFFFFF80;
										} else if ((((c)) & 0xF0) == 0xE0) {
											myPrivateRemaining = 2;
											(((c))) = (((c))) & 0x0F;
											checkMask = 0xFFFFF800;
										} else if (((((c))) & 0xF8) == 0xF0) {
											myPrivateRemaining = 3;
											(((c))) = (((c))) & 0x07;
											checkMask = 0xFFFF0000;
										} else if (((((c))) & 0xFC) == 0xF8) {
											myPrivateRemaining = 4;
											(((c))) = (((c))) & 0x03;
											checkMask = 0xFFE00000;
										} else if (((((c))) & 0xFE) == 0xFC) {
											myPrivateRemaining = 5;
											(((c))) = (((c))) & 0x01;
											checkMask = 0xFC000000;
										} else {
											throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
										}
										int myPrivateNext;
										do {
											myPrivateNext = (((myBytes)))[(((myPos)))++];
											if ((myPrivateNext & 0xC0) != 0x80) {
												throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
											}
											(((c))) = ((((c))) << 6) | (myPrivateNext & 0x3F);
											myPrivateRemaining--;
										} while (myPrivateRemaining != 0);
										if (((((c))) & checkMask) == 0) {
											throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
										}
									}
								}
								;
							}
							;
						} while ((c) == ' ' || (c) == '\t' || (c) == '\r' || (c) == '\n');
					}
					;
				} else {
					break;
				}
			} else if (c == '?') {
				if (foundClosingTag) {
					throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
				}
				if (mywPos == start) {
					if (isHeader) {
						throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
					}
					isHeader = true;
					{
						do {
							{
								if (((myPos)) == ((myEnd))) {
									throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
								}
								{
									(((c))) = (((myBytes)))[(((myPos)))++];
									if (((((c))) & 0x80) != 0) {
										int myPrivateRemaining;
										int checkMask;
										if (((((c))) & 0xE0) == 0xC0) {
											myPrivateRemaining = 1;
											(((c))) = (((c))) & 0x1F;
											checkMask = 0xFFFFFF80;
										} else if ((((c)) & 0xF0) == 0xE0) {
											myPrivateRemaining = 2;
											(((c))) = (((c))) & 0x0F;
											checkMask = 0xFFFFF800;
										} else if (((((c))) & 0xF8) == 0xF0) {
											myPrivateRemaining = 3;
											(((c))) = (((c))) & 0x07;
											checkMask = 0xFFFF0000;
										} else if (((((c))) & 0xFC) == 0xF8) {
											myPrivateRemaining = 4;
											(((c))) = (((c))) & 0x03;
											checkMask = 0xFFE00000;
										} else if (((((c))) & 0xFE) == 0xFC) {
											myPrivateRemaining = 5;
											(((c))) = (((c))) & 0x01;
											checkMask = 0xFC000000;
										} else {
											throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
										}
										int myPrivateNext;
										do {
											myPrivateNext = (((myBytes)))[(((myPos)))++];
											if ((myPrivateNext & 0xC0) != 0x80) {
												throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
											}
											(((c))) = ((((c))) << 6) | (myPrivateNext & 0x3F);
											myPrivateRemaining--;
										} while (myPrivateRemaining != 0);
										if (((((c))) & checkMask) == 0) {
											throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
										}
									}
								}
								;
							}
							;
						} while ((c) == ' ' || (c) == '\t' || (c) == '\r' || (c) == '\n');
					}
					;
				} else {
					foundClosingTag = true;
					if (!isHeader) {
						throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
					}
					break;
				}
			} else if (c == '>') {
				if (isHeader) {
					throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
				}
				foundEndOfTag = true;
				break;
			} else if (c == ' ' || c == '\t' || c == '\r' || c == '\n') {
				break;
			} else {
				{
					if ((c) < Character.MIN_SUPPLEMENTARY_CODE_POINT) {
						(myBuf)[(mywPos)++] = (char) (c);
					} else {
						(c) -= Character.MIN_SUPPLEMENTARY_CODE_POINT;
						(myBuf)[(mywPos)++] = (char) (Character.MIN_HIGH_SURROGATE + ((c) >>> 10));
						(myBuf)[(mywPos)++] = (char) (Character.MIN_LOW_SURROGATE + ((c) & 0x3FF));
					}
				}
				;
				{
					if ((myPos) == (myEnd)) {
						throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
					}
					{
						((c)) = ((myBytes))[((myPos))++];
						if ((((c)) & 0x80) != 0) {
							int myPrivateRemaining;
							int checkMask;
							if ((((c)) & 0xE0) == 0xC0) {
								myPrivateRemaining = 1;
								((c)) = ((c)) & 0x1F;
								checkMask = 0xFFFFFF80;
							} else if (((c) & 0xF0) == 0xE0) {
								myPrivateRemaining = 2;
								((c)) = ((c)) & 0x0F;
								checkMask = 0xFFFFF800;
							} else if ((((c)) & 0xF8) == 0xF0) {
								myPrivateRemaining = 3;
								((c)) = ((c)) & 0x07;
								checkMask = 0xFFFF0000;
							} else if ((((c)) & 0xFC) == 0xF8) {
								myPrivateRemaining = 4;
								((c)) = ((c)) & 0x03;
								checkMask = 0xFFE00000;
							} else if ((((c)) & 0xFE) == 0xFC) {
								myPrivateRemaining = 5;
								((c)) = ((c)) & 0x01;
								checkMask = 0xFC000000;
							} else {
								throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
							}
							int myPrivateNext;
							do {
								myPrivateNext = ((myBytes))[((myPos))++];
								if ((myPrivateNext & 0xC0) != 0x80) {
									throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
								}
								((c)) = (((c)) << 6) | (myPrivateNext & 0x3F);
								myPrivateRemaining--;
							} while (myPrivateRemaining != 0);
							if ((((c)) & checkMask) == 0) {
								throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
							}
						}
					}
					;
				}
				;
			}
		}
		if (start == mywPos) {
			throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
		}
		String strTag = new String(buf, start, mywPos - start);
		if (isEnd) {
			if (!foundEndOfTag) {
				{
					do {
						{
							if (((myPos)) == ((myEnd))) {
								throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
							}
							{
								(((c))) = (((myBytes)))[(((myPos)))++];
								if (((((c))) & 0x80) != 0) {
									int myPrivateRemaining;
									int checkMask;
									if (((((c))) & 0xE0) == 0xC0) {
										myPrivateRemaining = 1;
										(((c))) = (((c))) & 0x1F;
										checkMask = 0xFFFFFF80;
									} else if ((((c)) & 0xF0) == 0xE0) {
										myPrivateRemaining = 2;
										(((c))) = (((c))) & 0x0F;
										checkMask = 0xFFFFF800;
									} else if (((((c))) & 0xF8) == 0xF0) {
										myPrivateRemaining = 3;
										(((c))) = (((c))) & 0x07;
										checkMask = 0xFFFF0000;
									} else if (((((c))) & 0xFC) == 0xF8) {
										myPrivateRemaining = 4;
										(((c))) = (((c))) & 0x03;
										checkMask = 0xFFE00000;
									} else if (((((c))) & 0xFE) == 0xFC) {
										myPrivateRemaining = 5;
										(((c))) = (((c))) & 0x01;
										checkMask = 0xFC000000;
									} else {
										throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
									}
									int myPrivateNext;
									do {
										myPrivateNext = (((myBytes)))[(((myPos)))++];
										if ((myPrivateNext & 0xC0) != 0x80) {
											throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
										}
										(((c))) = ((((c))) << 6) | (myPrivateNext & 0x3F);
										myPrivateRemaining--;
									} while (myPrivateRemaining != 0);
									if (((((c))) & checkMask) == 0) {
										throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
									}
								}
							}
							;
						}
						;
					} while ((c) == ' ' || (c) == '\t' || (c) == '\r' || (c) == '\n');
				}
				;
				if (c != '>') {
					throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
				}
			}
			currentTag = null;
			foundClosingTag = false;
		} else {
			currentTag = strTag;
		}
		this.pos1 = myPos;
		this.wpos = mywPos;
		return strTag;
	}

	String readAttributeName() throws MarshallingException {
		if (foundClosingTag || foundEndOfTag) {
			return null;
		}
		int myPos = this.pos1;
		final int myEnd = this.end1;
		final byte[] myBytes = this.bytes;
		int c;
		{
			do {
				{
					if (((myPos)) == ((myEnd))) {
						throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
					}
					{
						(((c))) = (((myBytes)))[(((myPos)))++];
						if (((((c))) & 0x80) != 0) {
							int myPrivateRemaining;
							int checkMask;
							if (((((c))) & 0xE0) == 0xC0) {
								myPrivateRemaining = 1;
								(((c))) = (((c))) & 0x1F;
								checkMask = 0xFFFFFF80;
							} else if ((((c)) & 0xF0) == 0xE0) {
								myPrivateRemaining = 2;
								(((c))) = (((c))) & 0x0F;
								checkMask = 0xFFFFF800;
							} else if (((((c))) & 0xF8) == 0xF0) {
								myPrivateRemaining = 3;
								(((c))) = (((c))) & 0x07;
								checkMask = 0xFFFF0000;
							} else if (((((c))) & 0xFC) == 0xF8) {
								myPrivateRemaining = 4;
								(((c))) = (((c))) & 0x03;
								checkMask = 0xFFE00000;
							} else if (((((c))) & 0xFE) == 0xFC) {
								myPrivateRemaining = 5;
								(((c))) = (((c))) & 0x01;
								checkMask = 0xFC000000;
							} else {
								throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
							}
							int myPrivateNext;
							do {
								myPrivateNext = (((myBytes)))[(((myPos)))++];
								if ((myPrivateNext & 0xC0) != 0x80) {
									throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
								}
								(((c))) = ((((c))) << 6) | (myPrivateNext & 0x3F);
								myPrivateRemaining--;
							} while (myPrivateRemaining != 0);
							if (((((c))) & checkMask) == 0) {
								throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
							}
						}
					}
					;
				}
				;
			} while ((c) == ' ' || (c) == '\t' || (c) == '\r' || (c) == '\n');
		}
		;
		boolean foundEq = false;
		final char[] myBuf = this.buf;
		int mywPos = this.wpos;
		int start = mywPos;
		while (true) {
			if (c == '/' || c == '?') {
				if (mywPos != start) {
					throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
				}
				foundClosingTag = true;
				this.pos1 = myPos;
				this.wpos = mywPos;
				return null;
			}
			if (c == '>') {
				if (mywPos != start) {
					throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
				}
				foundEndOfTag = true;
				this.pos1 = myPos;
				this.wpos = mywPos;
				return null;
			}
			if (c == ' ' || c == '\t' || c == '\r' || c == '\n') {
				break;
			}
			if (c == '=') {
				foundEq = true;
				break;
			}
			{
				if ((c) < Character.MIN_SUPPLEMENTARY_CODE_POINT) {
					(myBuf)[(mywPos)++] = (char) (c);
				} else {
					(c) -= Character.MIN_SUPPLEMENTARY_CODE_POINT;
					(myBuf)[(mywPos)++] = (char) (Character.MIN_HIGH_SURROGATE + ((c) >>> 10));
					(myBuf)[(mywPos)++] = (char) (Character.MIN_LOW_SURROGATE + ((c) & 0x3FF));
				}
			}
			;
			{
				if ((myPos) == (myEnd)) {
					throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
				}
				{
					((c)) = ((myBytes))[((myPos))++];
					if ((((c)) & 0x80) != 0) {
						int myPrivateRemaining;
						int checkMask;
						if ((((c)) & 0xE0) == 0xC0) {
							myPrivateRemaining = 1;
							((c)) = ((c)) & 0x1F;
							checkMask = 0xFFFFFF80;
						} else if (((c) & 0xF0) == 0xE0) {
							myPrivateRemaining = 2;
							((c)) = ((c)) & 0x0F;
							checkMask = 0xFFFFF800;
						} else if ((((c)) & 0xF8) == 0xF0) {
							myPrivateRemaining = 3;
							((c)) = ((c)) & 0x07;
							checkMask = 0xFFFF0000;
						} else if ((((c)) & 0xFC) == 0xF8) {
							myPrivateRemaining = 4;
							((c)) = ((c)) & 0x03;
							checkMask = 0xFFE00000;
						} else if ((((c)) & 0xFE) == 0xFC) {
							myPrivateRemaining = 5;
							((c)) = ((c)) & 0x01;
							checkMask = 0xFC000000;
						} else {
							throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
						}
						int myPrivateNext;
						do {
							myPrivateNext = ((myBytes))[((myPos))++];
							if ((myPrivateNext & 0xC0) != 0x80) {
								throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
							}
							((c)) = (((c)) << 6) | (myPrivateNext & 0x3F);
							myPrivateRemaining--;
						} while (myPrivateRemaining != 0);
						if ((((c)) & checkMask) == 0) {
							throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
						}
					}
				}
				;
			}
			;
		}
		if (mywPos == start) {
			throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
		}
		if (!foundEq) {
			{
				do {
					{
						if (((myPos)) == ((myEnd))) {
							throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
						}
						{
							(((c))) = (((myBytes)))[(((myPos)))++];
							if (((((c))) & 0x80) != 0) {
								int myPrivateRemaining;
								int checkMask;
								if (((((c))) & 0xE0) == 0xC0) {
									myPrivateRemaining = 1;
									(((c))) = (((c))) & 0x1F;
									checkMask = 0xFFFFFF80;
								} else if ((((c)) & 0xF0) == 0xE0) {
									myPrivateRemaining = 2;
									(((c))) = (((c))) & 0x0F;
									checkMask = 0xFFFFF800;
								} else if (((((c))) & 0xF8) == 0xF0) {
									myPrivateRemaining = 3;
									(((c))) = (((c))) & 0x07;
									checkMask = 0xFFFF0000;
								} else if (((((c))) & 0xFC) == 0xF8) {
									myPrivateRemaining = 4;
									(((c))) = (((c))) & 0x03;
									checkMask = 0xFFE00000;
								} else if (((((c))) & 0xFE) == 0xFC) {
									myPrivateRemaining = 5;
									(((c))) = (((c))) & 0x01;
									checkMask = 0xFC000000;
								} else {
									throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
								}
								int myPrivateNext;
								do {
									myPrivateNext = (((myBytes)))[(((myPos)))++];
									if ((myPrivateNext & 0xC0) != 0x80) {
										throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
									}
									(((c))) = ((((c))) << 6) | (myPrivateNext & 0x3F);
									myPrivateRemaining--;
								} while (myPrivateRemaining != 0);
								if (((((c))) & checkMask) == 0) {
									throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
								}
							}
						}
						;
					}
					;
				} while ((c) == ' ' || (c) == '\t' || (c) == '\r' || (c) == '\n');
			}
			;
			if (c != '=') {
				throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
			}
		}
		this.pos1 = myPos;
		this.wpos = mywPos;
		return new String(buf, start, mywPos - start);
	}

	void appendQuotedHex() throws MarshallingException {
		int myPos = this.pos1;
		final int myEnd = this.end1;
		final byte[] myBytes = this.bytes;
		int val = 0;
		int cnt = 0;
		int c;
		outer: while (true) {
			{
				if ((myPos) == (myEnd)) {
					throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
				}
				{
					((c)) = ((myBytes))[((myPos))++];
					if ((((c)) & 0x80) != 0) {
						int myPrivateRemaining;
						int checkMask;
						if ((((c)) & 0xE0) == 0xC0) {
							myPrivateRemaining = 1;
							((c)) = ((c)) & 0x1F;
							checkMask = 0xFFFFFF80;
						} else if (((c) & 0xF0) == 0xE0) {
							myPrivateRemaining = 2;
							((c)) = ((c)) & 0x0F;
							checkMask = 0xFFFFF800;
						} else if ((((c)) & 0xF8) == 0xF0) {
							myPrivateRemaining = 3;
							((c)) = ((c)) & 0x07;
							checkMask = 0xFFFF0000;
						} else if ((((c)) & 0xFC) == 0xF8) {
							myPrivateRemaining = 4;
							((c)) = ((c)) & 0x03;
							checkMask = 0xFFE00000;
						} else if ((((c)) & 0xFE) == 0xFC) {
							myPrivateRemaining = 5;
							((c)) = ((c)) & 0x01;
							checkMask = 0xFC000000;
						} else {
							throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
						}
						int myPrivateNext;
						do {
							myPrivateNext = ((myBytes))[((myPos))++];
							if ((myPrivateNext & 0xC0) != 0x80) {
								throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
							}
							((c)) = (((c)) << 6) | (myPrivateNext & 0x3F);
							myPrivateRemaining--;
						} while (myPrivateRemaining != 0);
						if ((((c)) & checkMask) == 0) {
							throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
						}
					}
				}
				;
			}
			;
			switch (c) {
			case ';':
				break outer;
			case '0':
				val = val << 4;
				break;
			case '1':
				val = (val << 4) + 1;
				break;
			case '2':
				val = (val << 4) + 2;
				break;
			case '3':
				val = (val << 4) + 3;
				break;
			case '4':
				val = (val << 4) + 4;
				break;
			case '5':
				val = (val << 4) + 5;
				break;
			case '6':
				val = (val << 4) + 6;
				break;
			case '7':
				val = (val << 4) + 7;
				break;
			case '8':
				val = (val << 4) + 8;
				break;
			case '9':
				val = (val << 4) + 9;
				break;
			case 'A':
			case 'a':
				val = (val << 4) + 10;
				break;
			case 'B':
			case 'b':
				val = (val << 4) + 11;
				break;
			case 'C':
			case 'c':
				val = (val << 4) + 12;
				break;
			case 'D':
			case 'd':
				val = (val << 4) + 13;
				break;
			case 'E':
			case 'e':
				val = (val << 4) + 14;
				break;
			case 'F':
			case 'f':
				val = (val << 4) + 15;
				break;
			default:
				throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
			}
			cnt++;
		}
		if (cnt > 8 || cnt == 0) {
			throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
		}
		if (val < Character.MIN_CODE_POINT || val > Character.MAX_CODE_POINT) {
			throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
		}
		{
			if ((val) < Character.MIN_SUPPLEMENTARY_CODE_POINT) {
				(this.buf)[(this.wpos)++] = (char) (val);
			} else {
				(val) -= Character.MIN_SUPPLEMENTARY_CODE_POINT;
				(this.buf)[(this.wpos)++] = (char) (Character.MIN_HIGH_SURROGATE + ((val) >>> 10));
				(this.buf)[(this.wpos)++] = (char) (Character.MIN_LOW_SURROGATE + ((val) & 0x3FF));
			}
		}
		;
		this.pos1 = myPos;
	}

	void appendEscapedHex() throws MarshallingException {
		int myPos = this.pos1;
		final int myEnd = this.end1;
		final byte[] myBytes = this.bytes;
		int val = 0;
		int c;
		for (int cnt = 4; cnt != 0; cnt--) {
			{
				if ((myPos) == (myEnd)) {
					throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
				}
				{
					((c)) = ((myBytes))[((myPos))++];
					if ((((c)) & 0x80) != 0) {
						int myPrivateRemaining;
						int checkMask;
						if ((((c)) & 0xE0) == 0xC0) {
							myPrivateRemaining = 1;
							((c)) = ((c)) & 0x1F;
							checkMask = 0xFFFFFF80;
						} else if (((c) & 0xF0) == 0xE0) {
							myPrivateRemaining = 2;
							((c)) = ((c)) & 0x0F;
							checkMask = 0xFFFFF800;
						} else if ((((c)) & 0xF8) == 0xF0) {
							myPrivateRemaining = 3;
							((c)) = ((c)) & 0x07;
							checkMask = 0xFFFF0000;
						} else if ((((c)) & 0xFC) == 0xF8) {
							myPrivateRemaining = 4;
							((c)) = ((c)) & 0x03;
							checkMask = 0xFFE00000;
						} else if ((((c)) & 0xFE) == 0xFC) {
							myPrivateRemaining = 5;
							((c)) = ((c)) & 0x01;
							checkMask = 0xFC000000;
						} else {
							throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
						}
						int myPrivateNext;
						do {
							myPrivateNext = ((myBytes))[((myPos))++];
							if ((myPrivateNext & 0xC0) != 0x80) {
								throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
							}
							((c)) = (((c)) << 6) | (myPrivateNext & 0x3F);
							myPrivateRemaining--;
						} while (myPrivateRemaining != 0);
						if ((((c)) & checkMask) == 0) {
							throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
						}
					}
				}
				;
			}
			;
			switch (c) {
			case '0':
				val = val << 4;
				break;
			case '1':
				val = (val << 4) + 1;
				break;
			case '2':
				val = (val << 4) + 2;
				break;
			case '3':
				val = (val << 4) + 3;
				break;
			case '4':
				val = (val << 4) + 4;
				break;
			case '5':
				val = (val << 4) + 5;
				break;
			case '6':
				val = (val << 4) + 6;
				break;
			case '7':
				val = (val << 4) + 7;
				break;
			case '8':
				val = (val << 4) + 8;
				break;
			case '9':
				val = (val << 4) + 9;
				break;
			case 'A':
			case 'a':
				val = (val << 4) + 10;
				break;
			case 'B':
			case 'b':
				val = (val << 4) + 11;
				break;
			case 'C':
			case 'c':
				val = (val << 4) + 12;
				break;
			case 'D':
			case 'd':
				val = (val << 4) + 13;
				break;
			case 'E':
			case 'e':
				val = (val << 4) + 14;
				break;
			case 'F':
			case 'f':
				val = (val << 4) + 15;
				break;
			default:
				throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
			}
		}
		this.buf[this.wpos++] = (char) val;
		this.pos1 = myPos;
	}

	void appendQuotedChar() throws MarshallingException {
		int myPos = this.pos1;
		final int myEnd = this.end1;
		final byte[] myBytes = this.bytes;
		int c;
		{
			if ((myPos) == (myEnd)) {
				throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
			}
			{
				((c)) = ((myBytes))[((myPos))++];
				if ((((c)) & 0x80) != 0) {
					int myPrivateRemaining;
					int checkMask;
					if ((((c)) & 0xE0) == 0xC0) {
						myPrivateRemaining = 1;
						((c)) = ((c)) & 0x1F;
						checkMask = 0xFFFFFF80;
					} else if (((c) & 0xF0) == 0xE0) {
						myPrivateRemaining = 2;
						((c)) = ((c)) & 0x0F;
						checkMask = 0xFFFFF800;
					} else if ((((c)) & 0xF8) == 0xF0) {
						myPrivateRemaining = 3;
						((c)) = ((c)) & 0x07;
						checkMask = 0xFFFF0000;
					} else if ((((c)) & 0xFC) == 0xF8) {
						myPrivateRemaining = 4;
						((c)) = ((c)) & 0x03;
						checkMask = 0xFFE00000;
					} else if ((((c)) & 0xFE) == 0xFC) {
						myPrivateRemaining = 5;
						((c)) = ((c)) & 0x01;
						checkMask = 0xFC000000;
					} else {
						throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
					}
					int myPrivateNext;
					do {
						myPrivateNext = ((myBytes))[((myPos))++];
						if ((myPrivateNext & 0xC0) != 0x80) {
							throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
						}
						((c)) = (((c)) << 6) | (myPrivateNext & 0x3F);
						myPrivateRemaining--;
					} while (myPrivateRemaining != 0);
					if ((((c)) & checkMask) == 0) {
						throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
					}
				}
			}
			;
		}
		;
		if (c == '#') {
			{
				if ((myPos) == (myEnd)) {
					throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
				}
				{
					((c)) = ((myBytes))[((myPos))++];
					if ((((c)) & 0x80) != 0) {
						int myPrivateRemaining;
						int checkMask;
						if ((((c)) & 0xE0) == 0xC0) {
							myPrivateRemaining = 1;
							((c)) = ((c)) & 0x1F;
							checkMask = 0xFFFFFF80;
						} else if (((c) & 0xF0) == 0xE0) {
							myPrivateRemaining = 2;
							((c)) = ((c)) & 0x0F;
							checkMask = 0xFFFFF800;
						} else if ((((c)) & 0xF8) == 0xF0) {
							myPrivateRemaining = 3;
							((c)) = ((c)) & 0x07;
							checkMask = 0xFFFF0000;
						} else if ((((c)) & 0xFC) == 0xF8) {
							myPrivateRemaining = 4;
							((c)) = ((c)) & 0x03;
							checkMask = 0xFFE00000;
						} else if ((((c)) & 0xFE) == 0xFC) {
							myPrivateRemaining = 5;
							((c)) = ((c)) & 0x01;
							checkMask = 0xFC000000;
						} else {
							throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
						}
						int myPrivateNext;
						do {
							myPrivateNext = ((myBytes))[((myPos))++];
							if ((myPrivateNext & 0xC0) != 0x80) {
								throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
							}
							((c)) = (((c)) << 6) | (myPrivateNext & 0x3F);
							myPrivateRemaining--;
						} while (myPrivateRemaining != 0);
						if ((((c)) & checkMask) == 0) {
							throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
						}
					}
				}
				;
			}
			;
			if (c != 'x') {
				throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
			}
			this.pos1 = myPos;
			appendQuotedHex();
		} else {
			if (c == 'q') {
				{
					if ((myPos) == (myEnd)) {
						throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
					}
					{
						((c)) = ((myBytes))[((myPos))++];
						if ((((c)) & 0x80) != 0) {
							int myPrivateRemaining;
							int checkMask;
							if ((((c)) & 0xE0) == 0xC0) {
								myPrivateRemaining = 1;
								((c)) = ((c)) & 0x1F;
								checkMask = 0xFFFFFF80;
							} else if (((c) & 0xF0) == 0xE0) {
								myPrivateRemaining = 2;
								((c)) = ((c)) & 0x0F;
								checkMask = 0xFFFFF800;
							} else if ((((c)) & 0xF8) == 0xF0) {
								myPrivateRemaining = 3;
								((c)) = ((c)) & 0x07;
								checkMask = 0xFFFF0000;
							} else if ((((c)) & 0xFC) == 0xF8) {
								myPrivateRemaining = 4;
								((c)) = ((c)) & 0x03;
								checkMask = 0xFFE00000;
							} else if ((((c)) & 0xFE) == 0xFC) {
								myPrivateRemaining = 5;
								((c)) = ((c)) & 0x01;
								checkMask = 0xFC000000;
							} else {
								throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
							}
							int myPrivateNext;
							do {
								myPrivateNext = ((myBytes))[((myPos))++];
								if ((myPrivateNext & 0xC0) != 0x80) {
									throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
								}
								((c)) = (((c)) << 6) | (myPrivateNext & 0x3F);
								myPrivateRemaining--;
							} while (myPrivateRemaining != 0);
							if ((((c)) & checkMask) == 0) {
								throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
							}
						}
					}
					;
				}
				;
				if (c != 'u') {
					throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
				}
				{
					if ((myPos) == (myEnd)) {
						throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
					}
					{
						((c)) = ((myBytes))[((myPos))++];
						if ((((c)) & 0x80) != 0) {
							int myPrivateRemaining;
							int checkMask;
							if ((((c)) & 0xE0) == 0xC0) {
								myPrivateRemaining = 1;
								((c)) = ((c)) & 0x1F;
								checkMask = 0xFFFFFF80;
							} else if (((c) & 0xF0) == 0xE0) {
								myPrivateRemaining = 2;
								((c)) = ((c)) & 0x0F;
								checkMask = 0xFFFFF800;
							} else if ((((c)) & 0xF8) == 0xF0) {
								myPrivateRemaining = 3;
								((c)) = ((c)) & 0x07;
								checkMask = 0xFFFF0000;
							} else if ((((c)) & 0xFC) == 0xF8) {
								myPrivateRemaining = 4;
								((c)) = ((c)) & 0x03;
								checkMask = 0xFFE00000;
							} else if ((((c)) & 0xFE) == 0xFC) {
								myPrivateRemaining = 5;
								((c)) = ((c)) & 0x01;
								checkMask = 0xFC000000;
							} else {
								throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
							}
							int myPrivateNext;
							do {
								myPrivateNext = ((myBytes))[((myPos))++];
								if ((myPrivateNext & 0xC0) != 0x80) {
									throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
								}
								((c)) = (((c)) << 6) | (myPrivateNext & 0x3F);
								myPrivateRemaining--;
							} while (myPrivateRemaining != 0);
							if ((((c)) & checkMask) == 0) {
								throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
							}
						}
					}
					;
				}
				;
				if (c != 'o') {
					throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
				}
				{
					if ((myPos) == (myEnd)) {
						throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
					}
					{
						((c)) = ((myBytes))[((myPos))++];
						if ((((c)) & 0x80) != 0) {
							int myPrivateRemaining;
							int checkMask;
							if ((((c)) & 0xE0) == 0xC0) {
								myPrivateRemaining = 1;
								((c)) = ((c)) & 0x1F;
								checkMask = 0xFFFFFF80;
							} else if (((c) & 0xF0) == 0xE0) {
								myPrivateRemaining = 2;
								((c)) = ((c)) & 0x0F;
								checkMask = 0xFFFFF800;
							} else if ((((c)) & 0xF8) == 0xF0) {
								myPrivateRemaining = 3;
								((c)) = ((c)) & 0x07;
								checkMask = 0xFFFF0000;
							} else if ((((c)) & 0xFC) == 0xF8) {
								myPrivateRemaining = 4;
								((c)) = ((c)) & 0x03;
								checkMask = 0xFFE00000;
							} else if ((((c)) & 0xFE) == 0xFC) {
								myPrivateRemaining = 5;
								((c)) = ((c)) & 0x01;
								checkMask = 0xFC000000;
							} else {
								throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
							}
							int myPrivateNext;
							do {
								myPrivateNext = ((myBytes))[((myPos))++];
								if ((myPrivateNext & 0xC0) != 0x80) {
									throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
								}
								((c)) = (((c)) << 6) | (myPrivateNext & 0x3F);
								myPrivateRemaining--;
							} while (myPrivateRemaining != 0);
							if ((((c)) & checkMask) == 0) {
								throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
							}
						}
					}
					;
				}
				;
				if (c != 't') {
					throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
				}
				this.buf[this.wpos++] = '"';
			} else if (c == 'a') {
				{
					if ((myPos) == (myEnd)) {
						throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
					}
					{
						((c)) = ((myBytes))[((myPos))++];
						if ((((c)) & 0x80) != 0) {
							int myPrivateRemaining;
							int checkMask;
							if ((((c)) & 0xE0) == 0xC0) {
								myPrivateRemaining = 1;
								((c)) = ((c)) & 0x1F;
								checkMask = 0xFFFFFF80;
							} else if (((c) & 0xF0) == 0xE0) {
								myPrivateRemaining = 2;
								((c)) = ((c)) & 0x0F;
								checkMask = 0xFFFFF800;
							} else if ((((c)) & 0xF8) == 0xF0) {
								myPrivateRemaining = 3;
								((c)) = ((c)) & 0x07;
								checkMask = 0xFFFF0000;
							} else if ((((c)) & 0xFC) == 0xF8) {
								myPrivateRemaining = 4;
								((c)) = ((c)) & 0x03;
								checkMask = 0xFFE00000;
							} else if ((((c)) & 0xFE) == 0xFC) {
								myPrivateRemaining = 5;
								((c)) = ((c)) & 0x01;
								checkMask = 0xFC000000;
							} else {
								throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
							}
							int myPrivateNext;
							do {
								myPrivateNext = ((myBytes))[((myPos))++];
								if ((myPrivateNext & 0xC0) != 0x80) {
									throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
								}
								((c)) = (((c)) << 6) | (myPrivateNext & 0x3F);
								myPrivateRemaining--;
							} while (myPrivateRemaining != 0);
							if ((((c)) & checkMask) == 0) {
								throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
							}
						}
					}
					;
				}
				;
				if (c != 'm') {
					throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
				}
				{
					if ((myPos) == (myEnd)) {
						throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
					}
					{
						((c)) = ((myBytes))[((myPos))++];
						if ((((c)) & 0x80) != 0) {
							int myPrivateRemaining;
							int checkMask;
							if ((((c)) & 0xE0) == 0xC0) {
								myPrivateRemaining = 1;
								((c)) = ((c)) & 0x1F;
								checkMask = 0xFFFFFF80;
							} else if (((c) & 0xF0) == 0xE0) {
								myPrivateRemaining = 2;
								((c)) = ((c)) & 0x0F;
								checkMask = 0xFFFFF800;
							} else if ((((c)) & 0xF8) == 0xF0) {
								myPrivateRemaining = 3;
								((c)) = ((c)) & 0x07;
								checkMask = 0xFFFF0000;
							} else if ((((c)) & 0xFC) == 0xF8) {
								myPrivateRemaining = 4;
								((c)) = ((c)) & 0x03;
								checkMask = 0xFFE00000;
							} else if ((((c)) & 0xFE) == 0xFC) {
								myPrivateRemaining = 5;
								((c)) = ((c)) & 0x01;
								checkMask = 0xFC000000;
							} else {
								throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
							}
							int myPrivateNext;
							do {
								myPrivateNext = ((myBytes))[((myPos))++];
								if ((myPrivateNext & 0xC0) != 0x80) {
									throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
								}
								((c)) = (((c)) << 6) | (myPrivateNext & 0x3F);
								myPrivateRemaining--;
							} while (myPrivateRemaining != 0);
							if ((((c)) & checkMask) == 0) {
								throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
							}
						}
					}
					;
				}
				;
				if (c != 'p') {
					throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
				}
				this.buf[this.wpos++] = '&';
			} else if (c == 'g') {
				{
					if ((myPos) == (myEnd)) {
						throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
					}
					{
						((c)) = ((myBytes))[((myPos))++];
						if ((((c)) & 0x80) != 0) {
							int myPrivateRemaining;
							int checkMask;
							if ((((c)) & 0xE0) == 0xC0) {
								myPrivateRemaining = 1;
								((c)) = ((c)) & 0x1F;
								checkMask = 0xFFFFFF80;
							} else if (((c) & 0xF0) == 0xE0) {
								myPrivateRemaining = 2;
								((c)) = ((c)) & 0x0F;
								checkMask = 0xFFFFF800;
							} else if ((((c)) & 0xF8) == 0xF0) {
								myPrivateRemaining = 3;
								((c)) = ((c)) & 0x07;
								checkMask = 0xFFFF0000;
							} else if ((((c)) & 0xFC) == 0xF8) {
								myPrivateRemaining = 4;
								((c)) = ((c)) & 0x03;
								checkMask = 0xFFE00000;
							} else if ((((c)) & 0xFE) == 0xFC) {
								myPrivateRemaining = 5;
								((c)) = ((c)) & 0x01;
								checkMask = 0xFC000000;
							} else {
								throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
							}
							int myPrivateNext;
							do {
								myPrivateNext = ((myBytes))[((myPos))++];
								if ((myPrivateNext & 0xC0) != 0x80) {
									throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
								}
								((c)) = (((c)) << 6) | (myPrivateNext & 0x3F);
								myPrivateRemaining--;
							} while (myPrivateRemaining != 0);
							if ((((c)) & checkMask) == 0) {
								throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
							}
						}
					}
					;
				}
				;
				if (c != 't') {
					throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
				}
				this.buf[this.wpos++] = '>';
			} else if (c == 'l') {
				{
					if ((myPos) == (myEnd)) {
						throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
					}
					{
						((c)) = ((myBytes))[((myPos))++];
						if ((((c)) & 0x80) != 0) {
							int myPrivateRemaining;
							int checkMask;
							if ((((c)) & 0xE0) == 0xC0) {
								myPrivateRemaining = 1;
								((c)) = ((c)) & 0x1F;
								checkMask = 0xFFFFFF80;
							} else if (((c) & 0xF0) == 0xE0) {
								myPrivateRemaining = 2;
								((c)) = ((c)) & 0x0F;
								checkMask = 0xFFFFF800;
							} else if ((((c)) & 0xF8) == 0xF0) {
								myPrivateRemaining = 3;
								((c)) = ((c)) & 0x07;
								checkMask = 0xFFFF0000;
							} else if ((((c)) & 0xFC) == 0xF8) {
								myPrivateRemaining = 4;
								((c)) = ((c)) & 0x03;
								checkMask = 0xFFE00000;
							} else if ((((c)) & 0xFE) == 0xFC) {
								myPrivateRemaining = 5;
								((c)) = ((c)) & 0x01;
								checkMask = 0xFC000000;
							} else {
								throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
							}
							int myPrivateNext;
							do {
								myPrivateNext = ((myBytes))[((myPos))++];
								if ((myPrivateNext & 0xC0) != 0x80) {
									throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
								}
								((c)) = (((c)) << 6) | (myPrivateNext & 0x3F);
								myPrivateRemaining--;
							} while (myPrivateRemaining != 0);
							if ((((c)) & checkMask) == 0) {
								throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
							}
						}
					}
					;
				}
				;
				if (c != 't') {
					throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
				}
				this.buf[this.wpos++] = '<';
			} else {
				throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
			}
			{
				if ((myPos) == (myEnd)) {
					throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
				}
				{
					((c)) = ((myBytes))[((myPos))++];
					if ((((c)) & 0x80) != 0) {
						int myPrivateRemaining;
						int checkMask;
						if ((((c)) & 0xE0) == 0xC0) {
							myPrivateRemaining = 1;
							((c)) = ((c)) & 0x1F;
							checkMask = 0xFFFFFF80;
						} else if (((c) & 0xF0) == 0xE0) {
							myPrivateRemaining = 2;
							((c)) = ((c)) & 0x0F;
							checkMask = 0xFFFFF800;
						} else if ((((c)) & 0xF8) == 0xF0) {
							myPrivateRemaining = 3;
							((c)) = ((c)) & 0x07;
							checkMask = 0xFFFF0000;
						} else if ((((c)) & 0xFC) == 0xF8) {
							myPrivateRemaining = 4;
							((c)) = ((c)) & 0x03;
							checkMask = 0xFFE00000;
						} else if ((((c)) & 0xFE) == 0xFC) {
							myPrivateRemaining = 5;
							((c)) = ((c)) & 0x01;
							checkMask = 0xFC000000;
						} else {
							throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
						}
						int myPrivateNext;
						do {
							myPrivateNext = ((myBytes))[((myPos))++];
							if ((myPrivateNext & 0xC0) != 0x80) {
								throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
							}
							((c)) = (((c)) << 6) | (myPrivateNext & 0x3F);
							myPrivateRemaining--;
						} while (myPrivateRemaining != 0);
						if ((((c)) & checkMask) == 0) {
							throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
						}
					}
				}
				;
			}
			;
			if (c != ';') {
				throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
			}
			this.pos1 = myPos;
		}
	}

	String convertAttributeValue(char endChar) throws MarshallingException {
		int myPos = this.pos1;
		final int myEnd = this.end1;
		final byte[] myBytes = this.bytes;
		final char[] myBuf = this.buf;
		int mywPos = this.wpos;
		int start = mywPos;
		int slashCnt = 0;
		int c;
		outer: while (true) {
			{
				if ((myPos) == (myEnd)) {
					throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
				}
				{
					((c)) = ((myBytes))[((myPos))++];
					if ((((c)) & 0x80) != 0) {
						int myPrivateRemaining;
						int checkMask;
						if ((((c)) & 0xE0) == 0xC0) {
							myPrivateRemaining = 1;
							((c)) = ((c)) & 0x1F;
							checkMask = 0xFFFFFF80;
						} else if (((c) & 0xF0) == 0xE0) {
							myPrivateRemaining = 2;
							((c)) = ((c)) & 0x0F;
							checkMask = 0xFFFFF800;
						} else if ((((c)) & 0xF8) == 0xF0) {
							myPrivateRemaining = 3;
							((c)) = ((c)) & 0x07;
							checkMask = 0xFFFF0000;
						} else if ((((c)) & 0xFC) == 0xF8) {
							myPrivateRemaining = 4;
							((c)) = ((c)) & 0x03;
							checkMask = 0xFFE00000;
						} else if ((((c)) & 0xFE) == 0xFC) {
							myPrivateRemaining = 5;
							((c)) = ((c)) & 0x01;
							checkMask = 0xFC000000;
						} else {
							throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
						}
						int myPrivateNext;
						do {
							myPrivateNext = ((myBytes))[((myPos))++];
							if ((myPrivateNext & 0xC0) != 0x80) {
								throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
							}
							((c)) = (((c)) << 6) | (myPrivateNext & 0x3F);
							myPrivateRemaining--;
						} while (myPrivateRemaining != 0);
						if ((((c)) & checkMask) == 0) {
							throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
						}
					}
				}
				;
			}
			;
			if (c == '\\') {
				// do special slash processing
				// this has been taken out of the main loop to optimize the
				// common case
				slashCnt = 1;
				while (true) {
					{
						if ((myPos) == (myEnd)) {
							throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
						}
						{
							((c)) = ((myBytes))[((myPos))++];
							if ((((c)) & 0x80) != 0) {
								int myPrivateRemaining;
								int checkMask;
								if ((((c)) & 0xE0) == 0xC0) {
									myPrivateRemaining = 1;
									((c)) = ((c)) & 0x1F;
									checkMask = 0xFFFFFF80;
								} else if (((c) & 0xF0) == 0xE0) {
									myPrivateRemaining = 2;
									((c)) = ((c)) & 0x0F;
									checkMask = 0xFFFFF800;
								} else if ((((c)) & 0xF8) == 0xF0) {
									myPrivateRemaining = 3;
									((c)) = ((c)) & 0x07;
									checkMask = 0xFFFF0000;
								} else if ((((c)) & 0xFC) == 0xF8) {
									myPrivateRemaining = 4;
									((c)) = ((c)) & 0x03;
									checkMask = 0xFFE00000;
								} else if ((((c)) & 0xFE) == 0xFC) {
									myPrivateRemaining = 5;
									((c)) = ((c)) & 0x01;
									checkMask = 0xFC000000;
								} else {
									throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
								}
								int myPrivateNext;
								do {
									myPrivateNext = ((myBytes))[((myPos))++];
									if ((myPrivateNext & 0xC0) != 0x80) {
										throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
									}
									((c)) = (((c)) << 6) | (myPrivateNext & 0x3F);
									myPrivateRemaining--;
								} while (myPrivateRemaining != 0);
								if ((((c)) & checkMask) == 0) {
									throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
								}
							}
						}
						;
					}
					;
					if (c == 'u') {
						if (slashCnt == 1) {
							this.pos1 = myPos;
							this.wpos = mywPos;
							appendEscapedHex();
							myPos = this.pos1;
							mywPos = this.wpos;
							break;
						} else {
							myBuf[mywPos++] = '\\';
							myBuf[mywPos++] = 'u';
							break;
						}
					} else if (c == '\\') {
						if (slashCnt != 2) {
							slashCnt++;
						} else {
							myBuf[mywPos++] = '\\';
						}
					} else {
						while (slashCnt != 0) {
							myBuf[mywPos++] = '\\';
							slashCnt--;
						}
						if (c == '&') {
							this.pos1 = myPos;
							this.wpos = mywPos;
							appendQuotedChar();
							myPos = this.pos1;
							mywPos = this.wpos;
						} else if (c == endChar) {
							break outer;
						} else {
							{
								if ((c) < Character.MIN_SUPPLEMENTARY_CODE_POINT) {
									(myBuf)[(mywPos)++] = (char) (c);
								} else {
									(c) -= Character.MIN_SUPPLEMENTARY_CODE_POINT;
									(myBuf)[(mywPos)++] = (char) (Character.MIN_HIGH_SURROGATE + ((c) >>> 10));
									(myBuf)[(mywPos)++] = (char) (Character.MIN_LOW_SURROGATE + ((c) & 0x3FF));
								}
							}
							;
						}
						break;
					}
				}
			} else if (c == '&') {
				this.pos1 = myPos;
				this.wpos = mywPos;
				appendQuotedChar();
				myPos = this.pos1;
				mywPos = this.wpos;
			} else if (c == endChar) {
				break;
			} else {
				{
					if ((c) < Character.MIN_SUPPLEMENTARY_CODE_POINT) {
						(myBuf)[(mywPos)++] = (char) (c);
					} else {
						(c) -= Character.MIN_SUPPLEMENTARY_CODE_POINT;
						(myBuf)[(mywPos)++] = (char) (Character.MIN_HIGH_SURROGATE + ((c) >>> 10));
						(myBuf)[(mywPos)++] = (char) (Character.MIN_LOW_SURROGATE + ((c) & 0x3FF));
					}
				}
				;
			}
		}
		this.pos1 = myPos;
		this.wpos = mywPos;
		return new String(buf, start, mywPos - start);
	}

	String readAttributeValue() throws MarshallingException {
		int myPos = this.pos1;
		final int myEnd = this.end1;
		final byte[] myBytes = this.bytes;
		int c;
		{
			do {
				{
					if (((myPos)) == ((myEnd))) {
						throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
					}
					{
						(((c))) = (((myBytes)))[(((myPos)))++];
						if (((((c))) & 0x80) != 0) {
							int myPrivateRemaining;
							int checkMask;
							if (((((c))) & 0xE0) == 0xC0) {
								myPrivateRemaining = 1;
								(((c))) = (((c))) & 0x1F;
								checkMask = 0xFFFFFF80;
							} else if ((((c)) & 0xF0) == 0xE0) {
								myPrivateRemaining = 2;
								(((c))) = (((c))) & 0x0F;
								checkMask = 0xFFFFF800;
							} else if (((((c))) & 0xF8) == 0xF0) {
								myPrivateRemaining = 3;
								(((c))) = (((c))) & 0x07;
								checkMask = 0xFFFF0000;
							} else if (((((c))) & 0xFC) == 0xF8) {
								myPrivateRemaining = 4;
								(((c))) = (((c))) & 0x03;
								checkMask = 0xFFE00000;
							} else if (((((c))) & 0xFE) == 0xFC) {
								myPrivateRemaining = 5;
								(((c))) = (((c))) & 0x01;
								checkMask = 0xFC000000;
							} else {
								throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
							}
							int myPrivateNext;
							do {
								myPrivateNext = (((myBytes)))[(((myPos)))++];
								if ((myPrivateNext & 0xC0) != 0x80) {
									throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
								}
								(((c))) = ((((c))) << 6) | (myPrivateNext & 0x3F);
								myPrivateRemaining--;
							} while (myPrivateRemaining != 0);
							if (((((c))) & checkMask) == 0) {
								throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
							}
						}
					}
					;
				}
				;
			} while ((c) == ' ' || (c) == '\t' || (c) == '\r' || (c) == '\n');
		}
		;
		if (c != '"') {
			throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
		}
		this.pos1 = myPos;
		return convertAttributeValue('"');
	}

	String readEntityValue() throws MarshallingException {
		if (foundClosingTag) {
			return null;
		}
		String result = convertAttributeValue('<');
		foundStartOfTag = true;
		return result;
	}

	void demarshalHeader() throws MarshallingException {
		String tag = readTag();
		if (tag == null || !tag.equals("xml")) { //$NON-NLS-1$
		    StringBuilder sb = new StringBuilder();
			try {
				String contentsOfEnveloppe = new String(bytes, "UTF-8"); //$NON-NLS-1$
				sb.append("*** Internal Error ***\n"); //$NON-NLS-1$
				sb.append("==> Unable to demarshal modelled REST response. The body has the following content:\n"); //$NON-NLS-1$
				sb.append(contentsOfEnveloppe + "\n"); //$NON-NLS-1$
				sb.append("==> End-of-body-content\n"); //$NON-NLS-1$
			} catch (UnsupportedEncodingException e) {
				// never happens
			}
			throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL, sb.toString());
		}
		if (!isHeader) {
			throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
		}
		while (true) {
			String attr = readAttributeName();
			if (attr == null) {
				break;
			}
			readAttributeValue();
		}
		readTag();
		if (!isEnd) {
			throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
		}
	}

	EClass getType(String qual) throws MarshallingException {
		int idx = qual.indexOf(':');
		if (idx == -1) {
			throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
		}
		String nsPrefix = qual.substring(0, idx);
		String type = qual.substring(idx + 1);
		EFactory f = factories.get(nsPrefix);
		if (f == null) {
			f = EPackage.Registry.INSTANCE.getEFactory(nsPrefix);
			if (f == null) {
				throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
			}
		}
		EClass cls = (EClass) f.getEPackage().getEClassifier(type);
		if (cls == null) {
			throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL, qual + " " + type + " " + f.getEPackage()); //$NON-NLS-1$ //$NON-NLS-2$
		}
		cls = ItemUtil.getTargetClassIfFacade(cls);
		return cls;
	}

	EObject createType(String qual) throws MarshallingException {
		EClass cls = getType(qual);
		return cls.getEPackage().getEFactoryInstance().create(cls);
	}

	EStructuralFeature getFeature(EClass cls, String featureName) throws MarshallingException {
		EStructuralFeature feature = cls.getEStructuralFeature(featureName);
		if (feature != null)
			return feature;
		int idx = featureName.indexOf(':');
		if (idx == -1) {
			throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
		}
		String name = featureName.substring(idx + 1);
		for (EStructuralFeature eFeature : cls.getEAllStructuralFeatures()) {
			if (eFeature.getEAnnotations().size() > 0) {
				String nameAnnotationValue = MarshallerUtil.getAnnotation(eFeature, EXTENDED_METADATA_ANNOTATION, NAME_ANNOTATION_KEY);
				if (name.equals(nameAnnotationValue)) {
					return eFeature;
				}
			}
		}
		throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
	}

	Object demarshalAttribute(EAttribute attr) throws MarshallingException {
		String str = readEntityValue();
		EDataType attrType = attr.getEAttributeType();
		Class<?> target = attrType.getInstanceClass();
		// these checks are ordered from most to least likely as measured on
		// jazzdev
		if (target == UUID.class) {
			return UUID.valueOf(str);
		}
		if (target == int.class) {
			return Integer.valueOf(Integer.parseInt(str));
		}
		if (target == String.class) {
			return str;
		}
		if (target == boolean.class) {
			return Boolean.valueOf(str);
		}
		if (target == long.class) {
			return Long.valueOf(Long.parseLong(str));
		}
		if (target == Timestamp.class) {
			try {
				return DateUtils.parseTimeRFC3339(str);
			} catch (ParseException e) {
				throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL, e.getMessage(), e);
			}
		}
		if (Enumerator.class.isAssignableFrom(target)) {
			return attrType.getEPackage().getEFactoryInstance().createFromString(attrType, str);
		}
		if (target == BigDecimal.class) {
			return new BigDecimal(str);
		}
		if (target == Date.class) {
			try {
				return df.parse(str);
			} catch (ParseException e) {
				throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL, e.getMessage(), e);
			}
		}
		if (target == double.class) {
			return Double.valueOf(Double.parseDouble(str));
		}
		if (target == float.class) {
			return Float.valueOf(Float.parseFloat(str));
		}
		if (target == byte.class) {
			return Byte.valueOf(Byte.parseByte(str));
		}
		if (target == char.class) {
			return Character.valueOf((char) Integer.parseInt(str));
		}
		if (target == short.class) {
			return Short.valueOf(Short.parseShort(str));
		}
		if (target == Integer.class) {
			return Integer.valueOf(Integer.parseInt(str));
		}
		if (target == Boolean.class) {
			return Boolean.valueOf(str);
		}
		if (target == Long.class) {
			return Long.valueOf(Long.parseLong(str));
		}
		if (target == Double.class) {
			return Double.valueOf(Double.parseDouble(str));
		}
		if (target == Float.class) {
			return Float.valueOf(Float.parseFloat(str));
		}
		if (target == Byte.class) {
			return Byte.valueOf(Byte.parseByte(str));
		}
		if (target == Character.class) {
			return Character.valueOf((char) Integer.parseInt(str));
		}
		if (target == Short.class) {
			return Short.valueOf(Short.parseShort(str));
		}
		throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL, "Cannot handle attribute type " + target.toString()); //$NON-NLS-1$
	}

	EObject demarshalObject() throws MarshallingException {
		String tag = readTag();
		if (tag == null) {
			throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
		}
		if (isHeader) {
			throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
		}
		if (isEnd) {
			throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
		}
		String attr = readAttributeName();
		while (true) {
			if (attr == null) {
				throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
			}
			if (!attr.startsWith("xmlns:")) { //$NON-NLS-1$
				throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
			}
			if (attr.equals("xmlns:xsi")) { //$NON-NLS-1$
				readAttributeValue();
				break;
			}
			attr = attr.substring(6);
			String uri = readAttributeValue();
			EPackage p = EPackage.Registry.INSTANCE.getEPackage(uri);
			if (p == null) {
				if (isNamespaceToleranceEnabled()) {
					if ("repository".equals(uri)) { //$NON-NLS-1$
						p = handleMissingPackage(uri, "com.ibm.team.repository"); //$NON-NLS-1$;
					} else if ("compatibilitypack".equals(uri)) { //$NON-NLS-1$
						p = handleMissingPackage(uri, "com.ibm.team.compatibilitypack"); //$NON-NLS-1$;
					} else if ("auth".equals(uri)) { //$NON-NLS-1$
						p = handleMissingPackage(uri, "com.ibm.team.repository.auth"); //$NON-NLS-1$;
					} else {
						String overrideUriString = System.getProperty("com.ibm.team.repository.uriCorrection." + uri); //$NON-NLS-1$
						if (overrideUriString != null) {
							p = handleMissingPackage(uri, overrideUriString);
						}
					}
				}
				if (p == null) {
					throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL, "Cannot find package " + uri); //$NON-NLS-1$
				}
			}
			if (factories.put(attr, p.getEFactoryInstance()) != null) {
				throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
			}
			attr = readAttributeName();
		}
		EObject o = createType(tag);
		attr = readAttributeName();
		EClass cls = o.eClass();
		while (true) {
			while (attr != null) {
				EStructuralFeature sf = getFeature(cls, attr);
				if (sf == ITEM_ID_ATTR) {
					o.eSet(sf, UUID.valueOf(readAttributeValue()));
				} else {
					EReference r = (EReference) sf;
					((List<?>) o.eGet(r)).clear();
					if (readAttributeValue().length() != 0) {
						throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
					}
				}
				attr = readAttributeName();
			}
			tag = readTag();
			if (isEnd) {
				if (objectStack.isEmpty()) {
					break;
				}
				o = objectStack.remove(objectStack.size() - 1);
				cls = o.eClass();
			} else {
				EStructuralFeature f = getFeature(cls, tag);
				if (f instanceof EAttribute) {
					EAttribute a = (EAttribute) f;
					boolean isNull = false;
					attr = readAttributeName();
					while (attr != null) {
						String value = readAttributeValue();
						if (attr.equals("xsi:nil")) { //$NON-NLS-1$
							if (!value.equals("true")) { //$NON-NLS-1$
								throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
							}
							isNull = true;
						}
						attr = readAttributeName();
					}
					Object value;
					if (isNull) {
						value = null;
					} else {
						value = demarshalAttribute(a);
					}
					if (a.isMany()) {
						Object list = o.eGet(a);
						if (list instanceof EcoreEMap) {
							((EcoreEMap) list).addUnique((Entry) value);
						} else {
							((AbstractEList) list).addUnique(value);
						}
					} else {
						o.eSet(a, value);
					}
					tag = readTag();
					if (!isEnd) {
						throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
					}
					attr = null;
				} else {
					EReference ref = (EReference) f;
					attr = readAttributeName();
					UUID itemId = null;
					UUID stateId = null;
					String type = null;
					boolean isNull = false;
					boolean isHandle;
					if (MarshallerUtil.shouldConsiderAsContainment(ref)) {
						isHandle = false;
						while (attr != null) {
							if (attr.equals("xsi:type")) { //$NON-NLS-1$
								type = readAttributeValue();
							} else if (attr.equals("xsi:nil")) { //$NON-NLS-1$
								String val = readAttributeValue();
								if (!val.equals("true")) { //$NON-NLS-1$
									throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
								}
								isNull = true;
							} else {
								// attr is an item id or empty reference, will
								// be handled at the top of the main loop
								// after creating the object for this tag
								break;
							}
							attr = readAttributeName();
						}
					} else {
						isHandle = true;
						while (attr != null) {
							if (attr.equals("itemId")) { //$NON-NLS-1$
								String val = readAttributeValue();
								itemId = UUID.valueOf(val);
							} else if (attr.equals("stateId")) { //$NON-NLS-1$
								String val = readAttributeValue();
								stateId = UUID.valueOf(val);
							} else if (attr.equals("xsi:type")) { //$NON-NLS-1$
								type = readAttributeValue();
							} else if (attr.equals("xsi:nil")) { //$NON-NLS-1$
								String val = readAttributeValue();
								if (!val.equals("true")) { //$NON-NLS-1$
									throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
								}
								isNull = true;
							} else {
								// unexpected attribute on a handle
								throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
							}
							attr = readAttributeName();
						}
					}
					boolean doPush = false;
					Object value;
					if (isNull) {
						value = null;
						tag = readTag();
						if (!isEnd) {
							throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
						}
					} else {
						if (type != null) {
							value = createType(type);
						} else {
							EClass tcls = ref.getEReferenceType();
							tcls = ItemUtil.getTargetClassIfFacade(tcls);
							tcls = ItemUtil.getGeneralizedHelperClass(tcls, ref);
							EPackage pkg = tcls.getEPackage();
							value = pkg.getEFactoryInstance().create(tcls);
						}
						if (isHandle) {
							ItemHandle h = (ItemHandle) value;
							if (itemId != null) {
								h.setItemId(itemId);
							}
							if (stateId != null) {
								h.setStateId(stateId);
							}
							tag = readTag();
							if (!isEnd) {
								throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
							}
						} else {
							if (itemId != null) {
								((ItemHandle) value).setItemId(itemId);
							}
							doPush = true;
						}
					}
					if (ref.isMany()) {
						Object list = o.eGet(ref);
						if (list instanceof EcoreEMap) {
							((EcoreEMap) list).addUnique((Entry) value);
						} else {
							((AbstractEList) list).addUnique(value);
						}
					} else {
						o.eSet(ref, value);
					}
					if (doPush) {
						objectStack.add(o);
						o = (EObject) value;
						cls = o.eClass();
					}
				}
			}
		}
		if (readTag() != null) {
			throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL);
		}
		return o;
	}

	public EObject demarshal() throws MarshallingException {
		demarshalHeader();
		return demarshalObject();
	}

	/**
	 * 
	 * Tolerance is enabled by default, disable it by adding system variable:
	 * 
	 * <br>
	 * <code>-Dcom.ibm.team.repository.marshal.tolerance=false</code>
	 * 
	 * @return boolean indicating if namespace tolerance is enabled
	 */
	private boolean isNamespaceToleranceEnabled() {
		// While it is possible for there to be a race condition here, it
		// doesn't actually matter.
		String property = System.getProperty("com.ibm.team.repository.marshal.tolerance"); //$NON-NLS-1$
		if (property != null) {
			// Note that if the user provides a value of anything other than
			// "true"
			// it will disable tolerance.
			isNamespaceToleranceEnabled = Boolean.parseBoolean(property);
		}
		return isNamespaceToleranceEnabled;
	}

	private EPackage handleMissingPackage(String corruptName, String packageName) {
		if (!corruptedNamespaces.contains(packageName)) {
			// This is the first time, log it and set the corrupted flag.
			corruptedNamespaces.add(packageName);
			String message = NLS.bind(Messages.getServerString("WebServicesSAXXMLHandler.CorruptPackageName"), corruptName, packageName); //$NON-NLS-1$
			_logger.error(message);
		}
		EPackage p = EPackage.Registry.INSTANCE.getEPackage(packageName);
		if (p != null) {
			// We've already logged it. Add the corrupt name to the mapping with
			// the actual package to avoid a performance hit.
			EPackage.Registry.INSTANCE.put(corruptName, p);
		}
		return p;
	}
}
