/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010, 2015. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.repository.common.internal.marshal.impl;

import java.io.InputStream;
import java.io.OutputStream;
import java.io.Writer;
import java.lang.reflect.Array;
import java.nio.charset.Charset;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.xmlsoap.schemas.soap.envelope.Body;
import org.xmlsoap.schemas.soap.envelope.Envelope;
import org.xmlsoap.schemas.soap.envelope.EnvelopeFactory;
import org.xmlsoap.schemas.soap.envelope.Fault;

import com.ibm.team.repository.common.BacklevelClientException;
import com.ibm.team.repository.common.BacklevelClientRuntimeException;
import com.ibm.team.repository.common.IItem;
import com.ibm.team.repository.common.IItemHandle;
import com.ibm.team.repository.common.PermissionDeniedException;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.common.internal.marshal.MarshalFactory;
import com.ibm.team.repository.common.internal.marshal.Marshaller;
import com.ibm.team.repository.common.internal.marshal.MarshallerType;
import com.ibm.team.repository.common.internal.marshal.MarshallingException;
import com.ibm.team.repository.common.internal.marshal.MarshallingExceptionReasonCode;
import com.ibm.team.repository.common.internal.marshal.util.MarshallerUtil;
import com.ibm.team.repository.common.internal.util.IComponentElementDescriptor;
import com.ibm.team.repository.common.internal.util.ItemUtil;
import com.ibm.team.repository.common.model.ItemHandle;
import com.ibm.team.repository.common.serialize.internal.nls.Messages;
import com.ibm.team.repository.common.transport.HttpUtil.CharsetEncoding;
import com.ibm.team.repository.common.transport.internal.services.DataArg;
import com.ibm.team.repository.common.transport.internal.services.DataArgType;
import com.ibm.team.repository.common.transport.internal.services.PrimitiveDataArg;
import com.ibm.team.repository.common.transport.internal.services.Request;
import com.ibm.team.repository.common.transport.internal.services.Response;
import com.ibm.team.repository.common.transport.internal.services.ServicesFactory;
import com.ibm.team.repository.common.util.NLS;

/**
 * Class to marshal a Web Service request and response
 */
public class WebServicesMarshaller extends EObjectMarshaller {

    static private final Class ITEM_CLASS = IItem.class;

    static private final Class ITEM_HANDLE_CLASS = IItemHandle.class;

    public static final WebServicesMarshaller INSTANCE = new WebServicesMarshaller();
    
    public static final String WEB_SERVICVES_MARSHALLER_REGISTRATION_NAME = "Web Services"; //$NON-NLS-1$
    
    public static final int MARSHAL_AS_XML = 0;
    public static final int MARSHAL_AS_JSON = 1;

	private static final String EXCEPTION_WHILE_MARSHALING_EXCEPTION_MSG = "WebServicesMarshaller.ExceptionWhileMarshalingException"; //$NON-NLS-1$
	private static final String TYPE_UNKNOWN_TO_CLIENT_MSG = "WebServicesMarshaller.TypeUnknownToClient"; //$NON-NLS-1$

	@SuppressWarnings("serial")
	private static class OlderClientMarshallingException extends MarshallingException {
		private static final int OLDER_CLIENT_MARSHALLING_EXCEPTION_REASON = 1001;
		private static final MarshallingExceptionReasonCode INTERNAL_REASON_CODE =
			new MarshallingExceptionReasonCode(
				OLDER_CLIENT_MARSHALLING_EXCEPTION_REASON, "Internal to WebServicesMarshaller"); //$NON-NLS-1$
		private String unknownClassName;
		private String requiredClientVersion;

		public OlderClientMarshallingException(String message, String unknownClassName,
				String requiredClientVersion) {
			super(INTERNAL_REASON_CODE, message);
			this.unknownClassName = unknownClassName;
			this.requiredClientVersion = requiredClientVersion;
		}

		public String getUnknownClassName() {
			return unknownClassName;
		}

		public String getRequiredClientVersion() {
			return requiredClientVersion;
		}
	}

	private WebServicesMarshaller() {
        super();
        setSupportedType(WebServicesMarshaller.class);
        setMarshallerType(MarshallerType.WEB_SERVICES_LITERAL);
    }

    @Override
    public String getRegistrationName() {
        return WEB_SERVICVES_MARSHALLER_REGISTRATION_NAME;
    }

    @Override
    public boolean canMarshal(Class type) {
        return type == supportedType;
    }

    /**
	 * @param outputStream  
	 */
    protected void writeXMLHeader(OutputStream outputStream) {
        // don't write the XML header. Soap is used for transport. 
    }

    /**
     * Demarshal the input stream to Service Request object 
     * @param inputStream
     * @throws MarshallingException
     */
    public Request demarshalInputStreamToServiceRequest(InputStream inputStream) throws MarshallingException {
        return demarshalInputStreamToServiceRequest(inputStream, CharsetEncoding.UTF8.toCharset());
    }

    /**
     * Demarshal the input stream to a service request object using the specified
     * encoding.
     * 
     * @param inputStream The input stream. Must not be <code>null</code>.
     * @param charset The charset. Must not be <code>null</code>.
     * @return The request, or <code>null</code> if the input doesn't contain
     * a recognizable envelope.
     * 
     * @throws MarshallingException Thrown on error.
     */
    public Request demarshalInputStreamToServiceRequest(InputStream inputStream, Charset charset) throws MarshallingException {
        if (charset == null)
            throw new IllegalArgumentException("charset must not be null"); //$NON-NLS-1$

        Envelope envelope = (Envelope) demarshalInputStreamToObject(inputStream, charset);
        if (envelope != null) {
            return envelope.getBody().getRequest();
        }
        return null;
    }
    
    /**
     * Demarshal the input stream to service response
     * @param inputStream Input stream 
     * @throws MarshallingException
     */
    public Response demarshalInputStreamToServiceResponse(InputStream inputStream) throws MarshallingException {
        return demarshalInputStreamToServiceResponse(inputStream, CharsetEncoding.UTF8.toCharset());
    }

    /**
     * Demarshal the input stream to a service response object using the specified
     * encoding.
     * 
     * @param inputStream The input stream. Must not be <code>null</code>.
     * @param charset The charset. Must not be <code>null</code>.
     * @return The response, or <code>null</code> if the input doesn't contain
     * a recognizable envelope.
     * 
     * @throws MarshallingException Thrown on error.
     */
    public Response demarshalInputStreamToServiceResponse(InputStream inputStream, Charset charset) throws MarshallingException {
        if (charset == null)
            throw new IllegalArgumentException("charset must not be null"); //$NON-NLS-1$
        
        Envelope envelope = (Envelope) demarshalInputStreamToObject(inputStream, charset);
        if (envelope != null) {
            return envelope.getBody().getResponse();
        }
        return null;
    }

    /**
     * Marshal a service request to a Request object
     * @param methodName Method Name
     * @param parameterTypes Parameter Types
     * @param parameters Parameter Objects
     * @param marshalType One of XML or JSON
     * @param outputStream Output Stream
     * @param charset The charset encoding.
     */
    public Envelope marshalServiceRequest(String methodName, String interfaceName, 
            Class[] parameterTypes, Object[] parameters, int marshalType, 
            OutputStream outputStream, Charset charset)
            throws MarshallingException {

        if (parameterTypes.length != parameters.length) {
            throw new IllegalArgumentException("types and objects are not the same length"); //$NON-NLS-1$
        }
        Request request = ServicesFactory.eINSTANCE.createRequest();
        if (methodName != null) {
            request.setMethod(methodName);
        }
        if (interfaceName != null) {
            request.setInterface(interfaceName);
        }
        request.setMethod(methodName);

        for (int i = 0; i < parameterTypes.length; i++) {
            Object parameter = parameters[i];
            parameter = handleizeIfRequired(parameter, parameterTypes[i]);

            Marshaller marshaller = MarshalFactory.eINSTANCE.getMarshaller(parameterTypes[i]);
            if (marshaller == null) {
                if (parameter == null || EObject.class.isAssignableFrom(parameter.getClass()))
            		marshaller = MarshalFactory.eINSTANCE.getMarshaller(MarshallerType.EOBJECT_LITERAL);
                else {
                	marshaller = MarshalFactory.eINSTANCE.getMarshaller(MarshallerType.STRING_LITERAL);
                	parameter = parameter.toString();
                }
            }
            DataArg dataArg = marshaller.marshalObjectToDataArg(parameter);
            request.getParameters().add(dataArg);
        }

        // Create a Soap Envelope
        Envelope envelope = EnvelopeFactory.eINSTANCE.createEnvelope();
        Body body = EnvelopeFactory.eINSTANCE.createBody();
        body.setRequest(request);
        envelope.setBody(body);

        marshalObjectToOutputStream(envelope, outputStream, marshalType, charset);
        return envelope;
    }

    /**
     * Marshal a service request to a Request object
     * @param methodName Method Name
     * @param parameterTypes Parameter Types
     * @param parameters Parameter Objects
     * @param marshalType One of XML or JSON
     * @param outputStream Output Stream
     * @deprecated use {@link #marshalServiceRequest(String, String, Class[], Object[], int, OutputStream, Charset)
     */
    @Deprecated
    public Envelope marshalServiceRequest(String methodName, String interfaceName, 
            Class[] parameterTypes, Object[] parameters, int marshalType, 
            OutputStream outputStream)
            throws MarshallingException {

        return marshalServiceRequest(methodName, interfaceName, parameterTypes, parameters,
                marshalType, outputStream, CharsetEncoding.UTF8.toCharset());
    }

	/**
	 * Marshal the Web Service Response
	 */
	public Envelope marshalServiceResponse(String methodName, String interfaceName, 
			Class type, Object object, int marshalType, OutputStream outputStream,
			Charset charset) throws MarshallingException {
		return marshalServiceResponse(methodName, interfaceName, type, object,
				marshalType, outputStream, charset, null, false);
	}

	/**
	 * Marshal the Web Service Response 
	 */
	public Envelope marshalServiceResponse(String methodName, String interfaceName, 
			Class type, Object object, int marshalType, OutputStream outputStream,
			Charset charset, Map<String, String> clientConfig, boolean canThrowTRE)
			throws MarshallingException {

    	Response response = ServicesFactory.eINSTANCE.createResponse();
        if (methodName != null) {
            response.setMethod(methodName);
        }
        if (interfaceName != null) {
            response.setInterface(interfaceName);
        }

        // Create Data arg for object
        DataArg dataArg = null;
        if (object != null) {
            Marshaller marshaller = MarshalFactory.eINSTANCE.getMarshaller(type);
            if (marshaller == null) {
                marshaller = MarshalFactory.eINSTANCE.getMarshaller(
                        MarshallerType.EOBJECT_LITERAL);
            }
            dataArg = marshaller.marshalObjectToDataArg(object);
            response.setReturnValue(dataArg);
        } else {
            dataArg = ServicesFactory.eINSTANCE.createNullDataArg();
            dataArg.setType(DataArgType.NULL_LITERAL);
            response.setReturnValue(dataArg);
        }

        Body body = EnvelopeFactory.eINSTANCE.createBody();
        body.setResponse(response);
        Envelope envelope = EnvelopeFactory.eINSTANCE.createEnvelope();
        envelope.setBody(body);

        try {
			marshalObjectToOutputStream(envelope, outputStream, marshalType, charset, clientConfig);
		} catch (OlderClientMarshallingException marshallingException) {
			// We were about to marshal a type that the client does not know about.
			// Return an exception instead (TRE if the method allows it).  This will
			// be handled better in existing backlevel clients.
			Exception ex;
			if (canThrowTRE)
				ex = new BacklevelClientException(marshallingException.getMessage(),
						marshallingException.getUnknownClassName(),
						marshallingException.getRequiredClientVersion());
			else
				ex = new BacklevelClientRuntimeException(marshallingException.getMessage(), 
						marshallingException.getUnknownClassName(),
						marshallingException.getRequiredClientVersion());
			return marshalFaultServiceResponse(methodName, interfaceName, ex,
					marshalType, outputStream, charset, clientConfig);
		}
        return envelope;
    }

	/**
     * Marshal the Web Service Response
     * @deprecated use {@link #marshalServiceResponse(String, String, Class, Object, int, OutputStream, Charset)} 
     */
    @Deprecated
    public Envelope marshalServiceResponse(String methodName, String interfaceName, Class type,
            Object object, int marshalType, OutputStream outputStream) throws MarshallingException {
        
        return marshalServiceResponse(methodName, interfaceName, type, object,
                marshalType, outputStream, CharsetEncoding.UTF8.toCharset());
    }

	/**
	 * Marshal Fault service response
	 */
	public Envelope marshalFaultServiceResponse(String methodName, 
			String interfaceName, Throwable exception, int marshalType, 
			OutputStream outputStream, Charset charset) {
		return marshalFaultServiceResponse(methodName, interfaceName, exception,
				marshalType, outputStream, charset, null);
	}

    /**
     * Marshal Fault service response
     */
    public Envelope marshalFaultServiceResponse(String methodName, 
            String interfaceName, Throwable exception, int marshalType, 
            OutputStream outputStream, Charset charset,
            Map<String, String> clientConfig) {

        Response response = ServicesFactory.eINSTANCE.createResponse();
        if (methodName != null) {
            response.setMethod(methodName);
        }
        if (interfaceName != null) {
            response.setInterface(interfaceName);
        }

        // Set Fault
        Fault fault = EnvelopeFactory.eINSTANCE.createFault();
        fault.setFaultstring(MarshallerUtil.encodeExceptions(exception));

        // populate fault with exception data
        populateExtraDataInfo(fault, exception);
        
        Body body = EnvelopeFactory.eINSTANCE.createBody();
        body.setResponse(response);
        body.setFault(fault);

        Envelope envelope = EnvelopeFactory.eINSTANCE.createEnvelope();
        envelope.setBody(body);

        try {
            marshalObjectToOutputStream(envelope, outputStream, marshalType, charset, clientConfig);
		} catch (OlderClientMarshallingException marshallingException) {
			// We were about to marshal a type that the client does not know about.
			// Return an exception instead (TRE if the method allows it).  This will
			// be handled better in existing backlevel clients.
			String unknownClassName = marshallingException.getUnknownClassName();
			if (unknownClassName.contains("BacklevelClientException") || //$NON-NLS-1$
					unknownClassName.contains("BacklevelClientRuntimeException")) { //$NON-NLS-1$
				// Make sure to avoid infinite loop if somehow we have a client
				// that doesn't know about these exception classes
				return envelope;
			}
			Exception ex;
			String pattern = Messages.getClientString(EXCEPTION_WHILE_MARSHALING_EXCEPTION_MSG);
			String message = NLS.bind(pattern, exception.getMessage(), marshallingException.getMessage());
			if (exception instanceof TeamRepositoryException)
				ex = new BacklevelClientException(message, unknownClassName,
						marshallingException.getRequiredClientVersion());
			else
				// We never attempt to marshal data contained in non TRE's, so
				// this should never happen...
				ex = new BacklevelClientRuntimeException(message, unknownClassName,
						marshallingException.getRequiredClientVersion());
			return marshalFaultServiceResponse(methodName, interfaceName, ex,
					marshalType, outputStream, charset, clientConfig);
        } catch (MarshallingException e) {
            // ignore
        }
        return envelope;
    }

	@Override
	protected void marshalSingleValuedReference(EObject object,
			EReference reference, Writer writer, Map nsUriToNsPrefixMap,
			Map<String, String> clientConfig, int depth) throws MarshallingException {
		EObject objectToMarshal = (EObject) object.eGet(reference);
		checkForTypeUnknownToClient(objectToMarshal, clientConfig);
		super.marshalSingleValuedReference(object, reference, writer,
				nsUriToNsPrefixMap, clientConfig, depth);
	}

	@Override
	protected void marshalMultiValuedReference(EObject object,
			EReference reference, Writer writer, Map nsUriToNsPrefixMap,
			Map<String, String> clientConfig, int depth) throws MarshallingException {
		List<EObject> objectsToMarshal = (List<EObject>) object.eGet(reference);
		if (objectsToMarshal != null) {
			for (EObject objectToMarshal : objectsToMarshal) {
				checkForTypeUnknownToClient(objectToMarshal, clientConfig);
			}
		}
		super.marshalMultiValuedReference(object, reference, writer,
				nsUriToNsPrefixMap, clientConfig, depth);
	}

	// Throws an OlderClientMarshallingException if objectToMarshal is not
	// known for the given clientConfig
	private void checkForTypeUnknownToClient(EObject objectToMarshal, Map<String, String> clientConfig)
			throws MarshallingException {
		if (objectToMarshal != null) {
			String requiredVersion = requiredClientUpgradeVersion(objectToMarshal.eClass(), clientConfig);
			if (requiredVersion != null) {
				String msgPattern = Messages.getClientString(TYPE_UNKNOWN_TO_CLIENT_MSG);
				String msg = NLS.bind(msgPattern, requiredVersion);
				throw new OlderClientMarshallingException(msg,
						objectToMarshal.eClass().getName(), requiredVersion);
			}
		}
	}

	// Returns the earliest version that the client needs to know about the
	// given class.  Returns null if the client already knows about it.
	private String requiredClientUpgradeVersion(EClass eClass, Map<String, String> clientConfig) {
		String nsURI = eClass.getEPackage().getNsURI();
		IComponentElementDescriptor desc = getComponentDescriptor(nsURI);
		String clientVersion = getClientVersionForComponentIfDifferent(desc, clientConfig);
		if (clientVersion == null)
			return null;
		else
			return desc.getModelElementDescriptor(nsURI).requiredClientUpgradeVersion(
					eClass.getInstanceClassName(), clientVersion);
	}

	private void populateExtraDataInfo(Fault fault, Throwable throwable) {
        if (throwable instanceof PermissionDeniedException) {
            PermissionDeniedException pde = (PermissionDeniedException) throwable;
            Object processReportData = pde.getProcessReportData();
            if (processReportData != null && processReportData instanceof EObject) {
                fault.getExtraData().put(PermissionDeniedException.PROCESS_REPORT_DATA, (EObject) processReportData);
            }
        }
        
        if (throwable instanceof TeamRepositoryException) {
            TeamRepositoryException tre = (TeamRepositoryException) throwable;
            Object exceptionData = tre.getData();
            if (exceptionData != null && exceptionData instanceof EObject) {
                fault.getExtraData().put(TeamRepositoryException.EXCEPTION_DATA, (EObject)exceptionData);
            }
        } 
    }

	private void marshalObjectToOutputStream(Envelope envelope, OutputStream outputStream, 
			int marshalType, Charset charset) throws MarshallingException {
		marshalObjectToOutputStream(envelope, outputStream, marshalType,
				charset, null);
	}

	private void marshalObjectToOutputStream(Envelope envelope, OutputStream outputStream, 
			int marshalType, Charset charset, Map<String, String> clientConfig) throws MarshallingException {
        switch (marshalType) {
        case MARSHAL_AS_XML:
            marshalObjectToOutputStream(envelope, outputStream, charset, clientConfig);
            break;
        case MARSHAL_AS_JSON:
            EObjectJSONMarshaller eObjectJSONMarshaller = 
                (EObjectJSONMarshaller) MarshalFactory.eINSTANCE.getMarshaller(
                        MarshallerType.JSON_LITERAL);
            eObjectJSONMarshaller.marshalObjectToOutputStream(envelope, outputStream, charset);
            break;
        default:
        	throw new IllegalArgumentException("Unexpected value for marshalType: " + marshalType); //$NON-NLS-1$
        }
    }

    /**
     * Get the request parameters given a request object and parameter types
     */
    public Object[] getRequestParameters(Request request, Class[] parameterTypes) {
        return getMethodParameters(request.getParameters(), parameterTypes);
    }

    public Object getResponseReturnValue(Response response, Class parameterType) {
        return getReturnValue(response.getReturnValue(), parameterType);
    }

    /** Demarshal a JSON literal to get the given return value. 
     */
    public Object demarshalInputStreamToReturnValue(InputStream stream, Charset charset, Class type, int marshalType) throws MarshallingException {
        switch (marshalType) {
        case MARSHAL_AS_JSON:
            EObjectJSONMarshaller eObjectJSONMarshaller = 
                (EObjectJSONMarshaller) MarshalFactory.eINSTANCE.getMarshaller(
                        MarshallerType.JSON_LITERAL);
            
            Object ret = eObjectJSONMarshaller.demarshalInputStreamToObject(stream, charset);
            if (! (ret instanceof Envelope)) {
            	throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL, "Missing envelope"); //$NON-NLS-1$
			}
            Envelope envelope = (Envelope)ret;
            
            Body body = envelope.getBody();
            if (body == null) {
            	throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL, "Missing envelope body"); //$NON-NLS-1$
            }
            
            Response response = body.getResponse();
            if (response == null) {
            	throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL, "Missing body response"); //$NON-NLS-1$
            }

        	return getResponseReturnValue(response, type);
        }

        throw new IllegalArgumentException("Only know how to handle JSON"); //$NON-NLS-1$
    }
    
    /**
     * Get the method parameters
     */
    private Object[] getMethodParameters(List dataArgParameters, Class[] parameterTypes) {
        Object[] parameters = new Object[parameterTypes.length];
        for (int i = 0; i < dataArgParameters.size(); i++) {
            DataArg dataArg = (DataArg) dataArgParameters.get(i);
            parameters[i] = getMethodParameter(dataArg, parameterTypes[i]);
        }
        return parameters;
    }

    private Object getMethodParameter(DataArg dataArg, Class parameterType) {
        if (dataArg.getType().equals(DataArgType.NULL_LITERAL)) {
            return null;
        }
        Marshaller marshaller = MarshalFactory.eINSTANCE.getMarshaller(parameterType);
        if (marshaller == null) {
        	if (dataArg instanceof PrimitiveDataArg)
        		marshaller = MarshalFactory.eINSTANCE.getMarshaller(MarshallerType.STRING_LITERAL);
        	else 
        		marshaller = MarshalFactory.eINSTANCE.getMarshaller(
                    MarshallerType.EOBJECT_LITERAL);
        }
        return marshaller.demarshalDataArgToObject(dataArg, parameterType);
    }

    private Object getReturnValue(DataArg dataArg, Class parameterType) {

        if (dataArg.getType().equals(DataArgType.NULL_LITERAL)) {
            return null;
        }
        Marshaller marshaller = MarshalFactory.eINSTANCE.getMarshaller(parameterType);
        if (marshaller == null) {
            marshaller = MarshalFactory.eINSTANCE.getMarshaller(
                    MarshallerType.EOBJECT_LITERAL);
        }
        return marshaller.demarshalDataArgToObject(dataArg, parameterType);
    }

    /**
     * Return the argument, handle-ized, if so spec'd by the parameter type
     */
    private Object handleizeIfRequired(Object parameter, Class parameterType) {

        // if parameter is null, just return it.
        if (null == parameter)
            return null;

        if (parameterType.isArray()) {
            return handleizeArrayIfRequired(parameter, parameterType);
        }

        // if parameter type isn't a handle, just return it.
        if (!ITEM_HANDLE_CLASS.isAssignableFrom(parameterType))
            return parameter;

        // if parameter type is an item, just return it.
        if (ITEM_CLASS.isAssignableFrom(parameterType))
            return parameter;

        // at this point, the parameter type is a handle but not an item

        // check to see if we have an item or handle
        ItemHandle itemHandle = (ItemHandle) parameter;

        // if the parameter is a handle, return it
        // note that hasFullState is a misnomer:
        // it will return true for partial states created by AccessProfiles
        if (!itemHandle.hasFullState())
            return parameter;

        // the parameter is an item, but the parameter type is a handle,
        // handle-ize it
        return ItemUtil.newItemHandle(itemHandle, true);
    }

    /**
     * Return the argument elements, handle-ized, if so spec'd by the parameter type.
     * We've already determined the argument 'parameter' is an array
     */
    private Object handleizeArrayIfRequired(Object parameter, Class parameterType) {
        parameterType = parameterType.getComponentType();

        // if parameter type isn't a handle, just return it.
        if (!ITEM_HANDLE_CLASS.isAssignableFrom(parameterType))
            return parameter;

        // if parameter type is an item, just return it.
        if (ITEM_CLASS.isAssignableFrom(parameterType))
            return parameter;

        int length = Array.getLength(parameter);
        if (0 == length)
            return parameter;

        Object result = Array.newInstance(parameterType, length);

        // need to look at each item
        for (int i = 0; i < length; i++) {
            Object object = Array.get(parameter, i);

            object = handleizeIfRequired(object, parameterType);

            Array.set(result, i, object);
        }

        return result;
    }

}