/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010, 2017. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.repository.common.internal.marshal.util;

import java.io.IOException;
import java.io.LineNumberReader;
import java.io.PrintWriter;
import java.io.StringReader;
import java.io.StringWriter;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.StringTokenizer;

import org.eclipse.emf.common.util.EMap;
import org.eclipse.emf.ecore.EAnnotation;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.EcorePackage;
import org.xmlsoap.schemas.soap.envelope.Fault;

import com.ibm.team.repository.common.PermissionDeniedException;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.common.UUID;
import com.ibm.team.repository.common.internal.marshal.Marshaller;
import com.ibm.team.repository.common.internal.marshal.MarshallerType;
import com.ibm.team.repository.common.internal.util.ItemUtil;
import com.ibm.team.repository.common.serialize.IStackAdjuster;
import com.ibm.team.repository.common.serialize.LoggedException;
import com.ibm.team.repository.common.serialize.StackUtil;
import com.ibm.team.repository.common.serialize.internal.nls.Messages;
import com.ibm.team.repository.common.util.HashUtil;
import com.ibm.team.repository.common.util.NLS;

public class MarshallerUtil {
	/* By default, exceptions are serialized.  If we want to meet security
	 * requirements, we have to be configured to be secure by default.  But
	 * until that adoption is accepted, we preserve the current behavior by
	 * default.
	 * See com.ibm.team.jfs.app.error.ReportFactory.NEVER_SERIALIZE_EXCEPTION_PROPERTY 
	 **/

    /**
     * 
     */
    public static final String CAUSE = "cause:"; //$NON-NLS-1$
    /**
     * 
     */
    private static final String NEXT = "next:"; //$NON-NLS-1$
    /**
     * 
     */
    private static final String ERROR_CODE = "errorCode:"; //$NON-NLS-1$
    /**
     * 
     */
    private static final String SQLSTATE = "SQLState:"; //$NON-NLS-1$
    public static final String MARSHALLER_ANNOTATION = "http:///com/ibm/team/core/marshal"; //$NON-NLS-1$
    public static final String MARSHALLER_CONTAINMENT_ANNOTATION_KEY = "containment"; //$NON-NLS-1$
    
    private static final String STACK_TRACE_TOKEN = "Stack Trace follows:"; //$NON-NLS-1$

    public static boolean isComplexMarshaller(Marshaller marshaller) {
        return ((marshaller.getMarshallerType().getValue() & MarshallerType.COMPLEX_LITERAL.getValue()) != 0);
    }
    
    /**
     * Should the reference a containment?
     */
    public static boolean shouldConsiderAsContainment(EReference reference) {
        if(reference.isContainment()) {
            return true;
        }
        // Don't inline if it's ecore metadata
        if (reference.getEReferenceType().getEPackage() == EcorePackage.eINSTANCE && 
                reference.getEReferenceType() != EcorePackage.eINSTANCE.getEObject()) {
            return false;
        }
        String marshallerAnnotation = getAnnotation(reference, MARSHALLER_ANNOTATION, MARSHALLER_CONTAINMENT_ANNOTATION_KEY);
        boolean containment = (marshallerAnnotation != null && marshallerAnnotation.equals(String.valueOf(true)));
        if(!containment) {
            // this is to support shared reference contributor contained in a LoginResult class
            return !ItemUtil.isHandle(reference);
        }
        return containment; 
    }
    
    /**
     * Get Annotation given the source and key name
     * @param structuralFeature
     * @param sourceAnnotationName
     * @param detailsKey
     */
    public static String getAnnotation(EStructuralFeature structuralFeature, String sourceAnnotationName, String detailsKey) {
        EAnnotation annotation = structuralFeature.getEAnnotation(sourceAnnotationName);
        if(annotation != null) {
            return annotation.getDetails().get(detailsKey);
        }
        return null;
    }
    
    /**
     * Utility method to determine whether or not a feature should be marshaled. Features are marshaled if they are
     * neither transient nor volatile, and they have a non-default value.
     * @param object the object whose feature we are checking for marshaling
     * @param feature the feature
     * @return true if the feature should be  marshaled.
     */
    public static boolean shouldBeMarshalled(EObject object, EStructuralFeature feature) {
        // Transient or volatile should not be marshaled.
        if (feature.isTransient()) {
            return false;
        }

		// disabled provisional feature should not be marshaled.
		if (ItemUtil.isProvisionalAndDisabled(feature))
			return false;

        // don't marshal if the feature is not set
        return object.eIsSet(feature);
    }

    /**
     * Encodes an Exception (and any nested Exceptions) into a string
     * for marshalling.  Records each exception's class name, message
     * and stack trace.
     * <p>
     * The structure of the string is as follows:
     * <p>
     * <pre>
     * 1:   END-OF-MESSAGE-MARKER\n
     * 2:   Fully-qualified-name-of-exception-class\n
     *      Exception message\n
     *      ...
     *      Last line of exception messageEND-OF-MESSAGE-MARKER\n
     *      Number-of-stack-trace-elements\n
     *      First-stack-trace-element
     *      ...
     *      Last-stack-trace-element
     * </pre>
     * <p>
     * If there are nested exceptions, the structure repeats starting
     * at 2 above.  The end of message marker is a UUID value (hence
     * extremely unlikely to also appear in the value of an exception
     * message).
     * </p>
     * 
     * @param exception The exception to be encoded.
     * @return Encoded exception
     * @see MarshallerUtil#decodeExceptions(String, ClassLoader)
     */
    public static String encodeExceptions(Throwable exception) {
        String marker = UUID.generate().getUuidValue();
        StringWriter result = new StringWriter(1000);
        PrintWriter output = new PrintWriter(result, true);

        output.println(marker);
        boolean isLogged = false;
        if (exception instanceof LoggedException) {
        	exception = exception.getCause();
        	isLogged = true;
        }
        encodeExceptions(marker, output, exception, isLogged);

        output.close();
        String s = result.toString();
        return s;
    }

    /**
     * Encodes an Exception (and any nested Exceptions) into a string
     * for marshalling.
     */
    private static void encodeExceptions(String marker, PrintWriter output, Throwable exception, boolean isLogged) {
        // Ideally would use introspection to extract all useful info from
        // exception (i.e. SQLException.getErrorCode()).

        Class exClass = exception.getClass();
        while (true) {
            String className = exClass.getName();
            output.print(className);
            if (className.startsWith("java.lang.")) //$NON-NLS-1$
                break;
            output.print('>');
            exClass = exClass.getSuperclass();
        }
        output.println();

        String message = constructExceptionMessage(exception, isLogged);
        if (message != null) {
            output.print('>');  // Differentiates null from empty string
            output.print(message);
        }
        output.println(marker);

		StackTraceElement[] traceElements;
		if (shouldAvoidSerializingExceptions()) {
			traceElements = new StackTraceElement[0];
		} else {
			traceElements = exception.getStackTrace();
		}
        output.println(traceElements.length);

        for (StackTraceElement element : traceElements) {
            output.print(element.getClassName());
            output.print(',');
            output.print(element.getMethodName());
            output.print(',');
            output.print(element.getFileName());
            output.print(',');
            output.print(element.getLineNumber());
            output.println();
        }

        /*
         * Special support for SQLException (because we've been losing
         * detail during marshalling, and so users haven't been able to
         * submit sufficiently detailed bug reports).
         * 
         * Ideally we'd have an extension point allowing components to
         * provide a form of serialization support for exceptions/objects;
         * e.g. a "view" of an exception as a Map<String, Object>, with the
         * serializer (e.g. this code) taking care of encoding the key and
         * value.
         */

        if (exception instanceof java.sql.SQLException) {
            java.sql.SQLException se = (java.sql.SQLException) exception;

            output.print(SQLSTATE);
            if (se.getSQLState() != null)
                output.print(se.getSQLState());
            output.println();

            output.print(ERROR_CODE);
            output.println(se.getErrorCode());

            se = se.getNextException();

            if (se == null)
                return;

            output.println(NEXT);
            // nested exception, set isLogged to false
            encodeExceptions(marker, output, se, false);
            return;
        }
        else {
            Throwable cause = exception.getCause();
            if (cause == null)
                return;

            output.println(CAUSE);
            // nested exception, set isLogged to false
            encodeExceptions(marker, output, cause, false);
            return;
        }
    }

	/**
	 * Get the (possibly augmented) message from the exception.
	 * If we are stripping stack traces from the exceptions we're marshalling
	 * to the client, append a snippet to the first/topmost exception telling
	 * the user to look in the server logs for the specified marker for
	 * the full stack trace.
	 */
    private static String constructExceptionMessage(Throwable ex, boolean isLogged) {
        if (shouldAvoidSerializingExceptions() && isLogged) {
        	// Note: We're assuming this exception was logged by TeamServletLoggingHelper.logWarningWithClientDetails()
        	// which uses the exception stack's SHA-1 hash string as a correlation id, so we use the same id.
        	// TODO: Find a better way that doesn't make assumptions.
        	String correlationId = HashUtil.getStacktraceHash(ex);
        	if (ex.getMessage() == null) {
        		return NLS.bind(Messages.getClientString( "MarshallerUtil.DetailsInLogFile1"), correlationId); //$NON-NLS-1$
        	} else {
        		return NLS.bind(Messages.getClientString( "MarshallerUtil.DetailsInLogFile2"), ex.getMessage(), correlationId); //$NON-NLS-1$
        	}
        } else {
        	return ex.getMessage();
        }
	}

	/**
     * Decodes a Fault produced by
     * {@link #encodeExceptions(Throwable)} into a Throwable
     * and its nested Throwables, if any.
     * <p>
     * Any exceptions that can't be constructed (for whatever reason)
     * will be replaced with a RuntimeException with a message 
     * indicating the original exception type, the reason
     * it couldn't be constructed, and the original exception message.
     * 
     * @param fault
     *              A fault of the format produced by
     *              {@link #encodeExceptions(Throwable)}
     * @param classLoader
     *              ClassLoader to be used when attempting to create
     *              exceptions.
     * @return The Throwable encoded in <code>encodedException</code>
     */
    public static Throwable decodeFault(Fault fault, final ClassLoader classLoader) {
        Throwable throwable = decodeExceptions(fault.getFaultstring(), classLoader);
        if (throwable == null) {
            return null;
        }

        EMap<String, EObject> extraData = fault.getExtraData();
        populateExtraData(throwable, extraData);
        
        return throwable;
    }
    
    private static void populateExtraData(Throwable throwable, EMap<String, EObject> extraData) {
        if (extraData == null || extraData.isEmpty()) {
            return;
        }
        
        if (throwable instanceof PermissionDeniedException) {
            EObject value = extraData.get(PermissionDeniedException.PROCESS_REPORT_DATA);
            if (value != null) {
                ((PermissionDeniedException) throwable).setProcessReportData(value);
            }
        }
        
        if (throwable instanceof TeamRepositoryException) {
            EObject value = extraData.get(TeamRepositoryException.EXCEPTION_DATA);
            if (value != null) {
                ((TeamRepositoryException) throwable).setData(value);
            }
        } 
        
        // if the throwable is not a team repository exception, we don't have a
        // place holder to store the exception data. 
    }

    public static Throwable decodeExceptions(final String encodedException, final ClassLoader classLoader) {
    	return decodeExceptions(encodedException, classLoader, null);
    }
    
    /**
     * Decodes a String produced by
     * {@link #encodeExceptions(Throwable)} into a Throwable
     * and its nested Throwables, if any.
     * <p>
     * Any exceptions that can't be constructed (for whatever reason)
     * will be replaced with a RuntimeException with a message 
     * indicating the original exception type, the reason
     * it couldn't be constructed, and the original exception message.
     * 
     * @param encodedException
     *              A string of the format produced by
     *              {@link #encodeExceptions(Throwable)}
     * @param classLoader
     *              ClassLoader to be used when attempting to create
     *              exceptions.
     * @return The Throwable encoded in <code>encodedException</code>
     */
    public static Throwable decodeExceptions(final String encodedException, final ClassLoader classLoader, IStackAdjuster adj) {
        StringReader sr = new StringReader(encodedException);
        LineNumberReader input = new LineNumberReader(sr);
        Throwable throwable = null;

        // Get the current stack trace, for combining with the remote stack
        // trace (otherwise the local trace would be missing when printing
        // the exception chain).

        StackTraceElement[] localStackTrace = new Exception().fillInStackTrace().getStackTrace();

        try {
            // Read the string that is used to mark the end of
            // exception messages.
            String marker = input.readLine();

            if (marker == null)
                unparsable("Unable to read marker", input, encodedException); //$NON-NLS-1$

            throwable = decodeExceptions(marker, input, classLoader, localStackTrace, adj, encodedException);
        } catch (IOException ex) {
            unparsable(null, input, encodedException);
        }

        return throwable;
    }

    /**
     * Recursively decode the input into Throwables.
     * 
     * @param marker
     * @param input
     * @param localStackTrace 
     * @param classLoader 
     * @param encodedException
     * @return
     */
    private static Throwable decodeExceptions(String marker,
            LineNumberReader input, ClassLoader classLoader,
            StackTraceElement[] localStackTrace, IStackAdjuster adj,
            String encodedException) throws IOException {

        String typeHierarchy = input.readLine();
        if (typeHierarchy == null)
            return null;    // End of input

        String message = extractExceptionMessage(marker, input);
        StackTraceElement[] elems = extractStackTrace(input, localStackTrace, adj);

        String SQLState = null;
        int errorCode = 0;
        Throwable cause = null;
        SQLException next = null;
        String line;

        while ((line = input.readLine()) != null) {
            if (line.startsWith(SQLSTATE)) {
                SQLState = line.substring(SQLSTATE.length());
                continue;
            }
            else if (line.startsWith(ERROR_CODE)) {
                String tmp = line.substring(ERROR_CODE.length());
                errorCode = Integer.parseInt(tmp);
                continue;
            }

            if (line.startsWith(NEXT)) {
                try {
                    Throwable tmp = decodeExceptions(
                            marker, input, classLoader,
                            localStackTrace, adj, encodedException);
                    next = (SQLException) tmp;
                } catch (StackOverflowError error) {
                    /*
                     * Server-side exception's are too deeply nested to recreate
                     * (with this algorithm).  See bug 29325.
                     */
                }
                break;
            }
            else if (line.startsWith(CAUSE)) {
                try {
                    cause = decodeExceptions(
                            marker, input, classLoader,
                            localStackTrace, adj, encodedException);
                } catch (StackOverflowError error) {
                    /*
                     * Server-side exception's are too deeply nested to recreate
                     * (with this algorithm).  See bug 29325.
                     */
                    cause = error;
                }
                break;
            }
            else {
                // This spot shouldn't be reachable, assuming that the client
                // and server are compatible.
                break;
            }
        }

        /*
         * If we have a SQLState or a next, then this must be a SQLException
         * (or a sub-class).  Let's jump directly to constructing a SQLException
         * so that we can call the correct constructor.
         * Note that we're not providing special support for BatchUpdateException
         * or SQLWarning.
         */

        if (SQLState != null || next != null) {
            SQLException se = new SQLException(message, SQLState, errorCode, next);
            se.setStackTrace(elems);
            return se;
        }
        
        /*
         * Find the closest approximation we have the exception that was thrown
         * on the server.  We encode not just the name of the exception class,
         * but that of its base classes, separated by ">".
         */

        String[] typeNames = typeHierarchy.split(">"); //$NON-NLS-1$
        Throwable exception = null;
        String backupMessage = 
        		NLS.bind( Messages.getClientString( "MarshallerUtil.RemoteExceptionWas" ), typeNames[0], message ); //$NON-NLS-1$

        for (int i = 0; i < typeNames.length; i++) {
            String type = typeNames[i];
            try {
                exception = constructException(
                        type,
                        i == 0 ? message : backupMessage,
                        classLoader,
                        elems,
                        cause);
                if (exception == null)
                    continue;
                return exception;
            }
            catch (ClassNotFoundException e) {
                // Ignore, try the next type.
                continue;
            }
        }

        exception = new Throwable(backupMessage, cause);
        exception.setStackTrace(elems);
        return exception;
    }

    /**
     * @param input
     * @param localStackTrace
     * @return
     * @throws IOException
     */
    private static StackTraceElement[] extractStackTrace(
            LineNumberReader input, StackTraceElement[] localStackTrace, IStackAdjuster adj)
            throws IOException {

        // How many StackTraceElements?

        String line = input.readLine();
        if (line == null)
        	throw new IOException("Invalid input"); //$NON-NLS-1$
        int numElements = Integer.parseInt(line);
        StackTraceElement[] elems = new StackTraceElement[numElements + localStackTrace.length];

        // Read each element in the stack trace.
        
        for (int ndx = 0; ndx < numElements; ndx++) {
            line = input.readLine();
            if (line == null)
            	throw new IOException("Invalid input"); //$NON-NLS-1$
            String parts[] = line.split(","); //$NON-NLS-1$
            String className = parts[0];
            String methodName = parts[1];
            String fileName = parts[2];
            String lineNumber = parts[3];

            StackTraceElement elem = new StackTraceElement(
                    className, methodName, fileName,
                    Integer.parseInt(lineNumber));

            elems[ndx] = elem;
        }

        // Add the local stack trace elements

        for (int ndx = 0; ndx < localStackTrace.length; ndx++) {
            elems[numElements + ndx] = localStackTrace[ndx];
        }
        
        if (adj != null) {
        	elems = adj.adjustFrames(elems, numElements);
        }
        
        return elems;
    }

    /**
     * Get the exception message, which is possibly on several lines,
     * and is terminated by marker (with no additional newline added).
     * 
     * @param marker The string that appears at the end of the message
     * @param input
     * @return
     * @throws IOException
     */
    private static String extractExceptionMessage(String marker,
            LineNumberReader input) throws IOException {

        StringBuilder msg = new StringBuilder();

        while (true) {
            String line = input.readLine();
            if (line == null)
            	throw new IOException("Invalid input"); //$NON-NLS-1$

            if (msg.length() > 0)
                msg.append('\n');

            int indexOf = line.indexOf(marker);

            if (indexOf == 0) {
                // Original exception message ended with a newline.
                break;
            }
            else if (indexOf > 0) {
                // Original exception message did not end with a newline.
                msg.append(line, 0, indexOf);
                break;
            }

            msg.append(line);
        }

        if (msg.length() == 0)
            return null;

        msg.deleteCharAt(0);    // Delete the '>' at the start
        String message = msg.toString();
        return message;
    }

    private static void unparsable(String msg, LineNumberReader input, String encodedException) {
        if (msg == null)
            msg = "Unparsable encoded exception"; //$NON-NLS-1$
        msg += "\nAt line " + input.getLineNumber() + "\nEncoded exception: " + encodedException; //$NON-NLS-1$ //$NON-NLS-2$
        throw new RuntimeException(msg);
    }

    /**
     * Build the specified exception. 
     * @param exceptionClassName
     * @param message
     * @param classLoader
     * @param elems
     * @return
     * @throws ClassNotFoundException 
     */
    private static Throwable constructException(
            String exceptionType,
            String exceptionMesg,
            ClassLoader classLoader,
            StackTraceElement[] elems,
            Throwable cause) throws ClassNotFoundException {

        Class exceptionClass = Class.forName(exceptionType, true, classLoader);

        /* 
         * Pick the best constructor possible.  If there is a message,
         * we must set it; if we aren't able to do so, we'd prefer to
         * create an instance of a base class (e.g. Exception) that
         * does take a message, rather than lose the message.
         */

        Throwable result = null;

        do {
            result = constructException(exceptionClass,
                    new Class[] { String.class, Throwable.class },
                    new Object[] { exceptionMesg, cause });

            if (result != null)
                break;

            result = constructException(exceptionClass,
                    new Class[] { String.class },
                    new Object[] { exceptionMesg });

            if (result != null)
                break;

            if (exceptionMesg != null)
                return null;    // Find a base class

            result = constructException(exceptionClass,
                        new Class[] { Throwable.class },
                        new Object[] { cause });
            if (result != null)
                break;

            result = constructException(exceptionClass,
                    new Class[] { },
                    new Object[] { });

            if (result != null)
                break;
            return null;
        } while (false);

        if (cause != result.getCause()) {
            try {
                result.initCause(cause);
            } catch (RuntimeException e) {
                //somehow the cause is already initialized??
                cause = null;
            }
        }

        if (elems != null && elems.length > 0) {
            result.setStackTrace(elems);
        }

        return result;
    }

    /**
     * Attempts to execute the constructor with the specified parameters.
     * 
     * @param exceptionClass
     * @param parameterTypes
     * @param parameterValues
     * @return the new Exception, if successful, else null.
     */
    private static Throwable constructException(
            Class exceptionClass,
            Class[] parameterTypes,
            Object[] parameterValues) {

        Constructor exceptionConstructor = null;
        Throwable result = null;

        @SuppressWarnings({"unused"}) 
        Exception e = null;
        try {
            exceptionConstructor =
                exceptionClass.getConstructor(parameterTypes);
            
            result = (Throwable) exceptionConstructor.newInstance(parameterValues);
            return result;
        } catch (SecurityException ex) {
            e = ex;
        } catch (NoSuchMethodException ex) {
            e = ex;
        } catch (IllegalArgumentException ex) {
            e = ex;
        } catch (InstantiationException ex) {
            e = ex;
        } catch (IllegalAccessException ex) {
            e = ex;
        } catch (InvocationTargetException ex) {
            e = ex;
        } catch (RuntimeException ex) {
            e = ex; // E.g. IllegalArgumentException, or NullPointerException
        }

        return null;
    }
    
    /**
     * Encodes the stack trace in the provided exception into a string based format that
     * is safe to be marshalled.
     *
     * [Note: the stack trace cannot be encoded in a binary format because the Jazz
     * object marshaller can't handle binary data embedded in objects. See {@link XMLEscapeUtil}
     * for more details.]
     * 
     * @param exception the exception from which to encode the stack trace
     * @return the encoded stacktrace
     * @see #decodeStackTrace(String)
     */
	public static String encodeStackTrace(Throwable exception) {
		StringBuffer buffer = new StringBuffer();
			StackTraceElement[] traceElements = exception.getStackTrace();
			for (StackTraceElement element : traceElements) {
				String encoded = 
					nonNull(element.getClassName()) + ',' + 
					nonNull(element.getMethodName()) + ',' + 
					nonNull(element.getFileName()) + ',' + 
					element.getLineNumber() + ',';
				buffer.append(encoded);
				
			}
		return STACK_TRACE_TOKEN + buffer.toString();
	}
	
	/**
	 *  Decodes a stack trace encoded from {@link #encodeStackTrace(Throwable)} and returns an
	 *  exception with it's stack trace set to the decoded trace.
	 *  
	 * @param encodedStackTrace a buffer containing an encoded stacktrace
	 * @return an exception with the decoded stacktrace or <code>null</code> if the
	 * stacktrace cannot be decoded.
	 */
	public static Throwable decodeStackTrace(String encodedStackTrace) {
        Throwable stackTraceException = null;
        int stackTraceTextIndex = encodedStackTrace.indexOf(STACK_TRACE_TOKEN);
        String stackTraceText = null;
        if (stackTraceTextIndex != -1) {
			stackTraceText = encodedStackTrace.substring(stackTraceTextIndex
					+ STACK_TRACE_TOKEN.length(), encodedStackTrace.length());
			int endMessageIndex = stackTraceTextIndex;
			String exceptionMesg = "Undefined Exception"; //$NON-NLS-1$
			if(endMessageIndex != 0) {
				exceptionMesg = encodedStackTrace.substring(0,
					endMessageIndex);
			}
			ArrayList<StackTraceElement> traceElements = new ArrayList<StackTraceElement>();
			StringTokenizer tokenizer = new StringTokenizer(stackTraceText, ","); //$NON-NLS-1$
			while (tokenizer.hasMoreTokens()) {
				String className = tokenizer.nextToken();
				String methodName = tokenizer.nextToken();
				String fileName = tokenizer.nextToken();
				String lineNumber = tokenizer.nextToken();
				traceElements.add(new StackTraceElement(className, methodName,
						fileName, Integer.parseInt(lineNumber)));
			}
			stackTraceException = new Throwable(exceptionMesg);
			
        	// complete stack trace with local elements as well. otherwise, local trace is lost.
        	Exception e = new Exception();
        	e.fillInStackTrace();
        	StackTraceElement[] elements = e.getStackTrace();
        	List<StackTraceElement> completedTrace = new ArrayList<StackTraceElement>();
        	completedTrace.addAll(traceElements);
        	completedTrace.addAll(Arrays.asList(elements));
        	stackTraceException.setStackTrace(completedTrace.toArray(new StackTraceElement[completedTrace.size()]));
		}
        return stackTraceException;
	}

    private static String nonNull(String s) {
    	return s == null ? "unknown" : s; //$NON-NLS-1$
    }

	private static boolean shouldAvoidSerializingExceptions() {
		return StackUtil.shouldAvoidSerializingExceptions();
	}

    /**
     * This SQLException class exists because the JUnit view in Eclipse
     * doesn't know about SQLException.getNextException, and as a result
     * doesn't display the full sequence of nested exceptions, making it
     * harder to debug.
     */
    static class SQLException extends java.sql.SQLException {
		private static final long serialVersionUID = -6127147898252829787L;
		private final String SQLState;
        private final int vendorCode;
        private final SQLException next;

        public SQLException(
                String reason, String SQLState, int vendorCode, SQLException next) {
            super(reason);
            this.SQLState = "".equals(SQLState) ? null : SQLState; //$NON-NLS-1$
            this.vendorCode = vendorCode;
            this.next = next;
        }

        /* (non-Javadoc)
         * @see java.sql.SQLException#getErrorCode()
         */
        @Override
        public int getErrorCode() {
            return vendorCode;
        }

        /* (non-Javadoc)
         * @see java.sql.SQLException#getNextException()
         */
        @Override
        public SQLException getNextException() {
            return next;
        }

        /* (non-Javadoc)
         * @see java.sql.SQLException#getSQLState()
         */
        @Override
        public String getSQLState() {
            return SQLState;
        }

        /* (non-Javadoc)
         * @see java.lang.Throwable#getCause()
         */
        @Override
        public Throwable getCause() {
            return next;
        }

        /* (non-Javadoc)
         * @see java.lang.Throwable#toString()
         */
        @Override
        public String toString () {
            String msg = getLocalizedMessage();
            String name = "java.sql.SQLException"; //$NON-NLS-1$
            if (msg == null) {
                return name;
            } else {
                return new StringBuffer(name.length() + 2 + msg.length())
                    .append(name).append(": ").append(msg).toString(); //$NON-NLS-1$
            }
        }
    }
}
