/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010, 2012. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.repository.common.internal.marshal.util;

import java.io.BufferedWriter;
import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintStream;
import java.io.UnsupportedEncodingException;
import java.io.Writer;
import java.nio.charset.Charset;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.regex.Pattern;

import com.ibm.team.repository.common.internal.marshal.MarshallingException;
import com.ibm.team.repository.common.transport.HttpUtil.CharsetEncoding;

/**
 * Utility to construct XML elements corresponding to EStructuralFeatures
 */
public class XMLConstructionUtil {
    private static boolean printToSystemOut = false;
    public static final String XML_HEADER_FMT = "<?xml version=\"1.0\" encoding=\"%s\" ?>"; //$NON-NLS-1$
    public static final String XML_NAMESPACE_TAG = "xmlns:"; //$NON-NLS-1$
    public static final String XSI_NAMESPACE_TAG = "xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\""; //$NON-NLS-1$
    
    public static final String START_ELEMENT_TAG = "<"; //$NON-NLS-1$
    public static final String END_ELEMENT_TAG = ">"; //$NON-NLS-1$
    public static final String COMMENT_START_TAG = "<!--"; //$NON-NLS-1$
    public static final String COMMENT_END_TAG = "-->"; //$NON-NLS-1$
    
    public static final String EMPTY_STRING = "\"\""; //$NON-NLS-1$
    public static final String EQUAL_TAG = "="; //$NON-NLS-1$
    public static final String SLASH_TAG = "/"; //$NON-NLS-1$
    public static final String QUOTE_CHARACTER = "\""; //$NON-NLS-1$
    public static final String SPACE_CHARACTER = " "; //$NON-NLS-1$
    public static final String NEWLINE_CHARACTER = "\n"; //$NON-NLS-1$
    public static final String TAB_CHARACTER = "\t"; //$NON-NLS-1$
    
    public static final String ITEM_ID_ATTRIBUTE_NAME = "itemId"; //$NON-NLS-1$
    public static final String STATE_ID_ATTRIBUTE_NAME = "stateId"; //$NON-NLS-1$
    public static final String XSI_TYPE_NAME = "xsi:type"; //$NON-NLS-1$
    public static final String XSI_NIL = "xsi:nil=\"true\"/>"; //$NON-NLS-1$
    
    /*
     * Regex patterns in support of validating XML element names
     */
    private static final String nameChars = "^[a-zA-Z_][a-zA-Z0-9\\-_\\.]*"; //$NON-NLS-1$
    private static final String xmlChars = ".*(x|Xm|Ml|L).*"; //$NON-NLS-1$
    private static final Pattern nameCharsPattern = Pattern.compile(nameChars);
    private static final Pattern xmlCharsPattern = Pattern.compile(xmlChars);

    /**
     * Creates a new writer for the specified <code>outputStream</code> and
     * <code>charset</code>.
     * 
     * @param outputStream The output stream to create a writer for. Must not
     * be <code>null</code>.
     * 
     * @param charset The charset to use for the output stream encoding. May be
     * <code>null</code>.
     * 
     * @return The writer. Never <code>null</code>.
     * 
     * @throws UnsupportedEncodingException Thrown if the referenced charset
     * is unsupported.
     */
    public static Writer createWriter(OutputStream outputStream, Charset charset) 
        throws UnsupportedEncodingException
    {
        if (charset == null)
            charset = CharsetEncoding.UTF8.toCharset();

        if (outputStream == null)
            throw new IllegalArgumentException("outputStream must not be null"); //$NON-NLS-1$

        Writer writer =
            new BufferedWriter(new OutputStreamWriter(outputStream, charset));

        if (!printToSystemOut) {
            return writer;
        }

        return new SplitWriter(System.out, writer);
    }

    private static class SplitWriter extends Writer {
        private final PrintStream stream;
        private final Writer writer;
        public SplitWriter(PrintStream stream, Writer writer) {
            super();
            this.stream = stream;
            this.writer = writer;
        }

        @Override
        public Writer append(char c) throws IOException {
            stream.append(c);
            writer.append(c);
            return this;
        }

        @Override
        public Writer append(CharSequence csq, int start, int end) throws IOException {
            stream.append(csq, start, end);
            writer.append(csq, start, end);
            return this;
        }

        @Override
        public Writer append(CharSequence csq) throws IOException {
            stream.append(csq);
            writer.append(csq);
            return this;
        }

        @Override
        public void close() throws IOException {
            stream.flush(); // Don't close the PrintStream
            writer.close();
        }

        @Override
        public void flush() throws IOException {
            stream.flush();
            writer.flush();
        }

        @Override
        public void write(char[] cbuf, int off, int len) throws IOException {
            stream.append(new String(cbuf, off, len));
            writer.write(cbuf, off, len);
        }

        @Override
        public void write(char[] cbuf) throws IOException {
            stream.print(cbuf);
            writer.write(cbuf);
        }

        @Override
        public void write(int c) throws IOException {
            stream.write(c);
            writer.write(c);
        }

        @Override
        public void write(String str, int off, int len) throws IOException {
            stream.append(str, off, off + len);
            writer.write(str, off, len);
        }

        @Override
        public void write(String str) throws IOException {
            stream.print(str);
            writer.write(str);
        }
    }

    
    /**
     * Write XML Header
     * @param outputStream
     * @param charset
     */
    public static void writeXMLHeader(Writer writer, Charset charset) throws MarshallingException {
        write(writer, String.format(XML_HEADER_FMT, charset.name()));
    }
    
    /**
     * Write namespace information
     * @param outputStream 
     * @param referencedNamespacesHashMap
     */
    public static void writeNamespaceInformation(Writer writer, 
            Map referencedNamespacesHashMap) throws MarshallingException {
        
        StringBuffer namespaceBuffer = new StringBuffer();
        if(referencedNamespacesHashMap != null && referencedNamespacesHashMap.size() > 0) {
            Iterator i = referencedNamespacesHashMap.entrySet().iterator();
            while(i.hasNext()) {
                writeNewLineWithTabCharacters(namespaceBuffer, 1);
                Entry e = (Entry) i.next();
                String namespacePrefix = (String) e.getKey();
                String namespaceURI = (String) e.getValue();
                
                namespaceBuffer.append(XML_NAMESPACE_TAG);
                namespaceBuffer.append(namespacePrefix);
                namespaceBuffer.append(EQUAL_TAG);
                namespaceBuffer.append(QUOTE_CHARACTER);
                namespaceBuffer.append(namespaceURI);
                namespaceBuffer.append(QUOTE_CHARACTER);
                // write(writer, XML_NAMESPACE_TAG + namespacePrefix + EQUAL_TAG + QUOTE_CHARACTER + namespaceURI + QUOTE_CHARACTER);
            }
            
            // Add XSI Namespace tag
            writeNewLineWithTabCharacters(namespaceBuffer, 1);
            namespaceBuffer.append(XSI_NAMESPACE_TAG);
            write(writer, namespaceBuffer.toString());
        }
    }
    
    /**
     * Write comments
     * @param writer
     * @param comment
     */
    public static void writeComment(Writer writer, String comment) throws MarshallingException {
        StringBuffer commentBuffer = new StringBuffer();
        commentBuffer.append(COMMENT_START_TAG);
        commentBuffer.append(comment);
        commentBuffer.append(COMMENT_END_TAG);
        write(writer, commentBuffer.toString());
    }

	private static final int ELEMENT_STATIC_LENGTH = 
		(2 * START_ELEMENT_TAG.length()) + (2 * END_ELEMENT_TAG.length()) + SLASH_TAG.length();
    public static void writeElement(Writer writer, String elementName, 
            String value, int depth) throws MarshallingException {
        
        if(value != null) {
			// allow for 10% growth in XMLEscapeUtil.convert()
			int allocLength = ELEMENT_STATIC_LENGTH + (2 * elementName.length()) + (int) (value.length() * 1.1);
            StringBuffer elementBuffer = new StringBuffer(allocLength);
            writeNewLineWithTabCharacters(elementBuffer, depth);
            
            elementBuffer.append(START_ELEMENT_TAG);
            elementBuffer.append(elementName);
            elementBuffer.append(END_ELEMENT_TAG);

            elementBuffer.append(XMLEscapeUtil.convert(value));

            elementBuffer.append(START_ELEMENT_TAG);
            elementBuffer.append(SLASH_TAG);
            elementBuffer.append(elementName);
            elementBuffer.append(END_ELEMENT_TAG);
            
            write(writer, elementBuffer.toString());
        }
    }
    
    public static void writeElementName(Writer writer, String elementName) throws MarshallingException {
        write(writer, START_ELEMENT_TAG + elementName);
    }
    
    public static void writeEndElementTag(Writer writer) throws MarshallingException {
        write(writer, END_ELEMENT_TAG);
    }
    
    public static void writeEndElementName(Writer writer, String elementName, int depth) throws MarshallingException {
        StringBuffer endElementNameBuffer = new StringBuffer();
        writeNewLineWithTabCharacters(endElementNameBuffer, depth);
        endElementNameBuffer.append(START_ELEMENT_TAG);
        endElementNameBuffer.append(SLASH_TAG);
        endElementNameBuffer.append(elementName);
        endElementNameBuffer.append(END_ELEMENT_TAG);
        write(writer, endElementNameBuffer.toString());
    }
    
    /**
     * Write XML Attribute 
     */
    public static void writeAttribute(Writer writer, String attributeName, 
            String attributeValue) throws MarshallingException {
        StringBuffer attributeBuffer = new StringBuffer();
        writeAttributeToAttributeBuffer(attributeBuffer, attributeName, attributeValue);
        write(writer, attributeBuffer.toString());
    }
    
    private static void writeAttributeToAttributeBuffer(StringBuffer attributeBuffer, 
            String attributeName, String attributeValue) {
     
        attributeBuffer.append(SPACE_CHARACTER);
        attributeBuffer.append(attributeName);
        attributeBuffer.append(EQUAL_TAG);
        attributeBuffer.append(QUOTE_CHARACTER);
        
        attributeBuffer.append(XMLEscapeUtil.convert(attributeValue));
        
        attributeBuffer.append(QUOTE_CHARACTER);
    }
    
    public static void writeEmptyAttribute(String elementName, Writer writer) 
            throws MarshallingException {
        StringBuffer emptyAttributeBuffer = new StringBuffer();
        emptyAttributeBuffer.append(SPACE_CHARACTER);
        emptyAttributeBuffer.append(elementName);
        emptyAttributeBuffer.append(EQUAL_TAG);
        emptyAttributeBuffer.append(EMPTY_STRING);
        write(writer, emptyAttributeBuffer.toString());
    }
    
    public static void writeEmptyElement(Writer writer, String elementName) 
            throws MarshallingException {
        
        StringBuffer emptyReferenceBuffer = new StringBuffer();
        emptyReferenceBuffer.append(START_ELEMENT_TAG);
        emptyReferenceBuffer.append(elementName);
        emptyReferenceBuffer.append(END_ELEMENT_TAG);
        
        emptyReferenceBuffer.append(START_ELEMENT_TAG);
        emptyReferenceBuffer.append(SLASH_TAG);
        emptyReferenceBuffer.append(elementName);
        emptyReferenceBuffer.append(END_ELEMENT_TAG);
        
        write(writer, emptyReferenceBuffer.toString());
    }
    
    public static void writeNullElement(Writer writer, String elementName) 
            throws MarshallingException {
        StringBuffer nullReferenceBuffer = new StringBuffer();
        nullReferenceBuffer.append(START_ELEMENT_TAG);
        nullReferenceBuffer.append(elementName);
        nullReferenceBuffer.append(SPACE_CHARACTER);
        nullReferenceBuffer.append(XSI_NIL);
        write(writer, nullReferenceBuffer.toString());
    }
            
    
    /**
     * Write References
     */
    public static void writeReference(Writer writer, String referenceAttributeName, 
        String itemId, String stateId, String type) throws MarshallingException {
        
        StringBuffer referenceBuffer = new StringBuffer();
        writeReferenceToReferenceBuffer(referenceBuffer, referenceAttributeName, itemId, stateId, type);
        write(writer, referenceBuffer.toString());
    }
    
    private static void writeReferenceToReferenceBuffer(StringBuffer referenceBuffer,
            String referenceAttributeName, String itemId, String stateId, String type) {
        
        referenceBuffer.append(START_ELEMENT_TAG);
        referenceBuffer.append(referenceAttributeName);
        referenceBuffer.append(SPACE_CHARACTER);
        
        if(itemId != null) {
            writeAttributeToAttributeBuffer(referenceBuffer, ITEM_ID_ATTRIBUTE_NAME, itemId);
            referenceBuffer.append(SPACE_CHARACTER);
        }
        if(stateId != null) {
            writeAttributeToAttributeBuffer(referenceBuffer, STATE_ID_ATTRIBUTE_NAME, stateId);
            referenceBuffer.append(SPACE_CHARACTER);
        }
        if(type != null) {
            writeAttributeToAttributeBuffer(referenceBuffer, XSI_TYPE_NAME, type);
            referenceBuffer.append(SPACE_CHARACTER);
        }
        referenceBuffer.append(SLASH_TAG);
        referenceBuffer.append(END_ELEMENT_TAG);
    }
    
    
    public static void writeDataType(Writer writer, String name,
            String value, int depth) throws MarshallingException {
        
        StringBuffer dataTypeBuffer = new StringBuffer();
        writeNewLineWithTabCharacters(dataTypeBuffer, depth);
        
        dataTypeBuffer.append(START_ELEMENT_TAG);
        dataTypeBuffer.append(name);
        dataTypeBuffer.append(SPACE_CHARACTER);
        dataTypeBuffer.append(value);
        dataTypeBuffer.append(SPACE_CHARACTER);
        dataTypeBuffer.append(SLASH_TAG);
        dataTypeBuffer.append(END_ELEMENT_TAG);
        
        write(writer, dataTypeBuffer.toString());

    }
    
    public static void writeNewLineCharacter(Writer writer) throws MarshallingException {
        write(writer, NEWLINE_CHARACTER);
    }
    
    public static void writeTabCharacter(Writer writer) throws MarshallingException {
        write(writer, TAB_CHARACTER);
    }
    
    /**
     * Write New Line character and tabs based on the depth
     * @param outputStream
     * @param depth
     */
    public static void writeNewLineWithTabCharacters(Writer writer, int depth) throws MarshallingException {
        writeNewLineCharacter(writer);
        for(int i = 0; i < depth; i++) {
            writeTabCharacter(writer);
        }
    }
    
    /**
     * Write New Line character and tabs based on the depth
     * @param outputStream
     * @param depth
     */
    public static void writeNewLineWithTabCharacters(StringBuffer buffer, int depth) {
        writeNewLineCharacter(buffer);
        for(int i = 0; i < depth; i++) {
            writeTabCharacter(buffer);
        }
    }
    
    public static void writeNewLineCharacter(StringBuffer buffer) {
        buffer.append(NEWLINE_CHARACTER);
    }
    
    public static void writeTabCharacter(StringBuffer buffer) {
        buffer.append(TAB_CHARACTER);
    }
    
    /**
     * Returns whether or not the referenced string is suitable for use as an
     * XML element name. The definition of a name as described by 
     * http://www.w3.org/XML is: <p>
     *
     * A Name is a token beginning with a letter or one of a few punctuation
     * characters, and continuing with letters, digits, hyphens, underscores,
     * colons, or full stops, together known as name characters.] Names
     * beginning with the string " xml ", or with any string which would match
     * (('X'|'x') ('M'|'m') ('L'|'l')), are reserved for standardization in this
     * or future versions of this specification. <p>
     * 
     * Specifically:
     * 
     * <ul>
     * <li>
     * NameChar ::= Letter | Digit | '.' | '-' | '_' | ':' | CombiningChar | Extender
     * </li>
     * <li>  
     * Name ::= (Letter | '_' | ':') (NameChar)*
     * </li>
     * </ul> <p>
     * 
     * IMPORTANT NOTE: This routine does not tolerate the ":" character occurring
     * in any position in the referenced <code>name</code>.  Should it appear,
     * this routine returns false - ":" is reserved for separating this name
     * from its namespace.
     * 
     * @param name The name to test. May be <code>null</code>. If
     * <code>null</code> or empty, <code>false</code> is returned.
     * 
     * @return <code>true</code> if the rules described above can be
     * successfully described, <code>false</code> if not.
     */
    public static boolean validXMLName(String name) {
        if (name == null)
            return false;
        
        // The text "xml" cannot appear in the name.
        if (xmlCharsPattern.matcher(name).matches() == true)
            return false;
        
        return nameCharsPattern.matcher(name).matches();
    }
    
    /**
     * Write the data on to the stream using a writer
     * @param writer
     * @param output
     */
    private static void write(Writer writer, String output) throws MarshallingException {
        try {
            writer.write(output);
        } catch (IOException e) {
            throw new MarshallingException(e);
        }
    }
}