/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010, 2011. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.repository.common.internal.marshal.util;

import java.nio.BufferOverflowException;
import java.nio.CharBuffer;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.eclipse.emf.ecore.xml.type.internal.DataValue.XMLChar;

import com.ibm.team.repository.common.serialize.internal.nls.Messages;
import com.ibm.team.repository.common.util.NLS;

public class XMLEscapeUtil {

    private static final String NLS_INVALID_LOW_SURROGATE_CHARACTER = Messages.getServerString("XMLEscapeUtil.InvalidLowSurrogateChar"); //$NON-NLS-1$
    private static final String NLS_INVALID_HIGH_SURROGATE_CHARACTER = Messages.getServerString("XMLEscapeUtil.InvalidHighSurrogateChar"); //$NON-NLS-1$
    private static final String NLS_INVALID_XML_CHARACTER = Messages.getServerString("XMLEscapeUtil.InvalidXMLChar"); //$NON-NLS-1$

    private static final int MAX_UTF_MAPPABLE_CODEPOINT = 0x10FFFF;
    
    private static final char[] AMP = { '&', 'a', 'm', 'p', ';' };
    private static final char[] LESS = { '&', 'l', 't', ';' };
    private static final char[] GREATER = { '&', 'g', 't', ';' };
    private static final char[] QUOTE = { '&', 'q', 'u', 'o', 't', ';' };
    
    /*
     * Since we are encoding control chars as backslash-u, we need to escape
     * any legitimate backslash-u.
     */
    private static final char[] UNICODE_ESCAPE = { '\\', 'u' };
    private static final String ORIGINAL_UNICODE_LITERAL = "\\u"; //$NON-NLS-1$
    
    /*
     * Note that the only control characters allowed in XML character references
     * are 0x9, 0xA and 0xD.  Other control characters are encoded as Unicode as
     * acceptable to java.lang.String.  See Workitem 38591 .
     */
    private static final char[] NUL = { '\\', 'u', '0', '0', '0', '0' };
    private static final char[] SOH = { '\\', 'u', '0', '0', '0', '1' };
    private static final char[] STX = { '\\', 'u', '0', '0', '0', '2' };
    private static final char[] ETX = { '\\', 'u', '0', '0', '0', '3' };
    private static final char[] EOT = { '\\', 'u', '0', '0', '0', '4' };
    private static final char[] ENQ = { '\\', 'u', '0', '0', '0', '5' };
    private static final char[] ACK = { '\\', 'u', '0', '0', '0', '6' };
    private static final char[] BEL = { '\\', 'u', '0', '0', '0', '7' };
    private static final char[] BS = { '\\', 'u', '0', '0', '0', '8' };
    private static final char[] TAB = { '&', '#', 'x', '9', ';' };
    private static final char[] LF  = { '&', '#', 'x', 'A', ';' };
    private static final char[] VT = { '\\', 'u', '0', '0', '0', 'B' };
    private static final char[] FF = { '\\', 'u', '0', '0', '0', 'C' };
    private static final char[] CR  = { '&', '#', 'x', 'D', ';' };
    private static final char[] SO = { '\\', 'u', '0', '0', '0', 'E' };
    private static final char[] SI  = { '\\', 'u', '0', '0', '0', 'F' };
    private static final char[] DLE = { '\\', 'u', '0', '0', '1', '0' };
    private static final char[] DC1 = { '\\', 'u', '0', '0', '1', '1' };
    private static final char[] DC2 = { '\\', 'u', '0', '0', '1', '2' };
    private static final char[] DC3 = { '\\', 'u', '0', '0', '1', '3' };
    private static final char[] DC4 = { '\\', 'u', '0', '0', '1', '4' };
    private static final char[] NAK = { '\\', 'u', '0', '0', '1', '5' };
    private static final char[] SYN = { '\\', 'u', '0', '0', '1', '6' };
    private static final char[] ETB = { '\\', 'u', '0', '0', '1', '7' };
    private static final char[] CAN = { '\\', 'u', '0', '0', '1', '8' };
    private static final char[] EM  = { '\\', 'u', '0', '0', '1', '9' };
    private static final char[] SUB = { '\\', 'u', '0', '0', '1', 'A' };
    private static final char[] ESC = { '\\', 'u', '0', '0', '1', 'B' };
    private static final char[] FS  = { '\\', 'u', '0', '0', '1', 'C' };
    private static final char[] GS  = { '\\', 'u', '0', '0', '1', 'D' };
    private static final char[] RS  = { '\\', 'u', '0', '0', '1', 'E' };
    private static final char[] US  = { '\\', 'u', '0', '0', '1', 'F' };
    
    private static final char[] BACKSLASH = { '\\', 'u', '0', '0', '5', 'C' };
    
    private static final char[][] CONTROL_CHARACTERS = 
      new char [][]
      {
        NUL,
        SOH,
        STX,
        ETX,
        EOT,
        ENQ,
        ACK,
        BEL,
        BS,
        TAB,
        LF,
        VT,
        FF,
        CR,
        SO,
        SI,
        DLE,
        DC1,
        DC2,
        DC3,
        DC4,
        NAK,
        SYN,
        ETB,
        CAN,
        EM,
        SUB,
        ESC,
        FS,
        GS,
        RS,
        US,
      };
    

    private XMLEscapeUtil() {
    }

	public static String convert(String input) {
		try {
			// First try assuming a reasonable size increase (10%) will be enough
			return convert(input, (int) (input.length() * 1.1));
		} catch (BufferOverflowException e) {
			// If that's not enough, allocate as much as it could possibly need
			return convert(input, input.length() * 6);
		}
	}

    private static String convert(String input, int allocation) {
        boolean changed = false;
        int mappableLimit = MAX_UTF_MAPPABLE_CODEPOINT; 
        int inputLength = input.length();
        CharBuffer buffer = CharBuffer.allocate(allocation);
        int inputPos = 0;
        char ch = 0;
        while (inputLength-- > 0) {
            ch = input.charAt(inputPos++); 
            switch (ch) {
            case 0x0:
            case 0x1:
            case 0x2:
            case 0x3:
            case 0x4:
            case 0x5:
            case 0x6:
            case 0x7:
            case 0x8:
            case 0xB:
            case 0xC:
            case 0xE:
            case 0xF:
            case 0x10:
            case 0x11:
            case 0x12:
            case 0x13:
            case 0x14:
            case 0x15:
            case 0x16:
            case 0x17:
            case 0x18:
            case 0x19:
            case 0x1A:
            case 0x1B:
            case 0x1C:
            case 0x1D:
            case 0x1E:
            case 0x1F:
            	int pos;
            	for (pos = buffer.position(); pos != 0 && buffer.get(pos - 1) == '\\'; --pos);
            	int cnt = buffer.position() - pos;
            	buffer.position(pos);
            	for (;cnt != 0; --cnt) {
            		replaceChars(buffer, BACKSLASH);
            	}
                replaceChars(buffer, CONTROL_CHARACTERS[ch]);
                changed = true;
                break;
            case '&': 
                replaceChars(buffer, AMP);
                changed = true;
                break;
            case '<':
                replaceChars(buffer, LESS);
                changed = true;
                break;
            case '"':
                replaceChars(buffer, QUOTE);
                changed = true;
                break;
            case '\n':
                replaceChars(buffer, LF);
                changed = true;
                break;
            case '\r':
                replaceChars(buffer, CR);
                changed = true;
                break;
            case '\t':
                replaceChars(buffer, TAB);
                changed = true;
                break;
            case '>':
              if (inputPos >= 3 && input.charAt(inputPos - 2) == ']' && input.charAt(inputPos - 3) == ']') {
                replaceChars(buffer, GREATER);
                changed = true;
              } else {
            	  buffer.append(ch);
              }
              break;
            case 'u':
              /*
               * Because we escape some control characters as backslash-u, we need to
               * encode any literal backslash-u as back-back-u.
               */
              if (inputPos >= 2 && input.charAt(inputPos - 2) == '\\') {
                  replaceChars(buffer, UNICODE_ESCAPE);
                  changed = true;
              } else {
            	  buffer.append(ch);
              }
              break;
            default:
                if (!XMLChar.isValid(ch)) {
                    if (XMLChar.isHighSurrogate(ch)) {
                        char high = ch;
                        if (inputLength-- > 0) {
                            ch = input.charAt(inputPos++); 
                            if (XMLChar.isLowSurrogate(ch)) {
                                if (mappableLimit == MAX_UTF_MAPPABLE_CODEPOINT) {
                                    // Every codepoint is supported!
                                    buffer.append(high);
                                    buffer.append(ch);
                                } else {
                                    // Produce the supplemental character as an entity
                                    replaceChars(buffer, ("&#x" + Integer.toHexString(XMLChar.supplemental(high, ch)) + ";").toCharArray()); //$NON-NLS-1$ //$NON-NLS-2$
                                    changed = true;
                                }
                                break;
                            }
                            String msg = NLS.bind(NLS_INVALID_LOW_SURROGATE_CHARACTER, Integer.toHexString(ch), input);
                            throw new RuntimeException(msg);
                        } else {
                            String msg = NLS.bind(NLS_INVALID_HIGH_SURROGATE_CHARACTER, Integer.toHexString(ch), input);
                            throw new RuntimeException(msg);
                        }
                    } else {
                        String msg = NLS.bind(NLS_INVALID_XML_CHARACTER, Integer.toHexString(ch), input);
                        throw new RuntimeException(msg);
                    }
                } else {
                    // Normal (BMP) unicode code point. See if we know for a fact that the encoding supports it:
                    if (ch <= mappableLimit) {
                        buffer.append(ch);
                    } else {
                        // We not sure the encoding supports this code point, so we write it as a character entity reference.
                        replaceChars(buffer, ("&#x" + Integer.toHexString(ch) + ";").toCharArray()); //$NON-NLS-1$ //$NON-NLS-2$
                        changed = true;
                    }
                }
                break;
            }
      }
        
      String retVal = input;
      if (changed)
          retVal = new String(buffer.array(), 0, buffer.position());
      return retVal;
    }
    
    private static void replaceChars(CharBuffer buffer, char[] replacement) {
    	buffer.put(replacement, 0, replacement.length);
    }
    
    /**
     * Decodes any control characters that may have been encoded via
     * {@link #convert(String)}, or any escaped characters that we introduced
     * to normalize that encoding.
     * 
     * @param value The value to decode. Must not be <code>null</code>.
     * 
     * @return The decoded value.  Never <code>null</code>.  Will be
     * <code>value</code> if the input value contained no encodings.
     */
    public static String decodeControlChars(String value) {
        if (value == null)
            throw new IllegalArgumentException("value must not be null"); //$NON-NLS-1$
        
        if (value.indexOf(ORIGINAL_UNICODE_LITERAL) == -1)
            return value;
        
        try {
            int inputLength = value.length();
            int inputPos = 0;
            StringBuffer buf = new StringBuffer(inputLength);
            while (inputPos < inputLength) {
                char ch = value.charAt(inputPos++);
                if (ch == '\\' && (inputPos + 1 < inputLength)) {
                    if (value.charAt(inputPos) == '\\' && value.charAt(inputPos + 1) == 'u') {
                        buf.append(ORIGINAL_UNICODE_LITERAL);
                        inputPos += 2;
                    }
                    else if (value.charAt(inputPos) == 'u' && (inputPos + 4 < inputLength)) {
                        String numString = value.substring(inputPos + 1, inputPos + 5);
                        int cc = Integer.parseInt(numString, 16);
                        buf.append((char)cc);
                        inputPos += 5;
                    }
                    else
                        buf.append(ch);
                }
                else {
                    buf.append(ch);
                }
            }
            value = buf.toString();
        }
        catch (Throwable ex) {
            String message = NLS.bind(Messages.getServerString("XMLEscapeUtil.ErrorUnrecognizedControlSequence"), value); //$NON-NLS-1$
            _logger.error(message, ex);
        }

        return value;
    }
    
    private static Log _logger = LogFactory.getLog(XMLEscapeUtil.class);

}