/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010, 2012. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.repository.common.internal.marshal.util;

import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.EReference;

public class XMLNamespaceUtil {
    /**
     * Get the referenced namespaces
     */
    public static Map<String, String> getReferencedNamespaces(EObject eObject) {
        Map<String, String> referencedNamespaceHashMap = new HashMap<String, String>();
        addReferencedNamespaces(eObject, referencedNamespaceHashMap);
    
        // add referenced namespaces in EReferences present in the EObject
        getReferencedNamespacesInEReferences(eObject, referencedNamespaceHashMap);
        
        return referencedNamespaceHashMap;
    }
    
    private static void getReferencedNamespacesInEReferences(EObject eObject, 
            Map<String, String> referencedNamespaceHashMap) {
        
        Iterator i = eObject.eClass().getEAllReferences().iterator();
        while (i.hasNext()) {
            EReference reference = (EReference)i.next();
            if (MarshallerUtil.shouldBeMarshalled(eObject, reference)) {
                if (MarshallerUtil.shouldConsiderAsContainment(reference)) {
                    getReferencedNamespacedInContainmentReferences(eObject, reference, referencedNamespaceHashMap);
                } else {
                    getReferencedNamespacedInAssociationReferences(eObject, reference, referencedNamespaceHashMap);
                }
            }
        }
    }
    
    /**
     * Write Namespace information
     */
    private static void addReferencedNamespaces(EObject eObject, Map<String, String> nsPrefixToNsURIMap) {
        EPackage ePackage = eObject.eClass().getEPackage();
        String namespacePrefix = ePackage.getNsPrefix();
        if (!nsPrefixToNsURIMap.containsKey(namespacePrefix)) {
            nsPrefixToNsURIMap.put(namespacePrefix, ePackage.getNsURI());
        } else {
            // ns prefix contained in the map. 
            String nsURI = nsPrefixToNsURIMap.get(namespacePrefix);
            if (!ePackage.getNsURI().equals(nsURI)) {
                // Same ns prefix but different uri's
                
                // construct a new ns prefix and add it to the map
                String newNsPrefix = constructUniqueNsPrefix(namespacePrefix, nsPrefixToNsURIMap);
                nsPrefixToNsURIMap.put(newNsPrefix, ePackage.getNsURI());
            }
        }
    }
    
    private static String constructUniqueNsPrefix(String nsPrefix, Map<String, String> nsPrefixToNsURIMap) {
        int i = 1;
        while (true) {
            String newNsPrefix = nsPrefix + i;
            if (!nsPrefixToNsURIMap.containsKey(newNsPrefix)) {
                return newNsPrefix;
            }
            i++;
        }
    }
    
    private static void getReferencedNamespacedInAssociationReferences(EObject eObject, 
            EReference reference, Map<String, String> referencedNamespaceHashMap) {

        if (reference.isMany()) {
            List objectValues = (List) eObject.eGet(reference);
            if (objectValues != null) {
                Iterator i = objectValues.iterator();
                while (i.hasNext()) {
                    EObject objectValue = (EObject) i.next();
                    addReferencedNamespaces(objectValue, referencedNamespaceHashMap);
                }
            }
        } else {
            EObject objectValue = (EObject) eObject.eGet(reference);
            if (objectValue != null) {
                addReferencedNamespaces(objectValue, referencedNamespaceHashMap);
            }
        }
    }
    
    private static void getReferencedNamespacedInContainmentReferences(EObject eObject, 
            EReference reference, Map<String, String> referencedNamespaceHashMap) {

        if (reference.isMany()) {
            List objectValues = (List) eObject.eGet(reference);
            if (objectValues != null) {
                Iterator i = objectValues.iterator();
                while (i.hasNext()) {
                    EObject containedObject = (EObject) i.next();
                    addReferencedNamespaces(containedObject, referencedNamespaceHashMap);
                    getReferencedNamespacesInEReferences(containedObject, referencedNamespaceHashMap);
                }
            }
        } else {
            EObject containedObject = (EObject) eObject.eGet(reference);
            if (containedObject != null) {
                addReferencedNamespaces(containedObject, referencedNamespaceHashMap);
                getReferencedNamespacesInEReferences(containedObject, referencedNamespaceHashMap);
            }
        }
    }
    
    public static Map<String, String> constructNsURIToNsPrefixMap(Map<String, String> nsPrefixToNsURIMap) {
        Map<String, String> nsURIToNsPrefixMap = new HashMap<String, String>(nsPrefixToNsURIMap.size() * 2);
        for (Entry<String, String> entry : nsPrefixToNsURIMap.entrySet()) {
            nsURIToNsPrefixMap.put(entry.getValue(), entry.getKey());
        }
        return nsURIToNsPrefixMap;
    }
}