/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.repository.common.json;


import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.Reader;
import java.io.StringWriter;
import java.io.Writer;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

import com.ibm.team.repository.common.json.internal.Parser;
import com.ibm.team.repository.common.json.internal.Serializer;
import com.ibm.team.repository.common.json.internal.SerializerVerbose;

/**
 * Extension of ArrayList that only allows values which are JSON-able.  
 * See JSONObject for a list of valid values.
 * 
 * Instances of this class are not thread-safe.
 */
public class JSONArray extends ArrayList implements IJSONSerializable {

    private static final long serialVersionUID = 9076798781015779954L;

    /**
     * Create a new instance of this class.
     */
    public JSONArray() {
        super();
    }

    /**
     * Create a new instance of this class with the specified initial capacity.
     */
    public JSONArray(int initialCapacity) {
        super(initialCapacity);
    }

    /*
     * (non-Javadoc)
     * @see java.util.ArrayList#add(int, java.lang.Object)
     */
    @Override
    public void add(int index, Object element) {
        checkElement(element);
        super.add(index, element);
    }

    /*
     * (non-Javadoc)
     * @see java.util.ArrayList#add(java.lang.Object)
     */
    @Override
    public boolean add(Object element) {
        checkElement(element);
        return super.add(element);
    }

    /*
     * (non-Javadoc)
     * @see java.util.ArrayList#addAll(java.util.Collection)
     */
    @Override
    public boolean addAll(Collection collection) {
        checkElements(collection);
        return super.addAll(collection);
    }

    /*
     * (non-Javadoc)
     * @see java.util.ArrayList#addAll(int, java.util.Collection)
     */
    @Override
    public boolean addAll(int index, Collection collection) {
        checkElements(collection);
        return super.addAll(index, collection);
    }

    /*
     * (non-Javadoc)
     * @see java.util.ArrayList#set(int, java.lang.Object)
     */
    @Override
    public Object set(int index, Object element) {
        checkElement(element);
        return super.set(index, element);
    }

    /**
     * 
     */
    private void checkElement(Object element) {
        if (!JSONObject.isValidObject(element))
            throw new IllegalArgumentException("invalid type of element: " + element.getClass().getName()); //$NON-NLS-1$
    }

    /**
     * 
     */
    private void checkElements(Collection collection) {
        for (Iterator iter = collection.iterator(); iter.hasNext();) {
            final Object element = iter.next();
			if (!JSONObject.isValidObject(element))
                throw new IllegalArgumentException("invalid type of element: " + element.getClass().getName()); //$NON-NLS-1$
        }
    }
    
    /**
     * Convert this array into a stream of JSON text.
     */
    public void serialize(Writer writer) throws IOException {
        serialize(writer, false);
    }

    /**
     * Convert this array into a stream of JSON text, specifying verbosity.
     */
    public void serialize(Writer writer, boolean verbose) throws IOException {
        Serializer serializer;
        writer = new BufferedWriter(writer);

        if (verbose) {
            serializer = new SerializerVerbose(writer);
        } else {
            serializer = new Serializer(writer);
        }

        serializer.writeArray(this).flush();
    }

    /**
     * Convert a stream of JSON text into object form. 
     */
    static public JSONArray parse(Reader reader) throws IOException {
        reader = new BufferedReader(reader);
        Object obj = new Parser(reader).parse();
        if (!(obj instanceof JSONArray)) {
            final String msg = MessageFormat.format(
                    "The type \"{0}\" was encountered, rather than the expected type \"JSONArray\"", //$NON-NLS-1$
                    obj.getClass());
            throw new IOException(msg);
        }
        return (JSONArray)obj;
    }

    @Override
    public String toString() {
        return toString(true);
    }

    public String toString(boolean verbose) {
        StringWriter writer = new StringWriter();
        try {
            serialize(writer, verbose);
        } catch (IOException e) {
            return MessageFormat.format("Error serializing JSON object: {0}", e.getMessage()); //$NON-NLS-1$
        }
        return writer.toString();
    }
}
