/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.repository.common.json;


import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.Reader;
import java.io.StringWriter;
import java.io.Writer;
import java.text.MessageFormat;
import java.util.HashMap;

import com.ibm.team.repository.common.json.internal.Parser;
import com.ibm.team.repository.common.json.internal.Serializer;
import com.ibm.team.repository.common.json.internal.SerializerVerbose;

/**
 * Models a JSON Object.
 * 
 * Extension of HashMap that only allows String keys, and values which are JSON-able. 
 * JSON-able values are: null, and instances of String, Boolean, Number, JSONObject and JSONArray.
 *
 * Instances of this class are not thread-safe.
 */
public class JSONObject extends HashMap implements IJSONSerializable {

    private static final long serialVersionUID = -3269263069889337298L;

    /**
     * Return whether the object is a valid value for a property.
     */
    public static boolean isValidObject(Object object) {
        if (null == object)
            return true;
        return isValidType(object.getClass());
    }

    /**
     * Return whether the class is a valid type of value for a property.
     */
    public static boolean isValidType(Class clazz) {
        if (null == clazz)
            throw new IllegalArgumentException();

        if (String.class == clazz)
            return true;
        if (Boolean.class == clazz)
            return true;
        if (JSONObject.class == clazz)
            return true;
        if (JSONArray.class == clazz)
            return true;
        if (Number.class.isAssignableFrom(clazz))
            return true;

        return false;
    }

    /**
     * Convert a stream of JSON text into object form. 
     */
    static public JSONObject parse(Reader reader) throws IOException {
        reader = new BufferedReader(reader);
        Object obj = new Parser(reader).parse();
        if (!(obj instanceof JSONObject)) {
            final String msg = MessageFormat.format(
                    "The type \"{0}\" was encountered, rather than the expected type \"JSONObject\"", //$NON-NLS-1$
                    obj.getClass());
            throw new IOException(msg);
        }
        return (JSONObject)obj;
    }

    /**
     * Convert a stream of JSON text into some JSON object/array/value form. 
     */
    static public Object parseAny(Reader reader) throws IOException {
        reader = new BufferedReader(reader);
        return new Parser(reader).parse();
    }

    /**
     * Create a new instance of this class. 
     */
    public JSONObject() {
        super();
    }

    /**
     * Convert this object into a stream of JSON text.
     */
    public void serialize(Writer writer) throws IOException {
        serialize(writer, false);
    }

    /**
     * Convert this object into a stream of JSON text, specifying verbosity.
     */
    public void serialize(Writer writer, boolean verbose) throws IOException {
        Serializer serializer;
        writer = new BufferedWriter(writer);

        if (verbose) {
            serializer = new SerializerVerbose(writer);
        } else {
            serializer = new Serializer(writer);
        }

        serializer.writeObject(this).flush();
    }

    @Override
    public String toString() {
        return toString(true);
    }

    public String toString(boolean verbose) {
        StringWriter writer = new StringWriter();
        try {
            serialize(writer, verbose);
        } catch (IOException e) {
            return MessageFormat.format("Error serializing JSON object - exception message: {0}", e.getMessage()); //$NON-NLS-1$
        }
        return writer.toString();
    }

    /*
     * (non-Javadoc)
     * @see java.util.HashMap#put(java.lang.Object, java.lang.Object)
     */
    @Override
    public Object put(Object key, Object value) {
        if (null == key) {
            throw new IllegalArgumentException("key must not be null"); //$NON-NLS-1$
        }
        
        if (!(key instanceof String)) {
            throw new IllegalArgumentException("key must be a String"); //$NON-NLS-1$
        }

        if (!isValidObject(value)) {
            throw new IllegalArgumentException("invalid type of element: " + value.getClass().getName()); //$NON-NLS-1$ 
        }

        return super.put(key, value);
    }
}