/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/

package com.ibm.team.repository.common.json.internal;

import java.io.IOException;
import java.io.Reader;
import java.text.MessageFormat;

import com.ibm.team.repository.common.json.JSONArray;
import com.ibm.team.repository.common.json.JSONObject;

public class Parser {

    private final Tokenizer tokenizer;
    private Token lastToken;

    public Parser(Reader reader) throws IOException {
        super();
        this.tokenizer = new Tokenizer(reader);
    }

    public Object parse() throws IOException {
        lastToken = tokenizer.next();
        return parseValue();
    }

    public JSONObject parseObject() throws IOException {
        JSONObject result = new JSONObject();

        if (lastToken != Token.TokenBraceL)
            throw new IOException(MessageFormat.format("Expecting \"{\" at line/column: {0}", //$NON-NLS-1$ 
                    tokenizer.onLineCol()));
        lastToken = tokenizer.next();

        while (true) {
            if (lastToken == Token.TokenEOF)
                throw new IOException(MessageFormat.format("Unterminated object at line/column: {0}", //$NON-NLS-1$
                        tokenizer.onLineCol()));

            if (lastToken == Token.TokenBraceR) {
                lastToken = tokenizer.next();
                break;
            }

            if (!lastToken.isString())
                throw new IOException(MessageFormat.format("Expecting string key at line/column: {0}", //$NON-NLS-1$
                        tokenizer.onLineCol()));
            String key = lastToken.getString();

            lastToken = tokenizer.next();
            if (lastToken != Token.TokenColon)
                throw new IOException(MessageFormat.format("Expecting \":\" at line/column: {0}", //$NON-NLS-1$
                        tokenizer.onLineCol()));

            lastToken = tokenizer.next();
            Object val = parseValue();

            result.put(key, val);

            if (lastToken == Token.TokenComma) {
                lastToken = tokenizer.next();
            }

            else if (lastToken != Token.TokenBraceR) {
                throw new IOException(MessageFormat.format("Expecting either \",\" or \"}\" at line/column: {0}", //$NON-NLS-1$
                        tokenizer.onLineCol())); 
            }
        }

        return result;
    }

    public JSONArray parseArray() throws IOException {
        JSONArray result = new JSONArray();

        if (lastToken != Token.TokenBrackL)
            throw new IOException(MessageFormat.format("Expecting \"[\" at line/column: {0}", //$NON-NLS-1$
                    tokenizer.onLineCol()));
        lastToken = tokenizer.next();

        while (true) {
            if (lastToken == Token.TokenEOF)
                throw new IOException(MessageFormat.format("Unterminated array at line/column: {0}", //$NON-NLS-1$
                        tokenizer.onLineCol())); 

            if (lastToken == Token.TokenBrackR) {
                lastToken = tokenizer.next();
                break;
            }

            Object val = parseValue();

            result.add(val);

            if (lastToken == Token.TokenComma) {
                lastToken = tokenizer.next();
            } else if (lastToken != Token.TokenBrackR) {
                throw new IOException(MessageFormat.format("Expecting either \",\" or \"}\" at line/column: {0}", //$NON-NLS-1$
                        tokenizer.onLineCol())); 
            }
        }

        return result;
    }

    /**
     * 
     */
    public Object parseValue() throws IOException {
        if (lastToken == Token.TokenEOF)
            throw new IOException(MessageFormat.format("Expecting property value at line/column: {0}", //$NON-NLS-1$
                    tokenizer.onLineCol()));

        if (lastToken.isNumber()) {
            Object result = lastToken.getNumber();
            lastToken = tokenizer.next();
            return result;
        }

        if (lastToken.isString()) {
            Object result = lastToken.getString();
            lastToken = tokenizer.next();
            return result;
        }

        if (lastToken == Token.TokenFalse) {
            lastToken = tokenizer.next();
            return Boolean.FALSE;
        }

        if (lastToken == Token.TokenTrue) {
            lastToken = tokenizer.next();
            return Boolean.TRUE;
        }

        if (lastToken == Token.TokenNull) {
            lastToken = tokenizer.next();
            return null;
        }

        if (lastToken == Token.TokenBrackL)
            return parseArray();
        if (lastToken == Token.TokenBraceL)
            return parseObject();

        throw new IOException(MessageFormat.format("Invalid token at line/column: {0}", //$NON-NLS-1$
                tokenizer.onLineCol()));
    }
}