/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010, 2015. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/

package com.ibm.team.repository.common.json.internal;

import java.io.IOException;
import java.io.Writer;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import com.ibm.team.repository.common.json.JSONArray;
import com.ibm.team.repository.common.json.JSONObject;

public class Serializer {

    private final Writer writer;

    /**
     * Create a serializer on the specified output stream.
     * Note that you should create 
     */
    public Serializer(final Writer writer) {
        super();

        this.writer = writer;
    }

    public void flush() throws IOException {
        writer.flush();
    }

    public void close() throws IOException {
        writer.close();
    }

    public Serializer writeRawString(final String s) throws IOException {
        writer.write(s);
        return this;
    }

    public Serializer writeNull() throws IOException {
        writeRawString(String.valueOf((Object)null));
        return this;
    }

    public Serializer writeNumber(final Number value) throws IOException {
        if (null == value)
            return writeNull();

        if (value instanceof Float) {
        	Float casted_value = (Float) value;
        	if (casted_value.isNaN()|| casted_value.isInfinite())
        		return writeNull();
        }

        if (value instanceof Double) {
        	Double casted_value = (Double) value;
        	if (casted_value.isNaN()|| casted_value.isInfinite())
        		return writeNull();
        }

        return writeRawString(value.toString());
    }

    public Serializer writeBoolean(final Boolean value) throws IOException {
        if (null == value)
            return writeNull();

        writeRawString(value.toString());

        return this;
    }

    private String rightAlignedZero(final String s, final int len) {
        if (len == s.length())
            return s;

        StringBuffer sb = new StringBuffer(s);

        while (sb.length() < len) {
            sb.insert(0, '0');
        }

        return sb.toString();
    }

    public Serializer writeString(final String value) throws IOException {
        if (null == value)
            return writeNull();

        writer.write('"');

        char[] chars = value.toCharArray();

        for (int i = 0; i < chars.length; i++) {
            char c = chars[i];
            switch (c) {
            case '"':
                writer.write("\\\""); //$NON-NLS-1$
                break;
                /*
                 * See Workitem 59853
                 *     
            case '\'':
                writer.write("\\'"); //$NON-NLS-1$
                break;
                 */
            case '\\':
                writer.write("\\\\"); //$NON-NLS-1$
                break;
                
                /*
                 * See Workitem 59853
                 *     
            case 0:
                writer.write("\\0"); //$NON-NLS-1$
                break;
                 */
            case '\b':
                writer.write("\\b"); //$NON-NLS-1$
                break;
            case '\t':
                writer.write("\\t"); //$NON-NLS-1$
                break;
            case '\n':
                writer.write("\\n"); //$NON-NLS-1$
                break;
                /*
                 * See Workitem 59853
                 *     
            case 0x0B:
                writer.write("\\v"); //$NON-NLS-1$
                break;
                 */
            case '\f':
                writer.write("\\f"); //$NON-NLS-1$
                break;
            case '\r':
                writer.write("\\r"); //$NON-NLS-1$
                break;
            case '/':
                writer.write("\\/"); //$NON-NLS-1$
                break;
            default:
                if ((c >= 32) && (c <= 126)) {
                    writer.write(c);
                }
                /*
                 * See WorkItem 62700 
                else if (c < 255) {
                    writer.write("\\x"); //$NON-NLS-1$
                    writer.write(rightAlignedZero(Integer.toHexString(c), 2));
                }
                 */
                else {
                    writer.write("\\u"); //$NON-NLS-1$
                    writer.write(rightAlignedZero(Integer.toHexString(c), 4));
                }
            }
        }

        writer.write('"');

        return this;
    }

    /**
     * 
     */
    private Serializer write(final Object object) throws IOException {
        if (null == object)
            return writeNull();
        if (object instanceof Number)
            return writeNumber((Number) object);
        if (object instanceof String)
            return writeString((String) object);
        if (object instanceof Boolean)
            return writeBoolean((Boolean) object);
        if (object instanceof JSONObject)
            return writeObject((JSONObject) object);
        if (object instanceof JSONArray)
            return writeArray((JSONArray) object);

        throw new IOException(
                MessageFormat.format("Unsupported object for serialization: \"{0}\"", object)); //$NON-NLS-1$ 
    }

    /**
     * 
     */
    public Serializer writeObject(final JSONObject object) throws IOException {
        if (null == object)
            return writeNull();

        // write header
        writeRawString("{"); //$NON-NLS-1$
        indentPush();

        List propertyNames = getPropertyNames(object);
        for (Iterator iter = propertyNames.iterator(); iter.hasNext();) {
            Object key = iter.next();
            if (!(key instanceof String))
                throw new IOException(
                        MessageFormat.format("Invalid property name encountered while serializing: \"{0}\"", key)); //$NON-NLS-1$ 

            Object value = object.get(key);
            if (!JSONObject.isValidObject(value))
                throw new IOException(
                        MessageFormat.format("Invalid property value encountered while serializing: \"{0}\"", value)); //$NON-NLS-1$ 

            newLine();
            indent();
            //			writeRawString("\"");
            writeString((String) key);
            writeRawString(":"); //$NON-NLS-1$
            space();
            write(value);

            if (iter.hasNext())
                writeRawString(","); //$NON-NLS-1$
        }

        // write trailer
        indentPop();
        newLine();
        indent();
        writeRawString("}"); //$NON-NLS-1$

        return this;
    }

    /**
     * 
     */
    public Serializer writeArray(final JSONArray value) throws IOException {
        if (null == value)
            return writeNull();

        // write header
        writeRawString("["); //$NON-NLS-1$
        indentPush();

        for (Iterator iter = value.iterator(); iter.hasNext();) {
            Object element = iter.next();
            if (!JSONObject.isValidObject(element))
                throw new IOException(
                        MessageFormat.format("Invalid element encountered while serializing array: \"{0}\"", value)); //$NON-NLS-1$ 

            newLine();
            indent();
            write(element);

            if (iter.hasNext())
                writeRawString(","); //$NON-NLS-1$
        }

        // write trailer
        indentPop();
        newLine();
        indent();
        writeRawString("]"); //$NON-NLS-1$

        return this;
    }

    //---------------------------------------------------------------
    // pretty printing overridables
    //---------------------------------------------------------------

    /**
	 * @throws IOException  
	 */
    public void space() throws IOException {
    }

    /**
	 * @throws IOException  
	 */
    public void newLine() throws IOException {
    }

    /**
	 * @throws IOException  
	 */
    public void indent() throws IOException {
    }

    public void indentPush() {
    }

    public void indentPop() {
    }

    public List getPropertyNames(final Map map) {
        return new ArrayList(map.keySet());
    }
}