/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010, 2011. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/

package com.ibm.team.repository.common.serialize;

import java.io.Reader;
import java.io.Writer;
import java.util.Arrays;
import java.util.List;

import org.eclipse.emf.ecore.EObject;

import com.ibm.team.repository.common.IItem;
import com.ibm.team.repository.common.IItemType;
import com.ibm.team.repository.common.IManagedItem;
import com.ibm.team.repository.common.IVirtual;
import com.ibm.team.repository.common.IVirtualType;
import com.ibm.team.repository.common.serialize.internal.APPDeserializer;
import com.ibm.team.repository.common.serialize.internal.JSONDeserializer;
import com.ibm.team.repository.common.serialize.internal.UriSerializer;
import com.ibm.team.repository.common.serialize.internal.XMLDeserializer;
import com.ibm.team.repository.common.serialize.internal.nls.Messages;
import com.ibm.team.repository.common.transport.HttpUtil.MediaType;
import com.ibm.team.repository.common.util.NLS;

/**
 * The common interface for {@link IItem} deserialization by {@link MediaType}.
 */
public interface IDeserializer2 {

    /**
     * Deserializes a serialized item representation expressed in <code>reader</code> and
     * returns an {@link IManagedItem#isNewItem() new} item.
     * 
     * @param itemType The itemType of the item to create.
     * 
     * @param reader The reader containing the item representation. Must not be
     * <code>null</code>.
     * 
     * @return An item in the "new" state.  Never <code>null</code>. All fields 
     * for the type of item identified in the representation that that do not have values
     * serialized in the representation are set to their default values.
     * 
     * @throws SerializeException Thrown on error.
     */
    public IItem deserializeNewItem(IItemType itemType, Reader reader) throws SerializeException;
    
    /**
     * Deserializes an item representation expressed in <code>reader</code>
     * and returns a {@link IItem#isWorkingCopy() working copy} of the item
     * specified as <code>item</code>. 
     * 
     * <b>Note:</b> There is no attempt to merge any serialized reference lists
     * with those in the specified <code>item</code>.  Any serialized
     * reference lists present in <code>reader</code> replace those lists in
     * <code>item</code>.
     * 
     * @param item The item against which to serialize edits. Must not be <code>null</code>.
     * Must be a {@link IItem#isWorkingCopy() working copy}.
     * 
     * @param reader The reader containing the serialized item representation.
     * Must not be <code>null</code>.
     * 
     * @return A working copy of the item referenced in the argument list. Never
     * <code>null</code>.  Only the field values that are present
     * in the serialized representation are set on the working copy.  In effect,
     * the referenced item is updated with edits of the serialized values, and
     * returned.
     * 
     * @throws SerializeException Thrown on error.
     */
    public IItem deserializeWorkingCopy(IItem item, Reader reader) throws SerializeException;
    
    /**
     * Deserializes an item representation expressed in <code>reader</code>
     * and returns an immutable item.
     *
     * @param reader The reader containing the item representation. Must not be
     * <code>null</code>. The referenced representation is expected to have as a
     * minimum members describing the item's location and type, as well as its etag.  
     * 
     * @return An immutable item.  Never <code>null</code>. Fields
     * not available in the representation as set to their default values, but
     * not marked as "set".
     * 
     * @throws SerializeException Thrown on error.
     */
    public IItem deserializeItem(Reader reader) throws SerializeException;

    /**
     * Deserializes serialized item representations expressed in <code>reader</code>
     * and returns immutable items.
     *  
     * @param reader The reader containing the item representations. Must not be
     * <code>null</code>.   The referenced representations are expected to 
     * contain at minimum members describing the item's location and type, 
     * as well as its etag.  
     * 
     * @return A list of immutable items.  Never <code>null</code>. All fields 
     * for the type of item identified in the representation that that do not
     * have values serialized in the representation are set to their default values.
     * 
     * @throws SerializeException Thrown on error.
     */
    public List<IItem> deserializeItems(Reader reader) throws SerializeException;
    
    /**
     * Deserializes a virtual resource representation expressed in <code>reader</code>
     * and returns a resource instance.
     *
     * @param reader The reader containing the resource representation. Must not be
     * <code>null</code>. The referenced representation is expected to have as a
     * minimum members describing the item's location and type.
     * 
     * @return A virtual resource.  Never <code>null</code>. Fields
     * not available in the representation as set to their default values, but
     * not marked as "set".
     * 
     * @throws SerializeException Thrown on error.
     */
    public IVirtual deserializeVirtual(Reader reader) throws SerializeException;

    /**
     * Deserializes a virtual resource representation expressed in <code>reader</code>
     * and returns a resource instance.
     *
     * @param type The type of virtual represented in the serialized resource.
     * May be <code>null</code>.  If <code>null</code>, then this method behaves
     * identically to {@link #deserializeVirtual(Reader)}, in that the representation
     * must contain typing information.
     * 
     * @param reader The reader containing the resource representation. Must not be
     * <code>null</code>. The referenced representation is expected to have as a
     * minimum members describing the item's location and type.
     * 
     * @return A virtual resource.  Never <code>null</code>. Fields
     * not available in the representation as set to their default values, but
     * not marked as "set".
     * 
     * @throws SerializeException Thrown on error.
     */
    public IVirtual deserializeVirtual(IVirtualType type, Reader reader) throws SerializeException;
    
    /**
     * Deserializes serialized virtual resource representations expressed in 
     * <code>reader</code> and returns virtual items.
     *  
     * @param reader The reader containing the resource representations. Must not be
     * <code>null</code>.   The referenced representations are expected to 
     * contain at minimum members describing the item's location and type. 
     * 
     * @return A list of virtual resources.  Never <code>null</code>. All fields 
     * for the type of item identified in the representation that that do not
     * have values serialized in the representation are set to their default values.
     * 
     * @throws SerializeException Thrown on error.
     */
    public List<IVirtual> deserializeVirtuals(Reader reader) throws SerializeException;
    
    /**
     * Deserializes serialized virtual resource representations expressed in 
     * <code>reader</code> and returns virtual items.
     *  
     * @param type The type of virtuals represented in the serialized resource.
     * May be <code>null</code>.  If <code>null</code>, then this method behaves
     * identically to {@link #deserializeVirtual(Reader)}, in that the representation
     * must contain typing information.  If not <code>null</code>, then all of
     * the representations must be of this type.
     * 
     * @param reader The reader containing the resource representations. Must not be
     * <code>null</code>.   The referenced representations are expected to 
     * contain at minimum members describing the item's location and type. 
     * 
     * @return A list of virtual resources.  Never <code>null</code>. All fields 
     * for the type of item identified in the representation that that do not
     * have values serialized in the representation are set to their default values.
     * 
     * @throws SerializeException Thrown on error.
     */
    public List<IVirtual> deserializeVirtuals(IVirtualType type, Reader reader) throws SerializeException;
    
    /**
     * Returns the media type that this deserializer uses.
     * 
     * @return The media type that this deserializer was constructed with.
     * Never <code>null</code>.
     */
    public MediaType getMediaType();

    
    /**
     * Counterpart to {@link ISerializer2#serializeException(Throwable, Integer, Writer)}.
	 * The given stack adjuster may be null. 
     * @param classloader 
     * @throws SerializeException 
     */
    public Throwable deserializeException(Reader r, ClassLoader classloader, IStackAdjuster adj) throws SerializeException;
    
    /**
     * Factory for the creation of deserializers by media type.
     * 
     * @NoImplement Clients must not implement this interface.
     */
    public interface Factory {
        /**
         * Creates and returns a new deserializer instance.
         * 
         * @param mType The media type determining the deserializer to create.
         * Must not be <code>null</code>.
         *  
         * @param The serializer used to convert between objects and URIs. Must not
         * be <code>null</code>.
         * 
         * @return A deserializer instance for the specified media type.
         * May be <code>null</code>.
         * 
         * @throws UnsupportedOperationException Thrown if the referenced
         * media type is not recognized.
         */
        public IDeserializer2 newInstance(MediaType mType, IURISerializer uriSerializer);

		/**
		 * Returns whether or not the factory supports deserializing for the passed in media type
		 * 
		 * @return
		 */
		public boolean isSupportedMediaType(MediaType type);

    }

    /**
     * The static factory instance with which to create deserializers.
     */
    /**
     * 
     */
    public static final Factory FACTORY = new Factory() {
        
        /* (non-Javadoc)
         * @see com.ibm.team.repository.common.serialize.IDeserializer2.Factory#newInstance(com.ibm.team.repository.common.transport.HttpUtil.MediaType)
         */
        public IDeserializer2 newInstance(MediaType mType, IURISerializer uriSerializer) {

            if (uriSerializer == null)
                uriSerializer = new UriSerializer();
            
            if (mType == MediaType.XML)
                return new XMLDeserializer(mType, uriSerializer);
            else if (mType == MediaType.JSON || mType == MediaType.TEXT_PLAIN || mType == MediaType.ANY)
                return new JSONDeserializer(mType, uriSerializer);
            else if (mType == MediaType.APP)
                return new APPDeserializer(mType, uriSerializer);
            
            String msg = NLS.bind(Messages.getServerString("IDeserializer2.UnknownMediaType"), mType); //$NON-NLS-1$
            throw new UnsupportedOperationException(msg);
        }

		public boolean isSupportedMediaType(MediaType type) {
			if (type == null)
				return false;
			return getSupportedMediaTypes().contains(type);
		}

		public List<MediaType> getSupportedMediaTypes() {
			return Arrays.asList(new MediaType[] { MediaType.XML, MediaType.JSON, MediaType.TEXT_PLAIN, MediaType.ANY });
		}
    };

    /**
     * FOR INTERNAL TESTING USE ONLY
     */
    public Object deserialize(EObject object, Reader reader) throws SerializeException;

}
