/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010, 2014. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/

package com.ibm.team.repository.common.serialize;

import java.io.IOException;
import java.io.Writer;
import java.util.List;

import org.eclipse.emf.ecore.EObject;

import com.ibm.team.repository.common.IItem;
import com.ibm.team.repository.common.IVirtual;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.common.serialize.internal.APPSerializer;
import com.ibm.team.repository.common.serialize.internal.JSONSerializer;
import com.ibm.team.repository.common.serialize.internal.UriSerializer;
import com.ibm.team.repository.common.serialize.internal.XMLSerializer;
import com.ibm.team.repository.common.serialize.internal.nls.Messages;
import com.ibm.team.repository.common.transport.HttpUtil.MediaType;
import com.ibm.team.repository.common.util.NLS;

/**
 * The common interface for {@link IItem} serialization by {@link MediaType}.
 */
public interface ISerializer2 {

    /**
     * Serializes the referenced <code>item</code> to the specified <code>writer</code>.
     * 
     * @param item The item to serialize.  Must not be <code>null</code>.
     * 
     * @param writer The writer to serialize to. Must not be <code>null</code>.
     */
    public void serialize(IItem item, Writer writer) 
        throws TeamRepositoryException, IOException;

    /**
     * Serializes the referenced virtual <code>item</code> to the specified <code>writer</code>.
     * 
     * @param item The virtual resource to serialize.  Must not be <code>null</code>.
     * 
     * @param writer The writer to serialize to. Must not be <code>null</code>.
     */
    public void serialize(IVirtual item, Writer writer)
        throws TeamRepositoryException, IOException;
    
    /**
     * Serializes the referenced <code>items</code> to the specified <code>writer</code>.
     * 
     * @param items The items to serialize. Must not be <code>null</code>.
     * 
     * @param writer The writer to serialize to. Must not be <code>null</code>.
     *
     * @throws TeamRepositoryException Thrown on error.
     * @throws IOException Thrown on error.
     */
    public void serialize(List<IItem> items, Writer writer) 
        throws TeamRepositoryException, IOException;

    /**
     * Serializes the referenced virtual <code>items</code> to the specified <code>writer</code>.
     * 
     * @param items The virtual resources to serialize. Must not be <code>null</code>.
     * 
     * @param writer The writer to serialize to. Must not be <code>null</code>.
     *
     * @throws TeamRepositoryException Thrown on error.
     * @throws IOException Thrown on error.
     */
    public void serializeVirtuals(List<IVirtual> items, Writer writer) 
        throws TeamRepositoryException, IOException;

    /**
     * Serializes the referenced list of <code>items</code> to the specified
     * <code>writer</code>.
     * 
     * @param items The items to serialize.  Will be of type IItem or IVirtual.
     * Must not be <code>null</code>.
     * 
     * @param writer The writer to serialize to. Must not be <code>null</code>.
     * 
     * @throws TeamRepositoryException Thrown on error.
     * @throws IOException Thrown on error.
     */
    public void serializeList(List items, Writer writer)
        throws TeamRepositoryException, IOException;
    
    /**
     * Serializes the reference <code>exception</code> to the specified <code>writer</code>.
     * 
     * @param exception The exception to serialize. Must not be <code>null</code>.
     * 
     * @param errorCode The errorCode to serialize. May be <code>null</code>.
     * 
     * @param writer The writer to serialize to. Must not be <code>null</code>.
     * 
     * @throws TeamRepositoryException Thrown on error.
     * @throws IOException Thrown on error.
     */
    public void serializeException(Throwable exception, Integer errorCode, Writer writer)
        throws TeamRepositoryException, IOException;
        
    /**
     * Returns the media type that this deserializer uses.
     * 
     * @return The media type that this deserializer was constructed with.
     * Never <code>null</code>.
     */
    public MediaType getMediaType();
    
    /**
     * Returns whether or not the serialized representation will be formatted
     * for human readability.
     * 
     * @return <code>true</code> if the representation will be formatted,
     * <code>false</code> if not.
     */
    public boolean format();
    
    /**
     * Factory for creating instances of a serializer for the given media type.
     * @NoImplement Clients must not implement this interface.
     */
    interface Factory {

        /**
         * Creates and returns a new serializer instance.
         * 
         * @param mType The media type determining the serializer to create.
         * Must not be <code>null</code>.
         *  
         * @param repoUri Not used.
         *  
         * @param format Whether or not the serialized representation should
         * be formatted for human-readability.  <code>true</code> to 
         * "pretty-print" format (increasing white-space and verbosity), 
         * <code>false</code> to not include formatting characters.
         * 
         * @param uriSerializer A serializer used to transform object references to
         * URIs. May be <code>null</code>.
         * 
         * @return A serializer instance for the specified media type.
         * Never <code>null</code>.
         * 
         * @throws UnsupportedOperationException Thrown if the referenced
         * media type is not recognized.
         * 
         * @deprecated use newInstance(MediaType mType, boolean format, IURISerializer uriSerializer) instead
         */
        public ISerializer2 newInstance(MediaType mType, String repoUri, boolean format, IURISerializer uriSerializer); 

        /**
         * Creates and returns a new serializer instance.
         * 
         * @param mType The media type determining the serializer to create.
         * Must not be <code>null</code>.
         *  
         * @param format Whether or not the serialized representation should
         * be formatted for human-readability.  <code>true</code> to 
         * "pretty-print" format (increasing white-space and verbosity), 
         * <code>false</code> to not include formatting characters.
         * 
         * @param uriSerializer A serializer used to transform object references to
         * URIs. May be <code>null</code>.
         * 
         * @return A serializer instance for the specified media type.
         * Never <code>null</code>.
         * 
         * @throws UnsupportedOperationException Thrown if the referenced
         * media type is not recognized.
         */
        public ISerializer2 newInstance(MediaType mType, boolean format, IURISerializer uriSerializer); 
    }

    /**
     * The static factory instance with which to create serializers. 
     */
    public static final Factory FACTORY = new Factory() {

        public ISerializer2 newInstance(MediaType mType, String repoUri, boolean format, 
                IURISerializer uriSerializer)
        {
        	return newInstance(mType, format, uriSerializer);
        }

        public ISerializer2 newInstance(MediaType mType, boolean format, 
                IURISerializer uriSerializer)
        {
            if (uriSerializer == null)
                uriSerializer = new UriSerializer();
            
            if (mType == MediaType.XML)
                return new XMLSerializer(mType, format, uriSerializer);
            else if (mType == MediaType.JSON || mType == MediaType.TEXT_PLAIN || mType == MediaType.ANY)
                return new JSONSerializer(mType, format, uriSerializer);
            else if (mType == MediaType.APP)
                return new APPSerializer(mType, format, uriSerializer);
            
            String msg = NLS.bind(Messages.getServerString("ISerializer2.UnknownMediaType"), mType); //$NON-NLS-1$
            throw new UnsupportedOperationException(msg);
        }
    };

    /**
     * FOR INTERNAL TESTING USE ONLY
     */
    public void serialize(EObject object, Writer writer) 
        throws TeamRepositoryException, IOException;
    
}
