/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010, 2017. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.repository.common.serialize;

import java.lang.reflect.Method;
import java.util.UUID;

import org.apache.commons.logging.Log;

import com.ibm.team.repository.common.serialize.internal.AbstractSerializer;
import com.ibm.team.repository.common.serialize.internal.JSONSerializer;
import com.ibm.team.repository.common.serialize.internal.nls.Messages;
import com.ibm.team.repository.common.util.NLS;

/**
 * Utility classes for modifying stack traces.
 */
public abstract class StackUtil {
    
	/* By default, exceptions are serialized.  If we want to meet security
	 * requirements, we have to be configured to be secure by default.  But
	 * until that adoption is accepted, we preserve the current behavior by
	 * default. */
//	public static final String NEVER_SERIALIZE_EXCEPTION_PROPERTY = StackTraceUtil.NEVER_SERIALIZE_EXCEPTION_PROPERTY;
//	public static final String NEVER_SERIALIZE_EXCEPTION_DEFAULT = "true"; //$NON-NLS-1$
	public static final String NEVER_SERIALIZE_EXCEPTION_PROPERTY = "com.ibm.team.repository.neverSerializeExceptions"; //$NON-NLS-1$
	public static final String NEVER_SERIALIZE_EXCEPTION_DEFAULT = "true"; //$NON-NLS-1$

	
	/**
     * Shave the given stack trace down to the first invocation of the given 
     * fully qualified classname and method. 
     */
    public static StackTraceElement[] pruneDownTo(Method m, StackTraceElement[] trace) {
    	int pruneTo = 0;
    	
    	// Find the method to prune to
    	String className = m.getDeclaringClass().getCanonicalName();
    	String methodName = m.getName();
    	for ( ; pruneTo < trace.length; pruneTo++ ) {
    		StackTraceElement frame = trace[pruneTo];
    		if (frame.getClassName().equals(className)
    				&& frame.getMethodName().equals(methodName)) {
    			break;
    		}
    	}

    	if (pruneTo >= trace.length) {
    		return trace;
    	}

    	// Copy the frames into a newer, shorter stack
    	int start = 0;
    	int length = trace.length - pruneTo;
    	
    	StackTraceElement[] toReturn = new StackTraceElement[length];
    	System.arraycopy(trace, pruneTo, toReturn, start, length - start);
    	
    	return toReturn;
    }

	/**
     * Remove bottom-most frames of the stack, up to the first frame referencing 
	 * a class ending with the given classname. 
     */
    public static StackTraceElement[] pruneUpTo(String className, StackTraceElement[] trace) {
    	int pruneTo = trace.length - 1;
    	
    	// Find the method to prune to
    	for ( ; pruneTo >= 0; pruneTo-- ) {
    		StackTraceElement frame = trace[pruneTo];
    		if (frame.getClassName().endsWith(className)) {
    			break;
    		}
    	}

    	if (pruneTo < 0) {
    		return trace;
    	}

    	// Copy the frames into a newer, shorter stack
    	int size = pruneTo + 1;
    	
    	StackTraceElement[] toReturn = new StackTraceElement[size];
    	System.arraycopy(trace, 0, toReturn, 0, toReturn.length);
    	
    	return toReturn;
    }
    
    /** Append one stack trace to another. */
    public static StackTraceElement[] append(StackTraceElement[] trace, StackTraceElement[] toAppend) {
    	StackTraceElement[] toReturn = new StackTraceElement[trace.length + toAppend.length];
    	
		System.arraycopy(trace, 0, toReturn, 0, trace.length);
		System.arraycopy(toAppend, 0, toReturn, trace.length, toAppend.length);
		
		return toReturn;
    }
    
	
    public static boolean shouldAvoidSerializingExceptions() {
		// return StackTraceUtil.shouldAvoidSerializingExceptions();
    	return System.getProperty( NEVER_SERIALIZE_EXCEPTION_PROPERTY, NEVER_SERIALIZE_EXCEPTION_DEFAULT).equalsIgnoreCase("true"); //$NON-NLS-1$
	}
    
    /**
     * Format exception information omitting stack traces. Clients should call this method after checking {@link AbstractSerializer#shouldAvoidSerializingExceptions()}.
     * The message directs user to use a unique ID to locate information in the log file and shows names of the exceptions and their messages in the exceptions chain.
     * @param exception {@link Throwable} to format.
     * @param log Optional {@link Log} to log the formatted message to or <code>null</code>, in which case callers must ensure that the message is logged with the same ID that appears in the message.
     * @param identifyingUUID {@link UUID} Optional UUID to use as identifier for the error. If <code>null</code>, the method will generate it's own random id.
     * The identifier will be included in the message and written into the log file. Clients must ensure that the id displayed to the user is the same as the one
     * written into the log, so the error details can be easily found.
     * @return String formatted message.
     */
    public static final String formatNoStackTrace( Throwable exception, Log log, UUID identifyingUUID ){
    	// See also: com.ibm.team.jfs.app.util.ErrorUtils.formatNoStackTrace(Throwable, HttpServletRequest, Log, UUID)
		if( identifyingUUID == null )
			identifyingUUID = UUID.randomUUID();
		if( log != null ){
			String logPattern = Messages.getServerString(JSONSerializer.NLS_EXCEPTION_PLACEHOLDER_LOGGED_MESSAGE);
			String logMessage = NLS.bind(logPattern, identifyingUUID);
			log.error(logMessage, exception);
		}
		String userPattern = Messages.getClientString(JSONSerializer.NLS_EXCEPTION_PLACEHOLDER_USER_MESSAGE);
		String userMessage = NLS.bind(userPattern, exception.getMessage(), identifyingUUID); // print ID and message
		
		return userMessage;
    }
    
}
