/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010, 2016. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/

package com.ibm.team.repository.common.serialize.internal;

import org.eclipse.emf.ecore.EObject;

import com.ibm.team.repository.common.IItem;
import com.ibm.team.repository.common.serialize.IPrimitiveRegistry;
import com.ibm.team.repository.common.serialize.ISerializer2;
import com.ibm.team.repository.common.serialize.IURISerializer;
import com.ibm.team.repository.common.transport.HttpUtil.MediaType;

/**
 * The abstract base class for item serializers.
 */
public abstract class AbstractSerializer implements ISerializer2 {
	
    protected static final IPrimitiveRegistry PrimitiveReg = IPrimitiveRegistry.INSTANCE;

    private final MediaType _mType;
    private final boolean _format;
    
    protected final IURISerializer _uriSerializer;

    /**
     * Constructs an abstract serializer.
     * 
     * @param mType The media type for this serializer. Must not be <code>null</code>.
     * 
     * @param format Whether or not to format the representation to be more
     * easily read by humans.
     * 
     * @param The serializer used to convert between objects and URIs. Must not
     * be <code>null</code>.
     */
    AbstractSerializer(MediaType mType, boolean format, IURISerializer uriSerializer) {
        
        if (mType == null)
            throw new IllegalArgumentException("MediaType must not be null"); //$NON-NLS-1$
        if (uriSerializer == null)
            throw new IllegalArgumentException("URI Serializer must not be null"); //$NON-NLS-1$
        
        _mType = mType;
        _format = format;
        _uriSerializer = uriSerializer;
    }

    /* (non-Javadoc)
     * @see com.ibm.team.repository.common.serialize.IDeserializer2#getMediaType()
     */
    public MediaType getMediaType() {
        return _mType;
    }

    /* (non-Javadoc)
     * @see com.ibm.team.repository.common.serialize.ISerializer2#format()
     */
    public boolean format() {
        return _format;
    }
    
    @SuppressWarnings("nls")
	protected String encode( String text, boolean full ){
    	text = text.replaceAll("&", "&amp;"). 
					replaceAll("<", "&lt;").
					replaceAll(">", "&gt;");
    	if( full ){
			text = text.replaceAll("\"", "&quot;").
				    replaceAll("'", "&#x27;").
					replaceAll("/", "&#x2F;");
    	}
    	return text;
    }
    
   /**
    * Returns whether or not the referenced eObject should be converted to 
    * a handle, based on the class type of <code>expectedType</code>.
    *  
    * @param eObject The eObject to inspect. May be <code>null</code>.
    * Not required to be derived from {@link IItem}, although this routine
    * only makes sense if it is.  If <code>null</code> <code>false</code> is
    * returned.
    * 
    * @param expectedType The expected type of the referenced eObject.  If
    * the expected type derives from {@link IItem}, then the eObject shouldn't
    * be represented as a handle, else, it should.
    * 
    * @return Return <code>true</code> if the referenced <code>eObject</code> should
    * be represented as a handle, <code>false</code> if it should not.
    */
    protected boolean shouldMorphItemToHandle(EObject eObject, Class expectedType) {
       if (null == eObject)
           return false;

       if (!(eObject instanceof IItem))
           return false;

       /*
        * It's derived from Item, but check to see if expected type is.
        * If the expected type is an Item, don't morph it.
        */
       if (IItem.class.isAssignableFrom(expectedType))
           return false;

       return true;
   }

   /**
    * Returns a handle for the referenced object.
    * 
    * @param item The item to retrieve the handle for. May be <code>null</code>.
    * Must be derived from {@link IItem}.
    * 
    * @return The handle for the referenced item, or <code>null</code> if <code>null</code>
    * is passed as the item. 
    */
   public EObject morphItemToHandle(EObject item) {
       if (null == item)
          return null;

       return (EObject) ((IItem) item).getItemHandle();
   }
}
