/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010, 2014. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/

package com.ibm.team.repository.common.serialize.internal;

import java.io.PrintWriter;
import java.io.Reader;
import java.io.StringWriter;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EEnum;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.EcorePackage;

import com.ibm.team.repository.common.IContent;
import com.ibm.team.repository.common.IHelper;
import com.ibm.team.repository.common.IHelperType;
import com.ibm.team.repository.common.IItem;
import com.ibm.team.repository.common.IItemHandle;
import com.ibm.team.repository.common.IItemType;
import com.ibm.team.repository.common.IType;
import com.ibm.team.repository.common.IVirtual;
import com.ibm.team.repository.common.IVirtualType;
import com.ibm.team.repository.common.Location;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.common.UUID;
import com.ibm.team.repository.common.internal.marshal.util.MarshallerUtil;
import com.ibm.team.repository.common.internal.util.ItemUtil;
import com.ibm.team.repository.common.json.JSONArray;
import com.ibm.team.repository.common.json.JSONObject;
import com.ibm.team.repository.common.model.Item;
import com.ibm.team.repository.common.model.RepositoryPackage;
import com.ibm.team.repository.common.model.impl.VirtualImpl;
import com.ibm.team.repository.common.serialize.IDeserializer2;
import com.ibm.team.repository.common.serialize.IPrimitiveHandler;
import com.ibm.team.repository.common.serialize.IPrimitiveRegistry;
import com.ibm.team.repository.common.serialize.IStackAdjuster;
import com.ibm.team.repository.common.serialize.IURISerializer;
import com.ibm.team.repository.common.serialize.SerializeException;
import com.ibm.team.repository.common.serialize.internal.nls.Messages;
import com.ibm.team.repository.common.transport.HttpUtil.MediaType;
import com.ibm.team.repository.common.util.NLS;

/**
 * The implementation of the {@link IDeserializer2} interface for JSON
 * encodings.
  */
public class JSONDeserializer extends AbstractDeserializer {

    static private final String KEY_URI = JSONSerializer.KEY_URI;
    static private final String KEY_ETAG = JSONSerializer.KEY_ETAG;

    private IJSONDeserializationParticipant participant = new IJSONDeserializationParticipant() {
		public Throwable deserializeErrorObj(JSONDeserializer d, JSONObject json, Throwable t) {
			return t;
		}
	};
    
    /**
     * The constructor.
     * 
     * @param type The media type. Must not be <code>null</code>.
     * 
     * @param The serializer used to convert between objects and URIs. Must not
     * be <code>null</code>.
     */
    public JSONDeserializer(final MediaType type, IURISerializer uriSerializer) {
        super(type, uriSerializer);
    }

    /**
     * Add an object that is allowed to modify deserialized objects before 
     * they are returned. 
     * 
     * @param participant A non-<code>null</code> participant.
     * 
     * @throws IllegalArgumentException if the passed participant is 
     * 		<code>null</code>.
     */
    public void setParticipant(IJSONDeserializationParticipant participant) {
    	if (null == participant) {
    		throw new IllegalArgumentException("Participant may not be null."); //$NON-NLS-1$
    	}
    	
    	this.participant = participant;
    }
    
    /* (non-Javadoc)
     * @see com.ibm.team.repository.common.serialize.IDeserializer2#deserializeNewItem(java.io.Reader)
     */
    public IItem deserializeNewItem(final IItemType itemType, final Reader reader) throws SerializeException {
        if (reader == null)
            throw new IllegalArgumentException("reader must not be null"); //$NON-NLS-1$

        final JSONObject jsonObject = getObject(reader);

        final Item item = (Item)itemType.createItem();
        item.initNew();
        final EObject itemObj = (EObject)item.getAdapter(EObject.class);
        deserializeObject(itemObj, jsonObject);
        
        return item;
    }

    /* (non-Javadoc)
     * @see com.ibm.team.repository.common.serialize.IDeserializer2#deserializeWorkingCopy(com.ibm.team.repository.common.IItem, java.io.Reader)
     */
    public IItem deserializeWorkingCopy(final IItem item, final Reader reader) throws SerializeException {
        if (reader == null)
            throw new IllegalArgumentException("reader must not be null"); //$NON-NLS-1$

        if (!item.isWorkingCopy())
            throw new IllegalArgumentException("referenced item must be a working copy"); //$NON-NLS-1$
        
        final JSONObject jsonObject = getObject(reader);

        final EObject itemObj = (EObject)item.getAdapter(EObject.class);
        deserializeObject(itemObj, jsonObject);
        
        return item;
    }

    /* (non-Javadoc)
     * @see com.ibm.team.repository.common.serialize.IDeserializer2#deserializeItem(java.io.Reader)
     */
    public IItem deserializeItem(final Reader reader) throws SerializeException {
        
        if (reader == null)
            throw new IllegalArgumentException("reader must not be null"); //$NON-NLS-1$

        final JSONObject jsonObject = getObject(reader);
        final IItem item = toItem(jsonObject);
        return item;
    }

    /* (non-Javadoc)
     * @see com.ibm.team.repository.common.serialize.IDeserializer2#deserializeItems(java.io.Reader)
     */
    public List<IItem> deserializeItems(final Reader reader) throws SerializeException {
        
        if (reader == null)
            throw new IllegalArgumentException("reader must not be null"); //$NON-NLS-1$

        JSONArray jsonArray = null;
        try {
            jsonArray = JSONArray.parse(reader);
        }
        catch (final Exception e) {
            final String msg = NLS.bind(Messages.getServerString("JSONDeserializer.ErrorArrayParse"), e.getMessage()); //$NON-NLS-1$
            throw new SerializeException(msg, e);
        }
        
        final ArrayList<IItem> items = new ArrayList<IItem>(jsonArray.size());
        for (final Object obj : jsonArray) {
            if (!(obj instanceof JSONObject)) {
                final String msg = Messages.getServerString("JSONDeserializer.ErrorInvalidArrayMembers"); //$NON-NLS-1$
                throw new SerializeException(msg);
            }
            final JSONObject jsonObj = (JSONObject)obj;
            items.add(toItem(jsonObj));
        }
        
        return items;
    }
    
    /* (non-Javadoc)
     * @see com.ibm.team.repository.common.serialize.IDeserializer2#deserializeVirtual(java.io.Reader)
     */
    public IVirtual deserializeVirtual(final Reader reader) throws SerializeException {
        if (reader == null)
            throw new IllegalArgumentException("reader must not be null"); //$NON-NLS-1$

        return deserializeVirtual(null, reader);
    }

    /* (non-Javadoc)
     * @see com.ibm.team.repository.common.serialize.IDeserializer2#deserializeVirtual(com.ibm.team.repository.common.IVirtualType, java.io.Reader)
     */
    public IVirtual deserializeVirtual(final IVirtualType type, final Reader reader) throws SerializeException {
        if (reader == null)
            throw new IllegalArgumentException("reader must not be null"); //$NON-NLS-1$
        
        final JSONObject jsonObject = getObject(reader);
        final IVirtual item = toVirtual(type, jsonObject);
        return item;
    }

    /* (non-Javadoc)
     * @see com.ibm.team.repository.common.serialize.IDeserializer2#deserializeVirtuals(java.io.Reader)
     */
    public List<IVirtual> deserializeVirtuals(final Reader reader) throws SerializeException {
        if (reader == null)
            throw new IllegalArgumentException("reader must not be null"); //$NON-NLS-1$

        return deserializeVirtuals(null, reader);
    }

    /* (non-Javadoc)
     * @see com.ibm.team.repository.common.serialize.IDeserializer2#deserializeVirtuals(com.ibm.team.repository.common.IVirtualType, java.io.Reader)
     */
    public List<IVirtual> deserializeVirtuals(final IVirtualType type, final Reader reader) throws SerializeException {
        if (reader == null)
            throw new IllegalArgumentException("reader must not be null"); //$NON-NLS-1$

        JSONArray jsonArray = null;
        try {
            jsonArray = JSONArray.parse(reader);
        }
        catch (final Exception e) {
            final String msg = NLS.bind(Messages.getServerString("JSONDeserializer.ErrorParsingArray"), e.getMessage()); //$NON-NLS-1$
            throw new SerializeException(msg, e);
        }
        
        final ArrayList<IVirtual> items = new ArrayList<IVirtual>(jsonArray.size());
        for (final Object obj : jsonArray) {
            if (!(obj instanceof JSONObject)) {
                final String msg = Messages.getServerString("JSONDeserializer.ErrorInvalidArrayMembers"); //$NON-NLS-1$
                throw new SerializeException(msg);
            }
            final JSONObject jsonObj = (JSONObject)obj;
            items.add(toVirtual(type, jsonObj));
        }
        
        return items;
    }

    /* (non-Javadoc)
     * @see com.ibm.team.repository.common.serialize.IDeserializer2#deserialize(org.eclipse.emf.ecore.EObject, java.io.Reader)
     */
    public Object deserialize(final EObject object, final Reader reader) throws SerializeException {
        
        final JSONObject jsonObject = getObject(reader);
        
        deserializeObject(object, jsonObject);
        return object;
    }

    /**
     * Parses a JSON object from the referenced reader.
     * 
     * @param reader The reader to read from. Must not be <code>null</code>.
     * 
     * @return The parsed JSON object. Never <code>null</code>.
     * 
     * @throws SerializeException Thrown on parse or read errors.
     */
    static private JSONObject getObject(final Reader reader) throws SerializeException {
        JSONObject jsonObject = null;
        try {
            jsonObject = JSONObject.parse(reader);
        }
        catch (final Exception e) {
            final String msg = NLS.bind(Messages.getServerString("JSONDeserializer.ErrorObjectParsing"), e.getMessage()); //$NON-NLS-1$
            throw new SerializeException(msg, e);
        }
        return jsonObject;
    }

    /**
     * Deserializes the referenced json object into an immutable item, as
     * identified by that representation.
     * 
     * @param jsonObject The object containing the representation. Must not
     * be <code>null</code>.
     * 
     * @return The deserialized item. Never <code>null</code>. Immutable.
     * 
     * @throws SerializeException Thrown on error.
     */
    private IItem toItem(final JSONObject jsonObject) throws SerializeException {
        final Item item = getItem(jsonObject);
        final EObject itemObj = (EObject)item.getAdapter(EObject.class);
        deserializeObject(itemObj, jsonObject);
        item.makeImmutable();
        item.protect();
        return item;
    }

    /**
     * Deserializes the referenced json object into a virtual resource,
     * identified by that representation.
     *
     * @param type The type to convert the json object to.  May be 
     * <code>null</code>. If <code>null</code>, the json object representation
     * must contain typing information.
     * 
     * @param jsonObject The object containing the representation. Must not
     * be <code>null</code>.
     * 
     * @return The deserialized resource. Never <code>null</code>.
     * 
     * @throws SerializeException Thrown if type information is not available,
     * and on general error.
     */
    private IVirtual toVirtual(IVirtualType type, final JSONObject jsonObject)
        throws SerializeException
    {
        if (type == null) 
            type = getVirtualType(jsonObject);
        if (type == null) {
            final String msg = Messages.getServerString("JSONDeserializer.ErrorMissingType"); //$NON-NLS-1$
            throw new SerializeException(msg);
        }

        final IVirtual virtual = type.createVirtual();
        final EObject virtualObj = (EObject)virtual.getAdapter(EObject.class);
        deserializeObject(virtualObj, jsonObject);
        
        /*
         * Update uri and etag
         */
        VirtualImpl impl = (VirtualImpl)virtual.getAdapter(VirtualImpl.class);
        impl.setLocation((String)jsonObject.get(KEY_URI));
        final UUID etag = getETag(jsonObject);
        if (etag != null)
            impl.setEtag(etag.getUuidValue());
        
        return virtual;
    }

    /**
     * Returns an item from the serialized
     * JSON representation.  The returned item contains no properties referenced
     * in the serialized representation, other than location information.
     * 
     * @param jsonObject  The serialized representation of a JSON object. At
     * a minimum it must contain location information identifying an item handle.
     * Must not be <code>null</code>.
     * 
     * @return The deserialized item.  Never <code>null</code>. It contains
     * no properties other than item and state ids.  Neither "new" or a 
     * "working copy".
     * 
     * @throws SerializeException Thrown on error.
     */
    private Item getItem(final JSONObject jsonObject) throws SerializeException {
        
        final IItemHandle handle = getItemHandle(jsonObject);
        if (handle == null) {
            final String msg = Messages.getServerString("JSONDeserializer.ErrorMissingHandle"); //$NON-NLS-1$
            throw new SerializeException(msg);
        }

        /*
         * Create an uninitialized item.  Don't use "initNew" on the item, as
         * that will have the effect of making all properties appear as if they
         * been set.
         */
        final String namespaceURI = handle.getItemType().getNamespaceURI();
        final String typeName = handle.getItemType().getName();
        
        final EPackage ePackage = EPackage.Registry.INSTANCE.getEPackage(namespaceURI);
        final EClass eClass = (EClass)ePackage.getEClassifier(typeName);
        final EObject eObject = ePackage.getEFactoryInstance().create(eClass);
        final Item item = (Item)eObject; 

        /*
         * Set the item and state ids.
         * If the state was contained in the location information (from which
         * the handle was derived), use that in the new item, else use the etag
         * (if available). 
         */
        item.setItemId(handle.getItemId());
        if (handle.getStateId() != null)
            item.setStateId(handle.getStateId());
        else
            item.setStateId(getETag(jsonObject));
        
        return item;
    }

    /**
     * Returns the item handle encapsulated by the referenced JSON object.
     * 
     * @param jsonObject An object representing a serialized item. Must not
     * be <code>null</code>.  If an item handle is to be returned, must contain
     * a location member referencing an item handle.
     * 
     * @return The item handle of the serialized item. Will be <code>null</code>.
     * if <code>jsonObject</code> contains no location member.
     * 
     * @throws SerializeException Thrown on error.
     */
    private IItemHandle getItemHandle(final JSONObject jsonObject) throws SerializeException {
        
        final String locString = (String)jsonObject.get(KEY_URI);
        if (locString == null)
            return null;
        
        URI uri = null;
        try {
            uri = new URI(locString);
        } catch (URISyntaxException e) {
            final String msg = NLS.bind(Messages.getServerString("JSONDeserializer.InvalidLocationString"), //$NON-NLS-1$
                    locString, e.getMessage());
            throw new SerializeException(msg, e);
        }
        
        return _uriSerializer.getItemHandle(uri);
    }
    
    /**
     * Returns the virtual type encapsulated by the referenced JSON object.
     * 
     * @param jsonObject An object representing a serialized resource. Must not
     * be <code>null</code>.  If a virtual type is to be returned, must contain
     * a uri member referencing a virtual type.
     * 
     * @return The virtual type of the serialized resource. Will be <code>null</code>.
     * if <code>jsonObject</code> contains no uri member.
     * 
     * @throws SerializeException Thrown on error.
     */
    private IVirtualType getVirtualType(JSONObject jsonObject) throws SerializeException {
        
        IType type = getType(jsonObject);
        if (type != null) {
            if (type instanceof IVirtualType)
                return (IVirtualType)type;
        }
        
        return null;
    }
    
    /**
     * Returns the type encapsulated by the referenced JSON object.
     * 
     * @param jsonObject An object representing a serialized resource. Must not
     * be <code>null</code>.  If a type is to be returned, must contain
     * a uri member referencing a type.
     * 
     * @return The type of the serialized resource. Will be <code>null</code>.
     * if <code>jsonObject</code> contains no uri member.
     * 
     * @throws SerializeException Thrown on error.
     */
    private IType getType(JSONObject jsonObject) throws SerializeException {
        
        final String locString = (String)jsonObject.get(KEY_URI);
        if (locString == null)
            return null;
      
        URI uri = null;
        try {
            uri = new URI(locString);
        } catch (final URISyntaxException e) {
            final String msg = NLS.bind(Messages.getServerString("JSONDeserializer.InvalidLocationString"), //$NON-NLS-1$
                    locString, e.getMessage());
            throw new SerializeException(msg, e);
        }

        return _uriSerializer.getType(uri);
    }
    
    /**
     * Returns an {@link IContent} reference representing the specified JSON object.
     * 
     * @param jsonObject The JSON object containing a content reference.
     * Must not be <code>null</code>.
     * 
     * @return An {@link IContent} reference, or <code>null</code> if the specified
     * JSON object does not represent an {@link IContent}.
     * 
     * @throws SerializeException Thrown on URI parse errors.
     */
    private IContent toContent(JSONObject jsonObject) throws SerializeException {
        
        final String locString = (String)jsonObject.get(KEY_URI);
        if (locString == null)
            return null;
        
        URI uri = null;
        try {
            uri = new URI(locString);
        } catch (final URISyntaxException e) {
            final String msg = NLS.bind(Messages.getServerString("JSONDeserializer.InvalidLocationString"), //$NON-NLS-1$
                    locString, e.getMessage());
            throw new SerializeException(msg, e);
        }

        Location loc = null;
        try {
            loc = Location.location(uri);
        } catch (TeamRepositoryException e) {
            final String msg = NLS.bind(Messages.getServerString("JSONDeserializer.InvalidLocationString"), //$NON-NLS-1$
                    uri, e.getMessage());
            throw new SerializeException(msg, e);
        }
        
        if (loc.isContent())
            return loc.getContent();
        else
            return null;
    }
    
    /**
     * Returns the top-level etag member contained in the referenced JSON object.
     * 
     * @param jsonObject The object containing the etag. Must not be <code>null</code>.
     * 
     * @return The etag contained in the object. May be <code>null</code> if an
     * etag member is not contained in the JSON object.
     * 
     * @throws SerializationException Thrown if the referenced object contains
     * an etag member and if the specified value is syntactically incorrect.
     */
    static private UUID getETag(final JSONObject jsonObject) throws SerializeException {
        
        if (jsonObject == null)
            throw new IllegalArgumentException("jsonObject must not be null"); //$NON-NLS-1$
        
        UUID etagUuid = null;
        final String etag = (String)jsonObject.get(KEY_ETAG);
        if (etag != null) {
            try {
                etagUuid = UUID.valueOf(etag);
            } catch (final RuntimeException e2) {
                final String msg = NLS.bind(Messages.getServerString("JSONDeserializer.InvalidMemberSyntax"), KEY_ETAG, e2.getMessage()); //$NON-NLS-1$
                throw new SerializeException(msg, e2);
            }
        }
        
        return etagUuid;
    }
    
    /**
     * Populates the referenced <code>eObject</code> with properties contained
     * in the referenced <code>jsonObj</code>.
     * 
     * @param eObject The eObject in which to write property values.  Must not
     * be <code>null</code>.
     * 
     * @param jsonObj The JSON object containing the EObject representation.
     * Must not be <code>null</code>.
     * 
     * @throws SerializeException Thrown on error.
     */
    private void deserializeObject(final EObject eObject, final JSONObject jsonObj) throws SerializeException {
        final EClass eClass = eObject.eClass();
        final List<EStructuralFeature> sfs = eClass.getEAllStructuralFeatures();

        for (final EStructuralFeature sf : sfs) {
            if (!jsonObj.containsKey(sf.getName()))
                continue;
            
            /*
             * Don't set the immutable feature.  Setting will cause all
             * subsequent sets to fail.
             */
            if (ItemUtil.isImmutableFeature(sf))
                continue;
            
            final Object jsonValue = jsonObj.get(sf.getName());
            final Object itemValue = deserializeFeature(sf, jsonValue);
            eObject.eSet(sf, itemValue);
        }
    }
    
    /**
     * Try to deserialize the object. We assume that there are type hints 
     * embedded in the EMF object's serialized form. 
     * 
     */
    private EObject deserializeObject(final JSONObject object) throws SerializeException {
    	if (object == null) {
    		return null;
    	}
    	
    	String ePackageName = (String) object.get(JSONSerializer.KEY_EPACKAGE);
    	String eClassName = (String) object.get(JSONSerializer.KEY_ECLASS);

    	if (ePackageName == null || eClassName == null) {
    		throw new SerializeException();
    	}
    	
    	EPackage ePackage = EcorePackage.Registry.INSTANCE.getEPackage(ePackageName);
    	if (ePackage == null) {
    		throw new SerializeException(ePackageName);
    	}
    	
    	EClass eClass = (EClass) ePackage.getEClassifier(eClassName);
    	if (eClass == null) {
    		throw new SerializeException(eClassName);
    	}

    	EObject eObject;
		try {
			eObject = ePackage.getEFactoryInstance().create(eClass);
		} catch (IllegalArgumentException e) {
			throw new SerializeException(eClassName, e);
		}
    	
    	deserializeObject(eObject, object);

    	return eObject;
    }

    /**
     * Returns the value of the structural feature <code>sf</code>, as 
     * represented in the referenced <code>jsonValue</code>.
     * 
     * @param sf The structural feature whose value is encoded in the JSON
     * value. Must not be <code>null</code>.
     * 
     * @param jsonValue The JSON-encoded value for the structural feature.
     * May be <code>null</code>.
     * 
     * @return The value of the referenced structural feature, in the object
     * format expected by the eAttribute.  Will be <code>null</code> if 
     * <code>jsonValue</code> is null. 
     * 
     * @throws SerializeException Thrown on error.
     */
    private Object deserializeFeature(final EStructuralFeature sf, final Object jsonValue) throws SerializeException {

        if (sf == null)
            throw new IllegalArgumentException("sf must not be null"); //$NON-NLS-1$
        
        if (jsonValue == null)
            return null;
        
        final String featureName = sf.getName();
        final EClassifier featureType = sf.getEType();
        
        Object itemValue = null;
        if (sf instanceof EReference) {
            if (sf.isMany()) {
                if (!(jsonValue instanceof JSONArray)) {
                    final String msg = NLS.bind(Messages.getServerString("JSONDeserializer.ErrorExpectedArray"), featureName); //$NON-NLS-1$
                    throw new SerializeException(msg);
                }
                itemValue = deserializeReferenceList((JSONArray)jsonValue, (EClass)featureType);
            }
            else {
                if (!(jsonValue instanceof JSONObject)) {
                    final String msg = NLS.bind(Messages.getServerString("JSONDeserializer.ErrorExpectedArray"), featureName); //$NON-NLS-1$
                    throw new SerializeException(msg);
                }
                itemValue = deserializeReference((JSONObject)jsonValue, (EClass)featureType);
            }
        }
        else if (sf instanceof EAttribute) {
            if ((jsonValue instanceof JSONObject)) {
                final String msg = NLS.bind(Messages.getServerString("JSONDeserializer.UnexpectedObjectValue"), featureName); //$NON-NLS-1$
                throw new SerializeException(msg);
            }
            if (sf.isMany()) {
                if (!(jsonValue instanceof JSONArray)) {
                    final String msg = NLS.bind(Messages.getServerString("JSONDeserializer.ErrorExpectedArray"), featureName); //$NON-NLS-1$
                    throw new SerializeException(msg);
                }
                itemValue = deserializeAttributeList((JSONArray)jsonValue, featureType);
            }
            else {
                itemValue = deserializeAttribute(jsonValue, featureType);
            }
        }
        else {
            final String msg = NLS.bind(Messages.getServerString("JSONDeserializer.UnrecognizedFeature"), featureName); //$NON-NLS-1$
            throw new SerializeException(msg);
        }
        return itemValue;
    }
    
    
    /**
     * Translates the object referenced by <code>object</code> into the expected
     * type specified in <code>featureType</code>.
     * 
     * @param object The object to translate. Must not be <code>null</code>.
     * 
     * @param featureType The feature type to translate to. Must not be <code>null</code>.
     * 
     * @return The translated object. Never <code>null</code>.
     * 
     * @throws SerializeException On translation errors.
     */
    private Object deserializeAttribute(final Object object, final EClassifier featureType) throws SerializeException {
        
        final Class expectedType = featureType.getInstanceClass();
        
        // boolean
        if (object instanceof Boolean) {
            if ((Boolean.class == expectedType) || (Boolean.TYPE == expectedType)) {
                return object;
            }
            
            throw new SerializeException(NLS.bind(Messages.getServerString("JSONDeserializer.UnsupportedBoolean"), object)); //$NON-NLS-1$
        }
        
        // string, which includes primitives
        if (object instanceof String) {
            if (String.class == expectedType) {
                return object;
            }
            
            // if the value was untyped (Object[], or List), all we can
            // do is return the string version; conversion can be done
            // later for typed arrays given more information
            if (Object.class == expectedType) {
                return object;
            }

            // handle enums
            if (featureType instanceof EEnum) {
                final EEnum eEnum = (EEnum) featureType;
                return eEnum.getEEnumLiteral((String) object).getInstance();
            }
            
            // handle has primitive
            final IPrimitiveHandler primHandler = IPrimitiveRegistry.INSTANCE.getHandler(expectedType);
            if (null != primHandler) {
                return primHandler.stringToObject((String) object);
            }
        
            // give up
            throw new SerializeException(NLS.bind(Messages.getServerString("JSONDeserializer.UnsupportedPrimitive"), object)); //$NON-NLS-1$
        }
        
        // number
        if (object instanceof Number) {
            final Number number = (Number) object;
            
            if ((Byte.class == expectedType) || (Byte.TYPE == expectedType)) {
                return Byte.valueOf(number.byteValue());
            }
            if ((Short.class == expectedType) || (Short.TYPE == expectedType)) {
                return Short.valueOf(number.shortValue());
            }
            if ((Character.class == expectedType) || (Character.TYPE == expectedType)) {
                return Character.valueOf((char) number.shortValue());
            }
            if ((Integer.class == expectedType) || (Integer.TYPE == expectedType)) {
                return Integer.valueOf(number.intValue());
            }
            if ((Long.class == expectedType) || (Long.TYPE == expectedType)) {
                return Long.valueOf(number.longValue());
            }
            if ((Float.class == expectedType) || (Float.TYPE == expectedType)) {
                return new Float(number.floatValue());
            }
            if ((Double.class == expectedType) || (Double.TYPE == expectedType)) {
                return new Double(number.doubleValue());
            }
            
            throw new SerializeException(NLS.bind(Messages.getServerString("JSONDeserializer.UnsupportedNumber"), object)); //$NON-NLS-1$
        }
        
        throw new SerializeException(NLS.bind(Messages.getServerString("JSONDeserializer.UnexpectedPropertyValue"), object)); //$NON-NLS-1$
    }

    /**
     * Translates the objects contained <code>attrList</code> into a list of
     * object types as described by the referenced <code>featureType</code>.
     * 
     * @param attrList The list of objects to translate. Must not be <code>null</code>.
     * 
     * @param featureType The feature type to translate to. Must not be <code>null</code>.
     * 
     * @return The list of translated objects. Never <code>null</code>.
     * 
     * @throws SerializeException Thrown on translation errors.
     */
    private List deserializeAttributeList(final JSONArray attrList, final EClassifier featureType) throws SerializeException {
        
        if (attrList == null)
            throw new IllegalArgumentException("attrs must not be null"); //$NON-NLS-1$
        if (featureType == null)
            throw new IllegalArgumentException("featureType must not be null"); //$NON-NLS-1$
        
        final ArrayList retVal = new ArrayList(attrList.size());
        
        for (final Object attr : attrList) {
            if ((attr instanceof JSONObject) || (attr instanceof JSONArray))
                throw new SerializeException(Messages.getServerString("JSONDeserializer.InvalidAttributeList")); //$NON-NLS-1$

            retVal.add(deserializeAttribute(attr, featureType));
        }

        return retVal;
    }

    /**
     * Translates the objects referenced by <code>refList</code> into object
     * instances of type <code>eClass</code>.
     * 
     * @param refList The list of objects to translate. Must not be <code>null</code>.
     * 
     * @param eClass The class to translate each object in the list to. Must
     * not be <code>null</code>.
     * 
     * @return The list of translated objects. Never <code>null</code>.
     * 
     * @throws SerializeException Thrown on translation errors.
     */
    private List deserializeReferenceList(final JSONArray refList, final EClass eClass) throws SerializeException {
        
        if (refList == null)
            throw new IllegalArgumentException("refList must not be null"); //$NON-NLS-1$
        if (eClass == null)
            throw new IllegalArgumentException("eClass must not be null"); //$NON-NLS-1$
        
        final ArrayList refs = new ArrayList(refList.size());
        
        for (final Object ref : refList) {
            if (!(ref instanceof JSONObject))
                throw new SerializeException(Messages.getServerString("JSONDeserializer.InvalidReferenceList")); //$NON-NLS-1$

            refs.add(deserializeReference((JSONObject)ref, eClass));
        }

        return refs;
    }

    /**
     * Translates the object serialized in <code>jsonObj</code> to an object instance
     * of type <code>eClass</code>.
     * 
     * @param jsonObj The serialized object to translate. Must not be <code>null</code>.
     * 
     * @param eClass The class to translate the object to. Must not be <code>null</code>.
     * 
     * @return The translated object. Never <code>null</code>.
     * 
     * @throws SerializeException Thrown on translation errors.
     */
    private Object deserializeReference(final JSONObject jsonObj, final EClass eClass) throws SerializeException {
        
        if (jsonObj == null)
            throw new IllegalArgumentException("jsonObj must not be null"); //$NON-NLS-1$
        if (eClass == null)
            throw new IllegalArgumentException("eClass must not be null"); //$NON-NLS-1$
        
        /*
         * Check if the referenced item is a handle
         */
        final IItemHandle handle = getItemHandle(jsonObj);
        if (handle != null)
            return handle;
        
        /*
         * If it's a virtual...
         */
        final IVirtualType virtualType = getVirtualType(jsonObj);
        if (virtualType != null)
            return toVirtual(virtualType, jsonObj);
        
        /*
         * If it's a content reference...
         */
        final IContent content = toContent(jsonObj);
        if (content != null)
            return content;
        
        /*
         * If it's a helper
         */
        EObject newObj = null;
        EClass newClass = null;
        if (RepositoryPackage.eINSTANCE.getHelper().isSuperTypeOf(eClass)) {
            final IHelperType helperType = ItemUtil.helperTypeFor(eClass);
            final IHelper helper = helperType.createHelper();
            newObj = (EObject)helper.getAdapter(EObject.class);
            newClass = (EClass)helper.getAdapter(EClass.class);
        }
        else {
            /*
             * Otherwise it's some other eObject reference
             */
        	
        	if (jsonObj.containsKey(JSONSerializer.KEY_ECLASS)) {
        		// The object wants to override our expected type. 
        		// We're down with that. 
                String ePackageName = (String)jsonObj.get(JSONSerializer.KEY_EPACKAGE);
                String eClassName   = (String)jsonObj.get(JSONSerializer.KEY_ECLASS);
                
                EPackage ePackage = EcorePackage.Registry.INSTANCE.getEPackage(ePackageName);
                if (null == ePackage) 
                    throw new SerializeException(NLS.bind(Messages.getServerString("JSONDeserializer_03.ErrorMissingPackage"), ePackageName)); //$NON-NLS-1$
                
                EClassifier eClassifier = ePackage.getEClassifier(eClassName);
                if (null == eClassifier) 
                    throw new SerializeException(NLS.bind(Messages.getServerString("JSONDeserializer_03.ErrorMissingClass"), eClassName)); //$NON-NLS-1$

                if (!(eClassifier instanceof EClass)) {
                    throw new SerializeException(NLS.bind(Messages.getServerString("JSONDeserializer_03.ErrorWrongClassType"), eClassName)); //$NON-NLS-1$
                }
                 
                newClass = (EClass) eClassifier;
                newObj = ePackage.getEFactoryInstance().create(newClass);
        	}
        	else {
        		newObj = eClass.getEPackage().getEFactoryInstance().create(eClass);
        		newClass = eClass;
        	}
        }

        final List<EStructuralFeature> features = newClass.getEAllStructuralFeatures();
        
        for (final EStructuralFeature feature: features) {
            if (!jsonObj.containsKey(feature.getName()))
                continue;
            
            final Object jsonValue = jsonObj.get(feature.getName());
            if (jsonValue != null) {
                final Object itemValue = deserializeFeature(feature, jsonValue);
                newObj.eSet(feature, itemValue);
            }
        }

        return newObj;
    }
    
    @Override
    public Throwable deserializeException(Reader r, ClassLoader loader, IStackAdjuster adj) throws SerializeException {
    	return deserializeErrorObj(getObject(r), loader, adj);
    }
    
    /**
     * Counterpart to {@link JSONSerializer#formatErrorObj()}. 
     * @param loader 
     * @throws SerializeException (This exception is thrown by a subclass in
     * RTC.)
     */
    protected Throwable deserializeErrorObj(JSONObject jsonObject, ClassLoader loader, IStackAdjuster adj) throws SerializeException {
    	StringWriter result = new StringWriter(1000);
        PrintWriter output = new PrintWriter(result, true);

    	// Create the marker
    	String marker = UUID.generate().getUuidValue();

    	output.println(marker);

        deserializeErrorObjException(marker, jsonObject, output);

        output.flush();
        result.flush();
        output.close();
        
        Throwable toReturn = MarshallerUtil.decodeExceptions(result.toString(), loader, adj);

        // Handle exception-specific type fields. 
        if (toReturn instanceof TeamRepositoryException) {
        	TeamRepositoryException tre = (TeamRepositoryException) toReturn;
        	
        	Object allData = jsonObject.get(JSONSerializer.ERROR_ALLDATA);
        	if (allData != null && allData instanceof Map) {
        		HashMap<String, Object> fromJSON = (JSONObject) allData;
        		HashMap<String, Object> toSet = new HashMap<String, Object>(fromJSON.size());
        		
        		for (Map.Entry<String, Object> entry : fromJSON.entrySet()) {
        			EObject value;
        			try {
        				value = deserializeObject((JSONObject) entry.getValue());
        			} catch (SerializeException e) {
        				continue;
        			}
        			
        			toSet.put(entry.getKey(), value);
        		}
        		
        		tre.setSerializedData(toSet);
        	}
        }
        
        return participant.deserializeErrorObj(this, jsonObject, toReturn);
    }
    
    
    private static void deserializeErrorObjException(final String marker, JSONObject obj, PrintWriter out) {
    	// Print the classname
    	String className = getString(obj, JSONSerializer.ERROR_CLASSNAME_KEY, RuntimeException.class.getName());
    	out.println(className);
    	
    	// Print the message
    	String msg = getString(obj, JSONSerializer.ERROR_MSG_KEY, null);
    	if (msg != null) {
    	    out.print('>');
    	    out.print(msg);
    	}
    	out.println(marker);
    	
    	// Print the stack trace
    	JSONArray trace = getArray(obj, JSONSerializer.ERROR_TRACE_MARSHALL_KEY);
    	out.println(trace.size());
    	
    	for (Object elem : trace) {
    	    JSONObject frame;
    	    if (elem instanceof JSONObject) {
    	        frame = (JSONObject)elem;
			}
    	    else {
    	        frame = new JSONObject();
    	    }
            
            out.print(getString(frame, JSONSerializer.ERROR_TRACE_CLASSNAME, "UnknownClassName")); //$NON-NLS-1$
            out.print(","); //$NON-NLS-1$

            out.print(getString(frame, JSONSerializer.ERROR_TRACE_METHODNAME, "UnknownMethodName")); //$NON-NLS-1$
            out.print(","); //$NON-NLS-1$

            out.print(getString(frame, JSONSerializer.ERROR_TRACE_FILENAME, "UnknownFileName")); //$NON-NLS-1$
            out.print(","); //$NON-NLS-1$
            
            out.print(getInteger(frame, JSONSerializer.ERROR_TRACE_LINENUM, 0));
            out.println();
    	}
    	
    	// Append any "caused bys".
    	Object child = obj.get(JSONSerializer.NESTED_ERROR_KEY);
    	if (child != null && child instanceof JSONObject) {
    	    out.println(MarshallerUtil.CAUSE);
    	    deserializeErrorObjException(marker, (JSONObject)child, out);
    	}
    }
    
    private static String getString(JSONObject obj, String key, String defaultValue) {
        Object value = obj.get(key);
        if (value == null) {
            if (obj.containsKey(key)) { 
            	return null;
            }
            
            return defaultValue;
        }
        
        if (value instanceof String) {
            return (String)value;
        }
        
        return defaultValue;
    }
    
    private static Long getInteger(JSONObject obj, String key, long defaultValue) {
        Object value = obj.get(key);
        if (value == null) {
        	return defaultValue;
        }

        if (value instanceof Number) {
            return ((Number)value).longValue();
        }
        
        if (value instanceof String) {
            try {
            	return Long.getLong((String)value);
            } catch (NumberFormatException e) {
                return defaultValue;
            }
        }
        
        return defaultValue;
    }
    
    private static JSONArray getArray(JSONObject obj, String key) {
        Object value = obj.get(key);
        
        if (value instanceof JSONArray) {
			return (JSONArray) value;
		}
        
        return new JSONArray();
    }
}
