/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010, 2017. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/

package com.ibm.team.repository.common.serialize.internal;

import java.io.IOException;
import java.io.Writer;
import java.lang.reflect.Array;
import java.net.URI;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.common.util.Enumerator;
import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EStructuralFeature;

import com.ibm.team.repository.common.IAuditable;
import com.ibm.team.repository.common.IContent;
import com.ibm.team.repository.common.IItem;
import com.ibm.team.repository.common.IItemHandle;
import com.ibm.team.repository.common.ISimpleItem;
import com.ibm.team.repository.common.IVirtual;
import com.ibm.team.repository.common.Location;
import com.ibm.team.repository.common.PermissionDeniedException;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.common.internal.util.ItemUtil;
import com.ibm.team.repository.common.json.JSONArray;
import com.ibm.team.repository.common.json.JSONObject;
import com.ibm.team.repository.common.model.Auditable;
import com.ibm.team.repository.common.model.RepositoryPackage;
import com.ibm.team.repository.common.model.SimpleItem;
import com.ibm.team.repository.common.serialize.IPrimitiveHandler;
import com.ibm.team.repository.common.serialize.IPrimitiveRegistry;
import com.ibm.team.repository.common.serialize.IURISerializer;
import com.ibm.team.repository.common.serialize.LoggedException;
import com.ibm.team.repository.common.serialize.MessagePlaceholder;
import com.ibm.team.repository.common.serialize.SerializeException;
import com.ibm.team.repository.common.serialize.StackUtil;
import com.ibm.team.repository.common.serialize.internal.nls.Messages;
import com.ibm.team.repository.common.transport.HttpUtil.MediaType;
import com.ibm.team.repository.common.util.HashUtil;
import com.ibm.team.repository.common.util.NLS;

/**
 * A serializer for JSON representations.
 */
public class JSONSerializer extends AbstractSerializer {

	/**
	 * Allow options to be passed down into the serializer. 
	 */
	private static class SerializationContext {
		/**
		 * When set to true, we insert hints about EMF object types into 
		 * their JSON representations.
		 * 
		 * @see JSONSerializer#KEY_ECLASS
		 * @see JSONSerializer#KEY_EPACKAGE
		 */
		boolean include_etype = false;
	}
	
    public static final String KEY_URI = "uri"; //$NON-NLS-1$
    public static final String KEY_ETAG = "etag"; //$NON-NLS-1$
    public static final String KEY_ITEMID = "itemId"; //$NON-NLS-1$
    public static final String KEY_STATEID = "stateId"; //$NON-NLS-1$
    public static final String KEY_TYPE = "com.ibm.team.repository.typeName"; //$NON-NLS-1$
    
    /*
     * String identifiers for error formatting
     */
    public static final String ERROR_CLASSNAME_KEY = "errorClass"; //$NON-NLS-1$
    public static final String ERROR_CODE_KEY = "errorCode"; //$NON-NLS-1$
    public static final String ERROR_DATA = "errorData"; //$NON-NLS-1$
    
    /** Key for extended data in exceptions. */
    public static final String ERROR_ALLDATA = "errorAllData"; //$NON-NLS-1$
    
    /** Key for the contents of 
     * {@link PermissionDeniedException#getProcessReportData()} 
     */
    public static final String ERROR_DATA_PROCESS = "processData"; //$NON-NLS-1$

    /**
     * Key indicating that the EPackage of the enclosing JSON object. Note 
     * the "-1" indicating that we're version one of this serialization type.
     */
    public static final String KEY_EPACKAGE = "-hint-epackage-1"; //$NON-NLS-1$
    
    /**
     * Key indicating that the EClass of the enclosing JSON object. Note 
     * the "-1" indicating that we're version one of this serialization type.
     */
    public static final String KEY_ECLASS = "-hint-eclass-1"; //$NON-NLS-1$

    
    public static final String ERROR_MSG_KEY = "errorMessage"; //$NON-NLS-1$
    public static final String ERROR_TRACE_KEY = "errorTrace"; //$NON-NLS-1$
    public static final String ERROR_TRACE_MARSHALL_KEY = "errorTraceMarshall"; //$NON-NLS-1$
    public static final String ERROR_TRACE_CLASSNAME = "errorTraceClassName"; //$NON-NLS-1$
    public static final String ERROR_TRACE_METHODNAME = "errorTraceMethodName"; //$NON-NLS-1$
    public static final String ERROR_TRACE_FILENAME = "errorTraceFileName"; //$NON-NLS-1$
    public static final String ERROR_TRACE_LINENUM = "errorTraceLineNumber"; //$NON-NLS-1$
    public static final String NESTED_ERROR_KEY = "nestedError"; //$NON-NLS-1$

	public static final String NLS_EXCEPTION_PLACEHOLDER_LOGGED_MESSAGE = "JSONSerializer.ExceptionPlaceholderLoggedMessage"; //$NON-NLS-1$
	public static final String NLS_EXCEPTION_PLACEHOLDER_USER_MESSAGE = "JSONSerializer.ExceptionPlaceholderUserMessage"; //$NON-NLS-1$

    /**
     * Creates a JSONSerializer instance.
     * 
     * @param type The media type for this serializer instance. Must not be
     * <code>null</code>.
     * 
     * @param repoUri not used
     * 
     * @param format Whether or not the serialized representation should be
     * formatted for human readability.  <code>true</code> to format, <code>false</code>
     * to leave representation unformatted.
     * 
     * @param The serializer used to convert between objects and URIs. Must not
     * be <code>null</code>.
     * 
     * @deprecated use JSONSerializer(MediaType type, boolean format, IURISerializer uriSerializer) instead
     * Note - as of 12/7/2013, this internal constructor is invoked by RTC:
     * com.ibm.team.filesystem.client.internal.marshaling.ExceptionHandlingJSONSerializer
     * com.ibm.team.workitem.web.tests.runner.WebTestUtil
     */
    @Deprecated
	public JSONSerializer(MediaType type, String repoUri, boolean format, IURISerializer uriSerializer) {
        super(type, format, uriSerializer);
    }

    /**
     * Creates a JSONSerializer instance.
     * 
     * @param type The media type for this serializer instance. Must not be
     * <code>null</code>.
     * 
     * @param format Whether or not the serialized representation should be
     * formatted for human readability.  <code>true</code> to format, <code>false</code>
     * to leave representation unformatted.
     * 
     * @param The serializer used to convert between objects and URIs. Must not
     * be <code>null</code>.
     */
    public JSONSerializer(MediaType type, boolean format, IURISerializer uriSerializer) {
        super(type, format, uriSerializer);
    }

    /* (non-Javadoc)
     * @see com.ibm.team.repository.common.serialize.ISerializer2#serialize(com.ibm.team.repository.common.IItem, java.io.Writer)
     */
    @Override
	public void serialize(final IItem item, final Writer writer) throws TeamRepositoryException, IOException {
        if (item == null)
            throw new IllegalArgumentException("item must not be null"); //$NON-NLS-1$
        if (writer == null)
            throw new IllegalArgumentException("writer must not be null"); //$NON-NLS-1$

        final EObject eObject = (EObject)item.getAdapter(EObject.class);
        JSONObject obj = null;
        try {
            obj = toJSONObject(eObject, item.getClass(), new SerializationContext());
        } 
        catch (final SerializeException e) {
            final String msg = NLS.bind(Messages.getServerString("JSONSerializer.ErrorSerializingItem"), e.getMessage()); //$NON-NLS-1$
            throw new TeamRepositoryException(msg, e);
        }
        
        obj.serialize(writer, format());
    }

    
    @Override
	public void serialize(final IVirtual item, final Writer writer) throws TeamRepositoryException, IOException {
        if (item == null)
            throw new IllegalArgumentException("item must not be null");  //$NON-NLS-1$
        if (writer == null)
            throw new IllegalArgumentException("writer must not be null");  //$NON-NLS-1$

        final EObject eObject = (EObject)item.getAdapter(EObject.class);
        JSONObject obj = null;
        try {
            obj = toJSONObject(eObject, item.getClass(), new SerializationContext());
        } 
        catch (final SerializeException e) {
            final String msg = NLS.bind(Messages.getServerString("JSONSerializer.ErrorSerializingItem"), e.getMessage()); //$NON-NLS-1$
            throw new TeamRepositoryException(msg, e);
        }
        
        obj.serialize(writer, format());
    }

    /* (non-Javadoc)
     * @see com.ibm.team.repository.common.serialize.ISerializer2#serialize(java.util.Collection, java.io.Writer)
     */
    @Override
	public void serialize(final List<IItem> items, final Writer writer) throws TeamRepositoryException, IOException {
        if (items == null)
            throw new IllegalArgumentException("item must not be null");  //$NON-NLS-1$
        if (writer == null)
            throw new IllegalArgumentException("writer must not be null"); //$NON-NLS-1$
        
        JSONArray members = null;
        try {
            members = internalSerializeList(items, IItem.class, new SerializationContext());
        }
        catch (final SerializeException e) {
            final String msg = NLS.bind(Messages.getServerString("JSONSerializer.ErrorSerializingItem"), e.getMessage()); //$NON-NLS-1$
            throw new TeamRepositoryException(msg, e);
        }
        members.serialize(writer, format());
    }

    
    /* (non-Javadoc)
     * @see com.ibm.team.repository.common.serialize.ISerializer2#serializeVirtuals(java.util.List, java.io.Writer)
     */
    @Override
	public void serializeVirtuals(final List<IVirtual> items, final Writer writer) throws TeamRepositoryException, IOException {
        if (items == null)
            throw new IllegalArgumentException("item must not be null");  //$NON-NLS-1$
        if (writer == null)
            throw new IllegalArgumentException("writer must not be null"); //$NON-NLS-1$
        
        JSONArray members = null;
        try {
            members = internalSerializeList(items, IVirtual.class, new SerializationContext());
        }
        catch (final SerializeException e) {
            final String msg = NLS.bind(Messages.getServerString("JSONSerializer.ErrorSerializingItem"), e.getMessage()); //$NON-NLS-1$
            throw new TeamRepositoryException(msg, e);
        }
        members.serialize(writer, format());
    }

    /* (non-Javadoc)
     * @see com.ibm.team.repository.common.serialize.ISerializer2#serializeList(java.util.List, java.io.Writer)
     */
    @Override
	public void serializeList(List items, Writer writer) throws TeamRepositoryException, IOException {
        if (items == null)
            throw new IllegalArgumentException("item must not be null"); //$NON-NLS-1$
        if (writer == null)
            throw new IllegalArgumentException("writer must not be null"); //$NON-NLS-1$
        
        JSONArray members = null;
        try {
            members = internalSerializeList(items, null, new SerializationContext());
        }
        catch (final SerializeException e) {
            final String msg = NLS.bind(Messages.getServerString("JSONSerializer.ErrorSerializingItem"), e.getMessage()); //$NON-NLS-1$
            throw new TeamRepositoryException(msg, e);
        }
        members.serialize(writer, format());
    }

    /* (non-Javadoc)
     * @see com.ibm.team.repository.common.serialize.ISerializer2#serialize(org.eclipse.emf.ecore.EObject, java.io.Writer)
     */
    @Override
	public void serialize(EObject object, Writer writer) throws TeamRepositoryException, IOException {
        if (object == null)
            throw new IllegalArgumentException("item must not be null"); //$NON-NLS-1$
        if (writer == null)
            throw new IllegalArgumentException("writer must not be null"); //$NON-NLS-1$

        JSONObject obj = null;
        try {
            obj = toJSONObject(object, object.getClass(), new SerializationContext());
        } 
        catch (final SerializeException e) {
            final String msg = NLS.bind(Messages.getServerString("JSONSerializer.ErrorSerializingObject"), e.getMessage()); //$NON-NLS-1$
            throw new TeamRepositoryException(msg, e);
        }
        
        obj.serialize(writer, format());
    }

   /**
    * Serializes the referenced <code>eObject</code> into a JSON object.
    * 
    * @param eObject The object to serialize. May be <code>null</code>.  If
    * <code>null</code>, this function returns <code>null</code>.
    * 
    * @param expectedType The expected type of the object. Must not be
    * <code>null</code>.
    * 
    * @param ctx Describes any serialization parameters for this call. May not 
    * 		be <code>null</code>.  
    * 
    * @return The JSON object representing the serialized object. Will be <code>null</code>
    * if the referenced <code>eObject</code> is <code>null</code>, else never
    * <code>null</code>
    * 
    * @throws SerializeException Thrown on error.
    */
   private JSONObject toJSONObject(EObject eObject, Class expectedType, SerializationContext ctx) throws SerializeException {
       if (null == eObject)
           return null;
       
       final JSONObject jsonObject = new JSONObject();
       final boolean itemsToHandles = shouldMorphItemToHandle(eObject, expectedType);
       if (itemsToHandles)
           eObject = morphItemToHandle(eObject);
       final EClass eClass = eObject.eClass();
       
       if (ctx.include_etype) {
           // Add EMF type flags
           jsonObject.put(KEY_EPACKAGE, eClass.getEPackage().getNsURI());
           jsonObject.put(KEY_ECLASS, eClass.getName());
       }

       boolean isItemHandle = eObject instanceof IItemHandle;
       boolean isVirtual = eObject instanceof IVirtual;
       boolean isContent = eObject instanceof IContent;
       
       /*
        * Serialize a handle (or item) with location and etag information.
        */
       if (isItemHandle) {
           final IItemHandle handle = (IItemHandle)eObject;
           
           Location loc = null;
           try {
               loc = Location.location(_uriSerializer.getUri(handle));
           } catch (TeamRepositoryException e) {
               throw new SerializeException(e.getMessage(), e);
           }
           
           jsonObject.put(KEY_URI, loc.toRelativeUri().toString());

           jsonObject.put(KEY_ITEMID, handle.getItemId().getUuidValue());
           jsonObject.put(KEY_TYPE, ItemUtil.typeToName(handle.getItemType()));
           if (handle.getStateId() != null) {
               jsonObject.put(KEY_ETAG, handle.getStateId().getUuidValue());
               jsonObject.put(KEY_STATEID, handle.getStateId().getUuidValue());
           }

           /*
            * If this is a handle, just return - otherwise it's an item. 
            * If no stateId was available, it could be be because
            * this is a working copy.  Try to get the appropriate fields for an etag.
            */
           final IItem item = handle.getFullState();
           if (item == null) 
               return jsonObject;
           
           if (item.isWorkingCopy()) {
               if (item instanceof IAuditable) {
                   final Auditable audit = (Auditable)item;
                   if (audit.getWorkingCopyPredecessor() != null)
                       jsonObject.put(KEY_ETAG, audit.getWorkingCopyPredecessor().getUuidValue());
               }
               else if (item instanceof ISimpleItem) {
                   final SimpleItem simple = (SimpleItem)item;
                   if (simple.getPredecessor() != null)
                       jsonObject.put(KEY_ETAG, simple.getPredecessor().getUuidValue());
               }
           }
       }
       else if (isVirtual) {
           final IVirtual virtual = (IVirtual)eObject;
           jsonObject.put(KEY_TYPE, ItemUtil.typeToName(virtual.getVirtualType()));
           jsonObject.put(KEY_URI, getVirtualLocation(virtual));
           jsonObject.put(KEY_ETAG, virtual.getEtag());
       }
       else if (isContent) {
           final IContent content = (IContent)eObject;
           Location loc = Location.contentLocation(content, null);
           jsonObject.put(KEY_URI, loc.toRelativeUri().toString());
       }
       
       final List<EStructuralFeature> features = eClass.getEAllStructuralFeatures();
       for (final EStructuralFeature feature: features) {
           if (!eObject.eIsSet(feature)) 
               continue;
           
           /*
            * Non-persistent item attributes don't get serialized.
            */
           if (!ItemUtil.isPersistent(eClass, feature) && isItemHandle)
               continue;

           /*
            * Ignore state and item ids - those were generated with the handle.
            */
           if (isItemId(feature))
               continue;
           
           final Object value = eObject.eGet(feature);
           final String key = feature.getName();
           final Class valueExpectedType = feature.getEType().getInstanceClass();
           final Object serializedValue = serializeObject(value, valueExpectedType, ctx);

           jsonObject.put(key, serializedValue);
       }

       return jsonObject;
   }

   /**
    * Serializes and returns the specified <code>value</code> into an appropriate JSON
    * representation.
    * 
    * @param value The value to serialize. May be <code>null</code>.
    * 
    * @param expectedType The expected type of the value. May be <code>null</code>
    * if the expected type is unknown.
    * 
    * @param ctx Describes how the serialization should occur. May note be 
    * 		<code>null</code>.
    * 
    * @return The serialized object.  Will be <code>null</code> if <code>value</code>
    * is <code>null</code>.
    * 
    * @throws SerializeException Thrown on error.
    */
   private Object serializeObject(final Object value, final Class expectedType, SerializationContext ctx) throws SerializeException {
       if (null == value) 
           return null;
       
       final Class valueClass = value.getClass();

       if (value instanceof EObject) {
           Class type = expectedType;
           final EObject eValue = (EObject) value;
           if (type == null)
               type = eValue.getClass();
           return toJSONObject(eValue, type, ctx);
       }
       else if (value instanceof List) {
           return internalSerializeList((List) value, expectedType, ctx);
       }
       else if (valueClass.isArray()) {
           return serializeArray(value, expectedType, ctx);
       }
       else if (PrimitiveReg.isPrimitive(valueClass)) {
           return serializePrimitive(value, valueClass);
       }
       else if (value instanceof Enumerator) {
           return ((Enumerator) value).getLiteral();
       }
       
       throw new SerializeException(NLS.bind(Messages.getServerString("JSONSerializer.ErrorSerializingValue"), value)); //$NON-NLS-1$
   }
   
   /**
    * Serializes and returns the primitive value into the appropriate JSON object.
    *  
    * @param value The value to serialize. May be <code>null</code>.
    * 
    * @param valueClass The expected class of the value.  Must not be <code>null</code>.
    * 
    * @throws SerializeException Thrown on error.
    */
   private Object serializePrimitive(final Object value, final Class valueClass) throws SerializeException {
       if (null == value)
           return null;
       if (valueClass == Boolean.class)
           return value;
       if (valueClass == String.class)
           return value;
       if (Number.class.isAssignableFrom(valueClass))
           return value;

       final IPrimitiveHandler primHandler = IPrimitiveRegistry.INSTANCE.getHandler(valueClass);
       if (null == primHandler) {
           throw new SerializeException(NLS.bind(Messages.getServerString("JSONSerializer.MissingPrimitiveHandler"), valueClass.getName())); //$NON-NLS-1$
       }

       return primHandler.objectToString(value);
   }

   /**
    * Serializes and returns the specified list into a JSON array.
    * 
    * @param list The list to serialize. Must not be <code>null</code>.
    * 
    * @param expectedType The expected type of each item in the list.
    * Must not be <code>null</code>.
    * 
    * @return The JSON array representing the serialized list. Never <code>null</code>.
    * 
    * @throws SerializeException Thrown on error.
    */
   private JSONArray internalSerializeList(final List list, final Class expectedType, SerializationContext ctx) throws SerializeException {
       final JSONArray result = new JSONArray(list.size());
       
       for (final Object element : list) {
           final Object serializedElement = serializeObject(element, expectedType, ctx);
           result.add(serializedElement);
       }

      return result;
   }

   /**
    * Serializes and returns the specified array into a JSON array.
    * 
    * @param array The array to serialize. Must not be <code>null</code>.
    * 
    * @param expectedType The expected type of each item in the array.
    * Must not be <code>null</code>.
    * 
    * @return The JSON array representing the serialized array. Never <code>null</code>.
    * 
    * @throws SerializeException Thrown on error.
    */
   private JSONArray serializeArray(final Object array, final Class expectedType, SerializationContext ctx) throws SerializeException {
       final int length = Array.getLength(array);
     
       final JSONArray result = new JSONArray(length);

       for (int i=0; i<length; i++) {
           final Object element = Array.get(array, i);
           final Object serializedElement = serializeObject(element, expectedType, ctx);
           result.add(serializedElement);
       }

       return result;
   }

   /**
    * Returns whether or not the specified <code>feature</code> represents either
    * an item id or a state id.
    * 
    * @param feature The feature to query. Must not be <code>null</code>.
    * 
    * @return <code>true</code> is the specified <code>feature</code> is an
    * item or state id, false if it is not.
    * 
    * @throws SerializeException Thrown on error.
    */
   private boolean isItemId(final EStructuralFeature feature) {
       
       if (!(feature instanceof EAttribute))
           return false;
       
       if (feature.equals(RepositoryPackage.eINSTANCE.getItemHandle_ItemId()) ||
           feature.equals(RepositoryPackage.eINSTANCE.getItemHandle_StateId()))
               return true;
       
       return false;
   }
   
	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.ibm.team.repository.common.serialize.ISerializer2#serializeException(
	 * java.lang.Throwable, java.lang.Integer, java.io.Writer)
	 */
	@Override
	public void serializeException(Throwable exception, Integer errorCode, Writer writer) throws TeamRepositoryException, IOException {
		if (exception == null)
			throw new IllegalArgumentException("exception must not be null"); //$NON-NLS-1$
		if (writer == null)
			throw new IllegalArgumentException("writer must not be null"); //$NON-NLS-1$

		// unwrap LoggedException
       boolean isLogged = false;
       if (exception instanceof LoggedException) {
       	exception = exception.getCause();
       	isLogged = true;
       }
		// MessagePlaceholder - don't log/correlate it
		if (exception instanceof MessagePlaceholder) {
			isLogged = false;
		}

		JSONObject obj = new JSONObject();
		formatErrorObj(obj, exception, errorCode, isLogged);
		if (exception instanceof TeamRepositoryException) {
			if (((TeamRepositoryException) exception).getData() instanceof EObject) {
				EObject data = (EObject) ((TeamRepositoryException) exception).getData();
				try {
					obj.put(ERROR_DATA, toJSONObject(data, data.getClass(), new SerializationContext()));
				} catch (SerializeException e) {
					// Ignore errors in serializing the exception data
				}
			}
		}
		obj.serialize(writer, format());
	}

	/**
    * Recursively formats exceptions and nested exceptions into a JSON object
    * 
    * @param obj The JSON object to build. Must not be <code>null</code>. 
    * @param ex The exception to build the JSON object from. Must not be <code>null</code>.
    * @param errorCode The error code to serialize. May be <code>null</code>.
    */
	private void formatErrorObj(JSONObject obj, Throwable ex, Integer errorCode, boolean isLogged) {

		if (obj == null)
			throw new IllegalArgumentException("obj must not be null"); //$NON-NLS-1$
		if (ex == null)
			throw new IllegalArgumentException("exception must not be null"); //$NON-NLS-1$

		/*
		 * The main error object.
		 */
		obj.put(ERROR_CLASSNAME_KEY, ex.getClass().getCanonicalName());
		obj.put(ERROR_CODE_KEY, errorCode);
		obj.put(ERROR_MSG_KEY, encode(ex.getMessage(), true));

		if (StackUtil.shouldAvoidSerializingExceptions()) {
			/*
			 * Construct log correlation message, if appropriate
			 */
			if (isLogged) {
				// Correlation id handling copied from MarshallerUtil#constructExceptionMessage(Throwable, boolean)
				// Note: We're assuming this exception was logged by TeamServletLoggingHelper.logWarningWithClientDetails()
				// which uses the exception stack's SHA-1 hash string as a correlation id, so we use the same id.
				// TODO: Find a better way that doesn't make assumptions.
				String correlationId = HashUtil.getStacktraceHash(ex);
				String userPattern = Messages.getClientString(NLS_EXCEPTION_PLACEHOLDER_USER_MESSAGE);
				String userMessage = NLS.bind(userPattern, correlationId);
				String encodedUserMessage = encode(userMessage, true);
				obj.put(ERROR_TRACE_KEY, encodedUserMessage);
			}
		} else {
			/*
			 * Turn the stack trace into an array of strings, and an array of
			 * members.
			 */
			StackTraceElement stackTraceElems[] = ex.getStackTrace();

			JSONArray stackTrace = new JSONArray();
			JSONArray formattedStackTrace = new JSONArray();
			for (StackTraceElement elem : stackTraceElems) {
				stackTrace.add(elem.toString());

				JSONObject traceObj = new JSONObject();
				traceObj.put(ERROR_TRACE_CLASSNAME, elem.getClassName());
				traceObj.put(ERROR_TRACE_METHODNAME, elem.getMethodName());
				traceObj.put(ERROR_TRACE_FILENAME, elem.getFileName());
				traceObj.put(ERROR_TRACE_LINENUM, elem.getLineNumber());

				formattedStackTrace.add(traceObj);
			}
			obj.put(ERROR_TRACE_KEY, stackTrace);
			obj.put(ERROR_TRACE_MARSHALL_KEY, formattedStackTrace);
		}

		Throwable source = ex.getCause();
		if (source != null) {
			JSONObject nestedObj = new JSONObject();
			obj.put(NESTED_ERROR_KEY, nestedObj);
			// isLogged false - we don't want to put a log correlation msg on nested exceptions
			formatErrorObj(nestedObj, source, null, false);
		}

		serializeTeamRepositoryExceptionExtraData(obj, ex);
	}

	private void serializeTeamRepositoryExceptionExtraData(JSONObject obj, Throwable ex) {
		// Handle the special data fields  
		if (ex instanceof TeamRepositoryException) {
		   TeamRepositoryException tre = (TeamRepositoryException)ex;

		   JSONObject extraData = new JSONObject();

	       SerializationContext ctx = new SerializationContext();
	       ctx.include_etype = true;
	       for (Map.Entry<String, ?> entry : tre.getSerializableData().entrySet()) {
	    	   if (entry.getValue() instanceof EObject) {
		    	   try {
		    		   EObject value = (EObject) entry.getValue();
		    		   Class eClass = null;
		    		   if (value != null) {
		    			   eClass = value.getClass();
		    		   }
		    		   extraData.put(entry.getKey(), toJSONObject(value, eClass, ctx));
		    	   } catch (SerializeException e) {
		    	   }
	    	   }
	       }
		   
		   if (extraData.size() > 0) {
			   obj.put(ERROR_ALLDATA, extraData);
		   }
		}
	}
   
   /**
    * Get the location string for a virtual item. If the virtual specifies its
    * own location, it is used. Otherwise, the URI serializer is given a chance
    * to generate one.
    * 
    * @param virtual the virtual to get the location for
    * 
    * @return the location string for the virtual, or <code>null</code>
    * if one is not defined
    * 
    * @throws SerializeException Thrown on error.
    */
   private String getVirtualLocation(IVirtual virtual) throws SerializeException {
       String uriString = virtual.getLocation();
       if (uriString == null) {
           try {
               URI uri = _uriSerializer.getUri(virtual);
               if (uri != null) {
                   uriString = uri.toString();
               }
           } catch (IllegalArgumentException e) {
               // The default URI serializer throws IllegalArgumentException if
               // a virtual location is undefined; we just want to return null
           }
       }
       
       return uriString;
   }
}
