/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/

package com.ibm.team.repository.common.serialize.internal;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EPackage;

/**
 * Provides a thread-safe registry of XML Namespace URIs and prefixes.
 * Namespace URIs are registered with a preferred prefix.  If the preferred
 * prefix has already been registered, a new prefix will be chosen that does
 * not exist.  Thus, each registered namespace will have a unique prefix 
 * assigned to it.
 */
public class NamespaceRegistry {
	
	static public final NamespaceRegistry INSTANCE = new NamespaceRegistry();
	
	//--------------------------------------------------------------
	
	static public final String NSURI_XSD            = "http://www.w3.org/2001/XMLSchema"; //$NON-NLS-1$
	static public final String NSURI_XSI            = "http://www.w3.org/2001/XMLSchema-instance"; //$NON-NLS-1$
	
    static public final String NSURI_SERIALIZER     = "http://www.ibm.com/team/1.0.0/Serializer"; //$NON-NLS-1$
    static public final String NSURI_SERIALIZER_0_3 = "http://www.ibm.com/team/1.0.0/Serializer/0.3"; //$NON-NLS-1$
    static public final String NSURI_SERIALIZER_0_4 = "http://www.ibm.com/team/1.0.0/Serializer/0.4"; //$NON-NLS-1$
	
	//--------------------------------------------------------------
	
	private final Map    uri2prefix;
	private final Map    prefix2uri;
	private final Object lock;
    private int    nsIndex;
	
	/**
	 * Private to force instantiations through static method.
	 */
	private NamespaceRegistry() {
		super();
		
		uri2prefix = new HashMap();
		prefix2uri = new HashMap();
		lock       = new Object();
        nsIndex    = 0;
		
		registerDefaults();
	}

	/**
	 * Register a bunch of defaults.
	 */
	private void registerDefaults() {
        register(NSURI_SERIALIZER);
        register(NSURI_SERIALIZER_0_3);
        register(NSURI_SERIALIZER_0_4);
	}
	
	/**
	 * Register a prefix and uri.
	 * The prefix specified is the preferred prefix.  If null or
	 * empty, a prefix will be generated.  If the preferred 
	 * prefix has already been registered, 
	 * 
	 * @return the prefix the uri was actually registered under
	 */
    private String register(String uri) {
        if (null == uri) throw new IllegalArgumentException("uri must not be null"); //$NON-NLS-1$
        
        synchronized(lock) {
            String prefix = String.valueOf('n') + nsIndex;
            nsIndex++;
            
            uri2prefix.put(uri, prefix);
            prefix2uri.put(prefix, uri);
            
            return prefix;
        }
    }

	/**
	 * Return the uri associated with a prefix. 
	 */
	public String getURI(String prefix) {
		synchronized (lock) {
			return (String) prefix2uri.get(prefix);
		}
	}
	
	/**
	 * Return the prefix associated with an EPackage. 
	 */
	public String getPrefix(EPackage ePackage) {
		synchronized (lock) {
			String result = (String) uri2prefix.get(ePackage.getNsURI());
			if (null != result) return result;
			
			return register(ePackage.getNsURI());
		}
	}
	
	/**
	 * Return the prefix associated with an EClass. 
	 */
	public String getPrefix(EClass eClass) {
		return getPrefix(eClass.getEPackage());
	}
	
	/**
	 * Return the prefix associated with a nsURI. 
	 */
	public String getPrefix(String nsURI) {
		synchronized (lock) {
			return (String) uri2prefix.get(nsURI);
		}
	}
	
	/**
	 * Return the prefix associated with an EClass. 
	 */
	public String getPrefixedName(EClass eClass) {
		String prefix = getPrefix(eClass.getEPackage());
		
		StringBuffer sb = new StringBuffer();
		sb.append(prefix);
		sb.append(':');
		sb.append(eClass.getName());
		
		return sb.toString();
	}

	/**
	 * Return a list of all the prefixes known.
	 */
	public List getAllPrefixes() {
		return new ArrayList(prefix2uri.keySet());
	}
	
    /**
     * Return a list of all the prefixes known.
     */
    public List getAllURIs() {
        return new ArrayList(uri2prefix.keySet());
    }
    
}
