/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010, 2011. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/

package com.ibm.team.repository.common.serialize.internal;

import java.lang.reflect.Array;
import java.util.Iterator;
import java.util.List;

import org.eclipse.emf.common.util.Enumerator;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EStructuralFeature;

import com.ibm.team.repository.common.serialize.IPrimitiveRegistry;
import com.ibm.team.repository.common.serialize.SerializeException;
import com.ibm.team.repository.common.serialize.internal.nls.Messages;
import com.ibm.team.repository.common.util.NLS;

/**
 * Walks an object graph with a specified visitor.
 */
public class ObjectGraphWalker {

	private final Object                 root;
	private final ObjectGraphVisitorInfo info;
	
	/**
	 * Create a new instance of this class.
	 */
	static public ObjectGraphWalker newInstance(Object root) {
		return new ObjectGraphWalker(root);
	}
	
	/**
	 * private constructor.
	 */
	private ObjectGraphWalker(Object root) {
		super();
		
		this.root = root;
		this.info = new ObjectGraphVisitorInfo();
	}
	
	/**
	 * Walk the object graph with the specified visitor.
	 */
	public void walk(ObjectGraphVisitor visitor) throws SerializeException {
		walk(visitor, root, null, null, -1);
	}
	
	/**
	 * 
	 */
	private void walk(
			ObjectGraphVisitor visitor, 
			Object             object, 
			Object             container, 
			EStructuralFeature structuralFeature,
			int                index
			) throws SerializeException {
		
		info.object            = object; 
		info.container         = container;
		info.structuralFeature = structuralFeature;
		info.index             = index;

		// visit the first object
		if (null == object) {
			visitor.visit(info);
			return;
		}
		
		else if (IPrimitiveRegistry.INSTANCE.isPrimitive(object.getClass())) {
			visitor.visit(info);
			return;
		}
		
		else if (object instanceof EObject) {
			visitor.visit(info);
			walkEObject(visitor, (EObject) object);
		}
		
        else if (object instanceof Enumerator) {
            visitor.visit(info);
        }
        
		else if (object instanceof List) {
			walkList(visitor, (List) object, container, structuralFeature);
		}
		
        else if (object.getClass().isArray()) {
            walkArray(visitor, object, container, structuralFeature);
        }
        
		else {
			throw new SerializeException(NLS.bind(Messages.getServerString("ObjectGraphWalker.UnknownType"), object.getClass())); //$NON-NLS-1$
		}
		
	}
	
	/**
	 * 
	 */
	private void walkEObject(ObjectGraphVisitor visitor, EObject eObject) throws SerializeException {
		List sfs = Util.getStructuralFeatures(eObject.eClass(), null);
		
		for (Iterator iter=sfs.iterator(); iter.hasNext();) {
			EStructuralFeature sf = (EStructuralFeature) iter.next();
            
            if (sf.isTransient()) continue;
			
			Object value = eObject.eGet(sf);
			walk(visitor, value, eObject, sf, -1);
		}
	}
	
	/**
	 * 
	 */
	private void walkList(
			ObjectGraphVisitor visitor, 
			List               list, 
			Object             container, 
			EStructuralFeature structuralFeature
			) throws SerializeException {
		
		int index = 0;
		
		for (Iterator iter=list.iterator(); iter.hasNext();) {
			Object element = iter.next();
			walk(visitor, element, container, structuralFeature, index);
			
			index++;
		}
		
	}
    
    /**
     * 
     */
    private void walkArray(
            ObjectGraphVisitor visitor, 
            Object             array, 
            Object             container, 
            EStructuralFeature structuralFeature
            ) throws SerializeException {

        int length = Array.getLength(array);
        
        for (int index=0; index<length; index++) {
            Object element = Array.get(array, index);
            walk(visitor, element, container, structuralFeature, index);
        }
        
    }
    
}
