/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010, 2014. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/

package com.ibm.team.repository.common.serialize.internal;

import java.sql.Timestamp;
import java.text.ParseException;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import java.util.WeakHashMap;

import com.ibm.team.repository.common.UUID;
import com.ibm.team.repository.common.serialize.IPrimitiveHandler;
import com.ibm.team.repository.common.serialize.IPrimitiveRegistry;
import com.ibm.team.repository.common.serialize.SerializeException;
import com.ibm.team.repository.common.utils.DateUtils;

/**
 *
 */
public class PrimitiveRegistry implements IPrimitiveRegistry {

    private final Map primitives;
    private final Map nsPrimitives;
    
    /**
     * 
     */
    public PrimitiveRegistry() {
        super();

        primitives   = Collections.synchronizedMap(new WeakHashMap());
        nsPrimitives = Collections.synchronizedMap(new HashMap());
        addInstrinsicHandlers();
    }

    /**
     * 
     */
    private void addInstrinsicHandlers() {
        register(Byte.TYPE, PrimitiveHandlerByte.BYTE_INSTANCE);
        register(Byte.class, PrimitiveHandlerByte.BYTE_INSTANCE);
        register(Short.TYPE, PrimitiveHandlerShort.SHORT_INSTANCE);
        register(Short.class, PrimitiveHandlerShort.SHORT_INSTANCE);
        register(Integer.TYPE, PrimitiveHandlerInteger.INTEGER_INSTANCE);
        register(Integer.class, PrimitiveHandlerInteger.INTEGER_INSTANCE);
        register(Long.TYPE, PrimitiveHandlerLong.LONG_INSTANCE);
        register(Long.class, PrimitiveHandlerLong.LONG_INSTANCE);
        register(Float.TYPE, PrimitiveHandlerFloat.FLOAT_INSTANCE);
        register(Float.class, PrimitiveHandlerFloat.FLOAT_INSTANCE);
        register(Double.TYPE, PrimitiveHandlerDouble.DOUBLE_INSTANCE);
        register(Double.class, PrimitiveHandlerDouble.DOUBLE_INSTANCE);
        register(Boolean.TYPE, PrimitiveHandlerBoolean.BOOLEAN_INSTANCE);
        register(Boolean.class, PrimitiveHandlerBoolean.BOOLEAN_INSTANCE);
        register(Character.TYPE, PrimitiveHandlerCharacter.CHARACTER_INSTANCE);
        register(Character.class, PrimitiveHandlerCharacter.CHARACTER_INSTANCE);
        register(Void.TYPE, PrimitiveHandlerVoid.VOID_INSTANCE);
        register(String.class, PrimitiveHandlerString.STRING_INSTANCE);
        register(Date.class, PrimitiveHandlerDate.DATE_INSTANCE);
        register(Timestamp.class, PrimitiveHandlerTimestamp.TIMESTAMP_INSTANCE);
        register(UUID.class, PrimitiveHandlerRepoUUID.UUID_INSTANCE);
    }

    /**
     * 
     */
    public IPrimitiveHandler getHandler(final Class clazz) {
        return (IPrimitiveHandler) primitives.get(clazz);
    }

    public IPrimitiveHandler getHandler(final String[] xsiType) {
        String nsKey= xsiType[0] + ' ' + xsiType[1];
        
        return (IPrimitiveHandler) nsPrimitives.get(nsKey);
    }

    /**
     * 
     */
    public void register(final Class clazz, final IPrimitiveHandler handler) {
		primitives.put(clazz, handler);
        
        String[] xsiType = handler.getXsiType();
        String   nsKey   = xsiType[0] + ' ' + xsiType[1];
        
        nsPrimitives.put(nsKey, handler);
	}

    /**
     * 
     */
    public boolean isPrimitive(final Class clazz) {
        return primitives.containsKey(clazz);
    }

    /**
     *
     */
    static class PrimitiveHandlerByte implements IPrimitiveHandler {
        static final IPrimitiveHandler BYTE_INSTANCE = new PrimitiveHandlerByte();

        public String objectToString(final Object object) {
            return ((Byte) object).toString();
        }

        public Object stringToObject(final String string) {
            return Byte.decode(string);
        }

        public String[] getXsiType() {
            return new String[] { NamespaceRegistry.NSURI_XSD, "unsignedByte" }; //$NON-NLS-1$
        }

        public boolean isEmptyValid() {
            return false;
        }

    }

    /**
     *
     */
    static class PrimitiveHandlerShort implements IPrimitiveHandler {
        static final IPrimitiveHandler SHORT_INSTANCE = new PrimitiveHandlerShort();

        public String objectToString(final Object object) {
            return ((Short) object).toString();
        }

        public Object stringToObject(final String string) {
            return Short.decode(string);
        }

        public String[] getXsiType() {
            return new String[] { NamespaceRegistry.NSURI_XSD, "short" }; //$NON-NLS-1$
        }

        public boolean isEmptyValid() {
            return false;
        }

    }

    /**
     *
     */
    static class PrimitiveHandlerInteger implements IPrimitiveHandler {
        static final IPrimitiveHandler INTEGER_INSTANCE = new PrimitiveHandlerInteger();

        public String objectToString(final Object object) {
            return ((Integer) object).toString();
        }

        public Object stringToObject(final String string) {
            return Integer.decode(string);
        }

        public String[] getXsiType() {
            return new String[] { NamespaceRegistry.NSURI_XSD, "int" }; //$NON-NLS-1$
        }

        public boolean isEmptyValid() {
            return false;
        }

    }

    /**
     *
     */
    static class PrimitiveHandlerLong implements IPrimitiveHandler {
        static final IPrimitiveHandler LONG_INSTANCE = new PrimitiveHandlerLong();

        public String objectToString(final Object object) {
            return ((Long) object).toString();
        }

        public Object stringToObject(final String string) {
            return Long.decode(string);
        }

        public String[] getXsiType() {
            return new String[] { NamespaceRegistry.NSURI_XSD, "long" }; //$NON-NLS-1$
        }

        public boolean isEmptyValid() {
            return false;
        }

    }

    /**
     *
     */
    static class PrimitiveHandlerFloat implements IPrimitiveHandler {
        static final IPrimitiveHandler FLOAT_INSTANCE = new PrimitiveHandlerFloat();

        public String objectToString(final Object object) {
            return ((Float) object).toString();
        }

        public Object stringToObject(final String string) {
            return Float.valueOf(string);
        }

        public String[] getXsiType() {
            return new String[] { NamespaceRegistry.NSURI_XSD, "float" }; //$NON-NLS-1$
        }

        public boolean isEmptyValid() {
            return false;
        }

    }

    /**
     *
     */
    static class PrimitiveHandlerDouble implements IPrimitiveHandler {
        static final IPrimitiveHandler DOUBLE_INSTANCE = new PrimitiveHandlerDouble();

        public String objectToString(final Object object) {
            return ((Double) object).toString();
        }

        public Object stringToObject(final String string) {
            return Double.valueOf(string);
        }

        public String[] getXsiType() {
            return new String[] { NamespaceRegistry.NSURI_XSD, "double" }; //$NON-NLS-1$
        }

        public boolean isEmptyValid() {
            return false;
        }

    }

    /**
     *
     */
    static class PrimitiveHandlerBoolean implements IPrimitiveHandler {
        static final IPrimitiveHandler BOOLEAN_INSTANCE = new PrimitiveHandlerBoolean();

        public String objectToString(final Object object) {
            return ((Boolean) object).toString();
        }

        public Object stringToObject(final String string) {
            return Boolean.valueOf(string);
        }

        public String[] getXsiType() {
            return new String[] { NamespaceRegistry.NSURI_XSD, "boolean" }; //$NON-NLS-1$
        }

        public boolean isEmptyValid() {
            return false;
        }

    }

    /**
     *
     */
    static class PrimitiveHandlerCharacter implements IPrimitiveHandler {
        static final IPrimitiveHandler CHARACTER_INSTANCE = new PrimitiveHandlerCharacter();

        public String objectToString(final Object object) {
            return ((Character) object).toString();
        }

        public Object stringToObject(final String string) {
            return Character.valueOf(string.charAt(0));
        }

        public String[] getXsiType() {
            return new String[] { NamespaceRegistry.NSURI_XSD, "string" }; //$NON-NLS-1$
        }

        public boolean isEmptyValid() {
            return false;
        }

    }

    /**
     *
     */
    static class PrimitiveHandlerVoid implements IPrimitiveHandler {
        static final IPrimitiveHandler VOID_INSTANCE = new PrimitiveHandlerVoid();

        public String objectToString(final Object object) {
            return null;
        }

        public Object stringToObject(final String string) {
            return null;
        }

        public String[] getXsiType() {
            return new String[] { NamespaceRegistry.NSURI_XSD, "string" }; //$NON-NLS-1$
        }

        public boolean isEmptyValid() {
            return false;
        }

    }

    /**
     *
     */
    static class PrimitiveHandlerString implements IPrimitiveHandler {
        static final IPrimitiveHandler STRING_INSTANCE = new PrimitiveHandlerString();

        public String objectToString(final Object object) {
            return (String) object;
        }

        public Object stringToObject(final String string) {
            return string;
        }

        public String[] getXsiType() {
            return new String[] { NamespaceRegistry.NSURI_XSD, "string" }; //$NON-NLS-1$
        }

        public boolean isEmptyValid() {
            return true;
        }

    }

    /**
     *
     */
    static class PrimitiveHandlerDate implements IPrimitiveHandler {
        static final IPrimitiveHandler DATE_INSTANCE = new PrimitiveHandlerDate();

        public String objectToString(final Object object) {
            return DateUtils.formatTimeRFC3339((Date)object);
        }

        public Object stringToObject(final String string) throws SerializeException {
            try {
                Timestamp time = DateUtils.parseTimeRFC3339(string);
                return new Date(time.getTime());
            } 
            catch (ParseException e) {
                throw SerializeException.from(e);
            }
        }

        public String[] getXsiType() {
            return new String[] { NamespaceRegistry.NSURI_XSD, "dateTime" }; //$NON-NLS-1$
        }

        public boolean isEmptyValid() {
            return false;
        }

    }

    /**
     *
     */
    static class PrimitiveHandlerTimestamp implements IPrimitiveHandler {
        static final IPrimitiveHandler TIMESTAMP_INSTANCE = new PrimitiveHandlerTimestamp();

        public String objectToString(final Object object) {
            return DateUtils.formatTimeRFC3339((Date)object);
        }

        public Object stringToObject(final String string) throws SerializeException {
            try {
                return DateUtils.parseTimeRFC3339(string);
            } 
            catch (ParseException e) {
                throw SerializeException.from(e);
            }
        }

        public String[] getXsiType() {
            return new String[] { NamespaceRegistry.NSURI_XSD, "dateTime" }; //$NON-NLS-1$
        }

        public boolean isEmptyValid() {
            return false;
        }
    }

    /**
     *
     */
    static class PrimitiveHandlerRepoUUID implements IPrimitiveHandler {
        static final IPrimitiveHandler UUID_INSTANCE = new PrimitiveHandlerRepoUUID();

        public String objectToString(final Object object) {
            return ((UUID) object).getUuidValue();
        }

        public Object stringToObject(final String string) {
            return UUID.valueOf(string);
        }

        public String[] getXsiType() {
            return new String[] { NamespaceRegistry.NSURI_XSD, "string" }; //$NON-NLS-1$
        }

        public boolean isEmptyValid() {
            return false;
        }

    }

}
