/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010, 2013. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/

package com.ibm.team.repository.transport.auth;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * Utility class for trace logging. 
 */
public class Tracer {
    
    private final Log _logger;
    private final int _indent;
    private final int INDENT_DEFAULT = Integer.parseInt(System.getProperty("jazz.trace.indentLevel", "2")); //$NON-NLS-1$ //$NON-NLS-2$
    private final Class<?> _class;
    
    /**
     * Creates an instance of the tracer class using the default logger configured
     * for this class.
     */
    public Tracer() {
        _class = null;
        _logger = LogFactory.getLog(this.getClass());
        _indent = INDENT_DEFAULT;
    }
    
    /**
     * Creates an instance of the tracer class using the default logger configured
     * for this class.
     * 
     * @param clazz The class whose short-form classname is pre-pended to
     * enter/leave trace messages. May be <code>null</code>.
     */
    public Tracer(Class<?> clazz) {
        _class = clazz;
        _logger = LogFactory.getLog(this.getClass());
        _indent = INDENT_DEFAULT;
    }
    
    /**
     * Creates an instance of the tracer class using the default logger configured
     * for this class.
     * 
     * @param clazz The class whose short-form classname is pre-pended to
     * enter/leave trace messages. May be <code>null</code>.
     * @param log The log to append to. May be <code>null</code>.  If <code>null</code>
     * the logger configured for this class is used.
     */
    public Tracer(Class<?> clazz, Log log) {
        _class = clazz;
        if (log == null)
            _logger = LogFactory.getLog(this.getClass());
        else
            _logger = log;
        _indent = INDENT_DEFAULT;
    }
    
    /**
     * Resets the trace object to the values it was constructed with. 
     */
    public void reset() {
        setCallDepth(_indent);
    }
    
    /**
     * Prints debug-level messages to the configured logger.
     * The messages are indented accordingly to the level of recursion in the
     * call stack as influenced by the configured indent level.
     * 
     * @param message A format string as per "String.format(...)". May be
     * <code>null</code>.
     * 
     * @param args Optional args to be inserted in the format string.
     */
    public void debug(String message, Object... args) {
        if (message == null)
            return;
        if (_logger.isDebugEnabled()) {
            StringBuffer buf = new StringBuffer(127);
            incrementCallDepth();
            buf.append(getIndentString());
            decrementCallDepth();
            buf.append(String.format(message, args));
            _logger.debug(buf.toString());
        }
    }
    
    /**
     * Prints trace-level messages to the configured logger.
     * The messages are indented accordingly to the level of recursion in the
     * call stack as influenced by the configured indent level.
     * 
     * @param message A format string as per "String.format(...)". May be
     * <code>null</code>.
     * 
     * @param args Optional args to be inserted in the format string.
     */
    public void trace(String message, Object... args) {
        if (message == null)
            return;
        if (_logger.isTraceEnabled()) {
            StringBuffer buf = new StringBuffer(127);
            incrementCallDepth();
            buf.append(getIndentString());
            decrementCallDepth();
            buf.append(String.format(message, args));
            _logger.trace(buf.toString());
        }
    }

    /**
     * Prints trace-level messages to the configured logger prefixed with the term "Entering ".
     * The messages are indented accordingly to the level of recursion in the
     * call stack as influenced by the configured indent level.
     * 
     * @param message A format string in per "String.format(...)". May be
     * <code>null</code>.
     * 
     * @param args Optional args to be inserted in the format string.
     */
    public void enter(String message, Object... args) {
        if (message == null)
            return;
        if (_logger.isTraceEnabled()) {
            incrementCallDepth();
            StringBuffer buf = new StringBuffer(127);
            buf.append(getIndentString());
            buf.append("[ENTERING] "); //$NON-NLS-1$
            if (_class != null) {
                buf.append(_class.getSimpleName());
                buf.append('#');
            }
            buf.append(String.format(message, args));
            _logger.trace(buf.toString());
        }
    }
    
    /**
     * Prints trace-level messages to the configured logger prefixed with the term "Exiting ".
     * The messages are indented accordingly to the level of recursion in the
     * call stack as influenced by the configured indent level.
     * 
     * @param message A format string in per "String.format(...)". May be
     * <code>null</code>.
     * 
     * @param args Optional args to be inserted in the format string.
     */
    public void leave(String message, Object... args) {
        if (message == null)
            return;
        if (_logger.isTraceEnabled()) {
            StringBuffer buf = new StringBuffer(127);
            buf.append(getIndentString());
            buf.append("[EXITING]  "); //$NON-NLS-1$
            if (_class != null) {
                buf.append(_class.getSimpleName());
                buf.append('#');
            }
            buf.append(String.format(message, args));
            _logger.trace(buf.toString());
            decrementCallDepth();
        }
    }
    
    static private ThreadLocal<Integer> _callDepth = new ThreadLocal<Integer>();
    
    static private Integer getCallDepth() {
        Integer ndx = _callDepth.get();
        if (ndx == null) {
            ndx = 0;
            _callDepth.set(ndx);
        }
        return ndx;
    }
    
    private void setCallDepth(int depth) {
        _callDepth.set(depth);
    }
    
    private void incrementCallDepth() {
        int ndx = getCallDepth();
        ndx += _indent;
        setCallDepth(ndx);
    }
    
    private void decrementCallDepth() {
        int ndx = getCallDepth();
        if (ndx == 0)
            return;
        
        ndx -= _indent;
        setCallDepth(ndx);
    }
    
    private String getIndentString() {
        int ndx = getCallDepth();
        char[] chars = new char[ndx];
        for (int i = 0; i < ndx; i++)
            chars[i] = ' ';
        
        return new String(chars);
    }

}
