/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010, 2014. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/

package com.ibm.team.repository.transport.auth;

import java.text.MessageFormat;

import org.apache.commons.httpclient.Credentials;
import org.apache.commons.httpclient.Header;
import org.apache.commons.httpclient.HostConfiguration;
import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.HttpMethod;
import org.apache.commons.httpclient.HttpState;
import org.apache.commons.httpclient.HttpStatus;
import org.apache.commons.httpclient.UsernamePasswordCredentials;
import org.apache.commons.httpclient.auth.AuthScope;
import org.apache.commons.httpclient.methods.GetMethod;
import org.apache.commons.httpclient.methods.PostMethod;
import org.apache.commons.httpclient.params.HttpClientParams;

/**
 * Utilities in support of form-based authentication.
 * @noinstantiate This class is not intended to be instantiated by clients.
 * @noextend This class is not intended to be subclassed by clients.
 */
public class TransportAuthUtil {

    // No retry header
    static public final String X_NO_RETRY_HEADER = "X-com-ibm-team-repository.common.remoteaccess.noRetry";  //$NON-NLS-1$

    // name of custom header that authentication messages are stored in
    public static final String FORM_AUTH_HEADER = "X-com-ibm-team-repository-web-auth-msg"; //$NON-NLS-1$
    
    // auth header value when authentication is required
    public static final String FORM_AUTH_REQUIRED_MSG = "authrequired"; //$NON-NLS-1$
    
    // auth header value when authentication is required
    public static final String FORM_AUTH_FAILED_MSG = "authfailed"; //$NON-NLS-1$
    
    // submit credentials to this URI for authentication
    public static final String FORM_AUTH_URI = "j_security_check"; //$NON-NLS-1$
    
    // name of the field to use when submitting the username
    public static final String FORM_AUTH_USER_FIELD = "j_username"; //$NON-NLS-1$
    
    // name of the field to use when submitting the password
    public static final String FORM_AUTH_PASSWORD_FIELD = "j_password"; //$NON-NLS-1$
    
    // send a request to this URI to logout
    public static final String FORM_LOGOUT_URI = "/auth/logout"; //$NON-NLS-1$
    
    // URI the server redirects to when authentication fails
    public static final String FORM_AUTH_FAILED_URI = "/auth/authfailed"; //$NON-NLS-1$
    
    // URI the server redirects to when authentication is required
    public static final String FORM_AUTH_REQUESTED_URI = "/auth/authrequired"; //$NON-NLS-1$

    // HTTP constants
    public static final String FORM_AUTH = "FORM"; //$NON-NLS-1$
    public static final String BASIC_AUTH = "BASIC"; //$NON-NLS-1$
    public static final String LOCATION = "Location"; //$NON-NLS-1$
    public static final String CONTENT_TYPE_HEADER = "Content-Type"; //$NON-NLS-1$
    public static final String CONTENT_TYPE_FORM_URLENCODED = "application/x-www-form-urlencoded; charset=utf-8"; //$NON-NLS-1$

    public static final String SLASH = "/"; //$NON-NLS-1$
    
    private static final String NLS_UNEXPECTED_LOGIN_ERROR = "TransportAuthUtil.UnexpectedLoginError"; //$NON-NLS-1$
    private static final String NLS_UNEXPECTED_STATUS = "TransportAuthUtil.UnexpectedStatus"; //$NON-NLS-1$
    private static final String NLS_INVALID_CREDENTIALS = "TransportAuthUtil.InvalidCredentials"; //$NON-NLS-1$
    private static final String NLS_SERVLET_ERROR = "TransportAuthUtil.ServletError"; //$NON-NLS-1$
    private static final String NLS_SERVER_CONFIG_ERROR = "TransportAuthUtil.ServerConfigError"; //$NON-NLS-1$
    private static final String NLS_LOGIN_ERROR = "TransportAuthUtil.LoginError"; //$NON-NLS-1$
    private static final String NLS_DETERMINE_AUTH_ERROR = "TransportAuthUtil.DetermineAuthError"; //$NON-NLS-1$

    static private final Tracer _tracer = new Tracer(TransportAuthUtil.class);
    
    /**
     * Returns whether or not form-based authorization has been requested by
     * the http server as evidenced in the specified client method.
     * 
     * @param method The http method containing response data.  Must not be
     * <code>null</code>.  Assumed to contain the state of an HTTP response.
     * 
     * @return <code>true</code> if form-based authorization is requested by
     * the server, <code>false</code> if it is not. 
     */
    static public boolean formAuthRequested(HttpMethod method) {
        _tracer.enter("formAuthRequested"); //$NON-NLS-1$
        try {
	        Header locationHeader = method.getResponseHeader(LOCATION);
	        boolean retval = (hasAuthRequiredHeader(method)) ||
	                (locationHeader != null && locationHeader.getValue().matches("^.*" + FORM_AUTH_REQUESTED_URI + ".*$"));  //$NON-NLS-1$//$NON-NLS-2$
	        
	        _tracer.trace("formAuthRequested - %s", retval); //$NON-NLS-1$
	        return retval;
        } finally {
            _tracer.leave("formAuthRequested"); //$NON-NLS-1$
        }
    }
    
    static public boolean hasAuthRequiredHeader(HttpMethod method) {
        _tracer.enter("hasAuthRequiredHeader"); //$NON-NLS-1$
        try {
	    	Header formAuthMsg = method.getResponseHeader(FORM_AUTH_HEADER);
	    	boolean retval = (formAuthMsg != null && formAuthMsg.getValue().equals(FORM_AUTH_REQUIRED_MSG));
	    	
	    	_tracer.trace("hasAuthRequiredHeader - %s", retval); //$NON-NLS-1$
	    	return retval;
        } finally {
            _tracer.leave("hasAuthRequiredHeader"); //$NON-NLS-1$
        }
    }

    static public HttpMethod formBasedAuthenticate(String repoPath, HostConfiguration hostConfig, HttpClient httpClient, UsernamePasswordCredentials storedCredentials, HttpState privateState) throws TransportAuthException {
        _tracer.enter("formBasedAuthenticate"); //$NON-NLS-1$
        try {
	    	final int expectedRedirectResponse = 302;
	        final int expectedOkResponse = 200;
	
	        if (!repoPath.endsWith(SLASH))
	            repoPath += SLASH;
	        
	        String formAuthURI = repoPath + FORM_AUTH_URI;
	        _tracer.debug("formAuthURI: %s", formAuthURI); //$NON-NLS-1$
	        PostMethod postMethod = new PostMethod(formAuthURI);
	        try {
	            postMethod.addParameter(FORM_AUTH_USER_FIELD, storedCredentials.getUserName());
	            postMethod.addParameter(FORM_AUTH_PASSWORD_FIELD, storedCredentials.getPassword());
	            
	            /* When the username and/or password contain non-ascii characters,
	             * we must explicitly set the charset to UTF-8 in the Content-Type
	             * header.   Otherwise, the server assumes ISO-8859-1. */
	            postMethod.setRequestHeader(CONTENT_TYPE_HEADER,
	            		CONTENT_TYPE_FORM_URLENCODED);
	            
	            try {
	                int rc = httpClient.executeMethod(hostConfig, postMethod, privateState);
	                _tracer.debug("statusCode: %d", rc); //$NON-NLS-1$
	                if (rc == HttpStatus.SC_BAD_REQUEST) {
	                	// Tomcat will respond with SC_BAD_REQUEST when the j_security_check URL is visited
	                	// before an authenticated URL has been visited.  If we see this then visit an
	                	// authenticated URL and try again
	                    visitAuthenticatedURL(repoPath, httpClient, privateState);
	                	
	                	HttpMethod method = formBasedAuthenticate(repoPath, hostConfig, httpClient, storedCredentials, privateState);
	                	
	                	// After authenticating, we also have to visit the same URL again
	                    visitAuthenticatedURL(repoPath, httpClient, privateState);
	                	return method;
	                }
	                
	                Header authHeader = postMethod.getResponseHeader(FORM_AUTH_HEADER);
	                if (authHeader != null) {
	                    if (authHeader.getValue().equals(FORM_AUTH_REQUIRED_MSG)) {
	                        _tracer.trace("formBasedAuthenticate - throwing TransportAuthException - authHeader: %s", authHeader.getValue()); //$NON-NLS-1$
	                        throw new TransportAuthException(
	                                MessageFormat.format(Messages.getString(NLS_SERVER_CONFIG_ERROR), FORM_AUTH_URI)); 
	                    }
	                    else if (authHeader.getValue().equals(FORM_AUTH_FAILED_MSG)) {
	                        _tracer.trace("formBasedAuthenticate - throwing TransportAuthException - authHeader: %s", authHeader.getValue()); //$NON-NLS-1$
	                        throw new TransportAuthException(Messages.getString(NLS_INVALID_CREDENTIALS)); 
	                    }
	                    else {
	                        _tracer.trace("formBasedAuthenticate - throwing TransportAuthException - authHeader: %s", authHeader.getValue()); //$NON-NLS-1$
	                        throw new TransportAuthException(
	                                MessageFormat.format(Messages.getString(NLS_SERVLET_ERROR),  
	                                        FORM_AUTH_REQUIRED_MSG, FORM_AUTH_FAILED_MSG, authHeader.getValue()));
	                    }
	                }
	                else if (rc == expectedRedirectResponse) {
	                    String redirectURI = postMethod.getResponseHeader(LOCATION).getValue(); 
	                    if (redirectURI.matches("^.*" + FORM_AUTH_FAILED_URI + ".*$")) {  //$NON-NLS-1$ //$NON-NLS-2$
	                        // Login failed
	                        _tracer.trace("formBasedAuthenticate - throwing TransportAuthException - redirectUri: %s", redirectURI); //$NON-NLS-1$
	                        throw new TransportAuthException(Messages.getString(NLS_INVALID_CREDENTIALS));  
	                    }
	                }
	                else if (rc != expectedOkResponse) {
	                    // The form auth request should always respond with a 200 ok or 302 redirect code
	                    // Unexpected response: fail the login
	                    String msg = MessageFormat.format(Messages.getString(NLS_UNEXPECTED_STATUS), 
	                    		formAuthURI, expectedRedirectResponse, rc);
	                    _tracer.trace("formBasedAuthenticate - throwing TransportAuthException - %s", msg); //$NON-NLS-1$
	                    throw new TransportAuthException(msg);
	                }
	            } catch (Exception e) {
	                if (e.getClass().equals(TransportAuthException.class)) {
	                    _tracer.trace("formBasedAuthenticate - re-throwing TransportAuthException"); //$NON-NLS-1$
	                    throw (TransportAuthException)e;
	                }
	                else {
	                    String msg = MessageFormat.format(Messages.getString(NLS_UNEXPECTED_LOGIN_ERROR), e.getMessage()); 
	                    _tracer.trace("formBasedAuthenticate - throwing TransportAuthException - %s", msg); //$NON-NLS-1$
	                    throw new TransportAuthException(msg, e); 
	                }
	            }
	        } finally {
	            postMethod.releaseConnection();
	        }
	        return postMethod;
        } finally {
            _tracer.leave("formBasedAuthenticate"); //$NON-NLS-1$
        }
    }
    
    
    /**
     * Authenicate to the referenced server using form-based authentication.
     * 
     * @param server The server to authenticate to. Must not be <code>null</code>.
     * 
     * @param httpClient The http client communications package. Must not be <code>null</code>.
     * 
     * @param storedCredentials The credentials to authenticate with. Must not be <code>null</code>,
     * nor should its fields be null or empty.
     * 
     * @throws TransportAuthException Thrown on error.
     */
    static public void formBasedAuthenticate(String repoPath, HttpClient httpClient, 
            UsernamePasswordCredentials storedCredentials,
            HttpState privateState) throws TransportAuthException
    {
        _tracer.enter("formBasedAuthenticate"); //$NON-NLS-1$
        try {
        	formBasedAuthenticate(repoPath, null, httpClient, storedCredentials, privateState);
        } finally {
        	_tracer.leave("formBasedAuthenticate"); //$NON-NLS-1$
        }
    }
    
	/** 
	 * This is needed for Tomcat because it responds with SC_BAD_REQUEST when the j_security_check URL is visited
	 * before an authenticated URL has been visited. This same URL must also be visited after authenticating with j_security_check
	 * otherwise Tomcat will not consider the session authenticated
	 */
    public static void visitAuthenticatedURL(String repoPath, HttpClient httpClient, HttpState privateState) throws TransportAuthException {
    	visitAuthenticatedURLWithStatus(repoPath, httpClient, privateState);
    }
    
	/** 
	 * This is needed for Tomcat because it responds with SC_BAD_REQUEST when the j_security_check URL is visited
	 * before an authenticated URL has been visited. This same URL must also be visited after authenticating with j_security_check
	 * otherwise Tomcat will not consider the session authenticated
	 * 
	 * @since 3.0.1.1
	 */
    public static int visitAuthenticatedURLWithStatus(String repoPath, HttpClient httpClient, HttpState privateState) throws TransportAuthException {
        _tracer.enter("visitAuthenticatedURL"); //$NON-NLS-1$
        try {
	        if (!repoPath.endsWith(SLASH))
	            repoPath += SLASH;
	    	GetMethod authenticatedURL = new GetMethod(repoPath + "secure/authenticated/identity"); //$NON-NLS-1$
	    	_tracer.debug("Path: %s", authenticatedURL.getPath()); //$NON-NLS-1$
	    	
	    	authenticatedURL.setFollowRedirects(true);
	    	int statusCode;
	    	try {
	    		statusCode = httpClient.executeMethod(null, authenticatedURL, privateState);
	    	} catch (Exception e) {
	            String msg = MessageFormat.format(Messages.getString(NLS_UNEXPECTED_LOGIN_ERROR), e.getMessage()); 
	            _tracer.trace("visitAuthenticatedURL - throwning TransportAuthException: %s", msg); //$NON-NLS-1$
	            throw new TransportAuthException(msg, e); 
	    	} finally {
	    		authenticatedURL.releaseConnection();
	    	}
	    	
	    	_tracer.debug("HTTP status code after touching \"%s\": %d", //$NON-NLS-1$ 
	    	        authenticatedURL.getPath(), statusCode);
	    	
	        return statusCode;
        } finally {
            _tracer.leave("visitAuthenticatedURL"); //$NON-NLS-1$
        }
    }
    
    /**
     * Returns the credentials authorization method defined by the referenced
     * server.
     * 
     * @param httpClient The http client instance.  Must not be <code>null</code>.
     * 
     * @param repoPath The path to the repository to communicate with, as in /jazz.
     * Must not be <code>null</code>.
     * 
     * @return Returns "FORM" for form-based authenication, "BASIC" for basic
     * auth. Never <code>null</code>.
     * 
     * @throws TransportAuthException Thrown on error.
     */
    static public String determineAuthMethod(HttpClient httpClient, String repoPath) throws TransportAuthException {
        
        if (httpClient == null)
            throw new IllegalArgumentException("httpClient must not be null"); //$NON-NLS-1$
        if (repoPath == null)
            throw new IllegalArgumentException("repoPath must not be null"); //$NON-NLS-1$
        
        _tracer.enter("determineAuthMethod"); //$NON-NLS-1$
        try {
        if (!repoPath.endsWith(SLASH))
            repoPath += SLASH;
        repoPath += FORM_AUTH_URI;

        _tracer.debug("path: %s", repoPath); //$NON-NLS-1$
        GetMethod formAuth = new GetMethod(repoPath);
        String serverAuthMethod = null;
        formAuth.setFollowRedirects(false);
        try {
            /* 
             * if the FORM_AUTH_URI/j_security_check path exists (as evidenced by a 200 or 302 return code), 
        	 * then we know we are using FORM auth
        	 */
        	int rc;
            try {
                rc = httpClient.executeMethod(formAuth);
            }
            catch (Exception e) {
                String msg = MessageFormat.format(Messages.getString(NLS_LOGIN_ERROR), e.getMessage());
                _tracer.trace("determineAuthMethod - throwing TransportAuthException: %s", msg); //$NON-NLS-1$
                throw new TransportAuthException(msg, e);
            }
            _tracer.debug("status: %d", rc); //$NON-NLS-1$

            if (rc >= 500) {
            	/*
            	 * If the server returned a 500, there's something screwed up there.
            	 */
            	String msg = Messages.getString(NLS_DETERMINE_AUTH_ERROR);
                _tracer.trace("determineAuthMethod - throwing TransportAuthException: %s", msg); //$NON-NLS-1$
            	throw new TransportAuthException(msg);
            }
            else if (rc == 200) {
                // Tomcat and Jetty return a status code 200 if the server is using FORM auth
                serverAuthMethod = FORM_AUTH;
            }
            else if (rc == 302 && formAuth.getResponseHeader("location").getValue().matches("^.*" + FORM_AUTH_FAILED_URI + ".*$")) {  //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
                // WebSphere returns a status code 302 and redirects to the auth servlet if the server is using FORM authentication
                serverAuthMethod = FORM_AUTH;
            }
            else {
                serverAuthMethod = BASIC_AUTH;
            }
            
            _tracer.trace("determineAuthMethod - authMethod: %s", serverAuthMethod); //$NON-NLS-1$
            return serverAuthMethod;
        } finally {
            formAuth.releaseConnection();
        }
        } finally {
            _tracer.leave("determineAuthMethod"); //$NON-NLS-1$
        }
    }

    /**
     * Set the userid / password for the referenced httpClient session.
     * 
     * @param httpClient Must not be <code>null</code>.  Must be initialized. 
     * @param userid The userid. May be <code>null</code>.
     * @param password The password. May be <code>null</code>.
     */
    static public void setCredentials(HttpClient httpClient, AuthScope authScope, 
            String userid, String password)
    {
        _tracer.enter("setCredentials"); //$NON-NLS-1$
        try {
	        httpClient.getState().clearCredentials();
	
	        //RLS - Bug 8144
	        httpClient.getState().clearCookies();          
	
	    	if ((null == userid) || (null == password)) {
	            HttpClientParams httpClientParams = httpClient.getParams();
	            httpClientParams.setAuthenticationPreemptive(false);
	        }
	        else {
	            Credentials defaultcreds = new UsernamePasswordCredentials(userid, password);
	            httpClient.getState().setCredentials(authScope, defaultcreds);
	        }
        } finally {
        	_tracer.leave("setCredentials"); //$NON-NLS-1$
        }
    }

    /**
     * Returns whether or not the referenced credentials are defined
     * (non-null and non-empty).
     * 
     * @param storedCredentials The credentials. May be <code>null</code>.
     * 
     * @return <code>true</code> if the user/password is non-null/non-empty,
     * otherwise <code>false</code>.
     */
    static public boolean credDefined(UsernamePasswordCredentials storedCredentials) {
        _tracer.enter("credDefined"); //$NON-NLS-1$
        try {
	        String storedUsername = null;
	        String storedPassword = null;
	        if (storedCredentials != null) {
	            storedUsername = storedCredentials.getUserName();
	            storedPassword = storedCredentials.getPassword();
	        }
	        
	        boolean retval = (storedUsername != null && storedUsername.length() != 0 &&
	                storedPassword != null && storedPassword.length() != 0);
	        
	        _tracer.trace("credDefined: %s", retval); //$NON-NLS-1$
	        return retval;
        } finally {
            _tracer.leave("credDefined"); //$NON-NLS-1$
        }
    }

}
