/*******************************************************************************
 * Licensed Materials - Property of IBM
 * © Copyright IBM Corporation 2014, 2015. All Rights Reserved.
 *
 * Note to U.S. Government Users Restricted Rights:  Use,
 * duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.repository.transport.client;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.http.Header;
import org.apache.http.HttpRequest;
import org.apache.http.auth.AUTH;
import org.apache.http.auth.Credentials;
import org.apache.http.impl.auth.SPNegoScheme;
import org.apache.http.message.BufferedHeader;
import org.apache.http.protocol.HttpContext;
import org.apache.http.util.CharArrayBuffer;
import org.ietf.jgss.GSSContext;
import org.ietf.jgss.GSSException;
import org.ietf.jgss.GSSManager;
import org.ietf.jgss.GSSName;
import org.ietf.jgss.Oid;

import com.ibm.team.repository.common.transport.HttpUtil.CharsetEncoding;

/**
 * This SPNegoScheme adapts the one provided by Apache to work around a bug similar to what has been reported here:
 * http://www-01.ibm.com/support/docview.wss?uid=swg1IZ54545. The method generateGSSToken is identical to the one in the super class except for the constant
 * GSSContext.DEFAULT_LIFETIME which is replaced with GSSContext.INDEFINITE_LIFETIME. 
 * <p>
 * This issue has been reported to the IBM JRE team and is known to them.
 */
public class AdaptedSPNegoScheme extends SPNegoScheme {

    private final Log log = LogFactory.getLog(getClass());
    private final static String NEGOTIATE = ": Negotiate "; //$NON-NLS-1$
    
    /**
     * 
     */
    public AdaptedSPNegoScheme() {
        super();
    }

    /**
     * @param stripPort
     */
    public AdaptedSPNegoScheme(boolean stripPort) {
        super(stripPort);
    }

    protected byte[] generateGSSToken(final byte[] input, final Oid oid, final String authServer) throws GSSException {
        byte[] token = input;
        if (token == null) {
            token = new byte[0];
        }
        final GSSManager manager = getManager();
        final GSSName serverName = manager.createName("HTTP@" + authServer, GSSName.NT_HOSTBASED_SERVICE); //$NON-NLS-1$
        final GSSContext gssContext = manager.createContext(
                serverName.canonicalize(oid), oid, null, GSSContext.INDEFINITE_LIFETIME);
        gssContext.requestMutualAuth(true);
        gssContext.requestCredDeleg(true);
        return gssContext.initSecContext(token, 0, token.length);
    }
    
    /**
     * Because the Apache library uses the platform encoding to generate the token string, we run into problems. We need to fetch the auth header,
     * decode it with the platform string, and re-encode it with UTF-8. The concrete issue was reported in item 349426 for z/OS which runs
     * by default on Cp1047
     */
    @Override
    public Header authenticate(
            final Credentials credentials,
            final HttpRequest request,
            final HttpContext context) throws org.apache.http.auth.AuthenticationException {
        Header oldHeader = super.authenticate(credentials, request, context);
        String oldHeaderStr = oldHeader.toString();

        String startOfAuthenticationHeader;
        if (isProxy()) {
            startOfAuthenticationHeader = AUTH.PROXY_AUTH_RESP + NEGOTIATE;
        } else {
            startOfAuthenticationHeader = AUTH.WWW_AUTH_RESP + NEGOTIATE;
        }
        
        // we need to parse out the current string
        String oldTokenStr = oldHeaderStr.substring(startOfAuthenticationHeader.length()).trim();
        byte[] bytesInPlatform = oldTokenStr.getBytes();
        
        String newTokenstr = new String(bytesInPlatform, CharsetEncoding.UTF8.toCharset());
     
        if (log.isDebugEnabled()) {
            log.debug("Corrected response to use '" + newTokenstr + "'"); //$NON-NLS-1$ //$NON-NLS-2$
        }
        
        final CharArrayBuffer newBuffer = new CharArrayBuffer(32);
        newBuffer.append(startOfAuthenticationHeader);
        newBuffer.append(newTokenstr);
        return new BufferedHeader(newBuffer);
    
    }
}
