/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.repository.transport.client;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import org.apache.commons.httpclient.HttpMethod;
import org.apache.commons.httpclient.methods.RequestEntity;
import org.eclipse.core.runtime.OperationCanceledException;

/**
 * A {@link RequestEntity} instance with a variable sized copy buffer.
 */
public class BufferedRequestEntity implements RequestEntity {

    /** Content type to report to {@link #getContentType()} */
    private final String contentType;
    
    /** Content type to report to {@link #getContentLength()} */
    private final int contentSize;
    
    /** Size of the buffer to when copying */
    private final int bufferSize;
    
    /** The stream to write. */
    private final InputStream stream;
    
    /** The parent method */
    private final HttpMethod method;
    
    /**
     * True if we read from our input stream already
     */
    private boolean readStream;

    /**
     * Create a new request entity. 
     * 
     * @param stream The stream to write on {@link #writeRequest(OutputStream)}.
     *      May not be null.
     *      
     * @param size The size to return on {@link #getContentLength()}.  
     * 
     * @param contentType The text to return on {@link #getContentType()}.
     * @param bufferSize The size of buffer to use when reading from 
     *      <code>stream</code> in {@link #writeRequest(OutputStream)}. Must 
     *      be a positive integer.
     */
    public BufferedRequestEntity(InputStream stream, int size, 
            String contentType, int bufferSize, HttpMethod method) {
        
        // Validate
        if (stream == null) {
            throw new IllegalArgumentException("Stream may not be null"); //$NON-NLS-1$
        }
        
        if (bufferSize <= 0) {
            throw new IllegalArgumentException("bufferSize must be greater than zero"); //$NON-NLS-1$
        }
        
        // Store
        this.stream = stream;
        this.contentSize = size;
        this.contentType = contentType;
        this.bufferSize = bufferSize;
        this.method = method;
    }

    
    public void writeRequest(OutputStream out) throws IOException {
    	readStream = true;
    	
        byte[] buffer = new byte[bufferSize];

        int i = 0;
        
        try {
	        while ((i = stream.read(buffer)) >= 0) {
	            out.write(buffer, 0, i);
	        }
        } catch (OperationCanceledException e) {
        	method.abort();
        	throw e;
        }
    }

    
    public long getContentLength() {
        return contentSize;
    }

    
    public String getContentType() {
        return contentType;
    }


    public boolean isRepeatable() {
        return false;
    }
    
    public boolean canRepeat() {
    	return !readStream;
    }

}
