/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.repository.transport.client;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import org.apache.commons.httpclient.HttpMethod;
import org.apache.commons.httpclient.methods.RequestEntity;
import org.eclipse.core.runtime.OperationCanceledException;

import com.ibm.team.repository.common.util.NLS;

/**
 * A {@link RequestEntity} that will cache the contents of the stream 
 * so that small entities can be retried.
 */
class CachingRetryableRequestEntity implements RequestEntity {

    private final int contentSize;
	private final String contentType;
	private final byte[] buffer;

	/**
     * Create a new request entity. 
     * 
     * @param stream The stream to write on {@link #writeRequest(OutputStream)}.
     *      May not be null.
     *      
     * @param size The size to return on {@link #getContentLength()}.  Must not be negative
     * 
     * @param contentType The text to return on {@link #getContentType()}.
	 * @throws IOException 
     */
    public CachingRetryableRequestEntity(InputStream stream, int size, String contentType, HttpMethod method) throws IOException {
        
        // Validate
        if (stream == null) {
            throw new IllegalArgumentException("Stream may not be null"); //$NON-NLS-1$
        }
        
        if (size <= 0) {
            throw new IllegalArgumentException("size must be greater than zero"); //$NON-NLS-1$
        }
        
        // Store
        this.contentSize = size;
        this.contentType = contentType;
        this.buffer = initBuffer(stream, size, method);
    }
    
	private static byte[] initBuffer(InputStream stream, int size, HttpMethod method) throws IOException {
		byte[] buffer = new byte[size];
		
		int count, pos=0;
        try {
	        while ((count = stream.read(buffer, pos, buffer.length - pos)) > 0) {
	        	pos += count;
	        	
	        	if (pos > buffer.length) {
	        		throw new IllegalStateException(NLS.bind("The content size specified for the input stream was {0} but more bytes than that were in the input stream", size)); //$NON-NLS-1$
	        	}
	        }
        } catch (OperationCanceledException e) {
        	method.abort();
        	throw e;
        }
		
		return buffer;
	}

	public long getContentLength() {
		return contentSize;
	}

	public String getContentType() {
		return contentType;
	}

	public boolean isRepeatable() {
		return true;
	}

	public void writeRequest(OutputStream output) throws IOException {
		output.write(buffer);
	}

}
