/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010, 2013. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.repository.transport.client;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.cert.Certificate;
import java.security.cert.CertificateException;
import java.util.Enumeration;

import com.ibm.team.repository.common.UUID;

/**
 * Specialized storage, based on {@link KeyStore}, for managing trusted
 * certificates.
 */
public class CertificateStore {

    private final KeyStore keyStore;

    /**
     * Helper to convert a KeyStore password from a string to a character array,
     * allowing null.
     * 
     * @param s
     * @return
     */
    private static char[] string2chars(String s) {
        return (s == null) ? null : s.toCharArray();
    }

    /**
     * Get the underlying KeyStore.
     * 
     * @return
     */
    KeyStore getKeyStore() {
        return keyStore;
    }

    /**
     * Helper to create a {@link KeyStore} by reading it from a file.
     * 
     * @param file
     * @param password
     * @return
     * @throws NoSuchAlgorithmException
     * @throws CertificateException
     * @throws IOException
     * @throws KeyStoreException
     */
    static KeyStore load(File file, String password)
            throws NoSuchAlgorithmException, CertificateException, IOException,
            KeyStoreException {
        KeyStore ks = KeyStore.getInstance(KeyStore.getDefaultType());
        try {
			InputStream input = new FileInputStream(file);
			try {
				ks.load(input, string2chars(password));
			} finally {
				input.close();
			}
        } catch (IOException e) {
            // Return an empty initialized KeyStore
            ks.load(null, null);
        }
        return ks;
    }

    /**
     * Helper to write a {@link KeyStore} to a file.
     * 
     * @param ks
     * @param file
     * @param password
     * @throws KeyStoreException
     * @throws NoSuchAlgorithmException
     * @throws CertificateException
     * @throws IOException
     */
    static void store(KeyStore ks, File file, String password)
            throws KeyStoreException, NoSuchAlgorithmException,
            CertificateException, IOException {
        OutputStream output = new FileOutputStream(file);
        try {
        	ks.store(output, string2chars(password));
        } finally {
        	output.close();
        }
    }

    /**
     * Helper to compute a unique alias within the trust store for a specified
     * certificate.
     * 
     * @param cert
     * @return
     */
    static String computeAlias(Certificate cert) {
        // There appears to be no standard way to construct certificate aliases,
        // but this class never depends on looking up a certificate by its
        // computed alias, so just create an alias that's unique and be done.
        return UUID.generate().getUuidValue();
    }

    /**
     * Construct a new TrustStore initially containing no certificates.
     * 
     * @throws KeyStoreException
     * @throws NoSuchAlgorithmException
     * @throws CertificateException
     * @throws IOException
     */
    public CertificateStore() throws KeyStoreException,
            NoSuchAlgorithmException, CertificateException, IOException {
    	try {
            keyStore = KeyStore.getInstance(KeyStore.getDefaultType());    		
    	} catch (KeyStoreException e) {
    		// If the code above caused a KeyStoreException, then the JVM classpath is probably messed up.
    		throw new RuntimeException("KeyStoreException: ["+e.getLocalizedMessage()+"]. "  //$NON-NLS-1$//$NON-NLS-2$
    			+ "Likely cause is that the Java Cryptography Extension libraries are missing from the JRE classpath.  " //$NON-NLS-1$
    			+ "Make sure %JAVA_HOME%/lib/ext is specified in your JVM's java.ext.dirs system property.");//$NON-NLS-1$ 
    	}
        keyStore.load(null, null);
    }

    /**
     * Construct a new TrustStore initially containing certificates loaded from
     * the specified {@link KeyStore} file. Only certificates, and not any other
     * entries, are loaded.
     * 
     * @param file
     * @param password
     * @throws NoSuchAlgorithmException
     * @throws CertificateException
     * @throws KeyStoreException
     * @throws IOException
     */
    public CertificateStore(File file, String password)
            throws NoSuchAlgorithmException, CertificateException,
            KeyStoreException, IOException {
        this();
        loadCertificates(file, password);
    }

    /**
     * Does the trust store contain the specified certificate?
     * 
     * @param cert
     * @return
     * @throws KeyStoreException
     */
    public boolean containsCertificate(Certificate cert)
            throws KeyStoreException {
        return (keyStore.getCertificateAlias(cert) != null);
    }

    /**
     * Enter the specified certificate into the trust store.
     * 
     * @param cert
     * @throws KeyStoreException
     */
    public void enterCertificate(Certificate cert) throws KeyStoreException {
        if (!containsCertificate(cert)) {
            keyStore.setCertificateEntry(computeAlias(cert), cert);
        }
    }

    /**
     * Helper to copy all the certificate entries, and none of the other
     * entries, from a {@link KeyStore} into the trust store.
     * 
     * @param ks
     * @throws KeyStoreException
     */
    private void enterCertificates(KeyStore ks) throws KeyStoreException {
        Enumeration<String> aliases = ks.aliases();
        while (aliases.hasMoreElements()) {
            String alias = aliases.nextElement();
            if (ks.isCertificateEntry(alias)) {
                Certificate cert = ks.getCertificate(alias);
                enterCertificate(cert);
            }
        }
    }

    /**
     * Copy all of the certificates from the specified trust store into the
     * trust store.
     * 
     * @param ts
     * @throws KeyStoreException
     */
    public void copyCertificates(CertificateStore ts) throws KeyStoreException {
        enterCertificates(ts.keyStore);
    }

    /**
     * Load the specified {@link KeyStore} file and copy all of the certificates
     * it contains into the trust store. Only certificates, and not any other
     * entries, are loaded.
     * 
     * @param file
     * @param password
     * @throws NoSuchAlgorithmException
     * @throws CertificateException
     * @throws KeyStoreException
     * @throws IOException
     */
    public void loadCertificates(File file, String password)
            throws NoSuchAlgorithmException, CertificateException,
            KeyStoreException, IOException {
        KeyStore ks = load(file, password);
        enterCertificates(ks);
    }
}
