/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/

package com.ibm.team.repository.transport.client;

import org.eclipse.core.runtime.IStatus;

/**
 * Handler for acquiring an authenticated and authorized request token and
 * secret as a result of an unauthenticated proxied request attempt.
 * 
 * @DoImplement Clients implement this interface to provide their own
 * oauth authentication handler.
 * @since 0.7
 */
public interface IOAuthHandler {

    /**
     * The user has received a response requesting an OAuth authorization
     * for cross server communication from the origin server to a friend
     * server.
     * 
     * This method will perform the authentication with the friend server,
     * and may be required to also authorize the friend server to allow
     * oauth requests from the origin server on behalf of the logged in
     * user. The implementer should present dialogs as needed to allow entry
     * of user login and password, or to grant authorization to the friend
     * server.
     * <p>
     * This method may be called from a background thread. The implementer
     * of the method must ensure its thread-safety and avoid displaying the
     * message to the user multiple times if the method gets called
     * simultaneously by multiple threads. This can be achieved by
     * serializing the calls through the use of a lock. In that case special
     * care must be taken to avoid deadlocks; i.e., the
     * <code>challenge</code> method should avoid calling other code that
     * can acquire a lock of any sort. Nesting the acquisition of locks
     * increases the potential for deadlocks.
     * </p>
     * 
     * @param authorizeURL
     *            The redirect URL received from the response from the
     *            origin server
     * @return Response The response of the method. If the authorization was denied
     * or cancelled, it is acceptable to return <code>null</code>, otherwise
     * return a Response created with the appropriate status and callback URI.
     */
    IResponse authorize(String authorizeURL);

    /**
     * The response returned by {@link IOAuthHandler#authorize(String).
     */
    public interface IResponse {
        
        /**
         * Returns the callback URI that was referenced in {@link IOAuthHandler#authorize(String)}.
         * 
         * @return The callback URI.  Can be <code>null</code> if the authorization
         * was cancelled or denied.
         */
        public String getCallbackURL();
        
        /**
         * Returns the status of this response.
         * 
         * @return The status of this response. Can be <code>null</code> if the
         * authorization was cancelled or denied. 
         */
        public IStatus getStatus();
    }

}
