/*******************************************************************************
 * Licensed Materials - Property of IBM
 * © Copyright IBM Corporation 2010, 2014. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/

package com.ibm.team.repository.transport.client;

import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.Map;

import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.common.transport.TeamServiceException;

/**
 * The client-side interface to a remote REST service.  
 * Allows for invocation of http calls to arbitrary services.
 * @since 0.7
 */
public interface ITeamRawRestServiceClient {

    /**
     * Returns a server communications connection for the URI specified
     * by <code>relUri</code>. 
     * 
     * @param uri The URI to connect to. Must not be <code>null</code>. May be
     * relative or absolute.  If relative, the authority of the repository
     * to which this client bound is used.
     * 
     * @return The client connection. Never <code>null</code>.
     * 
     * @throws URISyntaxException Thrown if <code>uri</code> is relative and
     * an invalid syntax such that it cannot be constructed in the context of
     * this connection's repository connection. 
     */
    IRawRestClientConnection getConnection(URI uri) throws URISyntaxException;
    
    /**
     * A client-side connection to a server REST service.
     * 
     * Usage:<p> 
     * Clients get access to a connection via 
     * {@link ITeamRawRestServiceClient#getConnection(URI)}, possibly adding
     * additional request headers via {@link IRawRestClientConnection#addRequestHeader(String, String)},
     * and then calling one of the server contact methods prefixed by "do*" (e.g.
     * {@link IRawRestClientConnection#doGet()}).<p>
     * 
     * Once a "do*" method has been invoked, no other "do*" method may be 
     * invoked on the connection 
     */
    public interface IRawRestClientConnection {
        
        /**
         * A Response from one of this connections "do*" methods.
         */
        public static class Response {
            
            public enum StatusCode {
                OK,
                CREATED,
                NOT_FOUND,
                NOT_MODIFIED,
                ERROR
            }
            
            /**
             * Constructs a response with the specified arguments.
             * 
             * @param responseStream The input stream from which to read
             * a server response. May be <code>null</code>.
             * @param headers The HTTP headers of the response. Must not be <code>null</code>.
             * @param statusCode The status code of the response.  Expected to
             * be an HTTP status code, as documented by RFC2616.
             * @param statusText The status text or "reason phrase" of the response.
             * May be <code>null</code>.
             */
            public Response(InputStream responseStream, Map<String, String> headers, int statusCode,
                    String statusText)
            {
                _responseStream = responseStream;
                _headers = headers;
                _statusCode = statusCode;
                _statusText = statusText;
            }
            
            /**
             * Releases any resources/connections associated to this response. Equivalent to calling
             * "close" on the {@link InputStream} returned by {@link #getResponseStream()}.
             */
            public void close() {
                if (_responseStream != null) {
                    try {
                        _responseStream.close();
                    } catch (IOException e) {
                        // Ignored
                    }
                }
            }
            
            /**
             * Returns the response stream for this response.
             * 
             * @return The response stream. May be <code>null</code>.
             */
            public InputStream getResponseStream() {
                return _responseStream;
            }
            
            /**
             * Returns a header value from this response.
             * 
             * @param header
             *            The name of the header to return; must not be
             *            <code>null</code>.
             * @return The header value, or <code>null</code> if unset in
             *         response.
             */
            public String getResponseHeader(String header) {
                return _headers.get(header);
            }

            /**
             * Returns all of the headers associated to this response.
             * 
             * @return The headers for this response.  May be <code>null</code> if
             * the 
             */
            public Map<String, String> getAllResponseHeaders() {
                return _headers;
            }
            
            /**
             * Returns the status code for this response.  The possible values are in
             * the range of HTTP status codes as documented by RFC2616.
             * 
             * @return The status code.
             */
            public int getStatusCode() {
                return _statusCode;
            }
            
            /**
             * Returns the http status text ("reason phrase") for this response.
             * @return The HTTP status text for this response.  Will be <code>null</code>
             * if the status text cannot be determined.
             */
            public String getStatusText() {
                return _statusText;
            }
            
            private final InputStream _responseStream;
            private final Map<String, String> _headers;
            private final int _statusCode;
            private String _statusText;
        }
        
        /**
         * Adds a request header to the connection.  This method may not be called
         * on a connection on which a "do*" method has already been invoked.
         * 
         * @param name The header name.  May be <code>null</code>.  If <code>null</code>,
         * the request is ignored.
         * 
         * @param value The header value.  May be <code>null</code>.
         */
        public void addRequestHeader(String name, String value);
        
        /**
         * Returns an input stream representing the response from the URI
         * used to construct this connection, when an HTTP GET is
         * requested on the referenced location.
         * 
         * This method, or any other do* method, cannot be called twice.<p>
         * 
         * By default, this method requests server responses of type
         * "text/xml" from the server - this can be overridden by adding a
         * request header ({@link #addRequestHeader(String, String)}) of type
         * "accept", assuming the service that this connection is bound to
         * accepts accept headers.
         *
         * @return The response.  Never <code>null</code>.
         * 
         * @throws TeamServiceException Thrown on non-recoverable server I/O error.
         * @throws TeamRepositoryException Thrown on recoverable repository errors.
         */
        public Response doGet() throws TeamServiceException, TeamRepositoryException;

        /**
         * Returns an input stream representing the response from the URI
         * used to construct this connection, when an HTTP POST is
         * requested on the referenced location.
         * 
         * This method, or any other do* method, cannot be called twice.<p>
         * 
         * By default, this method requests server responses of type
         * "text/xml" from the server - this can be overridden by adding a
         * request header ({@link #addRequestHeader(String, String)}) of type
         * "accept", assuming the service that this connection is bound to
         * accepts accept headers.
         *
         * @param content
         *            input stream to read request body from
         * @param contentLength
         *            length of the stream supplied
         * @param contentType
         *            mime type
         * 
         * @return The response.  Never <code>null</code>.
         * 
         * @throws TeamServiceException Thrown on non-recoverable server I/O error.
         * @throws TeamRepositoryException Thrown on recoverable repository errors.
         */
        public Response doPost(InputStream content, long contentLength,
                String contentType) throws TeamServiceException, TeamRepositoryException;

        /**
         * Returns an input stream representing the response from the URI
         * used to construct this connection, when an HTTP PUT is
         * requested on the referenced location.
         * 
         * This method, or any other do* method, cannot be called twice.<p>
         * 
         * By default, this method requests server responses of type
         * "text/xml" from the server - this can be overridden by adding a
         * request header ({@link #addRequestHeader(String, String)}) of type
         * "accept", assuming the service that this connection is bound to
         * accepts accept headers.
         *
         * @param content
         *            input stream to read request body from
         * @param contentLength
         *            length of the stream supplied
         * @param contentType
         *            mime type
         * 
         * @return The response.  Never <code>null</code>.
         * 
         * @throws TeamServiceException Thrown on non-recoverable server I/O error.
         * @throws TeamRepositoryException Thrown on recoverable repository errors.
         */
        public Response doPut(InputStream content, long contentLength,
                String contentType) throws TeamServiceException, TeamRepositoryException;

        /**
         * Returns an input stream representing the response from the URI
         * used to construct this connection, when an HTTP DELETE is
         * requested on the referenced location.
         * 
         * This method, or any other do* method, cannot be called twice.<p>
         * 
         * By default, this method requests server responses of type
         * "text/xml" from the server - this can be overridden by adding a
         * request header ({@link #addRequestHeader(String, String)}) of type
         * "accept", assuming the service that this connection is bound to
         * accepts accept headers.
         *
         * @return The response.  Never <code>null</code>.
         * 
         * @throws TeamServiceException Thrown on non-recoverable server I/O error.
         * @throws TeamRepositoryException Thrown on recoverable repository errors.
         */
        public Response doDelete() throws TeamServiceException, TeamRepositoryException;

        /**
         * Returns an input stream representing the response from the URI
         * used to construct this connection, when an HTTP HEAD is
         * requested on the referenced location.
         * 
         * This method, or any other do* method, cannot be called twice.<p>
         * 
         * By default, this method requests server responses of type
         * "text/xml" from the server - this can be overridden by adding a
         * request header ({@link #addRequestHeader(String, String)}) of type
         * "accept", assuming the service that this connection is bound to
         * accepts accept headers.
         *
         * @return The response.  Never <code>null</code>.
         * 
         * @throws TeamServiceException Thrown on non-recoverable server I/O error.
         * @throws TeamRepositoryException Thrown on recoverable repository errors.
         */
        public Response doHead() throws TeamServiceException, TeamRepositoryException;
        
        /**
         * Returns the response associated to this connection.
         * 
         * The response returned will the same response returned by one of the
         * do* methods when that method completes successfully (without exception).  
         * However, when one of those methods fails due to an uncaught exception such
         * that the returned response object cannot be in the stack frame, 
         * this response captures the details of the transport-level error such as
         * any errors returned by the server, the http status code and reason phrase, and the 
         * associated response headers.
         * 
         * @return The response for this connection.  Will be <code>null</code> if
         * a do* method has not yet been invoked, or if an exception is thrown from
         * one of the do* methods that is not appropriate to mapping to a Response.
         */
        public Response getResponse();

        /**
         * Releases the connection created by a call to {@link #doGet()}.  It has
         * no effect if any of the "do*" methods have not been called.
         * 
         * To prevent leaking connections and sockets, you must either call
         * this method or {@link Response#close()} to release any resources
         * owned by this connection.
         */
        public void release();
        
        /**
         * This method returns the uri being invoked RESTfully
         * 
         * @return the uri string
         */
        public String getServiceUri();                
    }
}
