/*******************************************************************************
 * Licensed Materials - Property of IBM
 * © Copyright IBM Corporation 2010, 2014. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/

package com.ibm.team.repository.transport.client;

import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.util.Map;

import com.ibm.team.repository.common.IItem;
import com.ibm.team.repository.common.Location;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.common.transport.ITeamService;
import com.ibm.team.repository.common.transport.TeamServiceException;

/**
 * The client-side interface to a remote REST service.  This implementation
 * is returned by {@link ITeamService#getImplementation()} for REST services.
 */
public interface ITeamRestServiceClient {

    /**
     * Returns a server communications connection for the location specified
     * by <code>loc</code>.  Note that regardless of what server or service
     * is specified in <code>loc</code>, the server that this service client is bound to
     * is the server and service that this client connection is bound to.
     * 
     * @param loc The location to connect to.  May not be <code>null</code>.
     * May be relative or absolute - the server and service contained in the
     * location are not referenced.
     *  
     * @return The client connection. Never <code>null</code>.
     */
    IRestClientConnection getConnection(Location loc);
    
    /**
     * Returns a server communications connection for the relative URI specified
     * by <code>relUri</code>. Note that regardless of what is specified in
     * <code>relUri</code>, the server that this service client is bound to
     * is the server and service that this client connection is bound to.
     * 
     * @param relUri
     *            The relative URI to connect to. May not be <code>null</code>.
     * 
     * @return The client connection. Never <code>null</code>.
     */
    IRestClientConnection getConnection(URI relUri);
    
    /**
     * A client-side connection to a server REST service.
     * 
     * Usage:<p> 
     * Clients get access to a connection via 
     * {@link ITeamRestServiceClient#getConnection(Location)}, possibly add
     * additional request headers via {@link IRestClientConnection#addRequestHeader(String, String)},
     * and then call one of the server contact methods prefixed by "do*" (e.g.
     * {@link IRestClientConnection#doGet()}).<p>
     * 
     * After a "do*" method has been invoked the response can be unmarshaled
     * via one of the unmarshaling methods (the choice of unmarshaling methods
     * depends upon the nature of the service handling the call, and any accept
     * headers that may have been specified).<p>
     * 
     * Once a "do*" method has been invoked, no other "do*" method may be 
     * invoked on the connection - the connection should be 
     * {@link IRestClientConnection#release() released} and the connection 
     * disposed of.
     */
    public interface IRestClientConnection {
        
        /**
         * A Response from one of this connections "do*" methods.
         */
        public static class Response {
            
            public enum StatusCode {
                OK,
                CREATED,
                NOT_FOUND,
                NOT_MODIFIED,
                ERROR
            }
            
            /**
             * Constructs a response with the specified arguments.
             * 
             * @param responseStream The input stream from which to read
             * a server response. May be <code>null</code>.
             * @param headers The response headers associated to the response.
             * May be <code>null</code>.
             * @param httpStatusCode The HTTP status code of the response. Set to
             * -1 if unavailable.
             * @param httpStatusText The HTTP status text of the response. May
             * be <code>null</code>.
             * @param statusCode The status code of the response.  Expected to
             * be one of {@link StatusCode} values.
             */
            public Response(InputStream responseStream, Map<String, String> headers, 
                    int httpStatusCode, String httpStatusText, StatusCode statusCode) {
                _responseStream = responseStream;
                _headers = headers;
                _statusCode = statusCode;
                _httpStatusCode = httpStatusCode;
                _httpStatusText = httpStatusText;
            }
            
            /**
             * Returns the response stream for this response.
             * 
             * @return The response stream. May be <code>null</code>.
             */
            public InputStream getResponseStream() {
                return _responseStream;
            }
            
            /**
             * Releases any resources/connections associated to this response. Equivalent to calling
             * "close" on the {@link InputStream} returned by {@link #getResponseStream()}.
             */
            public void close() {
                if (_responseStream != null) {
                    try {
                        _responseStream.close();
                    } catch (IOException e) {
                        // Ignored
                    }
                }
            }
            
            /**
             * Returns a header value from this response.
             * 
             * @param header
             *            The name of the header to return; must not be
             *            <code>null</code>.
             * @return The header value, or <code>null</code> if unset in
             *         response.
             */
            public String getResponseHeader(String header) {
                return _headers.get(header);
            }

            /**
             * Returns all of the headers associated to this response.
             * 
             * @return The headers for this response.  May be <code>null</code> if
             * the 
             */
            public Map<String, String> getAllResponseHeaders() {
                return _headers;
            }
            
            /**
             * Returns the status code for this response.
             * 
             * @return The status code.
             */
            public StatusCode getStatusCode() {
                return _statusCode;
            }
            
            /**
             * Returns the http status code for this response.
             * 
             * @return The HTTP status code for this response.  Will be -1
             * if the status cannot be determined.
             */
            public int getHttpStatusCode() {
                return _httpStatusCode;
            }
            
            /**
             * Returns the http status text ("reason phrase") for this response.
             * @return The HTTP status text for this response.  Will be <code>null</code>
             * if the status text cannot be determined.
             */
            public String getHttpStatusText() {
                return _httpStatusText;
            }
            
            private final InputStream _responseStream;
            private final Map<String, String> _headers;
            private final StatusCode _statusCode;
            private final int _httpStatusCode;
            private final String _httpStatusText;
        }
        
        /**
         * Adds a request header to the connection.  This method may not be called
         * on a connection on which a "do*" method has already been invoked.
         * 
         * @param name The header name.  May be <code>null</code>.  If <code>null</code>,
         * the request is ignored.
         * 
         * @param value The header value.  May be <code>null</code>.
         */
        public void addRequestHeader(String name, String value);
        
        /**
         * Returns an input stream representing the response from the service
         * specified in {@link #boundAbsoluteUri()}, when an HTTP GET is
         * requested on the referenced location.
         * 
         * This method, or any other do* method, cannot be called twice.<p>
         * 
         * By default, this method requests server responses of type
         * "text/xml" from the server - this can be overridden by adding a
         * request header ({@link #addRequestHeader(String, String)}) of type
         * "accept", assuming the service that this connection is bound to
         * accepts accept headers.
         *
         * @return The response.  Never <code>null</code>.
         * 
         * @throws TeamServiceException Thrown on non-recoverable server I/O error.
         * @throws TeamRepositoryException Thrown on recoverable repository errors.
         */
        public Response doGet() throws TeamServiceException, TeamRepositoryException;

        /**
         * Returns an input stream representing the response from the service
         * specified in {@link #boundAbsoluteUri()}, when an HTTP OPTIONS is
         * requested on the referenced location.
         * 
         * This method, or any other do* method, cannot be called twice.<p>
         * 
         * By default, this method requests server responses of type
         * "text/xml" from the server - this can be overridden by adding a
         * request header ({@link #addRequestHeader(String, String)}) of type
         * "accept", assuming the service that this connection is bound to
         * accepts accept headers.
         *
         * @return The response.  Never <code>null</code>.
         * 
         * @throws TeamServiceException Thrown on non-recoverable server I/O error.
         * @throws TeamRepositoryException Thrown on recoverable repository errors.
         */
        public Response doOptions() throws TeamServiceException, TeamRepositoryException;

        /**
         * Returns an input stream representing the response from the service
         * specified in {@link #boundAbsoluteUri()}, when an HTTP POST is
         * requested on the referenced location.
         * 
         * This method, or any other do* method, cannot be called twice.<p>
         * 
         * By default, this method requests server responses of type
         * "text/xml" from the server - this can be overridden by adding a
         * request header ({@link #addRequestHeader(String, String)}) of type
         * "accept", assuming the service that this connection is bound to
         * accepts accept headers.
         *
         * @param content
         *            input stream to read request body from
         * @param contentLength
         *            length of the stream supplied
         * @param contentType
         *            mime type
         * 
         * @return The response.  Never <code>null</code>.
         * 
         * @throws TeamServiceException Thrown on non-recoverable server I/O error.
         * @throws TeamRepositoryException Thrown on recoverable repository errors.
         */
        public Response doPost(InputStream content, long contentLength,
                String contentType) throws TeamServiceException, TeamRepositoryException;

        /**
         * Returns an input stream representing the response from the service
         * specified in {@link #boundAbsoluteUri()}, when an HTTP PUT is
         * requested on the referenced location.
         * 
         * This method, or any other do* method, cannot be called twice.<p>
         * 
         * By default, this method requests server responses of type
         * "text/xml" from the server - this can be overridden by adding a
         * request header ({@link #addRequestHeader(String, String)}) of type
         * "accept", assuming the service that this connection is bound to
         * accepts accept headers.
         *
         * @param content
         *            input stream to read request body from
         * @param contentLength
         *            length of the stream supplied
         * @param contentType
         *            mime type
         * 
         * @return The response.  Never <code>null</code>.
         * 
         * @throws TeamServiceException Thrown on non-recoverable server I/O error.
         * @throws TeamRepositoryException Thrown on recoverable repository errors.
         */
        public Response doPut(InputStream content, long contentLength,
                String contentType) throws TeamServiceException, TeamRepositoryException;

        /**
         * Returns an input stream representing the response from the service
         * specified in {@link #boundAbsoluteUri()}, when an HTTP DELETE is
         * requested on the referenced location.
         * 
         * This method, or any other do* method, cannot be called twice.<p>
         * 
         * By default, this method requests server responses of type
         * "text/xml" from the server - this can be overridden by adding a
         * request header ({@link #addRequestHeader(String, String)}) of type
         * "accept", assuming the service that this connection is bound to
         * accepts accept headers.
         *
         * @return The response.  Never <code>null</code>.
         * 
         * @throws TeamServiceException Thrown on non-recoverable server I/O error.
         * @throws TeamRepositoryException Thrown on recoverable repository errors.
         */
        public Response doDelete() throws TeamServiceException, TeamRepositoryException;

        /**
         * Unmarshalls an input stream returned by one of the "do*" methods
         * into an {@link IItem} type.
         * 
         * @param response  An input stream returned by any of the "do*" methods.
         * Must not be <code>null</code>.
         * 
         * @return The item represented by the serialization contained in
         * <code>response</code>.
         * 
         * @throws TeamRepositoryException  Thrown on marshalling errors.
         */
        public IItem unmarshall(InputStream response) throws TeamRepositoryException;
        
        /**
         * Releases the connection created by a call to {@link #doGet()}.  It has
         * no effect if any of the "do*" methods have not been called.
         * 
         * To prevent leaking connections and sockets, you must either call
         * this method or {@link Response#close()} to release any resources
         * owned by this connection.
         */
        public void release();
        
        /**
         * Returns the absolute URI that this connection uses to contact the
         * server. Note that this URI may be different than what was passed to
         * {@link ITeamRestServiceClient#getConnection(Location)} or
         * {@link ITeamRestServiceClient#getConnection(URI)}.
         * 
         * @return The absolute URI that this connection uses to contact the
         * server. Never <code>null</code>.
         */
        public URI boundAbsoluteUri() throws TeamRepositoryException;
        
        /**
         * This method returns the service name being invoked,
         * eg, com.ibm.team.repository.common.service.IQueryService
         * 
         * @return the service name
         */
        public String getServiceUri();
        
        /**
         * Returns the response associated to this connection.
         * 
         * The response returned will the same response returned by one of the
         * do* methods when that method completes successfully (without exception).  
         * However, when one of those methods fails due to an uncaught exception such
         * that the returned response object cannot be in the stack frame, 
         * this response captures the details of the transport-level error such as
         * any errors returned by the server, the http status code and status text, and the 
         * associated response headers.
         * 
         * @return The response for this connection.  Will be <code>null</code> if
         * a do* method has not yet been invoked, or if an exception is thrown from
         * one of the do* methods that is not appropriate to mapping to a Response.
         */
        public Response getResponse();
        
    }
}
