/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010, 2014. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/

/**
 * 
 */
package com.ibm.team.repository.transport.client;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.Inet6Address;
import java.net.InetAddress;
import java.net.Socket;
import java.net.SocketAddress;
import java.net.SocketException;
import java.nio.channels.SocketChannel;

import org.apache.commons.httpclient.params.HttpConnectionParams;

import com.ibm.team.repository.common.transport.TeamServiceCallContextImpl;

/**
 * @since 0.7
 */
public class InterruptableSocket extends Socket {

	private Socket        _socket;
	private SocketChannel _socketChannel;
	private int receiveBufferSize = 8192;
	private int sendBufferSize = 8192;

	/**
	 * 
	 */
	static public Socket getSocketForThread(Thread thread) {
		return TeamServiceCallContextImpl.getSocketForThread(thread);
	}

	/**
	 * 
	 */
	static private void setSocketForCurrentThread(Socket socket) {
		TeamServiceCallContextImpl.setSocketForCurrentThread(socket);
	}

	/**
	 * 
	 */
	static public Socket newInterruptableSocket(Socket socket,
			SocketChannel socketChannel, HttpConnectionParams params) {
		return new InterruptableSocket(socket, socketChannel, params);
	}

	/**
	 * 
	 */
	private InterruptableSocket(Socket socket, SocketChannel socketChannel, HttpConnectionParams params) {
		super();

		_socket = socket;
		_socketChannel = socketChannel;
		
		// store the receive and send buffer size. This information will be used
		// if client is connecting to a IPV6 address using IBM JRE 1.5 or 1.6 on
		// Windows XP
		if (params != null) {
			if (params.isParameterSet(HttpConnectionParams.SO_SNDBUF)) {
				sendBufferSize = params.getSendBufferSize();
			}
			if (params.isParameterSet(HttpConnectionParams.SO_RCVBUF)) {
				receiveBufferSize = params.getReceiveBufferSize();
			}
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.net.Socket#close()
	 */
	@Override
	public synchronized void close() throws IOException {
		_socketChannel.close();
		_socket.close();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.net.Socket#connect(java.net.SocketAddress, int)
	 */
	@Override
	public void connect(SocketAddress endpoint, int timeout) throws IOException {
		try {
			setSocketForCurrentThread(this);
			_socket.connect(endpoint, timeout);
		} finally {
			setSocketForCurrentThread(null);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.net.Socket#connect(java.net.SocketAddress)
	 */
	@Override
	public void connect(SocketAddress endpoint) throws IOException {
		try {
			setSocketForCurrentThread(this);
			_socket.connect(endpoint);
		} finally {
			setSocketForCurrentThread(null);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.net.Socket#getInputStream()
	 */
	@Override
	public InputStream getInputStream() throws IOException {
		return new InterruptableSocketInputStream(this, _socket.getInputStream());
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.net.Socket#getOutputStream()
	 */
	@Override
	public OutputStream getOutputStream() throws IOException {
		return new InterruptableSocketOutputStream(this, _socket.getOutputStream());
	}
	
	/**
	 * 
	 */
	public boolean isChannelClosed() {
		return !_socketChannel.isOpen();
	}

	//-------------------------------------------------------------------
	// Delegated methods
	//-------------------------------------------------------------------
	
	/**
	 * @param bindpoint
	 * @throws IOException
	 * @see java.net.Socket#bind(java.net.SocketAddress)
	 */
	public void bind(SocketAddress bindpoint) throws IOException {
		_socket.bind(bindpoint);
	}

	/**
	 * @param obj
	 * @return
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	public boolean equals(Object obj) {
		return _socket.equals(obj);
	}

	/**
	 * @return
	 * @see java.net.Socket#getChannel()
	 */
	public SocketChannel getChannel() {
		return _socket.getChannel();
	}

	/**
	 * @return
	 * @see java.net.Socket#getInetAddress()
	 */
	public InetAddress getInetAddress() {
		return _socket.getInetAddress();
	}

	/**
	 * @return
	 * @throws SocketException
	 * @see java.net.Socket#getKeepAlive()
	 */
	public boolean getKeepAlive() throws SocketException {
		return _socket.getKeepAlive();
	}

	/**
	 * @return
	 * @see java.net.Socket#getLocalAddress()
	 */
	public InetAddress getLocalAddress() {
		return _socket.getLocalAddress();
	}

	/**
	 * @return
	 * @see java.net.Socket#getLocalPort()
	 */
	public int getLocalPort() {
		return _socket.getLocalPort();
	}

	/**
	 * @return
	 * @see java.net.Socket#getLocalSocketAddress()
	 */
	public SocketAddress getLocalSocketAddress() {
		return _socket.getLocalSocketAddress();
	}

	/**
	 * @return
	 * @throws SocketException
	 * @see java.net.Socket#getOOBInline()
	 */
	public boolean getOOBInline() throws SocketException {
		return _socket.getOOBInline();
	}

	/**
	 * @return
	 * @see java.net.Socket#getPort()
	 */
	public int getPort() {
		return _socket.getPort();
	}

	/**
	 * @return
	 * @throws SocketException
	 * @see java.net.Socket#getReceiveBufferSize()
	 */
	public int getReceiveBufferSize() throws SocketException {
		try {
			int receiveBuffSize = _socket.getReceiveBufferSize();
			return receiveBuffSize;
		} catch (SocketException e) {
			InetAddress remoteAddress = _socket.getInetAddress();
			if (remoteAddress instanceof Inet6Address) {
				// Machines trying to connect to a server using IPV6 will throw a 
				// Socket exception on IBM JRE 1.5 and 1.6. Use the receive buffer 
				// size from the HTTP Connection parameter.
				return receiveBufferSize;
			}
			throw e;
		}
	}

	/**
	 * @return
	 * @see java.net.Socket#getRemoteSocketAddress()
	 */
	public SocketAddress getRemoteSocketAddress() {
		return _socket.getRemoteSocketAddress();
	}

	/**
	 * @return
	 * @throws SocketException
	 * @see java.net.Socket#getReuseAddress()
	 */
	public boolean getReuseAddress() throws SocketException {
		return _socket.getReuseAddress();
	}

	/**
	 * @return
	 * @throws SocketException
	 * @see java.net.Socket#getSendBufferSize()
	 */
	public int getSendBufferSize() throws SocketException {
		try {
			int sendBuffSize = _socket.getSendBufferSize();
			return sendBuffSize;
		} catch (SocketException e) {
			InetAddress remoteAddress = _socket.getInetAddress();
			if (remoteAddress instanceof Inet6Address) {
				// Machines trying to connect to a server using IPV6 will throw a 
				// Socket exception on IBM JRE 1.5 and 1.6. Use the send buffer 
				// size from the HTTP Connection parameter.
				return sendBufferSize;
			}
			throw e;
		}
	}

	/**
	 * @return
	 * @throws SocketException
	 * @see java.net.Socket#getSoLinger()
	 */
	public int getSoLinger() throws SocketException {
		return _socket.getSoLinger();
	}

	/**
	 * @return
	 * @throws SocketException
	 * @see java.net.Socket#getSoTimeout()
	 */
	public int getSoTimeout() throws SocketException {
		return _socket.getSoTimeout();
	}

	/**
	 * @return
	 * @throws SocketException
	 * @see java.net.Socket#getTcpNoDelay()
	 */
	public boolean getTcpNoDelay() throws SocketException {
		return _socket.getTcpNoDelay();
	}

	/**
	 * @return
	 * @throws SocketException
	 * @see java.net.Socket#getTrafficClass()
	 */
	public int getTrafficClass() throws SocketException {
		return _socket.getTrafficClass();
	}

	/**
	 * @return
	 * @see java.lang.Object#hashCode()
	 */
	public int hashCode() {
		return _socket.hashCode();
	}

	/**
	 * @return
	 * @see java.net.Socket#isBound()
	 */
	public boolean isBound() {
		return _socket.isBound();
	}

	/**
	 * @return
	 * @see java.net.Socket#isClosed()
	 */
	public boolean isClosed() {
		return _socket.isClosed();
	}

	/**
	 * @return
	 * @see java.net.Socket#isConnected()
	 */
	public boolean isConnected() {
		return _socket.isConnected();
	}

	/**
	 * @return
	 * @see java.net.Socket#isInputShutdown()
	 */
	public boolean isInputShutdown() {
		return _socket.isInputShutdown();
	}

	/**
	 * @return
	 * @see java.net.Socket#isOutputShutdown()
	 */
	public boolean isOutputShutdown() {
		return _socket.isOutputShutdown();
	}

	/**
	 * @param data
	 * @throws IOException
	 * @see java.net.Socket#sendUrgentData(int)
	 */
	public void sendUrgentData(int data) throws IOException {
		_socket.sendUrgentData(data);
	}

	/**
	 * @param on
	 * @throws SocketException
	 * @see java.net.Socket#setKeepAlive(boolean)
	 */
	public void setKeepAlive(boolean on) throws SocketException {
		_socket.setKeepAlive(on);
	}

	/**
	 * @param on
	 * @throws SocketException
	 * @see java.net.Socket#setOOBInline(boolean)
	 */
	public void setOOBInline(boolean on) throws SocketException {
		_socket.setOOBInline(on);
	}

	/**
	 * @param connectionTime
	 * @param latency
	 * @param bandwidth
	 * @see java.net.Socket#setPerformancePreferences(int, int, int)
	 */
	public void setPerformancePreferences(int connectionTime, int latency,
			int bandwidth) {
		_socket.setPerformancePreferences(connectionTime, latency, bandwidth);
	}

	/**
	 * @param size
	 * @throws SocketException
	 * @see java.net.Socket#setReceiveBufferSize(int)
	 */
	public void setReceiveBufferSize(int size) throws SocketException {
		_socket.setReceiveBufferSize(size);
	}

	/**
	 * @param on
	 * @throws SocketException
	 * @see java.net.Socket#setReuseAddress(boolean)
	 */
	public void setReuseAddress(boolean on) throws SocketException {
		_socket.setReuseAddress(on);
	}

	/**
	 * @param size
	 * @throws SocketException
	 * @see java.net.Socket#setSendBufferSize(int)
	 */
	public void setSendBufferSize(int size) throws SocketException {
		_socket.setSendBufferSize(size);
	}

	/**
	 * @param on
	 * @param linger
	 * @throws SocketException
	 * @see java.net.Socket#setSoLinger(boolean, int)
	 */
	public void setSoLinger(boolean on, int linger) throws SocketException {
		_socket.setSoLinger(on, linger);
	}

	/**
	 * @param timeout
	 * @throws SocketException
	 * @see java.net.Socket#setSoTimeout(int)
	 */
	public void setSoTimeout(int timeout) throws SocketException {
		_socket.setSoTimeout(timeout);
	}

	/**
	 * @param on
	 * @throws SocketException
	 * @see java.net.Socket#setTcpNoDelay(boolean)
	 */
	public void setTcpNoDelay(boolean on) throws SocketException {
		_socket.setTcpNoDelay(on);
	}

	/**
	 * @param tc
	 * @throws SocketException
	 * @see java.net.Socket#setTrafficClass(int)
	 */
	public void setTrafficClass(int tc) throws SocketException {
		_socket.setTrafficClass(tc);
	}

	/**
	 * @throws IOException
	 * @see java.net.Socket#shutdownInput()
	 */
	public void shutdownInput() throws IOException {
		_socket.shutdownInput();
	}

	/**
	 * @throws IOException
	 * @see java.net.Socket#shutdownOutput()
	 */
	public void shutdownOutput() throws IOException {
		_socket.shutdownOutput();
	}

	//-------------------------------------------------------------------
	// InputStream implementation
	//-------------------------------------------------------------------
	
	static private class InterruptableSocketInputStream extends InputStream {

		private InputStream         _iStream;
		private InterruptableSocket _socket;

		public InterruptableSocketInputStream(InterruptableSocket socket, InputStream iStream) {
			super();

			_socket  = socket;
			_iStream = iStream;
		}

		@Override
		public int available() throws IOException {
			if (_socket.isChannelClosed()) throw new IOException("socket channel closed"); //$NON-NLS-1$
			
			try {
				setSocketForCurrentThread(_socket);
				return _iStream.available();
			}
			finally {
				setSocketForCurrentThread(null);
			}
		}

		@Override
		public void close() throws IOException {
			try {
				setSocketForCurrentThread(_socket);
				_iStream.close();
			}
			finally {
				setSocketForCurrentThread(null);
			}
		}

		@Override
		public synchronized void mark(int readlimit) {
			if (_socket.isChannelClosed()) return;
			
			try {
				setSocketForCurrentThread(_socket);
				_iStream.mark(readlimit);
			}
			finally {
				setSocketForCurrentThread(null);
			}
		}

		@Override
		public boolean markSupported() {
			try {
				setSocketForCurrentThread(_socket);
				return _iStream.markSupported();
			}
			finally {
				setSocketForCurrentThread(null);
			}
		}

		@Override
		public int read() throws IOException {
			if (_socket.isChannelClosed()) throw new IOException("socket channel closed"); //$NON-NLS-1$
			
			try {
				setSocketForCurrentThread(_socket);
				return _iStream.read();
			}
			finally {
				setSocketForCurrentThread(null);
			}
		}

		@Override
		public int read(byte[] b, int off, int len) throws IOException {
			if (_socket.isChannelClosed()) throw new IOException("socket channel closed"); //$NON-NLS-1$
			
			try {
				setSocketForCurrentThread(_socket);
				return _iStream.read(b, off, len);
			}
			finally {
				setSocketForCurrentThread(null);
			}
		}

		@Override
		public int read(byte[] b) throws IOException {
			if (_socket.isChannelClosed()) throw new IOException("socket channel closed"); //$NON-NLS-1$
			
			try {
				setSocketForCurrentThread(_socket);
				return _iStream.read(b);
			}
			finally {
				setSocketForCurrentThread(null);
			}
		}

		@Override
		public synchronized void reset() throws IOException {
			if (_socket.isChannelClosed()) throw new IOException("socket channel closed"); //$NON-NLS-1$
			
			try {
				setSocketForCurrentThread(_socket);
				_iStream.reset();
			}
			finally {
				setSocketForCurrentThread(null);
			}
		}

		@Override
		public long skip(long n) throws IOException {
			if (_socket.isChannelClosed()) throw new IOException("socket channel closed"); //$NON-NLS-1$
			
			try {
				setSocketForCurrentThread(_socket);
				return _iStream.skip(n);
			}
			finally {
				setSocketForCurrentThread(null);
			}
		}

	}

	//-------------------------------------------------------------------
	// OutputStream implementation
	//-------------------------------------------------------------------
	
	static private class InterruptableSocketOutputStream extends OutputStream {

		private OutputStream        _oStream;
		private InterruptableSocket _socket;

		public InterruptableSocketOutputStream(InterruptableSocket socket, OutputStream oStream) {
			super();

			_socket = socket;
			_oStream = oStream;
		}

		@Override
		public void close() throws IOException {
			try {
				setSocketForCurrentThread(_socket);
				_oStream.close();
			}
			finally {
				setSocketForCurrentThread(null);
			}
		}

		@Override
		public void flush() throws IOException {
			if (_socket.isChannelClosed()) throw new IOException("socket channel closed"); //$NON-NLS-1$
			
			try {
				setSocketForCurrentThread(_socket);
				_oStream.flush();
			}
			finally {
				setSocketForCurrentThread(null);
			}
		}

		@Override
		public void write(byte[] b, int off, int len) throws IOException {
			if (_socket.isChannelClosed()) throw new IOException("socket channel closed"); //$NON-NLS-1$
			
			try {
				setSocketForCurrentThread(_socket);
				_oStream.write(b, off, len);
			}
			finally {
				setSocketForCurrentThread(null);
			}
		}

		@Override
		public void write(byte[] b) throws IOException {
			if (_socket.isChannelClosed()) throw new IOException("socket channel closed"); //$NON-NLS-1$
			
			try {
				setSocketForCurrentThread(_socket);
				_oStream.write(b);
			}
			finally {
				setSocketForCurrentThread(null);
			}
		}

		@Override
		public void write(int b) throws IOException {
			if (_socket.isChannelClosed()) throw new IOException("socket channel closed"); //$NON-NLS-1$
			
			try {
				setSocketForCurrentThread(_socket);
				_oStream.write(b);
			}
			finally {
				setSocketForCurrentThread(null);
			}
		}

	}

}
