/*******************************************************************************
 * Licensed Materials - Property of IBM
 * © Copyright IBM Corporation 2010, 2014. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/

/**
 * 
 */
package com.ibm.team.repository.transport.client;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.Inet6Address;
import java.net.InetAddress;
import java.net.Socket;
import java.net.SocketAddress;
import java.net.SocketException;
import java.nio.channels.SocketChannel;

import org.apache.http.protocol.HttpContext;

import com.ibm.team.repository.common.transport.TeamServiceCallContextImpl;

/**
 * A logical derivation of {@link InterruptableSocket} in support of HttpClient 4.x
 */
public class InterruptableSocket2 extends Socket {

	private Socket _socket;
	private SocketChannel _socketChannel;
    static private final LogUtil _logger = new LogUtil(InterruptableSocket2.class);
	
    /*
     * The send/receive buffer size given when the socket implementation
     * throws an error.
     */	
	private int receiveBufferSize = 8192;
	private int sendBufferSize = 8192;

	/**
     * Returns the socket associated to the specified thread.
     * 
	 * @param thread The thread to query. May be <code>null</code>. 
	 * @return The socket associated to the specified thread. May be <code>null</code>.
	 */
	static public Socket getSocketForThread(Thread thread) {
	    _logger.debug("Entering getSocketForThread"); //$NON-NLS-1$
	    
		return TeamServiceCallContextImpl.getSocketForThread(thread);
	}

	/**
     * Returns an interruptable/closeable socket, effectively wrapping the referenced
     * socket.
     * 
	 * @param socket The socket to wrap.  Must not be <code>null</code>.
	 * @param socketChannel The channel to connect to. Must not be <code>null</code>.
	 * @param context The http context.  Ignored.
	 * @return The wrapped socket. Never <code>null</code>.
	 */
	static public Socket newInterruptableSocket(Socket socket,
			SocketChannel socketChannel, HttpContext context) {
		return new InterruptableSocket2(socket, socketChannel, context);
	}

	/**
	 * Creates an instance of the wrapped socket.
	 */
	private InterruptableSocket2(Socket socket, SocketChannel socketChannel, @SuppressWarnings("unused") HttpContext context) {
		super();
		_logger.debug("Entering InterruptableSocket2"); //$NON-NLS-1$
		
		_socket = socket;
		_socketChannel = socketChannel;
        _logger.debug("Created interruptable socket"); //$NON-NLS-1$
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.net.Socket#close()
	 */
	@Override
	public synchronized void close() throws IOException {
        _logger.debug("Entering close"); //$NON-NLS-1$
	    
		_socketChannel.close();
		_socket.close();
        _logger.debug("Closed interruptable socket"); //$NON-NLS-1$
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.net.Socket#connect(java.net.SocketAddress, int)
	 */
	@Override
	public void connect(SocketAddress endpoint, int timeout) throws IOException {
        _logger.debug("Entering connect"); //$NON-NLS-1$
        
		try {
			setSocketForCurrentThread(this);
			_socket.connect(endpoint, timeout);
            _logger.debug("Connecting interruptable socket to endpoint"); //$NON-NLS-1$
		} finally {
			setSocketForCurrentThread(null);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.net.Socket#connect(java.net.SocketAddress)
	 */
	@Override
	public void connect(SocketAddress endpoint) throws IOException {
        _logger.debug("Entering connect"); //$NON-NLS-1$
        
		try {
			setSocketForCurrentThread(this);
			_socket.connect(endpoint);
            _logger.debug("Connecting interruptable socket to endpoint"); //$NON-NLS-1$
		} finally {
			setSocketForCurrentThread(null);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.net.Socket#getInputStream()
	 */
	@Override
	public InputStream getInputStream() throws IOException {
        _logger.debug("Entering getInputStream"); //$NON-NLS-1$
        
		return new InterruptableSocketInputStream(this, _socket.getInputStream());
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.net.Socket#getOutputStream()
	 */
	@Override
	public OutputStream getOutputStream() throws IOException {
        _logger.debug("Entering getOutputStream"); //$NON-NLS-1$
        
		return new InterruptableSocketOutputStream(this, _socket.getOutputStream());
	}
	
	/**
	 * Returns whether or not the associated socket channel is closed. 
	 */
	public boolean isChannelClosed() {
        _logger.debug("Entering isChannelClosed"); //$NON-NLS-1$
        
		return !_socketChannel.isOpen();
	}

	//-------------------------------------------------------------------
	// Delegated methods
	//-------------------------------------------------------------------
	
	/**
	 * @param bindpoint
	 * @throws IOException
	 * @see java.net.Socket#bind(java.net.SocketAddress)
	 */
	public void bind(SocketAddress bindpoint) throws IOException {
		_socket.bind(bindpoint);
        _logger.debug("Binding interruptable socket to bindpoint"); //$NON-NLS-1$
	}

	/**
	 * @param obj
	 * @return
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	public boolean equals(Object obj) {
		return _socket.equals(obj);
	}

	/**
	 * @return
	 * @see java.net.Socket#getChannel()
	 */
	public SocketChannel getChannel() {
		return _socket.getChannel();
	}

	/**
	 * @return
	 * @see java.net.Socket#getInetAddress()
	 */
	public InetAddress getInetAddress() {
		return _socket.getInetAddress();
	}

	/**
	 * @return
	 * @throws SocketException
	 * @see java.net.Socket#getKeepAlive()
	 */
	public boolean getKeepAlive() throws SocketException {
		return _socket.getKeepAlive();
	}

	/**
	 * @return
	 * @see java.net.Socket#getLocalAddress()
	 */
	public InetAddress getLocalAddress() {
		return _socket.getLocalAddress();
	}

	/**
	 * @return
	 * @see java.net.Socket#getLocalPort()
	 */
	public int getLocalPort() {
		return _socket.getLocalPort();
	}

	/**
	 * @return
	 * @see java.net.Socket#getLocalSocketAddress()
	 */
	public SocketAddress getLocalSocketAddress() {
		return _socket.getLocalSocketAddress();
	}

	/**
	 * @return
	 * @throws SocketException
	 * @see java.net.Socket#getOOBInline()
	 */
	public boolean getOOBInline() throws SocketException {
		return _socket.getOOBInline();
	}

	/**
	 * @return
	 * @see java.net.Socket#getPort()
	 */
	public int getPort() {
		return _socket.getPort();
	}

	/**
	 * @return
	 * @throws SocketException
	 * @see java.net.Socket#getReceiveBufferSize()
	 */
	public int getReceiveBufferSize() throws SocketException {
		try {
			return _socket.getReceiveBufferSize();
		} catch (SocketException e) {
            _logger.debug("Handling SocketException in getReceiveBufferSize"); //$NON-NLS-1$
			InetAddress remoteAddress = _socket.getInetAddress();
			if (remoteAddress instanceof Inet6Address) {
				// Machines trying to connect to a server using IPV6 will throw a 
				// Socket exception on IBM JRE 1.5 and 1.6. Use the receive buffer 
				// size from the HTTP Connection parameter.
				return receiveBufferSize;
			}
			throw e;
		}
	}

	/**
	 * @return
	 * @see java.net.Socket#getRemoteSocketAddress()
	 */
	public SocketAddress getRemoteSocketAddress() {
		return _socket.getRemoteSocketAddress();
	}

	/**
	 * @return
	 * @throws SocketException
	 * @see java.net.Socket#getReuseAddress()
	 */
	public boolean getReuseAddress() throws SocketException {
		return _socket.getReuseAddress();
	}

	/**
	 * @return
	 * @throws SocketException
	 * @see java.net.Socket#getSendBufferSize()
	 */
	public int getSendBufferSize() throws SocketException {
		try {
			return _socket.getSendBufferSize();
		} catch (SocketException e) {
            _logger.debug("Handling SocketException in getSendBufferSize"); //$NON-NLS-1$
			InetAddress remoteAddress = _socket.getInetAddress();
			if (remoteAddress instanceof Inet6Address) {
				// Machines trying to connect to a server using IPV6 will throw a 
				// Socket exception on IBM JRE 1.5 and 1.6. Use the send buffer 
				// size from the HTTP Connection parameter.
				return sendBufferSize;
			}
			throw e;
		}
	}

	/**
	 * @return
	 * @throws SocketException
	 * @see java.net.Socket#getSoLinger()
	 */
	public int getSoLinger() throws SocketException {
		return _socket.getSoLinger();
	}

	/**
	 * @return
	 * @throws SocketException
	 * @see java.net.Socket#getSoTimeout()
	 */
	public int getSoTimeout() throws SocketException {
		return _socket.getSoTimeout();
	}

	/**
	 * @return
	 * @throws SocketException
	 * @see java.net.Socket#getTcpNoDelay()
	 */
	public boolean getTcpNoDelay() throws SocketException {
		return _socket.getTcpNoDelay();
	}

	/**
	 * @return
	 * @throws SocketException
	 * @see java.net.Socket#getTrafficClass()
	 */
	public int getTrafficClass() throws SocketException {
		return _socket.getTrafficClass();
	}

	/**
	 * @return
	 * @see java.lang.Object#hashCode()
	 */
	public int hashCode() {
		return _socket.hashCode();
	}

	/**
	 * @return
	 * @see java.net.Socket#isBound()
	 */
	public boolean isBound() {
		return _socket.isBound();
	}

	/**
	 * @return
	 * @see java.net.Socket#isClosed()
	 */
	public boolean isClosed() {
		return _socket.isClosed();
	}

	/**
	 * @return
	 * @see java.net.Socket#isConnected()
	 */
	public boolean isConnected() {
		return _socket.isConnected();
	}

	/**
	 * @return
	 * @see java.net.Socket#isInputShutdown()
	 */
	public boolean isInputShutdown() {
		return _socket.isInputShutdown();
	}

	/**
	 * @return
	 * @see java.net.Socket#isOutputShutdown()
	 */
	public boolean isOutputShutdown() {
		return _socket.isOutputShutdown();
	}

	/**
	 * @param data
	 * @throws IOException
	 * @see java.net.Socket#sendUrgentData(int)
	 */
	public void sendUrgentData(int data) throws IOException {
		_socket.sendUrgentData(data);
	}

	/**
	 * @param on
	 * @throws SocketException
	 * @see java.net.Socket#setKeepAlive(boolean)
	 */
	public void setKeepAlive(boolean on) throws SocketException {
		_socket.setKeepAlive(on);
	}

	/**
	 * @param on
	 * @throws SocketException
	 * @see java.net.Socket#setOOBInline(boolean)
	 */
	public void setOOBInline(boolean on) throws SocketException {
		_socket.setOOBInline(on);
	}

	/**
	 * @param connectionTime
	 * @param latency
	 * @param bandwidth
	 * @see java.net.Socket#setPerformancePreferences(int, int, int)
	 */
	public void setPerformancePreferences(int connectionTime, int latency,
			int bandwidth) {
		_socket.setPerformancePreferences(connectionTime, latency, bandwidth);
	}

	/**
	 * @param size
	 * @throws SocketException
	 * @see java.net.Socket#setReceiveBufferSize(int)
	 */
	public void setReceiveBufferSize(int size) throws SocketException {
		_socket.setReceiveBufferSize(size);
	}

	/**
	 * @param on
	 * @throws SocketException
	 * @see java.net.Socket#setReuseAddress(boolean)
	 */
	public void setReuseAddress(boolean on) throws SocketException {
		_socket.setReuseAddress(on);
	}

	/**
	 * @param size
	 * @throws SocketException
	 * @see java.net.Socket#setSendBufferSize(int)
	 */
	public void setSendBufferSize(int size) throws SocketException {
		_socket.setSendBufferSize(size);
	}

	/**
	 * @param on
	 * @param linger
	 * @throws SocketException
	 * @see java.net.Socket#setSoLinger(boolean, int)
	 */
	public void setSoLinger(boolean on, int linger) throws SocketException {
		_socket.setSoLinger(on, linger);
	}

	/**
	 * @param timeout
	 * @throws SocketException
	 * @see java.net.Socket#setSoTimeout(int)
	 */
	public void setSoTimeout(int timeout) throws SocketException {
		_socket.setSoTimeout(timeout);
	}

	/**
	 * @param on
	 * @throws SocketException
	 * @see java.net.Socket#setTcpNoDelay(boolean)
	 */
	public void setTcpNoDelay(boolean on) throws SocketException {
		_socket.setTcpNoDelay(on);
	}

	/**
	 * @param tc
	 * @throws SocketException
	 * @see java.net.Socket#setTrafficClass(int)
	 */
	public void setTrafficClass(int tc) throws SocketException {
		_socket.setTrafficClass(tc);
	}

	/**
	 * @throws IOException
	 * @see java.net.Socket#shutdownInput()
	 */
	public void shutdownInput() throws IOException {
		_socket.shutdownInput();
	}

	/**
	 * @throws IOException
	 * @see java.net.Socket#shutdownOutput()
	 */
	public void shutdownOutput() throws IOException {
		_socket.shutdownOutput();
	}

    static private void setSocketForCurrentThread(Socket socket) {
        TeamServiceCallContextImpl.setSocketForCurrentThread(socket);
    }

	//-------------------------------------------------------------------
	// InputStream implementation
	//-------------------------------------------------------------------
	
	static private class InterruptableSocketInputStream extends InputStream {

		private InputStream _iStream;
		private InterruptableSocket2 _socket;

		public InterruptableSocketInputStream(InterruptableSocket2 socket, InputStream iStream) {
			super();

			_socket  = socket;
			_iStream = iStream;
		}

		@Override
		public int available() throws IOException {
			if (_socket.isChannelClosed()) throw new IOException("socket channel closed"); //$NON-NLS-1$
			
			try {
				setSocketForCurrentThread(_socket);
				return _iStream.available();
			}
			finally {
				setSocketForCurrentThread(null);
			}
		}

		@Override
		public void close() throws IOException {
			try {
				setSocketForCurrentThread(_socket);
				_iStream.close();
			}
			finally {
				setSocketForCurrentThread(null);
			}
		}

		@Override
		public synchronized void mark(int readlimit) {
			if (_socket.isChannelClosed()) return;
			
			try {
				setSocketForCurrentThread(_socket);
				_iStream.mark(readlimit);
			}
			finally {
				setSocketForCurrentThread(null);
			}
		}

		@Override
		public boolean markSupported() {
			try {
				setSocketForCurrentThread(_socket);
				return _iStream.markSupported();
			}
			finally {
				setSocketForCurrentThread(null);
			}
		}

		@Override
		public int read() throws IOException {
			if (_socket.isChannelClosed()) throw new IOException("socket channel closed"); //$NON-NLS-1$
			
			try {
				setSocketForCurrentThread(_socket);
				return _iStream.read();
			}
			finally {
				setSocketForCurrentThread(null);
			}
		}

		@Override
		public int read(byte[] b, int off, int len) throws IOException {
			if (_socket.isChannelClosed()) throw new IOException("socket channel closed"); //$NON-NLS-1$
			
			try {
				setSocketForCurrentThread(_socket);
				return _iStream.read(b, off, len);
			}
			finally {
				setSocketForCurrentThread(null);
			}
		}

		@Override
		public int read(byte[] b) throws IOException {
			if (_socket.isChannelClosed()) throw new IOException("socket channel closed"); //$NON-NLS-1$
			
			try {
				setSocketForCurrentThread(_socket);
				return _iStream.read(b);
			}
			finally {
				setSocketForCurrentThread(null);
			}
		}

		@Override
		public synchronized void reset() throws IOException {
			if (_socket.isChannelClosed()) throw new IOException("socket channel closed"); //$NON-NLS-1$
			
			try {
				setSocketForCurrentThread(_socket);
				_iStream.reset();
			}
			finally {
				setSocketForCurrentThread(null);
			}
		}

		@Override
		public long skip(long n) throws IOException {
			if (_socket.isChannelClosed()) throw new IOException("socket channel closed"); //$NON-NLS-1$
			
			try {
				setSocketForCurrentThread(_socket);
				return _iStream.skip(n);
			}
			finally {
				setSocketForCurrentThread(null);
			}
		}

	}

	//-------------------------------------------------------------------
	// OutputStream implementation
	//-------------------------------------------------------------------
	
	static private class InterruptableSocketOutputStream extends OutputStream {

		private OutputStream        _oStream;
		private InterruptableSocket2 _socket;

		public InterruptableSocketOutputStream(InterruptableSocket2 socket, OutputStream oStream) {
			super();

			_socket = socket;
			_oStream = oStream;
		}

		@Override
		public void close() throws IOException {
			try {
				setSocketForCurrentThread(_socket);
				_oStream.close();
			}
			finally {
				setSocketForCurrentThread(null);
			}
		}

		@Override
		public void flush() throws IOException {
			if (_socket.isChannelClosed()) throw new IOException("socket channel closed"); //$NON-NLS-1$
			
			try {
				setSocketForCurrentThread(_socket);
				_oStream.flush();
			}
			finally {
				setSocketForCurrentThread(null);
			}
		}

		@Override
		public void write(byte[] b, int off, int len) throws IOException {
			if (_socket.isChannelClosed()) throw new IOException("socket channel closed"); //$NON-NLS-1$
			
			try {
				setSocketForCurrentThread(_socket);
				_oStream.write(b, off, len);
			}
			finally {
				setSocketForCurrentThread(null);
			}
		}

		@Override
		public void write(byte[] b) throws IOException {
			if (_socket.isChannelClosed()) throw new IOException("socket channel closed"); //$NON-NLS-1$
			
			try {
				setSocketForCurrentThread(_socket);
				_oStream.write(b);
			}
			finally {
				setSocketForCurrentThread(null);
			}
		}

		@Override
		public void write(int b) throws IOException {
			if (_socket.isChannelClosed()) throw new IOException("socket channel closed"); //$NON-NLS-1$
			
			try {
				setSocketForCurrentThread(_socket);
				_oStream.write(b);
			}
			finally {
				setSocketForCurrentThread(null);
			}
		}

	}

}
