/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.repository.transport.client;

import java.io.IOException;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.net.UnknownHostException;
import java.nio.channels.SocketChannel;

import org.apache.commons.httpclient.ConnectTimeoutException;
import org.apache.commons.httpclient.params.HttpConnectionParams;
import org.apache.commons.httpclient.protocol.ProtocolSocketFactory;

import com.ibm.team.repository.common.transport.TeamServiceCallContextImpl;

/**
 * @since 0.5
 */
public class InterruptableSocketFactory implements ProtocolSocketFactory {
	 
	public Socket createSocket(String host, int port) throws IOException,
			UnknownHostException {
		SocketChannel channel = SocketChannel.open();
		return doConnect(channel.socket(), channel, host, port);
	}

	public Socket createSocket(String host, int port, InetAddress localAddress, int localPort)
			throws IOException, UnknownHostException {
		SocketChannel channel = SocketChannel.open();
		Socket socket = channel.socket();
		bindAddress(socket, localAddress, localPort);

		return doConnect(socket, channel, host, port);
	}

	public Socket createSocket(String host, int port, InetAddress localAddress,
			int localPort, HttpConnectionParams params) throws IOException,
			UnknownHostException, ConnectTimeoutException {
		SocketChannel channel = SocketChannel.open();
		Socket socket = channel.socket();
		
        boolean socketException = true;
		try {
    		if (params.isParameterSet(HttpConnectionParams.TCP_NODELAY)) 
    			socket.setTcpNoDelay(params.getTcpNoDelay());
    		if (params.isParameterSet(HttpConnectionParams.SO_SNDBUF)) 
    			socket.setSendBufferSize(params.getSendBufferSize());
    		if (params.isParameterSet(HttpConnectionParams.SO_RCVBUF))
    			socket.setReceiveBufferSize(params.getReceiveBufferSize());
    		if (params.isParameterSet(HttpConnectionParams.SO_TIMEOUT))
    			socket.setSoTimeout(params.getSoTimeout());
    		
    		socketException = false;
		}
		finally {
            if (socketException && socket.getChannel() != null)
                socket.getChannel().close();
		}
		bindAddress(socket, localAddress, localPort);
		
		return doConnect(socket, channel, host, port, params);
	}
	
    /**
     * Returns an {@link InterruptableSocket} as a result of connecting the specified
     * <code>channel</code> to the referenced <code>remoteHost</code> and <code>remotePort</code>.<p/>
     * 
     * IMPORTANT NOTE: Any errors encountered while trying to establish a connection
     * results in the closing of the referenced <code>channel</code>.
     * 
     * @param socket The socket wrapped by the returned {@link InterruptableSocket}. Must
     * not be <code>null</code>. The socket is assumed to be associated to the specified
     * <code>channel</code>.
     * 
     * @param channel The channel to connect to. Must not be <code>null</code>. If any
     * errors are encountered while trying to establish a connection the channel is closed
     * {@link SocketChannel#close()}.
     * 
     * @param remoteHost The host to connect the channel to. Must not be <code>null</code>.
     * 
     * @param remotePort The port on the referenced host to connect to. Must not be <code>null</code>.
     * 
     * @return An {@link InterruptableSocket}. Never <code>null</code>.
     * 
     * @throws IOException Thrown on any errors encountered while trying to establish
     * a connection.  Prior to throwing the exception the referenced <code>channel</code>
     * is closed.
     * 
     * @since 0.7
     */
	protected Socket doConnect(Socket socket, SocketChannel channel, String remoteHost, int remotePort) throws IOException {
		return doConnect(socket, channel, remoteHost, remotePort, null);
	}
	
    /**
     * Returns an {@link InterruptableSocket} as a result of connecting the referenced
     * <code>channel</code> to the referenced <code>remoteHost</code> and <code>remotePort</code>.
     * 
     * IMPORTANT NOTE: Any errors encountered while trying to establish a connection
     * results in the closing of the referenced <code>channel</code>.
     * 
     * @param socket The socket wrapped by the returned {@link InterruptableSocket}. Must
     * not be <code>null</code>. The socket is assumed to be associated to the specified
     * <code>channel</code>.
     * 
     * @param channel The channel to connect to. Must not be <code>null</code>. If any
     * errors are encountered while trying to establish a connection the channel is closed
     * {@link SocketChannel#close()}.
     * 
     * @param remoteHost The host to connect the channel to. Must not be <code>null</code>.
     * 
     * @param remotePort The port on the referenced host to connect to. Must not be <code>null</code>.
     * 
     * @param params Additional http connection parameters to associate to the returned
     * socket. May be <code>null</code>.
     * 
     * @return An {@link InterruptableSocket}. Never <code>null</code>.
     * 
     * @throws IOException Thrown on any errors encountered while trying to establish
     * a connection.  Prior to throwing the exception the referenced <code>channel</code>
     * is closed.
     * 
     * @since 0.7
     */
	protected Socket doConnect(Socket socket, SocketChannel channel, String remoteHost, 
			int remotePort, HttpConnectionParams params) throws IOException {
		
		channel.configureBlocking(true);
		Thread currentThread = Thread.currentThread();
		TeamServiceCallContextImpl.registerConnectingSocket(channel, currentThread);
		boolean connectionFailed = true;
		try {
			InetSocketAddress remoteSocketAddress = new InetSocketAddress(remoteHost, remotePort);
			channel.connect(remoteSocketAddress);
			connectionFailed = false;
			return InterruptableSocket.newInterruptableSocket(socket, channel, params);
		} finally {
			TeamServiceCallContextImpl.unregisterConnectingSocket(currentThread);			
            if (connectionFailed)
                channel.close();
		}
	}
	
    /**
     * Binds the referenced socket to the specified address/port, if a <code>localAddress</code>
     * or <code>localPort</code> is specified.  If <code>localAddress</code> is specified
     * as <code>null</code> and <code>localPort</code> is specified as '0', no bind
     * takes place.
     * 
     * @param socket The socket to bind. Must not be <code>null</code>.
     * 
     * @param localAddress The local address to bind to.  Specify <code>null</code> to 
     * assign the wildcard address.
     * 
     * @param localPort The local port to bind to. Legal values must be specified
     * in the range of 0 to 65535.
     * 
     * @throws IOException Thrown on bind errors.
     * @throws IllegalArgumentException Thrown if the value specified for
     * <code>localPort</code> is out of range.
     */
    static private void bindAddress(Socket socket, InetAddress localAddress, int localPort) throws IOException {
        
        if (localAddress == null && localPort == 0)
            return;
        
        boolean bindFailed = true;
        try {
            socket.bind(new InetSocketAddress(localAddress, localPort));
            bindFailed = false;
        }
        finally {
            if (bindFailed && socket.getChannel() != null)
                socket.getChannel().close();
        }
    }

}