/*******************************************************************************
 * Licensed Materials - Property of IBM © Copyright IBM Corporation 2012, 2014. All Rights Reserved.
 * All Rights Reserved.
 *
 * Note to U.S. Government Users Restricted Rights: Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 ******************************************************************************/

package com.ibm.team.repository.transport.client;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * Utility class to facilitate logging.
 */
public class LogUtil {

    private final Log _logger;
    private final Class<?> _class;

    private static final int _indent;

    static {
        int indentProp = Integer.parseInt(System.getProperty("jazz.trace.indentLevel", "2")); //$NON-NLS-1$ //$NON-NLS-2$
        if (indentProp < 0) {
            indentProp = 2;
        }
        _indent = indentProp;
    }

    /**
     * Constructs an instance of this class.
     *
     * @param clazz The class to which this logger responds.
     */
    public LogUtil(Class<?> clazz) {
        if (clazz == null)
            throw new IllegalArgumentException("class must not be null"); //$NON-NLS-1$

        _class = clazz;
        _logger = LogFactory.getLog(_class);
    }

    /**
     * Returns the logger associated to this instance.
     * @return This instance's logger. Never <code>null</code>.
     */
    public Log getLog() {
        return _logger;
    }

    /**
     * Prints trace-level messages to the logger, prefixed by the simple name
     * of the class associated to this logger. The argument list is evaluated
     * only if "trace" logging is enabled.
     *
     * @param message A format string in per "String.format(...)". May be
     * <code>null</code>.
     *
     * @param args Optional args to be inserted in the format string. May
     * be <code>null</code>.
     */
    public void trace(String message, Object... args) {
        if (message == null || !_logger.isTraceEnabled())
            return;

        String msg = LogUtil.indentMessage(message);

        if (args == null || args.length == 0) {
            _logger.trace(msg);
        } else {
            try {
                _logger.trace(String.format(msg, args));
            }
            catch (Throwable ex) {
                _logger.trace(String.format("The arguments passed to the trace method with the msg \"%s\" caused the exception \"%s\": %s", msg, ex.getClass().getSimpleName(), ex.getMessage())); //$NON-NLS-1$
            }
        }
    }

    /**
     * Prints debug-level messages to the logger. The argument list is evaluated
     * only if "debug" logging is enabled.
     *
     * @param message A format string in per "String.format(...)". May be
     * <code>null</code>.
     *
     * @param args Optional args to be inserted in the format string. May
     * be <code>null</code>.
     */
    public void debug(String message, Object... args) {
        if (message == null || !_logger.isDebugEnabled())
            return;

        String msg = LogUtil.indentMessage(message);

        if (args == null || args.length == 0) {
            _logger.debug(msg);
        } else {
            try {
                _logger.debug(String.format(msg, args));
            }
            catch (Throwable ex) {
                _logger.debug(String.format("The arguments passed to the debug method with the msg \"%s\" caused the exception \"%s\": %s", msg, ex.getClass().getSimpleName(), ex.getMessage())); //$NON-NLS-1$
            }
        }
    }

    /**
     * Returns whether or not debug has been enabled for this logger.
     * 
     * @return <code>true</code> if debug is enabled/request, <code>false</code> is not.
     */
    public boolean isDebugEnabled() {
        return _logger.isDebugEnabled();
    }

    /**
     * Prints info-level messages to the logger. The argument list is evaluated
     * only if "info" logging is enabled.
     *
     * @param message A format string in per "String.format(...)". May be
     * <code>null</code>.
     *
     * @param args Optional args to be inserted in the format string. May
     * be <code>null</code>.
     */
    public void info(String message, Object... args) {
        if (message == null || !_logger.isInfoEnabled())
            return;

        String msg = LogUtil.indentMessage(message);

        if (args == null || args.length == 0) {
            _logger.info(msg);
        } else {
            try {
                _logger.info(String.format(msg, args));
            }
            catch (Throwable ex) {
                _logger.info(String.format("The arguments passed to the info method with the msg \"%s\" caused the exception \"%s\": %s", msg, ex.getClass().getSimpleName(), ex.getMessage())); //$NON-NLS-1$
            }
        }
    }

    /**
     * Prints warning-level messages to the logger. The argument list is evaluated
     * only if "warn" logging is enabled.
     *
     * @param message A format string in per "String.format(...)". May be
     * <code>null</code>.
     *
     * @param args Optional args to be inserted in the format string. May
     * be <code>null</code>.
     */
    public void warn(String message, Object... args) {
        if (message == null || !_logger.isWarnEnabled())
            return;

        String msg = LogUtil.indentMessage(message);

        if (args == null || args.length == 0) {
            _logger.warn(msg);
        } else {
            try {
                _logger.warn(String.format(msg, args));
            }
            catch (Throwable ex) {
                _logger.warn(String.format("The arguments passed to the warn method with the msg \"%s\" caused the exception \"%s\": %s", message, ex.getClass().getSimpleName(), ex.getMessage())); //$NON-NLS-1$
            }
        }
    }

    /**
     * Prints error-level messages to the logger. The argument list is evaluated
     * only if "error" logging is enabled.
     *
     * @param message A format string in per "String.format(...)". May be
     * <code>null</code>.
     *
     * @param args Optional args to be inserted in the format string. May be
     * <code>null</code>.
     */
    public void error(String message, Object... args) {
        if (message == null || !_logger.isErrorEnabled())
            return;

        String msg = LogUtil.indentMessage(message);

        if (args == null || args.length == 0) {
            _logger.error(msg);
        } else {
            try {
                _logger.error(String.format(msg, args));
            }
            catch (Throwable ex) {
                _logger.error(String.format("The arguments passed to the error method with the msg \"%s\" caused the exception \"%s\": %s", message, ex.getClass().getSimpleName(), ex.getMessage())); //$NON-NLS-1$
            }
        }
    }

    /**
     * Prints error-level messages to the logger. The argument list is evaluated
     * only if "fatal" logging is enabled.
     *
     * @param message A format string in per "String.format(...)". May be
     * <code>null</code>.
     *
     * @param args Optional args to be inserted in the format string. May be
     * <code>null</code>.
     */
    public void fatal(String message, Object... args) {
        if (message == null || !_logger.isFatalEnabled())
            return;

        String msg = LogUtil.indentMessage(message);

        if (args == null || args.length == 0) {
            _logger.fatal(msg);
        } else {
            try {
                _logger.fatal(String.format(msg, args));
            }
            catch (Throwable ex) {
                _logger.fatal(String.format("The arguments passed to the fatal method with the msg \"%s\" caused the exception \"%s\": %s", message, ex.getClass().getSimpleName(), ex.getMessage())); //$NON-NLS-1$
            }
        }
    }

    /**
     * Prints trace-level messages to the configured logger prefixed with the term "Entering ".
     * The messages are indented accordingly to the level of recursion in the
     * call stack as influenced by the configured indent level.
     *
     * @param message A format string in per "String.format(...)". May be
     * <code>null</code>.
     *
     * @param args Optional args to be inserted in the format string.
     */
    public void traceEnter(String message, Object... args) {
        if (message == null || !_logger.isTraceEnabled())
            return;

        LogUtil.incrementCallDepth();
        StringBuilder buf = new StringBuilder(127);
        buf.append(LogUtil.getIndentString());
        buf.append("[ENTERING] "); //$NON-NLS-1$
        if (args == null || args.length == 0) {
            buf.append(message);
        } else {
            try {
                buf.append(String.format(message, args));
            }
            catch (Throwable ex) {
                buf.append(String.format("The arguments passed to the trace method with the msg \"%s\" caused the exception \"%s\": %s", message, ex.getClass().getSimpleName(), ex.getMessage())); //$NON-NLS-1$
            }
        }
        _logger.trace(buf.toString());
    }

    /**
     * Prints trace-level messages to the configured logger prefixed with the term "Exiting ".
     * The messages are indented accordingly to the level of recursion in the
     * call stack as influenced by the configured indent level.
     *
     * @param message A format string in per "String.format(...)". May be
     * <code>null</code>.
     *
     * @param args Optional args to be inserted in the format string.
     */
    public void traceLeave(String message, Object... args) {
        if (message == null || !_logger.isTraceEnabled())
            return;

        StringBuilder buf = new StringBuilder(127);
        buf.append(LogUtil.getIndentString());
        buf.append("[EXITING]  "); //$NON-NLS-1$
        if (args == null || args.length == 0) {
            buf.append(message);
        } else {
            try {
                buf.append(String.format(message, args));
            }
            catch (Throwable ex) {
                buf.append(String.format("The arguments passed to the trace method with the msg \"%s\" caused the exception \"%s\": %s", message, ex.getClass().getSimpleName(), ex.getMessage())); //$NON-NLS-1$
            }
        }
        _logger.trace(buf.toString());
        LogUtil.decrementCallDepth();
    }

    static private ThreadLocal<Integer> _callDepth = new ThreadLocal<Integer>();

    /*
     * The initial calldepth value is set to < 0, so that the first
     * calls to logging don't start logging.
     */
    static private Integer getCallDepth() {
        Integer ndx = LogUtil._callDepth.get();
        if (ndx == null) {
            ndx = Integer.valueOf(0 - LogUtil._indent);
            LogUtil._callDepth.set(ndx);
        }
        return ndx;
    }

    private static void setCallDepth(int depth) {
        LogUtil._callDepth.set(Integer.valueOf(depth));
    }

    private static void incrementCallDepth() {
        int ndx = LogUtil.getCallDepth();
        ndx += LogUtil._indent;
        LogUtil.setCallDepth(ndx);
    }

    private static void decrementCallDepth() {
        int ndx = LogUtil.getCallDepth();
        if (ndx < 0)
            return;

        ndx -= LogUtil._indent;
        LogUtil.setCallDepth(ndx);
    }

    private static String indentMessage(String message) {
        if (LogUtil.getCallDepth() < 0)
            return message;

        LogUtil.incrementCallDepth();
        String msg = String.format("%s%s", LogUtil.getIndentString(), message); //$NON-NLS-1$
        LogUtil.decrementCallDepth();

        return msg;
    }

    private static String getIndentString() {
        int ndx = LogUtil.getCallDepth();
        if (ndx <= 0)
            return ""; //$NON-NLS-1$

        char[] chars = new char[ndx];
        for (int i = 0; i < ndx; i++) {
            chars[i] = ' ';
        }

        return new String(chars);
    }

}
