/*******************************************************************************
 * Licensed Materials - Property of IBM
 * © Copyright IBM Corporation 2010, 2014. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/

package com.ibm.team.repository.transport.client;

import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.Locale;
import java.util.Map;
import java.util.StringTokenizer;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.http.Header;
import org.apache.http.HttpStatus;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpRequestBase;
import org.eclipse.core.runtime.IStatus;

import com.ibm.team.repository.common.internal.util.LocationUtil;
import com.ibm.team.repository.common.transport.HttpUtil;
import com.ibm.team.repository.common.transport.IAbortableHttpRequest;
import com.ibm.team.repository.common.transport.TeamServiceCallContextImpl;
import com.ibm.team.repository.common.transport.TeamServiceException;

/**
 * Utilities for dealing with OAuth authorization challenges and protocol.
 */
public class OAuthUtil {

    /**
     * Handles OAuth challenges by invoking then referenced OauthHandler authorization
     * method, and then retrying the call with the appropriate OAuth parameters.<p/>
     * 
     * To be called after an unsuccessful {@link #executeMethod()} returns an
     * "unauthorized" status. 
     * 
     * @param status The status from the previous {@link #executeMethod()} call.
     * 
     * @param remoteTeamServer The remote team server with which we're communicating.
     * Must not be <code>null</code>.
     * 
     * @param method The RequestResponsePair from the previously failed {@link #executeMethod()}
     * call. Must not be <code>null</code>.  Assumed to be reset for reuse.
     * 
     * @param oauthHandler The oauth handler with which to do authentications.
     * May be <code>null</code>.
     * 
     * @return The status of a retry on the referenced <code>httpMethod</code>,
     * after the specified <code>oauthHandler</code> has been contacted.
     * The original status is returned if <code>null</code> is passed as the
     * oauth handler, if the referenced status is not {@link HttpStatus#SC_UNAUTHORIZED}.
     * 
     * @throws MalformedURLException Thrown on error.
     * @throws URIException Thrown on error.
     * @throws TeamServiceException Thrown on error.
     * @throws UnsupportedEncodingException Thrown on error.
     * @throws URISyntaxException Thrown on error.
     */
    public static int handleOAuthChallenge(int status, RemoteTeamServer remoteTeamServer, 
            RequestResponsePair method, IOAuthHandler oauthHandler) throws MalformedURLException, TeamServiceException, UnsupportedEncodingException, URISyntaxException
    {
        return handleOAuthChallenge(status, remoteTeamServer, method, oauthHandler, true);
    }
    
    /**
     * Handles OAuth challenges by invoking then referenced OauthHandler authorization
     * method, and then retrying the call with the appropriate OAuth parameters.<p/>
     * 
     * To be called after an unsuccessful {@link #executeMethod()} returns an
     * "unauthorized" status. 
     * 
     * @param status The status from the previous {@link #executeMethod()} call.
     * 
     * @param remoteTeamServer The remote team server with which we're communicating.
     * Must not be <code>null</code>.
     * 
     * @param method The RequestResponsePair from the previously failed {@link #executeMethod()}
     * call. Must not be <code>null</code>.  Assumed to be reset for reuse.
     * 
     * @param oauthHandler The oauth handler with which to do authentications.
     * May be <code>null</code>.
     * 
     * @param redirectUrl The URL to redirect to after successful authentication.
     * Must not be <code>null</code>.
     * 
     * @param allowRetries Set to <code>true</code> to allow retries on 401
     * responses, <code>false</code> to not retry.
     * 
     * @return The status of a retry on the referenced <code>httpMethod</code>,
     * after the specified <code>oauthHandler</code> has been contacted.
     * The original status is returned if <code>null</code> is passed as the
     * oauth handler, if the referenced status is not {@link HttpStatus#SC_UNAUTHORIZED}.
     * 
     * @throws MalformedURLException Thrown on error.
     * @throws URIException Thrown on error.
     * @throws TeamServiceException Thrown on error.
     * @throws UnsupportedEncodingException Thrown on error.
     * @throws URISyntaxException Thrown on error.
     */
    private static int handleOAuthChallenge(int status, RemoteTeamServer remoteTeamServer, 
            RequestResponsePair method, IOAuthHandler oauthHandler, boolean allowRetries) throws MalformedURLException, TeamServiceException, UnsupportedEncodingException, URISyntaxException
    {
        _logger.debug("Entering handleOAuthChallenge"); //$NON-NLS-1$
        
        IOAuthHandler.IResponse oauthAuthResponse = null;
        String oauthTokens[] = null;
        
        CloseableHttpResponse response = method.getResponse();
        HttpRequestBase request = method.getRequest();
        
        if (status == HttpStatus.SC_UNAUTHORIZED) {
            _logger.debug("Handling SC_UNAUTHORIZED"); //$NON-NLS-1$
            Header oauthUriHeader = response.getFirstHeader(X_OAUTH_AUTHORIZATION_URL);
            if (oauthUriHeader != null && oauthHandler != null) {
                String oauthUri = oauthUriHeader.getValue();
                if (oauthUri == null || oauthUri.length() == 0)
                    throw new IllegalArgumentException(String.format("Missing value for %s header", X_OAUTH_AUTHORIZATION_URL)); //$NON-NLS-1$
                oauthTokens = parseOAuthTokens(oauthUri);
                oauthAuthResponse = oauthHandler.authorize(oauthUri);
            }
        }
        
        /*
         * If the token info hasn't been set, either we're not handling an OAuth 
         * "SC_UNAUTHORIZED" challenge or the OAuth challenge wasn't handled correctly.
         */
        if (oauthAuthResponse == null || oauthAuthResponse.getCallbackURL() == null) {
            _logger.debug("No token info or callback URL - returning"); //$NON-NLS-1$
            return status;
        }
        
        if (oauthAuthResponse.getStatus() != null && !oauthAuthResponse.getStatus().isOK()) {
            _logger.debug("OAuth error status - returning"); //$NON-NLS-1$
            IStatus errorStatus = oauthAuthResponse.getStatus();
            throw new TeamServiceException(errorStatus.getMessage(), errorStatus.getException(), HttpStatus.SC_UNAUTHORIZED);
        }
        
        /*
         * If there was a secret in the parsed tokens above, then rebuild the original
         * URI path and resend.
         */
        String tokenQuery = String.format("%s=%s", AUTHORIZED_REQUEST_TOKEN_PARAM_NAME, //$NON-NLS-1$
                LocationUtil.encode(oauthTokens[0]));
        String secretQuery = null;
        if (oauthTokens[1] != null)
            secretQuery = String.format("%s=%s", REQUEST_TOKEN_SECRET_PARAM_NAME, LocationUtil.encode(oauthTokens[1])); //$NON-NLS-1$
        String providerQuery = null;
        if (oauthTokens[2] != null)
            providerQuery = String.format("%s=%s", OAUTH_PROVIDER, oauthTokens[2]); //$NON-NLS-1$

        /*
         * Build a URI with the new query parameters
         */
        String originalUri = request.getURI().toASCIIString();
        String newUri = stripOAuthParms(originalUri);
        
        Matcher matches = _queryPattern.matcher(newUri);
        matches.matches();
        String prefix = matches.group(1);
        String query = matches.group(2);
        String fragment = matches.group(3);
        if (fragment == null) {
            fragment = ""; //$NON-NLS-1$
        }
        StringBuffer queryBuf = new StringBuffer();
        if (query != null) {
            queryBuf.append(query);
            queryBuf.append('&');
        }            
        else
            queryBuf.append('?');
        
        queryBuf.append(tokenQuery);
        if (secretQuery != null) {
            queryBuf.append('&');
            queryBuf.append(secretQuery);
        }
        if (providerQuery != null) {
            queryBuf.append('&');
            queryBuf.append(providerQuery);
        }

        newUri = prefix + queryBuf.toString() + fragment;
        request.releaseConnection();
        request.setHeader(HttpUtil.CONNECTION, HttpUtil.CLOSE); 
        request.setURI(new URI(newUri));
        
        final HttpRequestBase cancellableMethod = request;
        TeamServiceCallContextImpl.setUpCall(new IAbortableHttpRequest() {
            @Override
            public String getURI() {
                return cancellableMethod.getURI().toString();
            }
            
            @Override
            public void abort() {
                cancellableMethod.abort();
            }
        });
        
        AuthenticationException authEx = null;
        try {
            _logger.debug("Reinvoking server call with oauth information"); //$NON-NLS-1$
            status = remoteTeamServer.executeMethod(null, method, false, true);
        } 
        catch (AuthenticationException e) {
            _logger.debug("Handling AuthenticationException"); //$NON-NLS-1$
            status = HttpStatus.SC_UNAUTHORIZED;
            authEx = e;
            if (allowRetries) {
                try {
                    status = OAuthUtil.handleOAuthChallenge(status, remoteTeamServer, method,
                            oauthHandler, false);
                }
                catch (Exception ex) {
                    int statusCode = -1;
                    if (ex instanceof TeamServiceException)
                        statusCode = ((TeamServiceException)ex).getStatusCode();
                    
                    throw new TeamServiceException(ex.getMessage(), ex, statusCode);
                }
            }
        }
        catch (IOException e) {
            throw new TeamServiceException(e.getMessage(), e);
        }
        finally {
            /*
             * Reset the path on the httpMethod method to what it was
             * before our edits, so no surprises to the caller.
             */
            request.setURI(new URI(originalUri));
            TeamServiceCallContextImpl.tearDownCall();
        }
        
        if (status == HttpStatus.SC_UNAUTHORIZED && authEx != null)
            throw authEx;

        return status;
    }

    /**
     * Returns the oauth token, and (optionally) the token secret and oauth provider
     * encoded URI values, from the referenced URI's query string.
     * 
     * This routine might be parsing a query string that was returned by a
     * 1.0a provider.  If true, then the secret and callback uri will not be
     * available.
     * 
     * @param uri The URI to parse. Must not be <code>null</code>.  May contain
     * a query string containing an encoded oauth callback (with its own
     * query string) and an oauth token. Optionally contains an oauth provider.
     * 
     * @return The oauth tokens in a three element array.  Never <code>null</code>.
     * The oauth token is element '0', never <code>null</code>, the optional secret element '1'
     * (will be <code>null</code> for a response from an OAuth1.0a provider).
     * The third element may contain an encoded URI of the provider that these 
     * tokens are are associated to, encoded.
     * 
     * @throws URISyntaxException 
     * @throws UnsupportedEncodingException 
     */
    static private String[] parseOAuthTokens(String uri) throws URISyntaxException, UnsupportedEncodingException {
        
        String oauthCallbackUri = null;
        String oauthSecret = null;
        String oauthToken = null;
        String oauthProviderUri = null;
        String[] tokens = new String[3];
        
        String queryString = new URI(uri).getRawQuery();
        StringTokenizer queryTokenizer = new StringTokenizer(queryString, "&"); //$NON-NLS-1$
        
        while (queryTokenizer.hasMoreTokens()) {
            StringTokenizer parmTokenizer = new StringTokenizer(queryTokenizer.nextToken(), "="); //$NON-NLS-1$
            if (parmTokenizer.countTokens() == 0)
                continue;

            String name = parmTokenizer.nextToken();
            if (name.toLowerCase(Locale.ROOT).equals(OAUTH_CALLBACK)) {
                oauthCallbackUri = LocationUtil.decode(parmTokenizer.nextToken());
            }
            else if (name.toLowerCase(Locale.ROOT).equals(AUTHORIZED_REQUEST_TOKEN_PARAM_NAME)) {
                oauthToken = LocationUtil.decode(parmTokenizer.nextToken());
            }
        }        

        if (oauthToken == null)
            throw new IllegalArgumentException(
                    String.format("Absence of oauth token in \"%s\" during OAuth parsing and redirection handling", uri)); //$NON-NLS-1$
        
        /*
         * If there's an oauth callback present, parse it for its token secret parameter
         */
        if (oauthCallbackUri != null) {
            queryString = new URI(oauthCallbackUri).getRawQuery();
            if (queryString != null && queryString.trim().length() > 0) {
                queryTokenizer = new StringTokenizer(queryString, "&"); //$NON-NLS-1$
                while (queryTokenizer.hasMoreTokens()) {
                    StringTokenizer parmTokenizer = new StringTokenizer(queryTokenizer.nextToken(), "="); //$NON-NLS-1$
                    if (parmTokenizer.countTokens() == 0)
                        continue;
        
                    String name = parmTokenizer.nextToken();
                    if (name.toLowerCase().equals(REQUEST_TOKEN_SECRET_PARAM_NAME)) {
                        oauthSecret = LocationUtil.decode(parmTokenizer.nextToken());
                    }
                    else if (name.toLowerCase().equals(OAUTH_PROVIDER)) {
                        oauthProviderUri = parmTokenizer.nextToken();
                    }
                }
            }
        }
        
        tokens[0] = oauthToken;
        tokens[1] = oauthSecret;
        tokens[2] = oauthProviderUri;
        return tokens;
    }
    
    
    /**
     * Returns a URI with oauth request and secret parameters removed
     * from its query, if they exist.
     * 
     * @param uri The URI to remove the query parameters from. Must not be
     * <code>null</code>.
     * 
     * @return The URI with the parameters stripped, if they existed.
     * Never <code>null</code>.
     */
    @SuppressWarnings("unchecked")
    static private String stripOAuthParms(String uri) {
    
        Matcher matches = _queryPattern.matcher(uri);
        
        matches.matches();
        String prefix = matches.group(1);
        String query = matches.group(2);
        String fragment = matches.group(3);
        if (fragment == null) {
            fragment = ""; //$NON-NLS-1$
        }
        if (query != null) {
            if (query.length() > 1)
                query = query.substring(1);
            else
                query = null; 
        }
        
        Map<String, String[]> queryParms = LocationUtil.parseQueryParameters(query);
        String queryString = LocationUtil.queryString(queryParms, new String[] {
                OAUTH_PROVIDER,
                AUTHORIZED_REQUEST_TOKEN_PARAM_NAME,
                REQUEST_TOKEN_SECRET_PARAM_NAME });

        StringBuffer buf = new StringBuffer(prefix);
        if (queryString.length() != 0) {
            buf.append('?');
            buf.append(queryString);
        }
        buf.append(fragment);
        
        return buf.toString();
    }

    public static final String OAUTH_CALLBACK_URL = "proxyCallback"; //$NON-NLS-1$
    private static final String X_OAUTH_AUTHORIZATION_URL = "X-jazz-web-oauth-url"; //$NON-NLS-1$
    private static final String OAUTH_CALLBACK="oauth_callback"; //$NON-NLS-1$
    
    /**
     * Name of the query string parameter authorized request token is retrieved from.
     * The JFS Server sets adds this parameter before redirecting the user to the application.
     */
    public final static String AUTHORIZED_REQUEST_TOKEN_PARAM_NAME = "oauth_token"; //$NON-NLS-1$

    /**
     * Cookie name used to store request token secret for JFS consumer applications
     */
    public final static String REQUEST_TOKEN_SECRET_PARAM_NAME = "request_token_secret"; //$NON-NLS-1$
    
    /**
     * Query string parameter whose value describes the originating URI that triggered
     * the OAuth authorization. 
     */
    public final static String OAUTH_PROVIDER = "oauth_provider"; //$NON-NLS-1$
    
    /*
     * Simple URI parser
     */
    static Pattern _queryPattern = Pattern.compile("^(.*?)([?][^#]*)?(#.*)?$"); //$NON-NLS-1$
    static private LogUtil _logger = new LogUtil(OAuthUtil.class);
}
