/*******************************************************************************
 * Licensed Materials - Property of IBM
 * © Copyright IBM Corporation 2010, 2014. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.repository.transport.client;

import java.io.FilterInputStream;
import java.io.IOException;
import java.io.InputStream;

import org.apache.http.HttpEntity;
import org.apache.http.util.EntityUtils;

import com.ibm.team.repository.transport.internal.nls.Messages;

/**
 *  An input stream wrapper for response streams that ensures close()
 *  closes the response stream and releases associated resources.
 *  
 *  The underlying response stream is allowed to be {@code null}. However, if it is {@code null},
 *  any method except for {@link #close()}, {@link #mark(int)}, and {@link #markSupported()} will result in an {@link IOException}.
 */
public class ReleasingAndAbortingInputStream extends FilterInputStream {

    private HttpEntity entity;
    static private LogUtil logger = new LogUtil(ReleasingAndAbortingInputStream.class);
	
    /**
     * If the referenced entity is <code>null</code>, then any method called on this object except for {@link #close()},
     * {@link #mark(int)}, and {@link #markSupported()}
     * will result in an {@link IOException}. It is recommended that callers do not construct this object with a {@code null}
     * response body stream.
     * @param entity The response entity to read from. May be <code>null</code>.
     * @throws IOException Thrown if an input stream cannot be created from the referenced entity.
     */
    public ReleasingAndAbortingInputStream(HttpEntity entity) throws IOException {
        super(entity != null ? entity.getContent() : null);
        logger.debug("Entering ReleasingAndAbortingInputStream"); //$NON-NLS-1$
        
        this.entity = entity;
    }
    
	/**
	 * @throws IOException if an I/O error occurs or the underlying input stream is {@code null}.
	 */
	@Override
	public int read() throws IOException {
        logger.debug("Entering read"); //$NON-NLS-1$
        
		try {
	        assertUnderlyingInputStreamIsNotNull();
			int bytesRead = super.read();
            
			return bytesRead;
		} catch (IOException e) {
		    close();
			throw e;
		}
	}
	
	/**
	 * @throws IOException if an I/O error occurs or the underlying input stream is {@code null}.
	 */
	@Override
	public int read(byte[] b, int off, int len) throws IOException {
        logger.debug("Entering read(b, off, len)"); //$NON-NLS-1$
        
		try {
	        assertUnderlyingInputStreamIsNotNull();
			int bytesRead = super.read(b, off, len);
			
			return bytesRead;
		} catch (IOException e) {
		    close();
			throw e;
		}
	}
	
	/**
	 * @throws IOException if an I/O error occurs or the underlying input stream is {@code null}.
	 */
	@Override
	public int read(byte[] b) throws IOException {
        logger.debug("Entering read(b)"); //$NON-NLS-1$
        
        try {
            assertUnderlyingInputStreamIsNotNull();
			int bytesRead = super.read(b);
			
			return bytesRead;
		} catch (IOException e) {
			close();
			throw e;
		}
	}
	
	/**
	 * @throws IOException if an I/O error occurs or the underlying input stream is {@code null}.
	 */
	@Override
	public long skip(long n) throws IOException {
        logger.debug("Entering skip"); //$NON-NLS-1$
        
	    try {
	        assertUnderlyingInputStreamIsNotNull();
	        return super.skip(n);
	    }
        catch (IOException e) {
            close();
            throw e;
        }
	}
	
	/**
	 * @throws IOException if an I/O error occurs or the underlying input stream is {@code null}.
	 */
	@Override
	public int available() throws IOException {
        logger.debug("Entering available"); //$NON-NLS-1$
        
	    try {
	        assertUnderlyingInputStreamIsNotNull();
	        return super.available();
	    }
	    catch (IOException e) {
            close();
            throw e;
	    }
	}

	/**
	 * If the underlying input stream is not {@code null}, calls {@link InputStream#mark(int)} on
	 * the underlying input stream. Else does nothing.
	 */
	@Override
	public synchronized void mark(int readlimit) {
        logger.debug("Entering mark"); //$NON-NLS-1$
        
		if(super.in != null) {
			super.mark(readlimit);
		}
	}
	
	/**
	 * @throws IOException if an I/O error occurs or the underlying input stream is {@code null}.
	 */
	@Override
	public synchronized void reset() throws IOException {
        logger.debug("Entering reset"); //$NON-NLS-1$
        
		try {
	        assertUnderlyingInputStreamIsNotNull();
		    super.reset();
		}
		catch (IOException e) {
            close();
            throw e;
		}
	}
	
	/**
	 * @return if the underlying input stream is not {@code null} and this stream type supports the mark and reset method; false otherwise.
	 */
	@Override
	public boolean markSupported() {
        logger.debug("Entering markSupported"); //$NON-NLS-1$
        
		if(super.in != null) {
			return super.markSupported();
		}
		return false;
	}
	
	/**
	 * If the underlying input stream is {@code null}, does nothing.
	 * Else, closes this input stream and releases any system resources associated with the stream.
	 */
	@Override
	public void close() throws IOException {
        logger.debug("Entering close"); //$NON-NLS-1$
        
	    EntityUtils.consumeQuietly(entity);
	}

    private void assertUnderlyingInputStreamIsNotNull() throws IOException {
		if(super.in == null) {
			throw new IOException(Messages.getClientString("ReleasingAndAbortingInputStream.NoResponseBody")); //$NON-NLS-1$
		}
	}
}
