/*******************************************************************************
 * Licensed Materials - Property of IBM
 * © Copyright IBM Corporation 2010, 2015. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/
/*
 *
 */
package com.ibm.team.repository.transport.client;

import java.io.BufferedInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URL;
import java.nio.charset.Charset;
import java.sql.Timestamp;
import java.util.List;
import java.util.Locale;
import java.util.zip.GZIPInputStream;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.http.Header;
import org.apache.http.HttpEntity;
import org.apache.http.HttpRequest;
import org.apache.http.HttpResponse;
import org.apache.http.HttpStatus;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.client.methods.HttpPut;
import org.apache.http.client.methods.HttpRequestBase;
import org.apache.http.client.protocol.HttpClientContext;
import org.apache.http.client.utils.DateUtils;
import org.apache.http.client.utils.HttpClientUtils;
import org.apache.http.entity.BufferedHttpEntity;
import org.apache.http.entity.ByteArrayEntity;
import org.apache.http.entity.ContentType;
import org.apache.http.entity.InputStreamEntity;
import org.apache.http.protocol.HTTP;
import org.eclipse.core.runtime.OperationCanceledException;
import org.xmlsoap.schemas.soap.envelope.Envelope;
import org.xmlsoap.schemas.soap.envelope.Fault;

import com.ibm.team.repository.common.ItemNotFoundException;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.common.internal.marshal.MarshalFactory;
import com.ibm.team.repository.common.internal.marshal.Marshaller;
import com.ibm.team.repository.common.internal.marshal.MarshallerType;
import com.ibm.team.repository.common.internal.marshal.MarshallingException;
import com.ibm.team.repository.common.internal.marshal.MarshallingExceptionReasonCode;
import com.ibm.team.repository.common.internal.marshal.impl.WebServicesMarshaller;
import com.ibm.team.repository.common.internal.marshal.util.MarshallerUtil;
import com.ibm.team.repository.common.transport.AbstractTeamService;
import com.ibm.team.repository.common.transport.ComponentConfiguration;
import com.ibm.team.repository.common.transport.ConnectionException;
import com.ibm.team.repository.common.transport.HttpUtil;
import com.ibm.team.repository.common.transport.HttpUtil.CharsetEncoding;
import com.ibm.team.repository.common.transport.HttpUtil.MediaType;
import com.ibm.team.repository.common.transport.IAbortableHttpRequest;
import com.ibm.team.repository.common.transport.IParameterWrapper;
import com.ibm.team.repository.common.transport.ITeamContentService;
import com.ibm.team.repository.common.transport.ITeamModelledRestService;
import com.ibm.team.repository.common.transport.ParameterWrapperHandler;
import com.ibm.team.repository.common.transport.ServiceMethodInvocationError;
import com.ibm.team.repository.common.transport.TeamContent;
import com.ibm.team.repository.common.transport.TeamServiceCallContextImpl;
import com.ibm.team.repository.common.transport.TeamServiceException;
import com.ibm.team.repository.common.transport.internal.services.Response;
import com.ibm.team.repository.common.util.NLS;
import com.ibm.team.repository.transport.internal.nls.Messages;

/**
 * @since 0.5
 */
public class RemoteTeamService extends AbstractTeamService {
    static private final Log logger = LogFactory.getLog(RemoteTeamService.class.getName());
	static private final String UserAgent = RemoteTeamService.class.getName();
	
    // gzip will be used when:
    //    the property value is set to the string "true"
    //    the property is not set at all
    static private final String UseGzipProperty = "com.ibm.team.repository.common.services.UseGzip"; //$NON-NLS-1$

    static private final String ContentUploadCachingMaximum = "com.ibm.team.repository.transport.client.contentUploadCacheMaximum"; //$NON-NLS-1$
    
    static private final boolean UseGzip = getUseGzip(UseGzipProperty);

    static private Method objectEquals;
    static private Method objectHashCode;
    static private Method objectToString;

    private final RemoteTeamServer remoteTeamServer;

    private final Object implementation;
    private final InvocationHandler invocationHandler;

    private final String serviceClassName;

    private final boolean isContentService;
    private final boolean isModelledRestService;
    
    private final String charsetEncoding = CharsetEncoding.UTF8.toString();

	static {
        initialize();
    }

    static private boolean getUseGzip(String property) {
        String value = System.getProperty(property, "true"); //$NON-NLS-1$
        return Boolean.valueOf(value).booleanValue();
    }

    static private void initialize() {
        try {
            objectEquals = Object.class.getMethod("equals", new Class[] { Object.class });  //$NON-NLS-1$
            objectHashCode = Object.class.getMethod("hashCode", new Class[] {}); //$NON-NLS-1$
            objectToString = Object.class.getMethod("toString", new Class[] {}); //$NON-NLS-1$
        } catch (NoSuchMethodException e) {
            throw new RuntimeException(e);
        }
    }
	
    
	/**
	 * @since 0.7
	 */
	public RemoteTeamService(RemoteTeamServer server, Class<?> serviceClass, 
            ClassLoader serviceClassLoader, String version, List<ComponentConfiguration> componentConfigurations) {

        super(serviceClass, serviceClassLoader, version, componentConfigurations);
        
        remoteTeamServer = server;
        invocationHandler = new ServiceInvocationHandler(this);

        implementation = Proxy.newProxyInstance(getServiceClassLoader(), 
                new Class[] { getServiceClass() }, invocationHandler);

        isContentService = implementation instanceof ITeamContentService;
        isModelledRestService = implementation instanceof ITeamModelledRestService;

        serviceClassName = getServiceClass().getName();

        validateURL(server.getRepositoryURL());
	}
	
	public Object getImplementation() {
        return implementation;
    }
    
	public Object invoke(Object proxy, final Method method, Object[] args) throws Throwable {
        if (null == remoteTeamServer) {
            throw new IllegalStateException("ITeamServer is disposed"); //$NON-NLS-1$
        }
        if (method.equals(objectEquals)) {
            return handleEquals(args[0]);
        }
        if (method.equals(objectHashCode)) {
            return handleHashCode();
        }
        if (method.equals(objectToString)) {
            return handleToString();
        }

        // note that REST services go a different route, so we don't need
        // to check for them here

        if (isContentService) {
            if (method.getName().equals("get")) { //$NON-NLS-1$
                return invokeContentGet((String) args[0]);
            } else if (method.getName().equals("put")) { //$NON-NLS-1$
                invokeContentPut((TeamContent) args[0]);
                return null;
            } else {
                throw new RuntimeException(NLS.bind("Method \"{0}\" is not supported because content services only support get and put.", method.getName())); //$NON-NLS-1$
            }
        } else if (isModelledRestService) {
            Object arg = (0 == args.length ? null : args[0]);
            return invokeModelledRestService(method, arg);
        }

        Method testMethod = getMethod(method.getName());
        if (null == testMethod) {
            throw new RuntimeException(NLS.bind("Method \"{0}\" is not supported because it was not found.", method.getName())); //$NON-NLS-1$
        }

        if (!method.toString().equals(testMethod.toString())) {
            throw new RuntimeException(NLS.bind("Method is not supported: {0}", method.getName())); //$NON-NLS-1$
        }

        return executeMethod(method, args);
    }
	
	private String getURL() {
		String repoPath = remoteTeamServer.getRepositoryPath();
        return repoPath + "service/" + serviceClassName; //$NON-NLS-1$
	}
	
    private Object invokeModelledRestService(Method method, Object argument) throws Throwable {

        if ((null != argument) && !(argument instanceof IParameterWrapper)) {
            throw new RuntimeException("parameter to method should be an instance of IParameterWrapper"); //$NON-NLS-1$
        }

        String methodName = method.getName();
        HttpRequestBase httpMethod = null;
        boolean isPost = methodName.startsWith("post"); //$NON-NLS-1$
        
        Class<?> parameterType = (method.getParameterTypes().length > 0) ? method.getParameterTypes()[0] : null;
        String queryString = ParameterWrapperHandler.createQueryString(parameterType, (IParameterWrapper) argument, isPost);
        
        if (methodName.startsWith("get") || methodName.startsWith("put")) { //$NON-NLS-1$ //$NON-NLS-2$
            String uriString = getURL() + HttpUtil.SLASH + lowerFirst(methodName.substring(3));
            // NOTE: because of problems with double encodings, we will not build the query string with the builder. 
            // Instead, we add any possible query String explicitly - see item 336158 
            if (queryString != null && !queryString.trim().isEmpty()) {
                if (queryString.charAt(0) == '?') {
                    uriString += queryString;
                }
                else {
                    uriString += "?" + queryString; //$NON-NLS-1$
                }
            }
            
            if (methodName.startsWith("get")) //$NON-NLS-1$
                httpMethod = new HttpGet(uriString);
            else
                httpMethod = new HttpPut(uriString);
        } 
        else if (methodName.startsWith("post")) { //$NON-NLS-1$
            methodName = lowerFirst(methodName.substring(4));
            httpMethod = new HttpPost(getURL() + HttpUtil.SLASH + methodName);
            isPost = true;
        } 
        else {
            throw new RuntimeException("method should start with 'get', 'put' or 'post'"); //$NON-NLS-1$
        }

        /*
         * The 5.0.2 and earlier versions of this class set a non-standard header given by the value
         * of the org.apache.commons.httpclient.params.HttpMethodParams.USER_AGENT constant.
         * We still that header, for back-compatibility, as well as the standard User-Agent header.
         */
        httpMethod.setHeader(HTTP.USER_AGENT, UserAgent);
        httpMethod.setHeader(HttpUtil.NON_STD_USER_AGENT_HEADER, UserAgent);
        httpMethod.setHeader(HttpUtil.ACCEPT, MediaType.XML.toString());
        httpMethod.setHeader(HttpUtil.ACCEPT_CHARSET, charsetEncoding);
        TransportUtils.setAcceptLanguageHeader(httpMethod);
        addUseridHeader(httpMethod);
        addServiceVersionHeader(httpMethod);

        if (isPost) {
            BufferedHttpEntity requestEntity = new BufferedHttpEntity(new ByteArrayEntity(
                    queryString.getBytes(charsetEncoding), ContentType.APPLICATION_FORM_URLENCODED));
            ((HttpPost)httpMethod).setEntity(requestEntity);
        }

        Throwable throwable = null;

        CloseableHttpResponse response = null;
        try {
        	response = executeCancelableHttpMethod(httpMethod, true, true).getResponse();
        } catch (Throwable e) {
    		if (e instanceof TeamServiceException && ((TeamServiceException) e).getCause() != null && ((TeamServiceException) e).isFromServer()) {
    			throwable = e.getCause();
    		} else {
    			throwable = e;
    		}
        } 
        
        if (null != throwable) {
            if (response != null) {
                HttpClientUtils.closeQuietly(response);
            }
            throw getAppropriateException(throwable, method.getExceptionTypes());
        }
        

        WebServicesMarshaller webServicesMarshaller = (WebServicesMarshaller) 
            MarshalFactory.eINSTANCE.getMarshaller(MarshallerType.WEB_SERVICES_LITERAL);
        Response serviceResponse = null;

        try {
        	InputStream responseStream = utilGetResponseBodyAsStream(response, httpMethod.getRequestLine().getUri());
            Envelope envelope = (Envelope) webServicesMarshaller.demarshalInputStreamToObject(responseStream);

            // envelope is null for cancelled requests
            if (null == envelope) {
                return thrownAwayResult(method.getReturnType());
            }

            serviceResponse = envelope.getBody().getResponse();
            Fault fault = envelope.getBody().getFault();

            if (fault != null) {
                throwable = constructExceptionFromFault(fault);
            } else if (serviceResponse == null) {
                throwable = new TeamServiceException("no SOAP response received"); //$NON-NLS-1$
            }
        } catch (Throwable e) {
            throwable = e;
        } finally {
            if (response != null) {
                HttpClientUtils.closeQuietly(response);
            }
        }

        if (null != throwable) {
            throw getAppropriateException(throwable, method.getExceptionTypes());
        }

        return webServicesMarshaller.getResponseReturnValue(serviceResponse, method.getReturnType());
    }
	
    private InputStream utilGetResponseBodyAsStream(HttpResponse method, String requestUri) throws MarshallingException, IOException {
        
      InputStream responseStream = method.getEntity().getContent();
      if (responseStream == null) {
          String message = NLS.bind(Messages.getServerString("ClientHttpUtil.EmptyResponse"), requestUri); //$NON-NLS-1$
          throw new MarshallingException(MarshallingExceptionReasonCode.CANNOT_DEMARSHAL_LITERAL, message);
      }
      return responseStream;
	}

	private TeamContent invokeContentGet(String uri) throws IOException, TeamRepositoryException {

		HttpGet method = new HttpGet(getURL() + "/" + uri); //$NON-NLS-1$

		method.setHeader(HTTP.USER_AGENT, UserAgent);
		method.setHeader(HttpUtil.ACCEPT_CHARSET, charsetEncoding);
		TransportUtils.setAcceptLanguageHeader(method);
		
		addUseridHeader(method);
        addServiceVersionHeader(method);
		
		boolean releaseConnection = true;
		CloseableHttpResponse response = null;

		try {
		    /*
		     * The content service is a REST service that responds with HTTP status
		     * codes for Jazz-level errors, rather than 200 status codes for Jazz errors
		     * as the Jazz RPC/SOAP messages do.
		     */
		    response = executeCancelableHttpMethod(method, false, true).getResponse();
		    
			int status = response.getStatusLine().getStatusCode();
            if (status == 404) {
                throwNotFoundException(response, method.getRequestLine().getUri());
            }

            Header header;
            
            header = response.getFirstHeader(HttpUtil.CONTENT_LENGTH);
            if (null == header) {
                throw new IOException(NLS.bind(Messages.getServerString("RemoteTeamService.MissingHeader"), HttpUtil.CONTENT_LENGTH)); //$NON-NLS-1$
            }
            long contentLen = Integer.parseInt(header.getValue());
			if (contentLen > Integer.MAX_VALUE) { 
			    throw new IOException(
					NLS.bind(Messages.getServerString("RemoteTeamService.ExceedsMaximumLength"), contentLen)); //$NON-NLS-1$
			} 

            header = response.getFirstHeader(HttpUtil.CONTENT_TYPE);
            if (null == header) {
                throw new IOException(NLS.bind(Messages.getServerString("RemoteTeamService.MissingHeader"), HttpUtil.CONTENT_TYPE)); //$NON-NLS-1$
            }
            String contentType = header.getValue();

			header = response.getFirstHeader(HttpUtil.LAST_MODIFIED);
			if (null == header) { 
			    throw new IOException(NLS.bind(Messages.getServerString("RemoteTeamService.MissingHeader"), HttpUtil.LAST_MODIFIED));  //$NON-NLS-1$
	        }
			Timestamp lastModified = new Timestamp(DateUtils.parseDate(header.getValue()).getTime());

			ReleasingAndAbortingInputStream in = new ReleasingAndAbortingInputStream(response.getEntity());
			TeamContent result = new TeamContent(uri)
			    .setInputStream(in)
			    .setSize((int) contentLen)
			    .setContentType(contentType)
			    .setLastModified(lastModified);

			releaseConnection = false;
			return result;
		}

		catch (IOException e) {
			throw e;
		}

		finally {
			if (releaseConnection && response != null) {
                HttpClientUtils.closeQuietly(response);
			}
		}
	}

    /**
     * Throws an appropriately constructed ItemNotFoundMethod based on the contents
     * of the specified response.  It is assumed that the method was used in
     * a previous request to the server, using REST services, that resulted in a 404
     * (i.e. REST-like status code, and SOAP-like body).
     * 
     * The response and its enclosed entity are closed/consumed.
     * 
     * @param response The server response. Must not be <code>null</code>.
     * 
     * @param requestUri The request URI.  Used in exception messaging.
     * 
     * @throws IOException Thrown on unexpected error.
     * @throws ItemNotFoundException Always thrown.
     */
    private void throwNotFoundException(CloseableHttpResponse response, String requestUri) throws IOException, ItemNotFoundException {
        WebServicesMarshaller webServicesMarshaller = (WebServicesMarshaller) 
        MarshalFactory.eINSTANCE.getMarshaller(MarshallerType.WEB_SERVICES_LITERAL);

        InputStream responseStream = null;

        try {
            if (response.getEntity() != null)
                responseStream = response.getEntity().getContent();
            
            Envelope envelope = null;
            if (responseStream != null) {
            	try {
            		envelope = (Envelope) webServicesMarshaller.demarshalInputStreamToObject(responseStream);
            	} catch (MarshallingException e) {
            		// ignore
            	}
            }
            
            if (envelope == null) {
                throw new ItemNotFoundException(requestUri);
            }
    
            Throwable throwable = null;
            Fault fault = envelope.getBody().getFault();
            if (fault != null)
                throwable = constructExceptionFromFault(fault);
            
            if (throwable == null)
                throw new ItemNotFoundException(requestUri);
            else if (throwable instanceof ItemNotFoundException)
                throw (ItemNotFoundException)throwable;
            else
                throw new ItemNotFoundException(requestUri, throwable);
        }
        finally {
            if (response != null) {
                HttpClientUtils.closeQuietly(response);
            }
        }
    }
    
    private int getCachedContentUploadSize() {
    	return Integer.getInteger(ContentUploadCachingMaximum, 1000000);
    }
    
    private void invokeContentPut(final TeamContent teamContent) throws IOException, TeamRepositoryException {

        HttpPut method = new HttpPut(getURL() + "/" + teamContent.getURI()); //$NON-NLS-1$
        addUseridHeader(method);
        addServiceVersionHeader(method);

        method.setHeader(HTTP.USER_AGENT, UserAgent);
        TransportUtils.setAcceptLanguageHeader(method);

        HttpEntity requestEntity;
        int contentSize = teamContent.getSize();
        
        // Unrecognized content type/encoding values are allowed, so we just use any given value
        // without any validation
        String contentType = teamContent.getContentType();
        if (contentType == null) {
            contentType = ContentType.DEFAULT_TEXT.toString();
        }

		if (contentSize > 0 && contentSize < getCachedContentUploadSize()) {
			// If the content being uploaded is of a known size, and is less than the caching threshold size
			// then use a request entity that caches the content so it can be retried.
		    InputStreamEntity inputStreamEntity = new InputStreamEntity(teamContent.getInputStream(), contentSize);
		    inputStreamEntity.setContentType(contentType);
		    requestEntity = new BufferedHttpEntity(inputStreamEntity);
		} else {
		    InputStreamEntity inputStreamEntity = new InputStreamEntity(new BufferedInputStream(
                teamContent.getInputStream(), 16384), contentSize);
            inputStreamEntity.setContentType(contentType);
			requestEntity = inputStreamEntity;
		}
        
        method.setEntity(requestEntity);

        String lastModified = DateUtils.formatDate(teamContent.getLastModified());
        method.addHeader(HttpUtil.LAST_MODIFIED, lastModified);

        CloseableHttpResponse response = null;        
        try {
            response = executeCancelableHttpMethod(method, true, false).getResponse();
        } catch (IOException e) {
            throw e;
        } finally {
            if (response != null) {
                HttpClientUtils.closeQuietly(response);
            }
        }
    }
    
    /**
	 * Execute the method using the new marshaller
	 * @param method
     *              Method to be executed
	 * @param parameters
     *              Parameters of the method
	 * @return
	 * @throws Throwable
	 */
    @SuppressWarnings("deprecation")
	private Object executeMethod(Method method, Object[] parameters) throws Throwable {

        // Get the method information
        String methodName = method.getName();
        String interfaceName = method.getDeclaringClass().getName();
        Class[] parameterTypes = method.getParameterTypes();

        // Marshal Service Request to Byte Output Stream
        WebServicesMarshaller webServicesMarshaller = (WebServicesMarshaller) 
            MarshalFactory.eINSTANCE.getMarshaller(MarshallerType.WEB_SERVICES_LITERAL);
        ByteArrayOutputStream byteArrayOutputStream = new ByteArrayOutputStream();

        Response serviceResponse = null;
        Throwable throwable = null;

        try {
            webServicesMarshaller.marshalServiceRequest(methodName, interfaceName, 
                    parameterTypes, parameters, WebServicesMarshaller.MARSHAL_AS_XML, byteArrayOutputStream);

            Envelope envelope = invokePost(webServicesMarshaller, 
                    byteArrayOutputStream);

            // envelope is null for canceled requests
            if (null == envelope) {
                return thrownAwayResult(method.getReturnType());
            }

            serviceResponse = envelope.getBody().getResponse();
            Fault fault = envelope.getBody().getFault();

            if (fault != null) {
                throwable = constructExceptionFromFault(fault);
            } else if (serviceResponse == null) {
                throwable = new TeamServiceException(Messages.getServerString("RemoteTeamService.NoResponse")); //$NON-NLS-1$
            } else if (!serviceResponse.getMethod().equals(methodName)) {
                throwable = new TeamServiceException(
                        NLS.bind(Messages.getServerString("RemoteTeamService.UnexpectedResponse"), //$NON-NLS-1$
                                methodName, serviceResponse.getMethod()));
            }
        } catch (Throwable e) {
            throwable = e;
        }

        if (null != throwable) {
            throw getAppropriateException(throwable, method.getExceptionTypes());
        }

        return webServicesMarshaller.getResponseReturnValue(serviceResponse, method.getReturnType());
    }
    
    private Envelope invokePost(WebServicesMarshaller webServicesMarshaller, 
            ByteArrayOutputStream byteArrayOutputStream) throws Throwable {

        BufferedHttpEntity requestEntity = new BufferedHttpEntity(
                new ByteArrayEntity(byteArrayOutputStream.toByteArray(), ContentType.create("text/xml", (Charset)null))); //$NON-NLS-1$
        HttpPost postMethod = new HttpPost(getURL());
        
        postMethod.setHeader(HTTP.USER_AGENT, UserAgent);
        TransportUtils.setAcceptLanguageHeader(postMethod);
        if (postMethod.getFirstHeader(HttpUtil.ACCEPT) == null)
            postMethod.setHeader(HttpUtil.ACCEPT, MediaType.XML.toString());
        if (postMethod.getFirstHeader(HttpUtil.ACCEPT_CHARSET) == null)
            postMethod.setHeader(HttpUtil.ACCEPT_CHARSET, charsetEncoding);
        addUseridHeader(postMethod);
        addMarshallingHeader(postMethod);
        addServiceVersionHeader(postMethod);

        if (UseGzip) {
            postMethod.setHeader(HttpUtil.ACCEPT_ENCODING, HttpUtil.GZIP);
        }

        postMethod.setEntity(requestEntity);

        CloseableHttpResponse response = null;
        HttpClientContext context = null;
        try {
           	RequestResponsePair method = executeCancelableHttpMethod(postMethod, true, true);
           	response = method.getResponse();
           	context = method.getContext();
            return demarshallEnvelope(webServicesMarshaller, response, postMethod.getRequestLine().getUri());
        } catch (MarshallingException e) {
            StringBuilder trace = new StringBuilder();

            trace.append(String.format("%n==> REQUEST URI: %s%n", postMethod.getRequestLine().toString())); //$NON-NLS-1$
            for (Header h : postMethod.getAllHeaders()) {
                trace.append(String.format("===> REQUEST HEADER: %s: %s%n", h.getName(), h.getValue())); //$NON-NLS-1$
            }
            
            if (context != null) {
                List<URI> redirectLocations = context.getRedirectLocations();
                if (redirectLocations != null) {
	                for (URI redirectLocation : redirectLocations) {
	                    trace.append(String.format("==> REDIRECTING GET: %s%n", redirectLocation)); //$NON-NLS-1$
	                }
                }
            }
            
            trace.append(String.format("==> RESPONSE: %s%n", response.getStatusLine())); //$NON-NLS-1$
            for (Header h : response.getAllHeaders()) {
                trace.append(String.format("===> RESPONSE HEADER: %s: %s%n", h.getName(), h.getValue())); //$NON-NLS-1$
            }

            logger.debug(trace);
            //System.out.println(trace);
            throw new MarshallingException(e);
        } finally {
            if (response != null) {
                HttpClientUtils.closeQuietly(response);
            }
        }
    }
    
    private RequestResponsePair executeCancelableHttpMethod(final HttpRequestBase request, boolean handle404, boolean wrapIOException)  throws IOException, TeamServiceException {
        int status = 500;
        AuthenticationException authEx = null;
        
        TeamServiceCallContextImpl.setUpCall(new IAbortableHttpRequest() {
            @Override
            public String getURI() {
                return request.getRequestLine().getUri();
            }
            
            @Override
            public void abort() {
                request.abort();
            }
        });
        

        RequestResponsePair method = new RequestResponsePair(request);
        try {
            status = remoteTeamServer.executeMethod(this, method, handle404, wrapIOException);
        }
        catch (AuthenticationException e) {
            status = HttpStatus.SC_UNAUTHORIZED;
            try {
                status = OAuthUtil.handleOAuthChallenge(status, remoteTeamServer, method,
                		remoteTeamServer.getOAuthHandler());
                // It was not an OAuth challenge
                if (status == HttpStatus.SC_UNAUTHORIZED)
                    authEx = e;
            }
            catch (Exception ex) {
                int statusCode = -1;
                if (ex instanceof TeamServiceException)
                    statusCode = ((TeamServiceException)ex).getStatusCode();

                throw new TeamServiceException(ex.getMessage(), ex, statusCode);
            }
        }
        finally {
            TeamServiceCallContextImpl.tearDownCall();
        }

        if (status == HttpStatus.SC_UNAUTHORIZED && authEx != null)
            throw authEx;
        
        return method;
    }
    
  
    private Envelope demarshallEnvelope(WebServicesMarshaller webServicesMarshaller, HttpResponse postMethod, String requestUri)
            throws IOException, MarshallingException
    {
        Header[] headers = postMethod.getHeaders(HttpUtil.CONTENT_ENCODING);
        if (null == headers) {
            headers = new Header[0];
        }
        
        boolean gzipped = false;
        for (int i = 0; i < headers.length; i++) {
            if (headers[i].getValue().equalsIgnoreCase(HttpUtil.GZIP)) {
                gzipped = true;
            }
        }

        InputStream responseStream = utilGetResponseBodyAsStream(postMethod, requestUri);
        if (gzipped) {
            responseStream = new GZIPInputStream(responseStream, 8192);
        }
        
        return (Envelope) webServicesMarshaller.demarshalInputStreamToObject(responseStream);
    }
    
    /**
     * Build an exception from a fault. 
     */
    private Throwable constructExceptionFromFault(Fault fault) {
        Throwable result = MarshallerUtil.decodeFault(fault, getServiceClassLoader());
        return result;
    }
    
    private Throwable getAppropriateException(Throwable exception, Class<?>[] declaredExceptionTypes) {
        Class<?> exceptionClass = exception.getClass();

        if (declaredExceptionTypes == null) {
            declaredExceptionTypes = new Class[0];
        }

        if (OperationCanceledException.class.isAssignableFrom(exceptionClass))
            return exception;

        if (ConnectionException.class.isAssignableFrom(exceptionClass))
            return exception;
        
        // TODO: temporary code to be removed - Bug 2171
        if (IllegalArgumentException.class.isAssignableFrom(exceptionClass)) {
            return exception;
        }
        if (Error.class.isAssignableFrom(exceptionClass)) {
            return new ServiceMethodInvocationError(exception);
        }
        if (RuntimeException.class.isAssignableFrom(exceptionClass)) {
            return new ServiceMethodInvocationError(exception);
        }

        for (int i = 0; i < declaredExceptionTypes.length; i++) {
            if (declaredExceptionTypes[i].isAssignableFrom(exceptionClass)) {
                return exception;
            }
        }

        // didn't find an appropriate exception. get the first declared one, and return it

        // first see if it declared any exceptions
        if (0 == declaredExceptionTypes.length) {
			// Special case TeamServiceExceptions which indicate transport errors
			String rteMessage = (exception instanceof TeamServiceException) ? 
			        Messages.getServerString("RemoteTeamService.CommError") : //$NON-NLS-1$
					Messages.getServerString("RemoteTeamService.UndeclaredException"); //$NON-NLS-1$
			return rte(rteMessage, exception);
		}

        exceptionClass = declaredExceptionTypes[0];

        Constructor<?> exceptionConstructor = null;
        Object[] exceptionArgs = null;

        // look for a constructor taking a String, Throwable
        try {
            exceptionConstructor = exceptionClass.getConstructor(new Class[] { 
                    String.class, Throwable.class });
            exceptionArgs = new Object[] { exception.getMessage(), exception };
        } catch (NoSuchMethodException e) {
        }

        // look for a constructor taking a Throwable
        if (null == exceptionConstructor) {
            try {
                exceptionConstructor = exceptionClass.getConstructor(new Class[] { 
                        Throwable.class });
                exceptionArgs = new Object[] { exception };
            } catch (NoSuchMethodException e) {
            }
        }

        // look for a constructor taking a String
        if (null == exceptionConstructor) {
            try {
                exceptionConstructor = exceptionClass.getConstructor(new Class[] { String.class });
                exceptionArgs = new Object[] { exception.getMessage() };
            } catch (NoSuchMethodException e) {
            }
        }

        // look for a constructor taking no args
        if (null == exceptionConstructor) {
            try {
                exceptionConstructor = exceptionClass.getConstructor(new Class[] {});
                exceptionArgs = new Object[] {};
            } catch (NoSuchMethodException e) {
            }
        }

        // didn't find an appropriate constructor, return a RuntimeException
        if (null == exceptionConstructor) {
            return rte(NLS.bind(Messages.getServerString("RemoteTeamService.MissingConstructor"),  //$NON-NLS-1$
                        exceptionClass.getName()), exception);
        }

        // construct the new exception
        Throwable throwable = null;
        try {
            return (Throwable) exceptionConstructor.newInstance(exceptionArgs);
        } catch (IllegalArgumentException e) {
            throwable = e;
        } catch (InstantiationException e) {
            throwable = e;
        } catch (IllegalAccessException e) {
            throwable = e;
        } catch (InvocationTargetException e) {
            throwable = e;
        }

        // couldn't construct it, return a RuntimeException
        return rte(NLS.bind(Messages.getServerString("RemoteTeamService.ExceptionError"), exception), throwable); //$NON-NLS-1$
    }
	
	private void addMarshallingHeader(HttpRequest method) {
        method.addHeader(HttpUtil.X_MARSHALLING_VERSION, Marshaller.MarshallingVersion);
    }
	
	private void addUseridHeader(HttpRequest method) {
        String userid = remoteTeamServer.getUserid();
        TransportUtils.addUseridHeader(userid, method);
    }
	
	private void addServiceVersionHeader(HttpRequest method) {
		String version = getVersion();
		if ((version != null) && !version.equals("")) //$NON-NLS-1$
			method.addHeader(HttpUtil.X_SERVICE_VERSION, version);
	}

    private RuntimeException rte(String message, Throwable cause) {
        return new RuntimeException(message, cause);
    }

    private String lowerFirst(String string) {
        return string.substring(0, 1).toLowerCase(Locale.ENGLISH) + string.substring(1);
    }

    private Object handleEquals(Object argument) {
        return (this == argument) ? Boolean.TRUE : Boolean.FALSE;
    }

    private Object handleHashCode() {
        return Integer.valueOf(hashCode());
    }

    private Object handleToString() {
        return String.format("%s{%s}", getClass().getName(), getURL()); //$NON-NLS-1$
    }

    @Override
    public String toString() {
        return String.format("%s{%s}", getServiceClass().getName(), getURL()); //$NON-NLS-1$
    }

    private static volatile String lastCheckedUrlString = null;
    private static void validateURL(String urlString) {
        if (urlString.equals(lastCheckedUrlString)) {
            return;
        }
        lastCheckedUrlString = urlString;

        URL urlObject = null;
        try {
            urlObject = new URL(urlString);
        }
        catch (MalformedURLException e) {
            String msg = NLS.bind(Messages.getServerString("RemoteTeamService.UnparseableUri"), urlString); //$NON-NLS-1$
            logger.warn(msg, e);
            return;
        }

        String protocol = urlObject.getProtocol();
        String host     = urlObject.getHost();
        int port        = urlObject.getPort();
//        String path     = urlObject.getPath();

        if (protocol == null || protocol.length() == 0) {
            String msg = NLS.bind(Messages.getServerString("RemoteTeamService.MissingProtocol"), urlString); //$NON-NLS-1$
            Throwable t = new Throwable(msg);
            logger.warn(msg, t);
        }

        if (host == null || host.length() == 0) {
            String msg = NLS.bind(Messages.getServerString("RemoteTeamService.MissingHost"), urlString); //$NON-NLS-1$
            Throwable t = new Throwable(msg);
            logger.warn(msg, t);
        }

//        if (path == null || path.equals("")) {
//            String msg = NLS.bind("URL does not include a path: {0}", urlString);
//            Throwable t = new Throwable(msg);
//            logger.warn(msg, t);
//        }

        if (port != -1) {
            return;
        }

        int defaultPort = urlObject.getDefaultPort();

        if (defaultPort == -1) {
            // I suppose if we're not going over IP, we may not actually need a port
            String msg = NLS.bind(Messages.getServerString("RemoteTeamService.MissingPort"), urlString); //$NON-NLS-1$
            Throwable t = new Throwable(msg);
            logger.warn(msg, t);
        }
        return;
    }
}
