/*******************************************************************************
 * Licensed Materials - Property of IBM © Copyright IBM Corporation 2014, 2015. All Rights Reserved.
 * All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights: Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 ******************************************************************************/

package com.ibm.team.repository.transport.client;

import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpRequestBase;
import org.apache.http.client.protocol.HttpClientContext;

import com.ibm.team.repository.transport.client.RemoteTeamServer.ChallengeType;

/**
 * A wrapper for an HTTP request/response pair. Introduced here for backwards compatibility for software that depends upon the behavior of Apache Httpclient v3,
 * where HttpMethod encapsulated both the request and the response.
 * 
 * Not thread-safe.
 */
public class RequestResponsePair {

    private HttpRequestBase request;
    private CloseableHttpResponse response;
    private HttpClientContext httpClientContext;
    private ChallengeType challengeType = ChallengeType.NA;

    /**
     * Constructs an instance wrapping the specified request. It is expected that callers will construct an instance with a request, and set the response after
     * the request has been sent.
     * 
     * @param request
     *            Associated to this pair. Must not be <code>null</code>.
     */
    public RequestResponsePair(HttpRequestBase request) {
        if (request == null)
            throw new IllegalArgumentException("request must not be null"); //$NON-NLS-1$
        this.request = request;
    }

    /**
     * Returns the response object associated to this pair.
     * 
     * @return The response. May be <code>null</code>.
     */
    public CloseableHttpResponse getResponse() {
        return response;
    }

    /**
     * Associate a response to this pair.
     * 
     * @param response
     *            The response to associate. May be <code>null</code>.
     */
    public void setResponse(CloseableHttpResponse response) {
        this.response = response;
    }

    /**
     * Sets the challenge type for 401 cases
     */
    public void setChallengeType(ChallengeType challengeType) {
        this.challengeType = challengeType;
    }
    
    public ChallengeType getChallengeType() {
        return this.challengeType;
    }
    
    /**
     * Returns the http client context associated to this pair.
     * 
     * @return The context. May be <code>null</code>.
     */
    public HttpClientContext getContext() {
        return httpClientContext;
    }

    /**
     * Associate an http client context to this pair.
     * 
     * @param httpClientContext
     *            The client context to associate. May be <code>null</code>.
     */
    public void setContext(HttpClientContext httpClientContext) {
        this.httpClientContext = httpClientContext;
    }

    /**
     * Returns the request associated to this pair.
     * 
     * @return The request. Never <code>null</code>.
     */
    public HttpRequestBase getRequest() {
        return request;
    }

}
