/*******************************************************************************
 * Licensed Materials - Property of IBM
 * © Copyright IBM Corporation 2010, 2015. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/

package com.ibm.team.repository.transport.client;

import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.http.Header;
import org.apache.http.HttpStatus;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpDelete;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpHead;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.client.methods.HttpPut;
import org.apache.http.client.methods.HttpRequestBase;
import org.apache.http.entity.BufferedHttpEntity;
import org.apache.http.entity.ContentType;
import org.apache.http.entity.InputStreamEntity;
import org.apache.http.protocol.HTTP;
import org.apache.http.util.EntityUtils;

import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.common.internal.IRepositoryRemoteService;
import com.ibm.team.repository.common.internal.util.LocationUtil;
import com.ibm.team.repository.common.model.ServerDescription;
import com.ibm.team.repository.common.transport.HostUtil;
import com.ibm.team.repository.common.transport.HttpUtil;
import com.ibm.team.repository.common.transport.HttpUtil.CharsetEncoding;
import com.ibm.team.repository.common.transport.HttpUtil.MediaType;
import com.ibm.team.repository.common.transport.IAbortableHttpRequest;
import com.ibm.team.repository.common.transport.TeamServiceCallContextImpl;
import com.ibm.team.repository.common.transport.TeamServiceException;
import com.ibm.team.repository.common.transport.UriUtil;
import com.ibm.team.repository.transport.client.ITeamRawRestServiceClient.IRawRestClientConnection.Response;

/**
 * Abstract base implementation shared by RestClientConnection and
 * RawRestClientConnection.
 * @since 0.7
 */
public abstract class RestClientConnectionBase {

    protected final RemoteTeamServer _remoteTeamServer;
    protected final String _repoUri;
    protected final URI _requestUri;
    protected final IOAuthHandler _oauthHandler;
    protected final String _proxyUriRoot;
    
    protected final HashMap<String, String> _headers = new HashMap<String, String>();
    protected boolean _hasBeenUsed = false;
    
    private URI _publicUri;
    private Response _response = null;    
    
    static private final String UserAgent = RestClientConnectionBase.class.getName();
    
    /*
     * This client sends the AJAX-y X-Requested-With header as the web ui proxy
     * called by this client expects the inclusion of that header to confirm that 
     * it is not called being called by a browser attack (or a non-AJAX client). 
     * workitem 164867
     */
    static private final String REQUESTED_WITH_HEADER = "X-Requested-With"; //$NON-NLS-1$
    
    /**
     * Constructs a rest client connection instance.
     * 
     * @param server
     *            The remote team server. Must not be <code>null</code>.
     * 
     * @param uri
     *            The URI of the resource to connect to. Must not
     *            be <code>null</code>.  May be absolute or relative.
     *            
     * @param handler The OAuth handler to use to react to OAuth authentication
     * challenges. May be <code>null</code>. 
     * @throws URISyntaxException 
     */
    protected RestClientConnectionBase(RemoteTeamServer server, URI uri, IOAuthHandler handler) throws URISyntaxException {
        if (server == null)
            throw new IllegalArgumentException("server must not be null"); //$NON-NLS-1$
        if (uri == null)
            throw new IllegalArgumentException("relUri must not be null"); //$NON-NLS-1$
            
        _repoUri = getRepoUri(server);
        
        if (uri.isAbsolute())
            _requestUri = uri;
        else
            _requestUri = new URI(getRepoUri(server) + UriUtil.trimLeadingSlash(uri.toString()));
        
        _remoteTeamServer = server; 
        _oauthHandler = handler;
        _proxyUriRoot = buildProxyUri(server);
    }

    
    public void addRequestHeader(String name, String value) {
        
        if (_hasBeenUsed)
            throw new IllegalStateException("Cannot add request headers to a connection has been previously run"); //$NON-NLS-1$
        
        if (name == null || name.length() == 0)
            return;
        
        if (value == null)
            value = ""; //$NON-NLS-1$
        
        _headers.put(name, value);
    }

    public void release() {
        if (_response == null)
            return;
        
        InputStream responseStream = _response.getResponseStream();
        if (responseStream != null) {
            try {
                responseStream.close();
            } catch (IOException e) {
                logger.warn(String.format("Unexpected IO error on connection release: %s", e.getMessage())); //$NON-NLS-1$
            }
        }
    }
    
    public Response doGet() throws TeamServiceException, TeamRepositoryException {
        return doMethod(new HttpGet());
    }
    
    public Response doHead() throws TeamServiceException, TeamRepositoryException {
        return doMethod(new HttpHead());
    }

    public Response doPost(InputStream content, long contentLength,
            String contentType) throws TeamServiceException, TeamRepositoryException {

        HttpPost postMethod = new HttpPost();

        BufferedHttpEntity requestEntity;
        try {
            requestEntity = new BufferedHttpEntity(
                    new InputStreamEntity(content, contentLength, ContentType.create(contentType)));
        } catch (IOException e) {
            throw new TeamServiceException(e.getMessage(), e);
        }
        postMethod.setEntity(requestEntity);

        return doMethod(postMethod);
    }

    public Response doPut(InputStream content, long contentLength,
            String contentType) throws TeamServiceException, TeamRepositoryException {

        HttpPut putMethod = new HttpPut();

        BufferedHttpEntity requestEntity;
        try {
            requestEntity = new BufferedHttpEntity(
                    new InputStreamEntity(content, contentLength, ContentType.create(contentType)));
        } catch (IOException e) {
            throw new TeamServiceException(e.getMessage(), e);
        }
        putMethod.setEntity(requestEntity);

        return doMethod(putMethod);
    }

    public Response doDelete() throws TeamServiceException, TeamRepositoryException {
        return doMethod(new HttpDelete());
    }
    
    /**
     * Helper for the various do* methods.
     * 
     * @param request The http method to use. Must not be <code>null</code>.
     * @return The response object. Never <code>null</code>.
     * @throws TeamServiceException Thrown on non-recoverable server I/O error.
     * @throws TeamRepositoryException Thrown on recoverable repository errors.
     */
    protected Response doMethod(HttpRequestBase request) throws TeamServiceException, TeamRepositoryException {
        
        if (_hasBeenUsed) 
            throw new IllegalStateException("do* called multiple times on the same connection"); //$NON-NLS-1$
        
        _hasBeenUsed = true;
        
        RequestResponsePair method = new RequestResponsePair(request);
        prepareMethod(method);
        
        int status = -1;
        TeamRepositoryException caughtException = null;
        
        try {
            executeMethod(method);
        } catch (TeamRepositoryException e) {
            caughtException = e;
            if (method.getResponse() != null)
                status = method.getResponse().getStatusLine().getStatusCode();
        }

        CloseableHttpResponse response = method.getResponse();
        if (response == null && caughtException != null) {
            throw caughtException;
        }
        if (response == null) {
            throw new TeamRepositoryException("Response is unexpectedly null - should never happen"); //$NON-NLS-1$
        }
        
        Header[] headers = response.getAllHeaders();
        Map<String, String> headersMap = new HashMap<String, String>(
                headers.length);
        for (Header header : headers) {
            headersMap.put(header.getName(), header.getValue());
        }

        /*
         * Only create the stream if there's a entity on the response
         */
        InputStream responseStream = null;
        try {
        	if (response.getEntity() != null)
        		responseStream = new ReleasingAndAbortingInputStream(response.getEntity());
        	
        } catch (IOException e) {
            EntityUtils.consumeQuietly(response.getEntity());
            throw new TeamServiceException(e.getMessage(), e);
        }
        
        String statusText = response.getStatusLine().getReasonPhrase();
        status = response.getStatusLine().getStatusCode();
        _response = new Response(responseStream, headersMap, status, statusText);
        
        if (caughtException != null)
            throw caughtException;
        
        return _response;
    }


    /**
     * Prepares an {@link RequestResponsePair} for execution.  Callers first
     * create an {@link RequestResponsePair} specialization, call {@link #prepareMethod()}
     * and then call {@link #executeMethod()}.
     * 
     * @throws TeamServiceException  Thrown on error.
     */
    protected void prepareMethod(RequestResponsePair method) throws TeamServiceException {

        HttpRequestBase request = method.getRequest();
        try {
            
            URI uri = bindProxyUri();
            /*
             * Strip off protocol, host, and port if present e.g. "https://jazz.net:9943"
             * to ensure our ProtocolSocketFactory is used
             * see: http://issues.apache.org/jira/browse/HTTPCLIENT-634
             * 
             * Per workitem 109606 I've pulled the parsing done by URI, as the 
             * quoting was incorrect
             */
            String formatString = "%s"; //$NON-NLS-1$
            if (uri.getQuery() != null) 
                formatString += "?%s"; //$NON-NLS-1$
            if (uri.getFragment() != null)
                formatString += "#%s"; //$NON-NLS-1$
            
            try {
                uri = new URI(String.format(formatString, uri.getRawPath(), 
                        uri.getRawQuery(), uri.getRawFragment()));
            } catch (URISyntaxException e) {
                logger.warn(String.format("URISyntaxException parsing URI: %s", e.getMessage())); //$NON-NLS-1$
            }
            request.setURI(uri);
            
            /*
             * If we're bound to the Web UI proxy, include a version header string
             * per its (internal) protocol.  The header value is not read by the proxy.
             */
            if (startsWithPath(uri, URI.create(_proxyUriRoot))) {
                request.addHeader("X-jazz-downstream-auth-client-level", null); //$NON-NLS-1$
            }
       } catch (IllegalStateException e) {
           throw new TeamServiceException(e.getMessage(), e);
       } catch (TeamRepositoryException e) {
           throw new TeamServiceException(e.getMessage(),e);
       }
     
       /*
        * Add in the requested headers, then fold in any of our proprietary headers.
        * The "X-Requested-With" header is to satisfy the requirements of the web UI proxy
        * ("proxy") that it is being called by a non-browser client.
        */
        Iterator<Entry<String, String>> iter = _headers.entrySet().iterator();
        while (iter.hasNext()) {
        	Entry<String, String> entry = iter.next();
            String key = entry.getKey();
            String value = entry.getValue();
            request.addHeader(key, value);
        }
        
        /*
         * The 5.0.2 and earlier versions of this class set a non-standard header given by the value
         * of the org.apache.commons.httpclient.params.HttpMethodParams.USER_AGENT constant.
         * We still that header, for back-compatibility, as well as the standard User-Agent header.
         */
       request.setHeader(HTTP.USER_AGENT, UserAgent);
       request.setHeader(HttpUtil.NON_STD_USER_AGENT_HEADER, UserAgent);
       request.setHeader(REQUESTED_WITH_HEADER, UserAgent);
       if (request.getFirstHeader(HttpUtil.ACCEPT) == null)
           request.setHeader(HttpUtil.ACCEPT, MediaType.XML.toString());
       if (request.getFirstHeader(HttpUtil.ACCEPT_CHARSET) == null)
           request.setHeader(HttpUtil.ACCEPT_CHARSET, CharsetEncoding.UTF8.toString());
       if (request.getFirstHeader(HttpUtil.ACCEPT_LANGUAGE) == null)
           TransportUtils.setAcceptLanguageHeader(request);
       
       String userid = _remoteTeamServer.getUserid();
       TransportUtils.addUseridHeader(userid, request);
    }
    
    /* (non-Javadoc)
     * @see com.ibm.team.repository.transport.client.ITeamRestServiceClient.IRestClientConnection#getResponse()
     */
    public Response getResponse() {
        return _response;
    }
    
    /**
     * Executes a prepared {@link RequestResponsePair}.  Callers first
     * create an {@link RequestResponsePair} specialization, call {@link #prepareMethod()}
     * and then call {@link #executeMethod()}.
     * 
     * @return The HTTP status code.
     * 
     * @throws TeamServiceException Thrown on non-recoverable server I/O error.
     * @throws TeamRepositoryException Thrown on recoverable repository errors.
     */
    protected void executeMethod(RequestResponsePair method) throws TeamServiceException, TeamRepositoryException {
        
        int status = 500;
        AuthenticationException authEx = null;
        
        final HttpRequestBase cancellableMethod = method.getRequest();
        TeamServiceCallContextImpl.setUpCall(new IAbortableHttpRequest() {
            
            @Override
            public String getURI() {
                return cancellableMethod.getURI().toString();
            }
            
            @Override
            public void abort() {
                cancellableMethod.abort();
            }
        });
        
        try {
            try {
                status = _remoteTeamServer.executeMethod(null, method, false, true);
            }
            catch (AuthenticationException e) {
                status = HttpStatus.SC_UNAUTHORIZED;
                try {
                    status = OAuthUtil.handleOAuthChallenge(status, _remoteTeamServer, method,
                            _oauthHandler);
                    // It was not an OAuth challenge
                    if (status == HttpStatus.SC_UNAUTHORIZED)
                        authEx = e;
                }
                catch (Exception ex) {
                    int statusCode = -1;
                    if (ex instanceof TeamServiceException)
                        statusCode = ((TeamServiceException)ex).getStatusCode();
                    throw new TeamServiceException(ex.getMessage(), ex, statusCode);
                }
            } catch (TeamServiceException exp) {
            	throw exp;
            } catch (Exception ex) {
                throw new TeamServiceException(ex.getMessage(), ex);
            }
    
            if (status == HttpStatus.SC_UNAUTHORIZED && authEx != null) {
                throw authEx;
            }
        }        
        finally {
            TeamServiceCallContextImpl.tearDownCall();
        }
    }

    /**
     * Returns an absolute URI for this connection, bound to the well-known 
     * proxy if the authority and application path for the connection's URI differs from that of
     * repository bound to this connection.
     * 
     * @return Returns a URI modified to point to the well-known proxy if
     * outside of the repository's authority; else the original URI is returned.
     * Never <code>null</code>.
     * 
     * @throws TeamRepositoryException
     */
    protected URI bindProxyUri() throws TeamRepositoryException {
        try {
            String uri = _requestUri.toString();
            URI repoUri = new URI(_repoUri);
            
            /*
             * If the request URI has the same host and context path of either the repository URI
             * or the public URI (if it can be determined), then no proxying is needed.
             */
            boolean equivalentApps = (HostUtil.authorityEquiv(repoUri, _requestUri) && startsWithPath(_requestUri, repoUri));
            if (!equivalentApps) {
                URI publicUri = getPublicUri();
                if (publicUri != null && HostUtil.authorityEquiv(publicUri, _requestUri) && startsWithPath(_requestUri, publicUri)) {
                    equivalentApps = true;
                }
            }
            if (!equivalentApps) {
                uri = String.format("%s?uri=%s", _proxyUriRoot, LocationUtil.encode(uri)); //$NON-NLS-1$
                return new URI(uri);
            }
            return _requestUri;
        } catch (URISyntaxException e) {
            throw new TeamRepositoryException(e);
        } catch (UnsupportedEncodingException e) {
            throw new TeamRepositoryException(e);
        }
    }

    /**
     * Returns whether or not the path of the referenced <code>baseUri</code>
     * starts with the path of the specified <code>compareUri</code>.
     * 
     * @param baseUri The base URI to compare against. Must not be <code>null</code>.
     * @param compareUri The URI to compare against the base. Must not be <code>null</code>.
     * @return <code>true</code> if the base URI path starts with the path of the comparing
     * URI, <code>false</code> if it does not.
     */
    private static boolean startsWithPath(URI baseUri, URI compareUri) {
        String comparePath = compareUri.getPath();
        if (comparePath == null)
            comparePath = ""; //$NON-NLS-1$
        if (!comparePath.endsWith(SLASH))
            comparePath += SLASH;

        String basePath = baseUri.getPath();
        if (basePath == null)
            basePath = ""; //$NON-NLS-1$
        if (!basePath.endsWith(SLASH))
            basePath += SLASH;
        
        if (basePath.startsWith(comparePath))
            return true;
        
        return false;
    }

    /**
     * Builds a proxy URI root from the referenced team server's repository
     * URI.  If the trailing "secure" segment is exists on the uri, it is removed.
     * 
     * @param server The team server. Must not be <code>null</code>.
     * @return The proxy URI.  Never <code>null</code> - never ends in '/'.
     */
    private static String buildProxyUri(RemoteTeamServer server) {
        
        String repoUri = server.getRepositoryURL();
        return repoUri + PROXY_CONTEXT_SEGMENT;
    }

    /**
     * Retrieve the repository URI for the referenced server.
     * 
     * @param server The team server. Must not be <code>null</code>.
     * @return The repository uri.  Never <code>null</code>.
     */
    private static String getRepoUri(RemoteTeamServer server) {
        
        String repoUri = server.getRepositoryURL();
        return repoUri;
    }

    /**
     * Try to determine the public URI for the repository. If it cannot be determined
     * (e.g. it may not be set yet), then <code>null</code> is returned.
     */
    private URI getPublicUri() throws TeamRepositoryException {
        if (_publicUri == null) {
            IRepositoryRemoteService repoService = (IRepositoryRemoteService) _remoteTeamServer.getServiceImplementation(IRepositoryRemoteService.class);
            if (repoService != null) {
                ServerDescription serverDescription = repoService.describe();
                String uri = serverDescription.getPublicUriRoot();
                if (uri != null) {
                    _publicUri = URI.create(uri);
                }
            }
        }
        
        return _publicUri;
    }
    
    private static final String PROXY_CONTEXT_SEGMENT = "proxy"; //$NON-NLS-1$
    private static final String SLASH = "/"; //$NON-NLS-1$
    static private final Log logger = LogFactory.getLog(RestClientConnectionBase.class);

}
