/*******************************************************************************
 * Licensed Materials - Property of IBM
 * © Copyright IBM Corporation 2014, 2015. All Rights Reserved.
 *
 * Note to U.S. Government Users Restricted Rights:  Use,
 * duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.repository.transport.client;

import java.net.Socket;
import java.text.MessageFormat;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

import javax.net.ssl.HostnameVerifier;
import javax.net.ssl.KeyManager;
import javax.net.ssl.SSLContext;
import javax.net.ssl.TrustManager;
import javax.net.ssl.X509TrustManager;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.http.conn.ssl.DefaultHostnameVerifier;
import org.apache.http.conn.ssl.NoopHostnameVerifier;
import org.apache.http.conn.ssl.SSLSocketFactory;
import org.apache.http.conn.ssl.X509HostnameVerifier;

/**
 * Convenience methods to simplify SSL configuration.
 *
 * WARNING: com.ibm.team.jfs.app.http.ssl.SSLUtils and com.ibm.team.repository.transport.client.SSLUtils are the
 * same. If you update either, update the other. This is because we don;t have any packages in common between
 * the SDK and client.
 *
 */
public class SSLUtils {

    private static final Log _log = LogFactory.getLog(SSLUtils.class);

    /**
     * System property name that controls whether secure or insecure validation of SSL certificates is
     * done. If true, secure validation (using a proper TrustManager and HostnameVerifier) is done (unless
     * overridden by other properties). If false, use a TrustManager that says every certificate is valid, and
     * a HostnameVerifier that doesn't care if the host and the certificate CN match.
     *
     */
    public static final String VALIDATE_CERTS_PROP = "com.ibm.team.clm.ssl.validateCertificates"; //$NON-NLS-1$

    /**
     * System property that controls whether a secure or insecure TrustManager is used for validating SSL
     * certificates. If not set, the default is the value of of the property specified by VALIDATE_CERTS_PROP.
     * If true, the TrustManager default for the JRE is used. If false, a TrustManager that trusts any certificate is used.
     */
    public static final String USE_SECURE_TRUST_MANAGER_PROP = "com.ibm.team.clm.ssl.useSecureTrustManager"; //$NON-NLS-1$

    /**
     * System property that controls what kind of HostnameVerifier is used to validate hostnames in certificates.
     * If true, uses a verifier similar to what a browser would use: as long as the real host name matches the
     * host name in the certificate, or one of the alternate names in the certificate, the host name is considered valid.
     * If false, a verifier that doesn't check anything is used.
     */
    public static final String USE_SECURE_HOSTNAME_VERIFIER_PROP = "com.ibm.team.clm.ssl.useSecureHostnameVerifier"; //$NON-NLS-1$

    /**
     * System property that allows setting the TLS/SSL protocol that should be used when connecting to a JTS.
     * If not set, a value from the DEFAULT_PROTOCOLS is used. Typically used when configuring
     * SP800-131a strict mode (the value would be "TLSv1.2" in that case).
     */
    public static final String CLIENT_SSL_PROTOCOL_PROPERTY = "com.ibm.team.repository.transport.client.protocol"; //$NON-NLS-1$

    private static final TrustingTrustManager trustingTM = new TrustingTrustManager();

    public static final List<String> DEFAULT_PROTOCOLS = Collections.unmodifiableList(Arrays.asList(new String[] {
            //"SSL_TLSv2", // only for IBM JVM: try TLSv1.2, 1.1, 1.0, SSLv3 //$NON-NLS-1$
            "SSL_TLS", // only for IBM JVM: try TLSv1, SSLv3 //$NON-NLS-1$
            //"TLSv1.2", // IBM: TLSv1.2 only, Oracle: maybe others, too //$NON-NLS-1$
            //"TLSv1.1", // IBM: TLSv1.1 only, Oracle: maybe others, too //$NON-NLS-1$
            //"TLSv1", // IBM: TLSv1 only, Oracle: maybe others, too //$NON-NLS-1$
            "TLS", // IBM: TLSv1 or SSLv3 Oracle: some version of TLS //$NON-NLS-1$
            "SSL" // IBM: SSLv3 only Oracle: SSLV3 or SSLv2//$NON-NLS-1$
    }));

    /**
     * Should we be validating SSL certificates? The default value is false.
     * @return true if certificates should be validated, false if they should not
     */
    public static boolean isValidateCertificates()  {
        String p = System.getProperty(VALIDATE_CERTS_PROP, "false"); //$NON-NLS-1$
        return Boolean.parseBoolean(p);
    }

    /**
     * Should a secure TrustManager be used to validate SSL certificates?
     * @return true if the default JRE TrustManger should be used, false if a naive TrustManager
     * (that trusts every certificate) should be used.
     */
    public static boolean isUseSecureTrustManager()  {
        String p = System.getProperty(USE_SECURE_TRUST_MANAGER_PROP);
        return p == null ? isValidateCertificates() : Boolean.parseBoolean(p);
    }

    /**
     * Should a HostnameVerifier be used to validate SSL certificates?
     * @return true if the server name should be verified with certificate names, false
     * if no verification is done
     */
    public static boolean isUseSecureHostnameVerifier()  {
        String p = System.getProperty(USE_SECURE_HOSTNAME_VERIFIER_PROP);
        return p == null ? isValidateCertificates() : Boolean.parseBoolean(p);
    }

    /**
     * Create a socket factory using a supported SSL protocol,
     * with the default TrustManager and HostnameVerifier.
     *
     * @return a {@link SSLSocketFactory}
     */
    public static SSLSocketFactory createSocketFactory() {
        return new SSLSocketFactory(SSLUtils.createSSLContext(SSLUtils.getDefaultTrustManager()),
                SSLUtils.getDefaultX509HostnameVerifier());
    }

    /**
     * Creates an SSL context.
     * The returned SSLContext will be created so that it is compatible with the
     * current security environment: uses a supported SSL protocol and the TrustManager returned
     * by getDefaultTrustManager().
     *
     * @return a {@link SSLContext}
     * @throws  RuntimeException if no suitable SSL protocol is configured
     */
    public static SSLContext createSSLContext() {
       return createSSLContext(getSSLProtocols(null), null, getDefaultTrustManager());
    }

    /**
     * Creates an SSL context.
     * The returned SSLContext will be created so that it is compatible with the
     * current security environment: uses a supported SSL protocol and the TrustManager provided.
     *
     * @param trustManager a TrustManager to use when validating certificates. null means use the TrustManager configured for the JRE.
     * @return a {@link SSLContext}
     * @throws  RuntimeException if no suitable SSL protocol is configured
     */
    public static SSLContext createSSLContext(TrustManager trustManager) {
       return createSSLContext(getSSLProtocols(null), null, trustManager);
    }

    /**
     * Creates an SSL context.
     * The returned SSLContext will be created so that it is compatible with the
     * current security environment: uses a supported SSL protocol from the list provided
     * and the TrustManager provided.
     *
     * @param protocols list of SSL protocols from which to choose
     * @param keyManagers KeyManagers
     * @param trustManager a TrustManager to use when validating certificates. null means use the TrustManager configured for the JRE.
     * @return a {@link SSLContext}
     * @throws  RuntimeException if no suitable SSL protocol is configured
     */
    public static SSLContext createSSLContext(List<String> protocols, KeyManager[] keyManagers, TrustManager trustManager) {
        logDebug("SSL Algorithm Priority: " + protocols); //$NON-NLS-1$

        SSLContext context = null;
        for(String protocol : getSSLProtocols(protocols)) {
            context = createSSLContext(protocol, keyManagers, trustManager);
            if(context != null) {
                return context;
            }
            logDebug(MessageFormat.format("Failed to create {0} context", protocol)); //$NON-NLS-1$
        }
        logDebug("No acceptable encryption algorithm found"); //$NON-NLS-1$
        throw new RuntimeException("No acceptable encryption algorithm found"); //$NON-NLS-1$
    }

    /**
     * Creates an SSL context.
     * The returned SSLContext will be created so that it is compatible with the
     * current security environment: uses a supported SSL protocol and the TrustManager provided.
     *
     * @param keyManagers KeyManagers
     * @param trustManager a TrustManager to use when validating certificates. null means use the TrustManager configured for the JRE.
     * @return a {@link SSLContext}
     * @throws  RuntimeException if no suitable SSL protocol is configured
     */
    public static SSLContext createSSLContext(KeyManager[] keyManagers, TrustManager trustManager) {
        return createSSLContext((List<String>) null, keyManagers, trustManager);
    }

    // Returns null when the given algorithm fails
    /**
     * Creates an SSL context.
     * The returned SSLContext will be created so that it is compatible with the
     * current security environment: uses the provided SSL protocol and TrustManager.
     *
     * @param protocol SSL protocol to use
     * @param keyManagers KeyManagers
     * @param trustManager a TrustManager to use when validating certificates. null means use the TrustManager configured for the JRE.
     * @return a {@link SSLContext}, or null if the SSLContext could not be created
     */
    public static SSLContext createSSLContext(String protocol, KeyManager[] keyManagers, TrustManager trustManager) {
        logDebug(MessageFormat.format("Creating {0} context", protocol)); //$NON-NLS-1$
        SSLContext context;
        try {
            context = SSLContext.getInstance(protocol);
            context.init(keyManagers, trustManager == null ? null : new TrustManager[] { trustManager }, null);

            /* Create a socket to ensure this algorithm is acceptable.  This will
             * correctly disallow certain configurations (such as SSL_TLS under FIPS) */
            Socket s = context.getSocketFactory().createSocket();
            s.close();
        } catch (Throwable t) {
            logDebug(t.getMessage(), t);
            return null;
        }
        return context;
    }


    /**
     * Return a list of the SSL protocols that are chosen from when creating SSLContexts. If the provided
     * list is null, the value of DEFAULT_PROTOCOLS is returned, or the value set by the CLIENT_SSL_PROTOCOL_PROPERTY
     * property, if set. Otherwise the provided list is just returned.
     *
     * @param algorithms A list of SSL protocols, or null.
     * @return a list of protocols
     */
    public static List<String> getSSLProtocols(List<String> protocols)  {
        if (protocols == null)  {
            String protocol = System.getProperty(CLIENT_SSL_PROTOCOL_PROPERTY);
            if (protocol != null)  {
                return Arrays.asList(new String[] { protocol });
            }
            return DEFAULT_PROTOCOLS;
        }
        return protocols;
    }


    /**
     * Return the trust manager to use, based on the settings of system properties. A null return
     * means that the default JRE configured trust manager will be used.
     *
     * @return a trust manager, or null
     */
    public static X509TrustManager getDefaultTrustManager()  {
        return isUseSecureTrustManager() ? getSecureTrustManager() : getTrustingTrustManager();
    }

    /**
     * Return a TrustManager that trusts all certificates.
     *
     * @return a trust manager
     */
    public static X509TrustManager getTrustingTrustManager()  {
        return trustingTM;
    }

    /**
     * Return the TrustManager configured for the JRE or null (which means use the JRE configured trust manager).
     *
     * @return a trust manager, or null
     */
    public static X509TrustManager getSecureTrustManager()  {
        return null;    // use the JRE TrustManager
    }

    /**
     * Convenience method to get the default HostnameVerifier as a {@link DefaultHostnameVerifier} or {@link NoopHostnameVerifier}.
     *
     * @return the default HostnameVerifier to use, based on the settings of the system properties.
     */
    public static HostnameVerifier getDefaultX509HostnameVerifier2()  {
        return isUseSecureHostnameVerifier() ? new DefaultHostnameVerifier() : new NoopHostnameVerifier();
    }

    
    /**
     * Convenience method to get the default HostnameVerifier as a {@link X509HostnameVerifier}.
     *
     * @return the default HostnameVerifier to use, based on the settings of the system properties.
     * @deprecated Newer code should use {@link #getDefaultHostnameVerifier2()}
     * @since 6.0.1
     */
    public static X509HostnameVerifier getDefaultX509HostnameVerifier()  {
        return isUseSecureHostnameVerifier() ? getSecureX509HostNameverifier() : getTrustingX509HostnameVerifier();
    }

    /**
     * @return the default HostnameVerifier to use, based on the settings of the system properties.
     * @deprecated Newer code should use {@link #getDefaultHostnameVerifier2()}
     * @since 6.0.1 
     */
    public static HostnameVerifier getDefaultHostnameVerifier()  {
        return getDefaultX509HostnameVerifier();
    }

    /**
     * @return the default HostnameVerifier to use, based on the settings of the system properties.
     */
    public static HostnameVerifier getDefaultHostnameVerifier2()  {
        return getDefaultX509HostnameVerifier2();
    }

    /**
     * @return a HostnameVerifier that does not verify the host name
     * @deprecated Newer code should use {@link #getTrustingX509HostnameVerifier2()}
     */
    public static X509HostnameVerifier getTrustingX509HostnameVerifier()  {
        return SSLSocketFactory.ALLOW_ALL_HOSTNAME_VERIFIER;
    }
    
    /**
     * @return a HostnameVerifier that does not verify the host name
     */
    public static HostnameVerifier getTrustingX509HostnameVerifier2()  {
        return new NoopHostnameVerifier();
    }

    /**
     * Returns a HostnameVerifier that verifies host names with names in a certificate much like a browser
     * does: as long as the real host name matches the
     * host name in the certificate, or one of the alternate names in the certificate, the host name is considered verified.
     * @return
     * @deprecated Newer code should use {@link #getSecureX509HostNameverifier2()}
     */
    public static X509HostnameVerifier getSecureX509HostNameverifier()  {
        return SSLSocketFactory.BROWSER_COMPATIBLE_HOSTNAME_VERIFIER;
    }

    public static HostnameVerifier getSecureX509HostNameverifier2()  {
        return new DefaultHostnameVerifier();
    }
    
    private static void logDebug(String message)  {
        if (_log.isDebugEnabled())  {
            _log.debug(message);
        }
    }

    private static void logDebug(String message, Throwable t)  {
        if (_log.isDebugEnabled())  {
            _log.debug(message, t);
        }
    }


    // Implements an all-trusting TrustManager
    private static class TrustingTrustManager implements X509TrustManager  {
        private static final java.security.cert.X509Certificate[] emptyArray = new java.security.cert.X509Certificate[0];

        @Override
        public void checkClientTrusted(java.security.cert.X509Certificate[] certs, String authType) {
            /* Ignore Method Call */
        }

        @Override
        public void checkServerTrusted(java.security.cert.X509Certificate[] certs, String authType) {
            /* Ignore Method Call */
        }

        @Override
        public java.security.cert.X509Certificate[] getAcceptedIssuers() {
            return emptyArray;
        }
    }

}
