/*******************************************************************************
 * Licensed Materials - Property of IBM
 * © Copyright IBM Corporation 2010, 2014. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.repository.transport.client;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.net.UnknownHostException;
import java.nio.ByteBuffer;
import java.security.cert.Certificate;
import java.security.cert.CertificateEncodingException;
import java.security.cert.CertificateException;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;

import javax.net.ssl.HandshakeCompletedEvent;
import javax.net.ssl.HandshakeCompletedListener;
import javax.net.ssl.KeyManager;
import javax.net.ssl.SSLContext;
import javax.net.ssl.SSLPeerUnverifiedException;
import javax.net.ssl.SSLSocket;
import javax.net.ssl.TrustManager;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.http.HttpHost;
import org.apache.http.conn.socket.LayeredConnectionSocketFactory;
import org.apache.http.conn.ssl.BrowserCompatHostnameVerifier;
import org.apache.http.protocol.HttpContext;

import com.ibm.team.repository.common.transport.ICertificateValidator;
import com.ibm.team.repository.common.transport.ICertificateValidator.Trust;
import com.ibm.team.repository.common.util.NLS;
import com.ibm.team.repository.transport.internal.nls.Messages;

/**
 * A variant of {@link SecureInterruptableSocketFactory} in support of HttpClient 4.x
 */
public class SecureInterruptableSocketFactory2 extends InterruptableSocketFactory2 implements LayeredConnectionSocketFactory {

    static private final LogUtil _logger = new LogUtil(SecureInterruptableSocketFactory2.class);
    
    public static final String SSL_TLS = SSLContextUtil.SSL_TLS;
    public static final String TLS = SSLContextUtil.TLS;
    public static final String SSL = SSLContextUtil.SSL;
    
    private final static Log LOG = LogFactory.getLog(SecureInterruptableSocketFactory2.class);
    
    private final static ConcurrentMap<String, X509Certificate> hostnameToCertificate = new ConcurrentHashMap<String, X509Certificate>();
    protected volatile X509Certificate [] lastKnownServerCertPath;
    
    protected final HandshakeCompletedListener handshakeListener = new HandshakeCompletedListener() {
    	public void handshakeCompleted(HandshakeCompletedEvent event) {
            _logger.debug("Handshake completed"); //$NON-NLS-1$
    	    
			try {
				Certificate [] certs = event.getPeerCertificates();
	    		if (certs == null || certs.length == 0) {
	    			return;
	    		}
	    		if (certs instanceof X509Certificate[]) {
	    			lastKnownServerCertPath = (X509Certificate[])certs;
	    			return;
	    		}
	    		X509Certificate [] certPath = new X509Certificate[certs.length];
	    		for (int i = certs.length - 1; i >= 0; --i) {
	    			if (!(certs[i] instanceof X509Certificate)) {
	    				return;
	    			}
	    			certPath[i] = (X509Certificate)certs[i];
	    		}
	    		lastKnownServerCertPath = certPath;
			} catch (SSLPeerUnverifiedException e) {
				LOG.trace(e);
			}
    	}
    };

    // SSL context -- use its socket factory to wrap unencrypted sockets
    private final SSLContext sslContext;
    private final ICertificateValidator validator;

    public SecureInterruptableSocketFactory2(TrustManager trustManager) {
        this(null, trustManager, null);
    }
    
    public SecureInterruptableSocketFactory2(KeyManager[] keyManagers, TrustManager trustManager) {
        this(keyManagers, trustManager, null);
    }
    
    public SecureInterruptableSocketFactory2(ICertificateValidator validator) {
        this(null, new ValidatingX509TrustManager(validator), validator);
    }
    
    public SecureInterruptableSocketFactory2(KeyManager[] keyManager, ICertificateValidator validator) {
        this(keyManager, new ValidatingX509TrustManager(validator), validator);
    }
    
    private SecureInterruptableSocketFactory2(KeyManager[] keyManager, TrustManager trustManager, ICertificateValidator validator) {
    	super();
        sslContext = SSLContextUtil.createSSLContext(keyManager, trustManager);
        this.validator = validator;
        _logger.debug("Created factory"); //$NON-NLS-1$
    }

    @Override
    public Socket createLayeredSocket(Socket socket, String target, int port, HttpContext context) throws IOException, UnknownHostException {
        return createEncryptedSocket(socket, target, port);
    }

    @Override
    public Socket connectSocket(int connectTimeout, Socket sock, HttpHost host, InetSocketAddress remoteAddress, InetSocketAddress localAddress, HttpContext context)
            throws IOException
    {
        Socket connectedSocket = super.connectSocket(connectTimeout, sock, host, remoteAddress, localAddress, context);
        return createEncryptedSocket(connectedSocket, host.getHostName(), remoteAddress.getPort());
    }
    
	private void handleCertificateHostnameMismatch(String host, int port, SSLSocket sslSocket, IOException exp) throws SSLPeerUnverifiedException, IOException {
        _logger.debug("Handling certificate hostname mismatch"); //$NON-NLS-1$
        
		X509Certificate certificate = null;
		Certificate[] certificates = sslSocket.getSession().getPeerCertificates();
		if (certificates != null && certificates.length > 0) {
			if (certificates[0] instanceof X509Certificate) {
				certificate = (X509Certificate)certificates[0];
			}
		}
		
		String hostnameKey = host + '-' + port;

		// Check for previously accepted mismatches
		X509Certificate savedCertificate = hostnameToCertificate.get(hostnameKey);
		if (savedCertificate != null) {
			if (savedCertificate.equals(certificate)) {
                _logger.debug("Saved certificate found"); //$NON-NLS-1$
				// If the saved certificate matches the certificate
				// that was found then everything is good.
				return;
			}
		}
		
		if (isPermanentlyAccepted(hostnameKey, certificate)) {
            _logger.debug("Certificate permanently accepted"); //$NON-NLS-1$
			// Cache the certificate in the in-memory cache to
			// avoid going back to the drive
			hostnameToCertificate.put(hostnameKey, certificate);
			return;
		}
		
		if (validator != null) {
            _logger.debug("Validating certificate acceptance"); //$NON-NLS-1$
            
			CertificateException exception = new CertificateHostnameMismatchException(NLS.bind(Messages.getClientString("SecureInterruptableSocketFactory.ServerCertifiateHostnameMismatch"),host)); //$NON-NLS-1$
			Trust trust = validator.validate(certificate, exception);
			switch (trust) {
			case ACCEPT_CONNECTION: {
				// Continue to allow the connection
				break;
			}
			case ACCEPT_PERMANENT: {
				// Store the certificate and continue
				storePermanentlyAccepted(hostnameKey, certificate);
				break;
			}
			case ACCEPT_SESSION: {
				// Store the certificate and continue
				hostnameToCertificate.put(hostnameKey, certificate);
				break;
			}
			case REJECT: {
				throw exp;
			}
			}
		} else {
			// Do nothing because we used to allow mismatched hostnames we should continue
			// to allow mismatched hostnames to avoid adoption headaches
		}
	}
	
	private String getHostnameCertStoreDirectory() {
		String userHome = System.getProperty("user.home"); //$NON-NLS-1$
		return userHome + "/.jazzhostnamecerts"; //$NON-NLS-1$
	}
	
	private boolean isPermanentlyAccepted(String host, X509Certificate certificate) throws IOException {
		String certFilename = (getHostnameCertStoreDirectory() + '/' + host).replace('/', File.separatorChar); 
		
		File file = new File(certFilename);
		if (file.exists() == false)
			return false;
		
		X509Certificate savedCertificate;
		FileInputStream input = new FileInputStream(file);
		try {
			try {
				CertificateFactory cf = CertificateFactory.getInstance("X.509"); //$NON-NLS-1$
				savedCertificate = (X509Certificate)cf.generateCertificate(input);
			} catch (CertificateException exp) {
				// The certificate could not be loaded, so just return that it was not accepted
				LOG.warn(NLS.bind("The certificate in file \"{0}\" could not be read.", file.getAbsolutePath()), exp); //$NON-NLS-1$
				return false;
			}
		} finally {
			input.close();
		}
		
		return savedCertificate.equals(certificate);
	}
	
	private void storePermanentlyAccepted(String host, X509Certificate certificate) throws IOException {
	    _logger.debug("Storing permanently accepted certificate"); //$NON-NLS-1$
	    
		String dirName = getHostnameCertStoreDirectory();
		
		File dir = new File(dirName);
		if (dir.exists() == false) {
			if (dir.mkdirs() == false) {
				LOG.warn(NLS.bind("A directory \"{0}\" could not be created to permanently store the accepted certificate", dir.getAbsolutePath())); //$NON-NLS-1$
				return;
			}
		}
		
		File file = new File(dir, host);
		FileOutputStream output = new FileOutputStream(file, false);
		try {
			ByteBuffer buffer;
			try {
				buffer = ByteBuffer.wrap(certificate.getEncoded());
			} catch (CertificateEncodingException e) {
				// The certificate could not be saved, so log and return
				LOG.warn(NLS.bind("A certificate could not be saved to file \"{0}\".", file.getAbsolutePath()), e); //$NON-NLS-1$
				return;
			}
			while (buffer.hasRemaining()) {
				output.getChannel().write(buffer);
			}
		} finally {
			output.close();
		}
	}
	
    /**
     * Returns an encrypted socket for the specified parameters.<p/>
     * 
     * NOTE: If there is an error creating the encrypted socket, the channel for the referenced 
     * <code>unencryptedSocket</code> is closed.
     * 
     * @param unencryptedSocket The socket unencrypted socket to wrap. Must not be <code>null</code>.
     * @param host The hostname to bind the socket to. Must not be <code>null</code>.
     * @param port The host port to bind the socket to.
     * @return The encrypted socket. Never <code>null</code>.
     * @throws IOException Thrown if there are errors creating the encrypted socket.
     */
    private Socket createEncryptedSocket(Socket unencryptedSocket, String host, int port) throws IOException, UnknownHostException {
        _logger.debug("Creating encrypted socket"); //$NON-NLS-1$
        
        boolean encryptedSocketFailure = true;
        try {
            Socket socket = createSocket(unencryptedSocket, host, port, true);
            encryptedSocketFailure = false;
            return socket;
        }
        finally {
            if (encryptedSocketFailure && unencryptedSocket.getChannel() != null) {
                _logger.debug("Failure creating encrypted socket"); //$NON-NLS-1$
                unencryptedSocket.getChannel().close();
            }
        }
    }
    
    private Socket createSocket(Socket socket, String host, int port, boolean autoClose) throws IOException, UnknownHostException {
        
        SSLSocket sslSocket = (SSLSocket)sslContext.getSocketFactory().createSocket(socket, host, port, autoClose);
        sslSocket.addHandshakeCompletedListener(handshakeListener);
        
        // The hostname verifying code below needs to have a valid SSLSessions started
        // with the certificates already exchanged, so calling startHandshake() should 
        // setup the socket correctly.  The javadoc for startHandshake() says that the
        // initial setup in synchronous, so calling this before any other socket read 
        // should block until there is a valid session.
        _logger.debug("Starting handshake"); //$NON-NLS-1$
        sslSocket.startHandshake();
        
        BrowserCompatHostnameVerifier verifier = new BrowserCompatHostnameVerifier();
        try {
            _logger.debug("Verifying hostname"); //$NON-NLS-1$
            verifier.verify(host, sslSocket);
        } catch (IOException exp) {
            // An IOException would be thrown if the certificate does 
            // not match the hostname
            handleCertificateHostnameMismatch(host, port, sslSocket, exp);
        }
        
        return sslSocket;
    }

    public X509Certificate [] getLastKnownServerCertificatePath() {
    	return lastKnownServerCertPath;
    }
    
}
