/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.repository.transport.client;

import java.io.File;
import java.io.IOException;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.cert.Certificate;
import java.security.cert.CertificateException;

/**
 * @since 0.7
 */
public class SharedTrustStoreProvider implements ITrustStoreProvider {

    // In-memory trust store of all certificates explicitly accepted by the
    // certificate validator during this session. The validator will not be
    // called again during this session for any of these certificates. These may
    // include expired, not yet valid, or otherwise untrusted certificates.
    // These are kept distinctly, rather than merged into the runtime trust
    // store, because the base trust manager will never accept expired, etc.
    // certificates, even if from a trusted source.
    private static CertificateStore sessionCerts;

    // In-memory trust store of all permanently trusted certificates, assembled
    // from a number of key store files. These are provided to the base trust
    // manager as the basis for its decision making.
    private static CertificateStore runtimeCerts;

    // Location and password of the user's private trust store for this application.
    private static final String userTrustStoreLocation;
    private static final String userTrustStorePassword;

    static {
        String userHome = System.getProperty("user.home"); //$NON-NLS-1$
        userTrustStoreLocation = userHome + "/.jazzcerts".replace('/', File.separatorChar); //$NON-NLS-1$
        userTrustStorePassword = "ibmrationaljazz"; //$NON-NLS-1$
    }
	
    /**
     * Build the in-memory session trust store.
     * 
     * @return
     * @throws IOException
     * @throws CertificateException
     * @throws NoSuchAlgorithmException
     * @throws KeyStoreException
     */
    private static synchronized void buildSessionTrustStore() throws KeyStoreException, NoSuchAlgorithmException, CertificateException, IOException {
        if (sessionCerts == null) {
            sessionCerts = new CertificateStore();
        }
    }

    /**
     * Build the in-memory runtime trust store, a key store containing all
     * permanently trusted certificates, assembled form multiple on-disk
     * locations. These are provided to the base trust manager as the basis for
     * its decision making.
     * 
     * @return
     * @throws IOException
     * @throws CertificateException
     * @throws NoSuchAlgorithmException
     * @throws KeyStoreException
     */
    private static synchronized void buildRuntimeTrustStore() throws KeyStoreException, NoSuchAlgorithmException, CertificateException, IOException {
        if (runtimeCerts != null) {
            return;
        }
        
        CertificateStore store = loadSystemCertificates();
        
        // Application-specific keystore for the current user
        addCertificatesFromStore(store, userTrustStoreLocation, userTrustStorePassword);

        runtimeCerts = store;
    }
    
    public static CertificateStore loadSystemCertificates() throws KeyStoreException, NoSuchAlgorithmException, CertificateException, IOException {
    	CertificateStore store = new CertificateStore();
    	
    	String javaHome = System.getProperty("java.home"); //$NON-NLS-1$
    	
    	// JRE keystore override
        String file = System.getProperty("javax.net.ssl.trustStore"); //$NON-NLS-1$
        String password = System.getProperty("javax.net.ssl.trustStorePassword"); //$NON-NLS-1$
        addCertificatesFromStore(store, file, password);
        
        // JRE Signer CA keystore
        file = (javaHome + "/lib/security/cacerts").replace('/', File.separatorChar); //$NON-NLS-1$
        password = null;
        addCertificatesFromStore(store, file, password);
        
        // JRE Secure Site CA keystore
        file =  (javaHome + "/lib/security/jssecacerts").replace('/', File.separatorChar); //$NON-NLS-1$
        password = null;
        addCertificatesFromStore(store, file, password);
        
        return store;
    }
    
    public static void addCertificatesFromStore(CertificateStore store, String file, String password) {
    	if (file == null || file.length() < 1)
    		return;
    	
        File f = new File(file);
        if (f.canRead()) {
            // Load trust store into the runtime trust store
            try {
                store.loadCertificates(f, password);
            } catch (Exception e) {
                // Discard errors
            }
        }
    }
    
    public SharedTrustStoreProvider() {
        try {
            if (sessionCerts == null) {
                buildSessionTrustStore();
            }

            if (runtimeCerts == null) {
                buildRuntimeTrustStore();
            }

        } catch (KeyStoreException e) {
        } catch (NoSuchAlgorithmException e) {
        } catch (CertificateException e) {
        } catch (IOException e) {
        }    	
    }
	
	public CertificateStore getRuntimeTrustStore() {
		return runtimeCerts;
	}

	public CertificateStore getSessionTrustStore() {
		return sessionCerts;
	}
	
    /**
     * Install a certificate in the user's application-specific on-disk key
     * store, if possible.
     * 
     * @param cert
     * @throws NoSuchAlgorithmException
     * @throws CertificateException
     * @throws KeyStoreException
     * @throws IOException
     */
    public void installCertificate(Certificate cert) throws NoSuchAlgorithmException, CertificateException, KeyStoreException, IOException {
        File  f = new File(userTrustStoreLocation);
        KeyStore ks = CertificateStore.load(f, userTrustStorePassword);
        ks.setCertificateEntry(CertificateStore.computeAlias(cert), cert);
        CertificateStore.store(ks, f, userTrustStorePassword);
    }
	

}
