/*******************************************************************************
 * Licensed Materials - Property of IBM
 * © Copyright IBM Corporation 2010, 2014. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/

package com.ibm.team.repository.transport.client;

import java.io.InputStream;
import java.net.URI;
import java.net.URISyntaxException;

import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.common.transport.TeamServiceException;

/**
 * @since 0.7
 */
public class TeamRawRestServiceClient implements IInternalRawRestServiceClient {

    private final RemoteTeamServer _remoteTeamServer;
    private IOAuthHandler _oauthHandler = null;
    
    /**
     * Constructs an {@link ITeamRawRestServiceClient} instance.
     * 
     * @param remoteTeamServer The remote team server. Must not be <code>null</code>.
     */
    public TeamRawRestServiceClient(RemoteTeamServer remoteTeamServer)
    {
        super();
        
        if (remoteTeamServer == null)
            throw new IllegalArgumentException("remoteTeamServer must not be null"); //$NON-NLS-1$
        
        _remoteTeamServer = remoteTeamServer; 
    }
    
    synchronized public IOAuthHandler setOAuthHandler(IOAuthHandler oauthHandler) {
        IOAuthHandler priorHandler = _oauthHandler;
        _oauthHandler = oauthHandler;
        return priorHandler;
    }
    
    /* (non-Javadoc)
     * @see com.ibm.team.repository.transport.client.ITeamRawRestServiceClient#getConnection(java.net.URI)
     */
    public IRawRestClientConnection getConnection(URI uri) throws URISyntaxException {
        if (uri == null)
            throw new IllegalArgumentException("relative URI must not be null"); //$NON-NLS-1$
        
        return new RawRestClientConnection(_remoteTeamServer, uri, _oauthHandler);
    }
    
    

    /**
     * The {@link IRawRestClientConnection} implementation.
     */
    static public class RawRestClientConnection extends RestClientConnectionBase implements IRawRestClientConnection {
    	private final String serviceURI;

        /**
         * Constructs a rest client connection instance.
         * 
         * @param server
         *            The remote team server. Must not be <code>null</code>.
         * 
         * @param uri
         *            The URI of the resource to connect to. Must not
         *            be <code>null</code>.  May be absolute or relative.
         *            
         * @param handler The OAuth handler to use to react to OAuth authentication
         * challenges. May be <code>null</code>. 
         * @throws URISyntaxException 
         */
        public RawRestClientConnection(RemoteTeamServer server, URI uri, IOAuthHandler handler) throws URISyntaxException {
            super(server, uri, handler);
            serviceURI = uri.toString();
        }

        /* (non-Javadoc)
         * @see com.ibm.team.repository.transport.client.ITeamRawRestServiceClient.ITeamRawRestClientConnection#addRequestHeader(java.lang.String, java.lang.String)
         */
        public void addRequestHeader(String name, String value) {
            super.addRequestHeader(name, value);
        }               

        /* (non-Javadoc)
         * @see com.ibm.team.repository.transport.client.ITeamRawRestServiceClient.ITeamRawRestClientConnection#doGet()
         */
        public Response doGet() throws TeamServiceException, TeamRepositoryException {
            return super.doGet();
        }
        
        /* (non-Javadoc)
         * @see com.ibm.team.repository.transport.client.ITeamRawRestServiceClient.ITeamRawRestClientConnection#doGet()
         */
        public Response doHead() throws TeamServiceException, TeamRepositoryException {
            return super.doHead();
        }

        /* (non-Javadoc)
         * @see com.ibm.team.repository.transport.client.ITeamRawRestServiceClient.ITeamRawRestClientConnection#doPost(java.io.InputStream, long, java.lang.String)
         */
        public Response doPost(InputStream content, long contentLength,
                String contentType) throws TeamServiceException, TeamRepositoryException
        {
            return super.doPost(content, contentLength, contentType);
        }

        /* (non-Javadoc)
         * @see com.ibm.team.repository.transport.client.ITeamRawRestServiceClient.ITeamRawRestClientConnection#doPut(java.io.InputStream, long, java.lang.String)
         */
        public Response doPut(InputStream content, long contentLength,
                String contentType) throws TeamServiceException, TeamRepositoryException
        {
            return super.doPut(content,contentLength, contentType);
        }

        /* (non-Javadoc)
         * @see com.ibm.team.repository.transport.client.ITeamRawRestServiceClient.ITeamRawRestClientConnection#doDelete()
         */
        public Response doDelete() throws TeamServiceException, TeamRepositoryException {
            return super.doDelete();
        }
        
        /* (non-Javadoc)
         * @see com.ibm.team.repository.transport.client.RestClientConnectionBase#release()
         */
        public void release() {
            super.release();
        }

        public String getServiceUri() {
        	return serviceURI;
        }        
    }

}
