/*******************************************************************************
 * Licensed Materials - Property of IBM
 * © Copyright IBM Corporation 2010, 2015. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/

package com.ibm.team.repository.transport.client;

import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;

import org.apache.commons.httpclient.HttpMethod;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.http.Header;
import org.apache.http.HttpStatus;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpDelete;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpOptions;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.client.methods.HttpPut;
import org.apache.http.client.methods.HttpRequestBase;
import org.apache.http.entity.BufferedHttpEntity;
import org.apache.http.entity.ContentType;
import org.apache.http.entity.InputStreamEntity;
import org.apache.http.protocol.HTTP;
import org.apache.http.util.EntityUtils;

import com.ibm.team.repository.common.IItem;
import com.ibm.team.repository.common.ItemNotFoundException;
import com.ibm.team.repository.common.Location;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.common.internal.marshal.MarshalFactory;
import com.ibm.team.repository.common.internal.marshal.Marshaller;
import com.ibm.team.repository.common.internal.marshal.MarshallerType;
import com.ibm.team.repository.common.internal.marshal.MarshallingException;
import com.ibm.team.repository.common.transport.HttpUtil;
import com.ibm.team.repository.common.transport.HttpUtil.CharsetEncoding;
import com.ibm.team.repository.common.transport.HttpUtil.MediaType;
import com.ibm.team.repository.common.transport.IAbortableHttpRequest;
import com.ibm.team.repository.common.transport.ITeamService;
import com.ibm.team.repository.common.transport.TeamServiceCallContextImpl;
import com.ibm.team.repository.common.transport.TeamServiceException;
import com.ibm.team.repository.common.util.NLS;
import com.ibm.team.repository.transport.client.ITeamRestServiceClient.IRestClientConnection.Response.StatusCode;
import com.ibm.team.repository.transport.internal.nls.Messages;

/**
 * The implementation of the {@link ITeamRestServiceClient} interface.
  */
public class TeamRestServiceClient implements ITeamRestServiceClient {

    static private final String UserAgent = TeamRestServiceClient.class.getName();
    
    private final RemoteTeamServer _remoteTeamServer;
    private final String _serviceBinding;

    /**
     * Constructs an {@link ITeamRestServiceClient} instance.
     * 
     * @param remoteTeamServer The remote team server. Must not be <code>null</code>.
     * 
     * @param serviceClass The remote team service this client binds to. Must
     * not be <code>null</code>.
     * 
     * @param serviceClassLoader  The service class loader.  May be <code>null</code> -
     * it is ignored in this context.
     */
    public TeamRestServiceClient(RemoteTeamServer remoteTeamServer, 
            Class<?> serviceClass, ClassLoader serviceClassLoader)
    {
        super();
        
        if (remoteTeamServer == null)
            throw new IllegalArgumentException("remoteTeamServer must not be null"); //$NON-NLS-1$
        if (serviceClass == null)
            throw new IllegalArgumentException("serviceClass must not be null"); //$NON-NLS-1$
        
        _remoteTeamServer   = remoteTeamServer; 
        _serviceBinding = serviceClass.getName();
    }
    
    /**
     * Constructs an {@link ITeamRestServiceClient} instance.
     * 
     * @param remoteTeamServer The remote team server. Must not be <code>null</code>.
     * 
     * @param serviceBinding The remote team service this client binds to. Must be the 
     * fully qualified name of a remote team service. Must not be <code>null</code>.
     * 
     * @param serviceClassLoader  The service class loader.  May be <code>null</code> -
     * it is ignored in this context.
     */
    public TeamRestServiceClient(RemoteTeamServer remoteTeamServer, 
    		String serviceBinding, ClassLoader serviceClassLoader)
    {
    	super();
    	
    	if (remoteTeamServer == null)
    		throw new IllegalArgumentException("remoteTeamServer must not be null"); //$NON-NLS-1$
    	if (serviceBinding == null)
    		throw new IllegalArgumentException("serviceBinding must not be null"); //$NON-NLS-1$
    	
    	_remoteTeamServer   = remoteTeamServer; 
    	_serviceBinding = serviceBinding;
    }

    /* (non-Javadoc)
     * @see com.ibm.team.repository.common.transport.ITeamRestServiceClient#getConnection(com.ibm.team.repository.common.Location)
     */
    public IRestClientConnection getConnection(Location loc) {
        if (loc == null)
            throw new IllegalArgumentException("location must not be null"); //$NON-NLS-1$
        
        return new RestClientConnection(_remoteTeamServer, _serviceBinding, loc);
    }

    /* (non-Javadoc)
     * @see com.ibm.team.repository.common.transport.ITeamRestServiceClient#getConnection(java.net.URI)
     */
    public IRestClientConnection getConnection(URI relUri) {
        if (relUri == null)
            throw new IllegalArgumentException("relative URI must not be null"); //$NON-NLS-1$
        
        return new RestClientConnection(_remoteTeamServer, _serviceBinding, relUri);
    }

    /**
     * The {@link IRestClientConnection} implementation.
     */
    static public class RestClientConnection implements IRestClientConnection {

        private final RemoteTeamServer _remoteTeamServer;
        private final String _repoUrl;
        private final String _serviceBinding;
        private final Location _loc;
        private final URI _relUri;
        
        private final HashMap<String, String> _headers = new HashMap<String, String>();
        
        protected boolean _hasBeenUsed = false;
        
        private Response _response = null;
        
        /**
         * Constructs a rest client connection instance.
         * 
         * @param server The remote team server. Must not be <code>null</code>.
         * 
         * @param serviceBinding
         *            The fully qualified name of a service class this connection 
         *            binds to. Must not be <code>null</code>.
         * 
         * @param loc The location of the resource to connect to. Must not
         * be <code>null</code>.
         */
        public RestClientConnection(RemoteTeamServer server, String serviceBinding,
                Location loc) {
            if (server == null)
                throw new IllegalArgumentException("remoteTeamServer must not be null"); //$NON-NLS-1$
            if (serviceBinding == null)
                throw new IllegalArgumentException("serviceBinding must not be null"); //$NON-NLS-1$
            if (loc == null)
                throw new IllegalArgumentException("loc must not be null"); //$NON-NLS-1$
                
            _remoteTeamServer = server; 
            _repoUrl = server.getRepositoryURL();
            _serviceBinding = serviceBinding;
            _loc = loc;
            _relUri = null;
        }

        /**
         * Constructs a rest client connection instance.
         * 
         * @param server
         *            The remote team server. Must not be <code>null</code>.
         * 
         * @param serviceBinding
         *            The fully qualified name of a service class this connection 
         *            binds to. Must not be <code>null</code>.
         * 
         * @param relUri
         *            The relative URI of the resource to connect to. Must not
         *            be <code>null</code>.
         */
        public RestClientConnection(RemoteTeamServer server,
                String serviceBinding, URI relUri) {
            if (server == null)
                throw new IllegalArgumentException("server must not be null"); //$NON-NLS-1$
            if (serviceBinding == null)
                throw new IllegalArgumentException("serviceClass must not be null"); //$NON-NLS-1$
            if (relUri == null)
                throw new IllegalArgumentException("relUri must not be null"); //$NON-NLS-1$
                
            _remoteTeamServer = server; 
            _repoUrl = server.getRepositoryURL();
            _serviceBinding = serviceBinding;
            _loc = null;
            _relUri = relUri;
        }

        /* (non-Javadoc)
         * @see com.ibm.team.repository.common.transport.ITeamRestServiceClient.IRestClientConnection#addRequestHeader(java.lang.String, java.lang.String)
         */
        public void addRequestHeader(String name, String value) {
            
            if (_hasBeenUsed)
                throw new IllegalStateException("Cannot add request headers to a connection has been previously run"); //$NON-NLS-1$
            
            if (name == null || name.length() == 0)
                return;
            
            if (value == null)
                value = ""; //$NON-NLS-1$
            
            _headers.put(name, value);
        }

        /* (non-Javadoc)
         * @see com.ibm.team.repository.common.transport.ITeamRestServiceClient.IRestClientConnection#doGet()
         */
        public Response doGet() throws TeamServiceException, TeamRepositoryException {
            return doMethod(new HttpGet());
        }
        
        /* (non-Javadoc)
         * @see com.ibm.team.repository.transport.client.ITeamRestServiceClient.IRestClientConnection#doOptions()
         */
        public Response doOptions() throws TeamServiceException, TeamRepositoryException {
            return doMethod(new HttpOptions());
        }

        /* (non-Javadoc)
         * @see com.ibm.team.repository.common.transport.ITeamRestServiceClient.IRestClientConnection#doPost(java.io.InputStream, long, java.lang.String)
         */
        public Response doPost(InputStream content, long contentLength,
                String contentType) throws TeamServiceException, TeamRepositoryException {

            HttpPost postMethod = new HttpPost();

            BufferedHttpEntity requestEntity;
            try {
                requestEntity = new BufferedHttpEntity(
                        new InputStreamEntity(content, contentLength, ContentType.create(contentType)));
            } catch (IOException e) {
                throw new TeamServiceException(e.getMessage(), e);
            }
            postMethod.setEntity(requestEntity);

            return doMethod(postMethod);
        }

        /* (non-Javadoc)
         * @see com.ibm.team.repository.common.transport.ITeamRestServiceClient.IRestClientConnection#doPut(java.io.InputStream, long, java.lang.String)
         */
        public Response doPut(InputStream content, long contentLength,
                String contentType) throws TeamServiceException, TeamRepositoryException {

            HttpPut putMethod = new HttpPut();

            BufferedHttpEntity requestEntity;
            try {
                requestEntity = new BufferedHttpEntity(
                        new InputStreamEntity(content, contentLength, ContentType.create(contentType)));
            } catch (IOException e) {
                throw new TeamServiceException(e.getMessage(), e);
            }
            putMethod.setEntity(requestEntity);
            
            return doMethod(putMethod);
        }

        
        /* (non-Javadoc)
         * @see com.ibm.team.repository.common.transport.ITeamRestServiceClient.IRestClientConnection#doDelete()
         */
        public Response doDelete() throws TeamServiceException, TeamRepositoryException {

            return doMethod(new HttpDelete());
        }

        /**
         * Helper for the various do* methods.
         * 
         * @param request The http method to use. Must not be <code>null</code>.
         * @return The response object. Never <code>null</code>.
         * @throws TeamServiceException Thrown on non-recoverable server I/O error.
         * @throws TeamRepositoryException Thrown on recoverable repository errors.
         */
        protected Response doMethod(HttpRequestBase request) throws TeamServiceException, TeamRepositoryException {
            
            if (_hasBeenUsed) 
                throw new IllegalStateException("do* called multiple times on the same connection"); //$NON-NLS-1$
            
            _hasBeenUsed = true;
            
            RequestResponsePair method = new RequestResponsePair(request);
            prepareMethod(method);
            
            int status = -1;
            TeamRepositoryException caughtException = null;
            
            try {
                executeMethod(method);
            } catch (TeamRepositoryException e) {
                caughtException = e;
                if (method.getResponse() != null)
                    status = method.getResponse().getStatusLine().getStatusCode();
            }

            CloseableHttpResponse response = method.getResponse();
            if (response == null && caughtException != null) {
                throw caughtException;
            }
            if (response == null) {
                throw new TeamRepositoryException("Response is unexpectedly null - should never happen"); //$NON-NLS-1$
            }
            
            Header[] headers = response.getAllHeaders();
            Map<String, String> headersMap = new HashMap<String, String>(
                    headers.length);
            for (Header header : headers) {
                headersMap.put(header.getName(), header.getValue());
            }

            /*
             * Only create the stream if there's a entity on the response
             */
            InputStream responseStream = null;
            try {
                if (response.getEntity() != null)
                    responseStream = new ReleasingAndAbortingInputStream(response.getEntity());

            } catch (IOException e) {
                EntityUtils.consumeQuietly(response.getEntity());
                throw new TeamServiceException(e.getMessage(), e);
            }
            
            String statusText = response.getStatusLine().getReasonPhrase();
            status = response.getStatusLine().getStatusCode();
            _response = new Response(responseStream, headersMap, status, statusText, 
                    httpStatusToStatusCode(status));
            
            if (caughtException != null)
                throw caughtException;
            
            return _response;
        }

        /**
         * Returns a StatusCode for the referenced HTTP response status.
         * 
         * @param retVal A value from the domain of HTTP response status.
         * 
         * @return The status code. Never <code>null</code>.
         */
        private StatusCode httpStatusToStatusCode(int retVal) {
            StatusCode statusCode;
            switch (retVal) {
            case HttpStatus.SC_OK:
                statusCode = StatusCode.OK;
                break;
            case HttpStatus.SC_CREATED:
                statusCode = StatusCode.CREATED;
                break;
            case HttpStatus.SC_NOT_FOUND:
                statusCode = StatusCode.NOT_FOUND;
                break;
            case HttpStatus.SC_NOT_MODIFIED:
                statusCode = StatusCode.NOT_MODIFIED;
                break;
            default:
                statusCode = StatusCode.ERROR;
                break;
            }
            
            return statusCode;
        }

        /* (non-Javadoc)
         * @see com.ibm.team.repository.common.transport.ITeamRestServiceClient.IRestClientConnection#release()
         */
        public void release() {

            if (_response == null)
                return;
            
            InputStream responseStream = _response.getResponseStream();
            if (responseStream != null) {
                try {
                    responseStream.close();
                } catch (IOException e) {
                    logger.warn(String.format("Unexpected IO error on connection release: %s", e.getMessage())); //$NON-NLS-1$
                }
            }
        }

        /* (non-Javadoc)
         * @see com.ibm.team.repository.common.transport.ITeamRestServiceClient.IRestClientConnection#unmarshall(java.io.InputStream)
         */
        public IItem unmarshall(InputStream response) throws TeamRepositoryException {
            
            if (response == null)
                throw new IllegalArgumentException("response must not be null"); //$NON-NLS-1$
            
            Marshaller marshaller = MarshalFactory.eINSTANCE
                    .getMarshaller(MarshallerType.EOBJECT_LITERAL);
            try {
                return (IItem) marshaller.demarshalInputStreamToObject(response);
            } catch (MarshallingException e) {
                throw new TeamRepositoryException(e);
            }
        }

        /**
         * Returns the absolute location that this connection uses to contact the
         * server.  Note that this location may be different than what was passed
         * to {@link ITeamRestServiceClient#getConnection(Location)}, if the client
         * is bound to a different server and service than what is referenced in
         * the specified Location. 
         * 
         * @return The location that this connection uses to contact the server.
         * Never <code>null</code>. 
         */
        private Location boundLocation() {
            return Location.location(_loc, _repoUrl, _serviceBinding); 
        }

        /* (non-Javadoc)
         * @see com.ibm.team.repository.common.transport.ITeamRestServiceClient.IRestClientConnection#boundAbsoluteUri()
         */
        public URI boundAbsoluteUri() throws TeamRepositoryException {
            if (_loc != null) {
                return boundLocation().toAbsoluteUri();
            } else {
                // TODO improve this, use constants, use URI class more
                final String SERVICE_BINDING_PREFIX = Location.SERVICE_BINDING_PREFIX + Location.SLASH;
                String s = this._repoUrl + SERVICE_BINDING_PREFIX
                        + this._serviceBinding + Location.SLASH + HttpUtil.trimLeadingSlash(this._relUri.toString());
                try {
                    return new URI(s);
                } catch (URISyntaxException e) {
                    throw new TeamRepositoryException(
                            NLS.bind(Messages.getServerString("TeamRestServiceClient.UriSyntaxError"), s), e); //$NON-NLS-1$
                }
            }
        }

        /**
         * Prepares an {@link HttpMethod} for execution.  Callers first
         * create an {@link HttpMethod} specialization and set the
         * {@link #_httpMethod} variable, call {@link #prepareMethod()}
         * and then call {@link #executeMethod()}.
         * 
         * @throws TeamServiceException  Thrown on error.
         */
        private void prepareMethod(RequestResponsePair method) throws TeamServiceException {
            
            HttpRequestBase request = method.getRequest();
            
            try {
                URI uri = boundAbsoluteUri();
                /*
                 * Strip off protocol, host, and port if present e.g. "https://jazz.net:9943"
                 * to ensure our ProtocolSocketFactory is used
                 * see: http://issues.apache.org/jira/browse/HTTPCLIENT-634
                 * 
                 * Per workitem 109606 I've pulled the parsing done by URI, as the 
                 * quoting was incorrect
                 */
                String formatString = "%s"; //$NON-NLS-1$
                if (uri.getQuery() != null) 
                    formatString += "?%s"; //$NON-NLS-1$
                if (uri.getFragment() != null)
                    formatString += "#%s"; //$NON-NLS-1$
                
                try {
                    uri = new URI(String.format(formatString, uri.getRawPath(), 
                            uri.getRawQuery(), uri.getRawFragment()));
                } catch (URISyntaxException e) {
                    logger.warn(String.format("URISyntaxException parsing URI: %s", e.getMessage())); //$NON-NLS-1$
                }
                request.setURI(uri);
           } catch (IllegalStateException e) {
               throw new TeamServiceException(e.getMessage());
           } catch (TeamRepositoryException e) {
               throw new TeamServiceException(e.getMessage());
           }

			/*
			 * Add in the request headers, and then fold in our proprietary headers.
			 */
			Iterator<Entry<String, String>> iter = _headers.entrySet().iterator();
			while (iter.hasNext()) {
				Entry<String, String> entry = iter.next();
				String key = entry.getKey();
				String value = entry.getValue();
				request.addHeader(key, value);
			}

           /*
            * The 5.0.2 and earlier versions of this class set a non-standard header given by the value
            * of the org.apache.commons.httpclient.params.HttpMethodParams.USER_AGENT constant.
            * We still that header, for back-compatibility, as well as the standard User-Agent header.
            */
           request.setHeader(HTTP.USER_AGENT, UserAgent);
           request.setHeader(HttpUtil.NON_STD_USER_AGENT_HEADER, UserAgent);
           if (request.getFirstHeader(HttpUtil.ACCEPT) == null)
               request.setHeader(HttpUtil.ACCEPT, MediaType.XML.toString());
           if (request.getFirstHeader(HttpUtil.ACCEPT_CHARSET) == null)
               request.setHeader(HttpUtil.ACCEPT_CHARSET, CharsetEncoding.UTF8.toString());
           if (request.getFirstHeader(HttpUtil.ACCEPT_LANGUAGE) == null)
               TransportUtils.setAcceptLanguageHeader(request);

           addServiceVersionHeader(request, _serviceBinding);
           String userid = _remoteTeamServer.getUserid();
           TransportUtils.addUseridHeader(userid, request);
       }
        
    	private void addServiceVersionHeader(HttpRequestBase method, String serviceName) {
    		ITeamService teamService = _remoteTeamServer.getServiceByName(serviceName);
    		if(teamService == null) return;
    		String version = teamService.getVersion();
    		if ((version != null) && version.length() > 0)
    			method.addHeader(HttpUtil.X_SERVICE_VERSION, version);
    	}        

        /**
         * Executes a prepared {@link HttpMethod}.  Callers first
         * create an {@link HttpMethod} specialization and set the
         * {@link #_httpRequest} variable, call {@link #prepareMethod()}
         * and then call {@link #executeMethod()}.
         * 
         * @return The HTTP status code.
         * 
         * @throws TeamServiceException Thrown on non-recoverable server I/O error.
         * @throws TeamRepositoryException Thrown on recoverable repository errors.
         */
        protected int executeMethod(RequestResponsePair method) throws TeamServiceException, TeamRepositoryException {
            
            int status = 500;
            
            final HttpRequestBase cancellableMethod = method.getRequest();
            TeamServiceCallContextImpl.setUpCall(new IAbortableHttpRequest() {
                
                @Override
                public String getURI() {
                    return cancellableMethod.getURI().toString();
                }
                
                @Override
                public void abort() {
                    cancellableMethod.abort();
                }
            });
            
            try {
                try {
                    status = _remoteTeamServer.executeMethod(null, method, false, true);
                    if (status == 404) {
                        throw new ItemNotFoundException(method.getRequest().getURI().toString());
                    }
                } catch (IOException e) {
                    throw new TeamServiceException(e.getMessage());
                } catch (TeamRepositoryException e) {
                    throw e;
                }
            }
            finally {
                TeamServiceCallContextImpl.tearDownCall();
            }
             
            return status; 
                
        }

       /* (non-Javadoc)
        * @see java.lang.Object#toString()
        */
       @Override
       public String toString() {
            return (_loc != null) ? _loc.toString() : _relUri.toString();
        }

       /* (non-Javadoc)
        * @see com.ibm.team.repository.transport.client.ITeamRestServiceClient.IRestClientConnection#getServiceUri()
        */
       public String getServiceUri() {
           return _serviceBinding;
       }

       /* (non-Javadoc)
        * @see com.ibm.team.repository.transport.client.ITeamRestServiceClient.IRestClientConnection#getResponse()
        */
       public Response getResponse() {
           return _response;
       }
       
       static private final Log logger = LogFactory.getLog(TeamRestServiceClient.class);
    }

}
