/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/

package com.ibm.team.repository.transport.client;

import com.ibm.team.repository.common.transport.ICertificateValidator;
import com.ibm.team.repository.common.transport.ITeamServer;
import com.ibm.team.repository.common.transport.LenientCertificateValidator;

/**
 * Contains methods to create an ITeamServer and to validate URLs to them.
 * 
 * An ITeamServer is created with a URL, which may have the form:
 * <ul>
 * <li>starts with "local-server:" for a local team server (used by servers)
 * <li>starts with "local:"for a local proxy team server (used by 'local' server testing)
 * <li>a valid http: or https: URL
 * </ul>
 */
public class TeamServerFactory {

	/**
	 * 
	 */
	public static final TeamServerFactory INSTANCE = new TeamServerFactory();

	/**
	 * URL used to create a local ITeamServer that sits behind a web server. 
	 */
	public static final String URL_LOCAL = TeamServerFactoryImpl.URL_LOCAL;
	
	/**
	 * URL used to create a local proxy ITeamServer for testing.
	 */
	public static final String URL_LOCAL_PROXY = TeamServerFactoryImpl.URL_LOCAL_PROXY;
	
	/**
	 * URL is valid.
	 */
	public static final int URL_VALID             = TeamServerFactoryImpl.URL_VALID;
	
	/**
	 * URL is invalid, but can't say why. 
	 */
	public static final int URL_INVALID_UNKNOWN   = TeamServerFactoryImpl.URL_INVALID_UNKNOWN;
	
	/**
	 * URL has invalid protocol.
	 */
	public static final int URL_INVALID_PROTOCOL  = TeamServerFactoryImpl.URL_INVALID_PROTOCOL;
	
	/**
	 * URL has invalid host.
	 */
	public static final int URL_INVALID_HOST      = TeamServerFactoryImpl.URL_INVALID_HOST;
	
	/**
	 * URL has invalid port.
	 */
	public static final int URL_INVALID_PORT      = TeamServerFactoryImpl.URL_INVALID_PORT;
	
	public static String SERVER_PORT = TeamServerFactoryImpl.SERVER_PORT;
	
	/**
	 *
	 */
	private TeamServerFactory() {
		super();
	}

	/**
	 * Validates whether a given URL is valid or not, structurally.
	 * Does not check to see if there is a real repository on the other end.
	 * Returns one of URL_VALID or one of the URL_INVALID_* constants,
	 * depending on whether the URL appears to be valid, and if it's not, if
	 * it can be determined why it's not valid.
	 */
	public int validateURL(String urlString) {
		return TeamServerFactoryImpl.INSTANCE.validateURL(urlString);
	}
	
	/**
	 * Create a new ITeamServer based on the specified URL
	 */
	public ITeamServer newTeamServerFromURL(String url) {
	    // TODO remove this method, or make it strict rather than lenient wrt SSL certificates
		return TeamServerFactoryImpl.INSTANCE.newTeamServerFromURL(url, LenientCertificateValidator.INSTANCE);
	}

    /**
     * Create a new ITeamServer based on the specified URL
     */
    public ITeamServer newTeamServerFromURL(String url, ICertificateValidator handler) {
        return TeamServerFactoryImpl.INSTANCE.newTeamServerFromURL(url, handler);
    }
}
