/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010, 2012. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/

package com.ibm.team.repository.transport.client;

import java.net.MalformedURLException;
import java.net.URL;
import java.util.List;

import com.ibm.team.repository.common.IServerContainerProvider;
import com.ibm.team.repository.common.transport.HttpUtil;
import com.ibm.team.repository.common.transport.ICertificateValidator;
import com.ibm.team.repository.common.transport.IServerContainerDescriptor;
import com.ibm.team.repository.common.transport.ITeamServer;
import com.ibm.team.repository.common.transport.ServerContainerProviderRegistry;

/**
 * Contains methods to create an ITeamServer and to validate URLs to them.
 * 
 * An ITeamServer is created with a URL, which may have the form:
 * <ul>
 * <li>starts with "local-server:" for a local team server (used by servers)
 * <li>starts with "local:"for a local proxy team server (used by 'local' server testing)
 * <li>a valid http: or https: URL
 * </ul>
 */
public class TeamServerFactoryImpl {

	/**
	 * 
	 */
	public static final TeamServerFactoryImpl INSTANCE = new TeamServerFactoryImpl();

	/**
	 * URL used to create a local ITeamServer that sits behind a web server. 
	 */
	public static final String URL_LOCAL = "local-server:/"; //$NON-NLS-1$
	
	/**
	 * URL used to create a local proxy ITeamServer for testing.
	 */
	public static final String URL_LOCAL_PROXY = "local:/"; //$NON-NLS-1$

	/**
	 * URL is valid.
	 */
	public static final int URL_VALID             = 0;
	
	/**
	 * URL is invalid, but can't say why. 
	 */
	public static final int URL_INVALID_UNKNOWN   = 1;
	
	/**
	 * URL has invalid protocol.
	 */
	public static final int URL_INVALID_PROTOCOL  = 2;
	
	/**
	 * URL has invalid host.
	 */
	public static final int URL_INVALID_HOST      = 3;
	
	/**
	 * URL has invalid port.
	 */
	public static final int URL_INVALID_PORT      = 4;
	
	public static final String SERVER_PORT = System
	.getProperty("server.port", "9080"); //$NON-NLS-1$ //$NON-NLS-2$

	public static final String SERVER_PORT_SECURE = System
	.getProperty("server.port.secure", "9443");	 //$NON-NLS-1$ //$NON-NLS-2$
	
	/**
	 *
	 */
	private TeamServerFactoryImpl() {
		super();		
	}

	/**
	 * Validates whether a given URL is valid or not, structurally.
	 * Does not check to see if there is a real repository on the other end.
	 * Returns one of URL_VALID or one of the URL_INVALID_* constants,
	 * depending on whether the URL appears to be valid, and if it's not, if
	 * it can be determined why it's not valid.
	 */
	public int validateURL(String urlString) {
		
		if (null == urlString) return URL_INVALID_UNKNOWN;
		if (0 == urlString.length()) return URL_INVALID_UNKNOWN;
		
		if (!urlString.endsWith(HttpUtil.SLASH)) urlString += HttpUtil.SLASH;
		
		if (urlString.startsWith(URL_LOCAL)) return URL_VALID;
		if (urlString.startsWith(URL_LOCAL_PROXY)) return URL_VALID;
		
		URL url;
		try {
			url = new URL(urlString);
		}
		catch (MalformedURLException e) {
			return URL_INVALID_UNKNOWN;
		}

		String urlProtocol = url.getProtocol();
		String urlHost     = url.getHost();
		if ((null == urlProtocol) || (0 == urlProtocol.length())) {
			return URL_INVALID_PROTOCOL;
		}
		
		if (!urlProtocol.equals(HttpUtil.HTTP) && !urlProtocol.equals(HttpUtil.HTTPS)) {
			return URL_INVALID_PROTOCOL;
		}		

		if ((null == urlHost) || (0 == urlHost.length())) {
			return URL_INVALID_HOST;
		}

//		if (urlPort <= 0) return URL_INVALID_PORT;
		
		return URL_VALID;
	}
	
	public String startLocalServer(boolean secure) {
        List<IServerContainerDescriptor> descriptors = ServerContainerProviderRegistry.INSTANCE.getDescriptors();
        boolean empty = descriptors.isEmpty();
        
        if (empty == true) {
            // If we found no descriptors but are in local mode we have a problem
            throw new RuntimeException("No Server Container Descriptors were found on the classpath.  Please check that the Jetty bundles are included in your launch or packaging."); //$NON-NLS-1$              
        }

        for (IServerContainerDescriptor descriptor : descriptors) {
            try {
                IServerContainerProvider provider = descriptor.getProvider();
                provider.start();
            } catch (Exception e) {
                e.printStackTrace();
            }                   
        }
        
        if (secure) {
        	return "https://localhost:"+SERVER_PORT_SECURE+"/jazz/secure/"; //$NON-NLS-1$ //$NON-NLS-2$
        } else {
        	return "http://localhost:"+SERVER_PORT+"/jazz/"; //$NON-NLS-1$ //$NON-NLS-2$
        }
	}

	/**
     * Start the local-mode server. Normally this method is only called
     * internally by the transport layer client to start the local-mode server
     * (in response to a "local://" URI) for testing. The only external callers
     * should be tests that need to run in local mode but are not using the
     * transport layer client, e.g. REST tests.
     * 
     * @return the URI of the local-mode server.
     */
	public String startLocalServer() {
		return startLocalServer(false);
	}

	/**
	 * Create a new ITeamServer based on the specified URL
	 */
	public ITeamServer newTeamServerFromURL(String url, ICertificateValidator handler) {
		if (URL_VALID != validateURL(url)) return null;
		
		if (!url.endsWith(HttpUtil.SLASH)) url += HttpUtil.SLASH;

		if (url.startsWith(URL_LOCAL_PROXY)) {
			url = startLocalServer(url.endsWith("/secure/")); //$NON-NLS-1$
		}
		
		return new RemoteTeamServer(url, handler);
	}
}
