/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.repository.transport.client;

import java.io.IOException;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;

import javax.net.ssl.TrustManager;
import javax.net.ssl.TrustManagerFactory;
import javax.net.ssl.X509TrustManager;

import com.ibm.team.repository.common.transport.ICertificateValidator;

/**
 * A trust manager that will call a registered {@link ICertificateValidator} in
 * the event that a problematic (e.g. expired, not yet valid) or untrusted
 * certificate is presented by a server, and react appropriately. This trust
 * manager will rely on multiple key stores, and manage one of its own. The
 * managed key store and the session-accepted key store are shared by all trust
 * manager instances.
 * @since 0.5
 */
public class ValidatingX509TrustManager implements X509TrustManager {

    // The JRE-provided trust manager used to validate certificates presented by
    // a server.
    private X509TrustManager baseTrustManager;

    // The registered certificate validator, may be null, called when the base
    // trust manager rejects a certificate presented by a server.
    private ICertificateValidator validator;
    
    private ITrustStoreProvider trustStoreProvider;


    /**
     * Initialize the base X509 trust manager that will be used to evaluate
     * certificates presented by the server against the runtime trust store.
     * 
     * @throws KeyStoreException
     * @throws NoSuchAlgorithmException
     * @throws CertificateException 
     * @throws IOException 
     */
    private void initializeBaseTrustManager() throws KeyStoreException, NoSuchAlgorithmException, CertificateException, IOException {
        TrustManagerFactory factory = TrustManagerFactory.getInstance(TrustManagerFactory.getDefaultAlgorithm());
        factory.init(trustStoreProvider.getRuntimeTrustStore().getKeyStore());
        TrustManager[] managers = factory.getTrustManagers();
        for (TrustManager manager : managers) {
            if (manager instanceof X509TrustManager) {
                // Take the first X509TrustManager we find
                baseTrustManager = (X509TrustManager) manager;
                break;
            }
        }
        if (baseTrustManager == null) {
            throw new IllegalStateException("Couldn't find JRE's X509TrustManager"); //$NON-NLS-1$
        }
    }


    /**
     * Construct a new ValidatingX509TrustManager.
     * 
     * @param validator
     *            certificate validator to consult regarding problematic
     *            certificates, or <code>null</code> to always reject them.
     */
    public ValidatingX509TrustManager(ICertificateValidator validator) {
    	this(validator, new SharedTrustStoreProvider());
    }
    
    /**
     * Construct a new ValidatingX509TrustManager.
     * 
     * @param validator
     *            certificate validator to consult regarding problematic
     *            certificates, or <code>null</code> to always reject them.
     * @param trustStoreProvider
     *            the helper class that will provide the storage for permanent and
     *            temporary storage
     * @since 0.7
     */
    public ValidatingX509TrustManager(ICertificateValidator validator, ITrustStoreProvider trustStoreProvider) {
        this.validator = validator;
        this.trustStoreProvider = trustStoreProvider;
        
        try {
            // Initialize the base trust manager and the key store it will use
			initializeBaseTrustManager();
		} catch (KeyStoreException e) {
		} catch (NoSuchAlgorithmException e) {
		} catch (CertificateException e) {
		} catch (IOException e) {
		}

    }

    /**
     * Delegate to base trust manager.
     * 
     * @see X509TrustManager#getAcceptedIssuers()
     */
    public X509Certificate[] getAcceptedIssuers() {
        return baseTrustManager.getAcceptedIssuers();
    }

    /**
     * Delegate to base trust manager.
     * 
     * @see X509TrustManager#checkClientTrusted(X509Certificate[], String)
     */
    public void checkClientTrusted(X509Certificate[] chain, String authType) throws CertificateException {
        baseTrustManager.checkClientTrusted(chain, authType);
    }

    /**
     * Consult base trust manager and use registered certificate validator to
     * evaluate any untrusted or otherwise problematic certificates.
     * 
     * @see X509TrustManager#checkServerTrusted(X509Certificate[], String)
     */
    public void checkServerTrusted(X509Certificate[] chain, String authType) throws CertificateException {
        X509Certificate cert = chain[0];

        // Has the certificate been OK'd for the session?
        try {
            if (trustStoreProvider.getSessionTrustStore().containsCertificate(cert)) {
                return;
            } 
        } catch (KeyStoreException e) {
            // Ignore; proceed to try base trust manager
        }

        try {
            // Rely on the base trust manager to check the certificate against
            // the assembled runtime key store
            baseTrustManager.checkServerTrusted(chain, authType);
        } catch (CertificateException certEx) {
            // TODO handle CertificateExpiredException specially
            // TODO handle CertificateNotYetValidExceptoin specially

            // Done if there isn't a validator to consult
            if (validator == null) {
                throw certEx; // Rejected!
            }

            // Ask the registered certificate validator to rule on the certificate
            ICertificateValidator.Trust disposition = validator.validate(cert, certEx);
            switch (disposition) {
            case REJECT:
                throw certEx;

            case ACCEPT_CONNECTION:
                return;

            case ACCEPT_SESSION:
                try {
                	trustStoreProvider.getSessionTrustStore().enterCertificate(cert);
                } catch (KeyStoreException e) {
                }
                return;

            case ACCEPT_PERMANENT:
                try {
                	trustStoreProvider.getSessionTrustStore().enterCertificate(cert);
                	trustStoreProvider.installCertificate(cert);
                } catch (KeyStoreException e) {
                } catch (NoSuchAlgorithmException e) {
                } catch (IOException e) {
                }
                return;
            }
        }
    }
}
